@@ -40,7 +40,9 @@
 #include <map>
 #include <memory>
 #include <ostream>
+#include <type_traits>
 #include <unordered_map>
+#include <variant>
 
 #include "src/base/macros.h"
 #include "src/base/memory.h"
@@ -148,7 +150,7 @@ struct JumpOptimizationInfo {
   std::map<int, int> align_pos_size;
 
   int farjmp_num = 0;
-  // For collecting stage, should contains all far jump informatino after
+  // For collecting stage, should contains all far jump information after
   // collecting.
   std::vector<JumpInfo> farjmps;
 
@@ -204,7 +206,7 @@ enum class BuiltinCallJumpMode {
   // 1) we encode the target as an offset from the code range which is not
   // always available (32-bit architectures don't have it),
   // 2) serialization of RelocInfo::RUNTIME_ENTRY is not implemented yet.
-  // TODO(v8:11527): Address the resons above and remove the kForMksnapshot in
+  // TODO(v8:11527): Address the reasons above and remove the kForMksnapshot in
   // favor of kPCRelative or kIndirect.
   kForMksnapshot,
 };
@@ -215,15 +217,11 @@ struct V8_EXPORT_PRIVATE AssemblerOptions {
   // module. This flag allows this reloc info to be disabled for code that
   // will not survive process destruction.
   bool record_reloc_info_for_serialization = true;
-  // Recording reloc info can be disabled wholesale. This is needed when the
-  // assembler is used on existing code directly (e.g. JumpTableAssembler)
-  // without any buffer to hold reloc information.
-  bool disable_reloc_info_for_patching = false;
   // Enables root-relative access to arbitrary untagged addresses (usually
   // external references). Only valid if code will not survive the process.
   bool enable_root_relative_access = false;
   // Enables specific assembler sequences only used for the simulator.
-  bool enable_simulator_code = false;
+  bool enable_simulator_code = USE_SIMULATOR_BOOL;
   // Enables use of isolate-independent constants, indirected through the
   // root array.
   // (macro assembler feature).
@@ -250,9 +248,19 @@ struct V8_EXPORT_PRIVATE AssemblerOptions {
   // Whether to emit code comments.
   bool emit_code_comments = v8_flags.code_comments;
 
+  bool is_wasm = false;
+
   static AssemblerOptions Default(Isolate* isolate);
 };
 
+// Wrapper around an optional Zone*. If the zone isn't present, the
+// AccountingAllocator* may be used to create a fresh one.
+//
+// This is useful for assemblers that want to Zone-allocate temporay data,
+// without forcing all users to have to create a Zone before using the
+// assembler.
+using MaybeAssemblerZone = std::variant<Zone*, AccountingAllocator*>;
+
 class AssemblerBuffer {
  public:
   virtual ~AssemblerBuffer() = default;
@@ -267,8 +275,8 @@ class AssemblerBuffer {
 
 // Describes a HeapObject slot containing a pointer to another HeapObject. Such
 // a slot can either contain a direct/tagged pointer, or an indirect pointer
-// (i.e. an index into an indirect pointer table, which then contains the
-// actual pointer to the object) together with a specific IndirectPointerTag.
+// (i.e. an index into a pointer table, which then contains the actual pointer
+// to the object) together with a specific IndirectPointerTag.
 class SlotDescriptor {
  public:
   bool contains_direct_pointer() const {
@@ -292,7 +300,7 @@ class SlotDescriptor {
     return SlotDescriptor(tag);
   }
 
-  static SlotDescriptor ForMaybeIndirectPointerSlot(IndirectPointerTag tag) {
+  static SlotDescriptor ForTrustedPointerSlot(IndirectPointerTag tag) {
 #ifdef V8_ENABLE_SANDBOX
     return ForIndirectPointerSlot(tag);
 #else
@@ -300,8 +308,13 @@ class SlotDescriptor {
 #endif
   }
 
+  static SlotDescriptor ForCodePointerSlot() {
+    return ForTrustedPointerSlot(kCodeIndirectPointerTag);
+  }
+
  private:
-  SlotDescriptor(IndirectPointerTag tag) : indirect_pointer_tag_(tag) {}
+  explicit SlotDescriptor(IndirectPointerTag tag)
+      : indirect_pointer_tag_(tag) {}
 
   // If the tag is null, this object describes a direct pointer slot.
   IndirectPointerTag indirect_pointer_tag_;
@@ -369,15 +382,17 @@ class V8_EXPORT_PRIVATE AssemblerBase : public Malloced {
 
   int pc_offset() const { return static_cast<int>(pc_ - buffer_start_); }
 
-  int pc_offset_for_safepoint() {
-#if defined(V8_TARGET_ARCH_MIPS64) || defined(V8_TARGET_ARCH_LOONG64)
-    // MIPS and LOONG need to use their own implementation to avoid trampoline's
-    // influence.
-    UNREACHABLE();
-#else
-    return pc_offset();
+  Address current_pc() const { return reinterpret_cast<Address>(pc_); }
+
+  void skip_bytes(int num_bytes) { pc_ += num_bytes; }
+
+// MIPS, LOONG, and RISC-V need to use their own implementations to avoid the
+// influence of branch trampolines. They provide their implementations in the
+// architecture-specific assembler subclasses.
+#if !defined(V8_TARGET_ARCH_MIPS64) && !defined(V8_TARGET_ARCH_LOONG64) && \
+    !defined(V8_TARGET_ARCH_RISCV32) && !defined(V8_TARGET_ARCH_RISCV64)
+  int pc_offset_for_safepoint() const { return pc_offset(); }
 #endif
-  }
 
   uint8_t* buffer_start() const { return buffer_->start(); }
   int buffer_size() const { return buffer_->size(); }
@@ -401,39 +416,63 @@ class V8_EXPORT_PRIVATE AssemblerBase : public Malloced {
 
   // Record an inline code comment that can be used by a disassembler.
   // Use --code-comments to enable.
-  V8_INLINE void RecordComment(const char* comment) {
+  V8_INLINE void RecordComment(
+      const char* comment,
+      const SourceLocation& loc = SourceLocation::Current()) {
     // Set explicit dependency on --code-comments for dead-code elimination in
     // release builds.
     if (!v8_flags.code_comments) return;
     if (options().emit_code_comments) {
-      code_comments_writer_.Add(pc_offset(), std::string(comment));
+      std::string comment_str(comment);
+      if (loc.FileName()) {
+        comment_str += " - " + loc.ToString();
+      }
+      code_comments_writer_.Add(pc_offset(), comment_str);
     }
   }
 
-  V8_INLINE void RecordComment(std::string comment) {
+  V8_INLINE void RecordComment(
+      std::string comment,
+      const SourceLocation& loc = SourceLocation::Current()) {
     // Set explicit dependency on --code-comments for dead-code elimination in
     // release builds.
     if (!v8_flags.code_comments) return;
     if (options().emit_code_comments) {
-      code_comments_writer_.Add(pc_offset(), std::move(comment));
+      std::string comment_str(comment);
+      if (loc.FileName()) {
+        comment_str += " - " + loc.ToString();
+      }
+      code_comments_writer_.Add(pc_offset(), comment_str);
     }
   }
 
 #ifdef V8_CODE_COMMENTS
   class CodeComment {
    public:
-    V8_NODISCARD CodeComment(Assembler* assembler, const std::string& comment)
+    // `comment` can either be a value convertible to std::string, or a function
+    // that returns a value convertible to std::string which is invoked lazily
+    // when code comments are enabled.
+    template <typename CommentGen>
+    V8_NODISCARD CodeComment(
+        Assembler* assembler, CommentGen&& comment,
+        const SourceLocation& loc = SourceLocation::Current())
         : assembler_(assembler) {
-      if (v8_flags.code_comments) Open(comment);
+      if (!v8_flags.code_comments) return;
+      if constexpr (std::is_invocable_v<CommentGen>) {
+        Open(comment(), loc);
+      } else {
+        Open(comment, loc);
+      }
     }
     ~CodeComment() {
-      if (v8_flags.code_comments) Close();
+      if (!v8_flags.code_comments) return;
+      Close();
     }
     static const int kIndentWidth = 2;
 
    private:
     int depth() const;
-    void Open(const std::string& comment);
+    void Open(const std::string& comment, const SourceLocation& loc);
     void Close();
     Assembler* assembler_;
   };
@@ -454,14 +493,14 @@ class V8_EXPORT_PRIVATE AssemblerBase : public Malloced {
  protected:
   // Add 'target' to the {code_targets_} vector, if necessary, and return the
   // offset at which it is stored.
-  int AddCodeTarget(Handle<Code> target);
-  Handle<Code> GetCodeTarget(intptr_t code_target_index) const;
+  int AddCodeTarget(IndirectHandle<Code> target);
+  IndirectHandle<Code> GetCodeTarget(intptr_t code_target_index) const;
 
   // Add 'object' to the {embedded_objects_} vector and return the index at
   // which it is stored.
   using EmbeddedObjectIndex = size_t;
-  EmbeddedObjectIndex AddEmbeddedObject(Handle<HeapObject> object);
-  Handle<HeapObject> GetEmbeddedObject(EmbeddedObjectIndex index) const;
+  EmbeddedObjectIndex AddEmbeddedObject(IndirectHandle<HeapObject> object);
+  IndirectHandle<HeapObject> GetEmbeddedObject(EmbeddedObjectIndex index) const;
 
   // The buffer into which code and relocation info are generated.
   std::unique_ptr<AssemblerBuffer> buffer_;
@@ -488,21 +527,17 @@ class V8_EXPORT_PRIVATE AssemblerBase : public Malloced {
   // by the pc offset associated with each request).
   void RequestHeapNumber(HeapNumberRequest request);
 
+  void AllocateAndInstallRequestedHeapNumbers(LocalIsolate* isolate);
+  virtual void PatchInHeapNumberRequest(Address pc,
+                                        Handle<HeapNumber> object) = 0;
+
   bool ShouldRecordRelocInfo(RelocInfo::Mode rmode) const {
     DCHECK(!RelocInfo::IsNoInfo(rmode));
-    if (options().disable_reloc_info_for_patching) return false;
     if (RelocInfo::IsOnlyForSerializer(rmode) &&
         !options().record_reloc_info_for_serialization &&
-        !v8_flags.debug_code) {
+        !v8_flags.debug_code && !v8_flags.slow_debug_code) {
       return false;
     }
-    if (RelocInfo::IsOnlyForDisassembler(rmode)) {
-#ifdef ENABLE_DISASSEMBLER
-      return true;
-#else
-      return false;
-#endif  // ENABLE_DISASSEMBLER
-    }
     return true;
   }
 
@@ -515,19 +550,20 @@ class V8_EXPORT_PRIVATE AssemblerBase : public Malloced {
   // guaranteed to fit in the instruction's offset field. We keep track of the
   // code handles we encounter in calls in this vector, and encode the index of
   // the code handle in the vector instead.
-  std::vector<Handle<Code>> code_targets_;
+  std::vector<IndirectHandle<Code>> code_targets_;
 
   // If an assembler needs a small number to refer to a heap object handle
   // (for example, because there are only 32bit available on a 64bit arch), the
   // assembler adds the object into this vector using AddEmbeddedObject, and
   // may then refer to the heap object using the handle's index in this vector.
-  std::vector<Handle<HeapObject>> embedded_objects_;
+  std::vector<IndirectHandle<HeapObject>> embedded_objects_;
 
   // Embedded objects are deduplicated based on handle location. This is a
   // compromise that is almost as effective as deduplication based on actual
   // heap object addresses maintains GC safety.
-  std::unordered_map<Handle<HeapObject>, EmbeddedObjectIndex,
-                     Handle<HeapObject>::hash, Handle<HeapObject>::equal_to>
+  std::unordered_map<IndirectHandle<HeapObject>, EmbeddedObjectIndex,
+                     IndirectHandle<HeapObject>::hash,
+                     IndirectHandle<HeapObject>::equal_to>
       embedded_objects_map_;
 
   const AssemblerOptions options_;
@@ -575,7 +611,12 @@ class V8_EXPORT_PRIVATE V8_NODISCARD CpuFeatureScope {
 };
 
 #ifdef V8_CODE_COMMENTS
+#if V8_SUPPORTS_SOURCE_LOCATION
+// We'll get the function name from the source location, no need to pass it in.
+#define ASM_CODE_COMMENT(asm) ASM_CODE_COMMENT_STRING(asm, "")
+#else
 #define ASM_CODE_COMMENT(asm) ASM_CODE_COMMENT_STRING(asm, __func__)
+#endif
 #define ASM_CODE_COMMENT_STRING(asm, comment) \
   AssemblerBase::CodeComment UNIQUE_IDENTIFIER(asm_code_comment)(asm, comment)
 #else
