@@ -5,13 +5,14 @@
 #ifndef V8_COMPILER_HEAP_REFS_H_
 #define V8_COMPILER_HEAP_REFS_H_
 
+#include <optional>
 #include <type_traits>
 
-#include "src/base/optional.h"
 #include "src/ic/call-optimization.h"
 #include "src/objects/elements-kind.h"
 #include "src/objects/feedback-vector.h"
 #include "src/objects/instance-type.h"
+#include "src/objects/object-list-macros.h"
 #include "src/utils/boxed-float.h"
 #include "src/zone/zone-compact-set.h"
 
@@ -22,7 +23,6 @@ class CFunctionInfo;
 namespace internal {
 
 class BytecodeArray;
-class CallHandlerInfo;
 class FixedDoubleArray;
 class FunctionTemplateInfo;
 class HeapNumber;
@@ -60,27 +60,38 @@ class PropertyAccessInfo;
 // distinct semantics for private class fields (in which private field
 // accesses must throw when storing a field which does not exist, or
 // adding/defining a field which already exists).
-enum class AccessMode { kLoad, kStore, kStoreInLiteral, kHas, kDefine };
+enum class AccessMode : uint8_t {
+  kLoad,
+  kStore,
+  kStoreInLiteral,
+  kHas,
+  kDefine
+};
 
 inline bool IsAnyStore(AccessMode mode) {
   return mode == AccessMode::kStore || mode == AccessMode::kStoreInLiteral ||
          mode == AccessMode::kDefine;
 }
 
+inline bool IsDefiningStore(AccessMode mode) {
+  return mode == AccessMode::kStoreInLiteral || mode == AccessMode::kDefine;
+}
+
 enum class OddballType : uint8_t {
   kNone,     // Not an Oddball.
   kBoolean,  // True or False.
   kUndefined,
   kNull,
-  kUninitialized,
-  kOther  // Oddball, but none of the above.
 };
 
 enum class HoleType : uint8_t {
   kNone,  // Not a Hole.
-  kGeneric,
-  kPropertyCell,
-  kHashTable,
+
+#define FOR_HOLE(Name, name, Root) k##Name,
+  HOLE_LIST(FOR_HOLE)
+#undef FOR_HOLE
+
+      kGeneric = kTheHole,
 };
 
 enum class RefSerializationKind {
@@ -101,6 +112,7 @@ enum class RefSerializationKind {
   BACKGROUND_SERIALIZED(JSGlobalObject)                                       \
   BACKGROUND_SERIALIZED(JSGlobalProxy)                                        \
   BACKGROUND_SERIALIZED(JSTypedArray)                                         \
+  BACKGROUND_SERIALIZED(JSPrimitiveWrapper)                                   \
   /* Subtypes of Context */                                                   \
   NEVER_SERIALIZED(NativeContext)                                             \
   /* Subtypes of FixedArray */                                                \
@@ -109,7 +121,6 @@ enum class RefSerializationKind {
   /* Subtypes of String */                                                    \
   NEVER_SERIALIZED(InternalizedString)                                        \
   /* Subtypes of FixedArrayBase */                                            \
-  NEVER_SERIALIZED(BytecodeArray)                                             \
   BACKGROUND_SERIALIZED(FixedArray)                                           \
   NEVER_SERIALIZED(FixedDoubleArray)                                          \
   /* Subtypes of Name */                                                      \
@@ -117,12 +128,13 @@ enum class RefSerializationKind {
   NEVER_SERIALIZED(Symbol)                                                    \
   /* Subtypes of JSReceiver */                                                \
   BACKGROUND_SERIALIZED(JSObject)                                             \
+  BACKGROUND_SERIALIZED(JSProxy)                                              \
   /* Subtypes of HeapObject */                                                \
   NEVER_SERIALIZED(AccessorInfo)                                              \
   NEVER_SERIALIZED(AllocationSite)                                            \
   NEVER_SERIALIZED(ArrayBoilerplateDescription)                               \
   BACKGROUND_SERIALIZED(BigInt)                                               \
-  NEVER_SERIALIZED(CallHandlerInfo)                                           \
+  NEVER_SERIALIZED(BytecodeArray)                                             \
   NEVER_SERIALIZED(Cell)                                                      \
   NEVER_SERIALIZED(Code)                                                      \
   NEVER_SERIALIZED(Context)                                                   \
@@ -132,6 +144,7 @@ enum class RefSerializationKind {
   BACKGROUND_SERIALIZED(FixedArrayBase)                                       \
   NEVER_SERIALIZED(FunctionTemplateInfo)                                      \
   NEVER_SERIALIZED(HeapNumber)                                                \
+  NEVER_SERIALIZED(ContextCell)                                               \
   BACKGROUND_SERIALIZED(JSReceiver)                                           \
   BACKGROUND_SERIALIZED(Map)                                                  \
   NEVER_SERIALIZED(Name)                                                      \
@@ -197,6 +210,7 @@ HEAP_BROKER_OBJECT_LIST_BASE(BACKGROUND_SERIALIZED_REF_TRAITS,
 template <>
 struct ref_traits<Object> {
   using ref_type = ObjectRef;
+  using data_type = ObjectData;
   // Note: While a bit awkward, this artificial ref serialization kind value is
   // okay: smis are never-serialized, and we never create raw non-smi
   // ObjectRefs (they would at least be HeapObjectRefs instead).
@@ -204,8 +218,8 @@ struct ref_traits<Object> {
       RefSerializationKind::kNeverSerialized;
 };
 
-// For types used in ReadOnlyRoots, but which don't have a corresponding Ref
-// type, use HeapObjectRef.
+// For types which don't have a corresponding Ref type, use the next best
+// existing Ref.
 template <>
 struct ref_traits<Oddball> : public ref_traits<HeapObject> {};
 template <>
@@ -225,7 +239,7 @@ struct ref_traits<PropertyArray> : public ref_traits<HeapObject> {};
 template <>
 struct ref_traits<ByteArray> : public ref_traits<HeapObject> {};
 template <>
-struct ref_traits<ExternalPointerArray> : public ref_traits<HeapObject> {};
+struct ref_traits<TrustedFixedArray> : public ref_traits<HeapObject> {};
 template <>
 struct ref_traits<ClosureFeedbackCellArray> : public ref_traits<HeapObject> {};
 template <>
@@ -256,8 +270,40 @@ struct ref_traits<RegisteredSymbolTable> : public ref_traits<HeapObject> {};
 template <>
 struct ref_traits<WasmNull> : public ref_traits<HeapObject> {};
 #endif  // V8_ENABLE_WEBASSEMBLY
+template <>
+struct ref_traits<Smi> : public ref_traits<Object> {};
+template <>
+struct ref_traits<Boolean> : public ref_traits<HeapObject> {};
+template <>
+struct ref_traits<JSWrappedFunction> : public ref_traits<JSFunction> {};
 
-// Wrapper around heap refs which works roughly like a base::Optional, but
+template <class... T>
+struct ref_traits<Union<T...>> {
+  // There's no good way in C++ to find a common base class, so just test a few
+  // common cases.
+  static constexpr bool kAllJSReceiverRef =
+      (std::is_base_of_v<JSReceiverRef, typename ref_traits<T>::ref_type> &&
+       ...);
+  static constexpr bool kAllHeapObjectRef =
+      (std::is_base_of_v<JSReceiverRef, typename ref_traits<T>::ref_type> &&
+       ...);
+
+  using ref_type = std::conditional_t<
+      kAllJSReceiverRef, JSReceiverRef,
+      std::conditional_t<kAllHeapObjectRef, HeapObjectRef, ObjectRef>>;
+  using data_type = std::conditional_t<
+      kAllJSReceiverRef, JSReceiverData,
+      std::conditional_t<kAllHeapObjectRef, HeapObjectData, ObjectData>>;
+
+  static constexpr RefSerializationKind ref_serialization_kind =
+      ((ref_traits<T>::ref_serialization_kind ==
+        RefSerializationKind::kNeverSerialized) &&
+       ...)
+          ? RefSerializationKind::kNeverSerialized
+          : RefSerializationKind::kBackgroundSerialized;
+};
+
+// Wrapper around heap refs which works roughly like a std::optional, but
 // doesn't use extra storage for a boolean, but instead uses a null data pointer
 // as a sentinel no value.
 template <typename TRef>
@@ -282,19 +328,21 @@ class OptionalRef {
 
   OptionalRef() = default;
   // NOLINTNEXTLINE
-  OptionalRef(base::nullopt_t) : OptionalRef() {}
+  OptionalRef(std::nullopt_t) : OptionalRef() {}
 
   // Allow implicit upcasting from OptionalRefs with compatible refs.
-  template <typename SRef, typename = typename std::enable_if<
-                               std::is_convertible<SRef*, TRef*>::value>::type>
+  template <typename SRef>
   // NOLINTNEXTLINE
-  V8_INLINE OptionalRef(OptionalRef<SRef> ref) : data_(ref.data_) {}
+  V8_INLINE OptionalRef(OptionalRef<SRef> ref)
+    requires std::is_convertible_v<SRef*, TRef*>
+      : data_(ref.data_) {}
 
   // Allow implicit upcasting from compatible refs.
-  template <typename SRef, typename = typename std::enable_if<
-                               std::is_convertible<SRef*, TRef*>::value>::type>
+  template <typename SRef>
   // NOLINTNEXTLINE
-  V8_INLINE OptionalRef(SRef ref) : data_(ref.data_) {}
+  V8_INLINE OptionalRef(SRef ref)
+    requires std::is_convertible_v<SRef*, TRef*>
+      : data_(ref.data_) {}
 
   constexpr bool has_value() const { return data_ != nullptr; }
   constexpr explicit operator bool() const { return has_value(); }
@@ -346,7 +394,7 @@ class V8_EXPORT_PRIVATE ObjectRef {
     CHECK_NOT_NULL(data_);
   }
 
-  Handle<Object> object() const;
+  IndirectHandle<Object> object() const;
 
   bool equals(ObjectRef other) const;
 
@@ -369,19 +417,23 @@ class V8_EXPORT_PRIVATE ObjectRef {
   bool IsTheHole() const;
   bool IsPropertyCellHole() const;
   bool IsHashTableHole() const;
+  bool IsPromiseHole() const;
   bool IsNullOrUndefined() const;
+  bool IsUndefinedContextCell() const;
 
-  base::Optional<bool> TryGetBooleanValue(JSHeapBroker* broker) const;
+  std::optional<bool> TryGetBooleanValue(JSHeapBroker* broker) const;
   Maybe<double> OddballToNumber(JSHeapBroker* broker) const;
 
   bool should_access_heap() const;
+  bool is_read_only() const;
+
+  ObjectData* data() const;
 
   struct Hash {
     size_t operator()(ObjectRef ref) const { return ref.hash_value(); }
   };
 
  protected:
-  ObjectData* data() const;
   ObjectData* data_;  // Should be used only by object() getters.
 
  private:
@@ -434,9 +486,10 @@ class HeapObjectType {
 
   using Flags = base::Flags<Flag>;
 
-  HeapObjectType(InstanceType instance_type, Flags flags,
-                 OddballType oddball_type, HoleType hole_type)
+  HeapObjectType(InstanceType instance_type, ElementsKind elements_kind,
+                 Flags flags, OddballType oddball_type, HoleType hole_type)
       : instance_type_(instance_type),
+        elements_kind_(elements_kind),
         oddball_type_(oddball_type),
         hole_type_(hole_type),
         flags_(flags) {
@@ -451,12 +504,14 @@ class HeapObjectType {
   HoleType hole_type(JSHeapBroker* broker) const { return hole_type_; }
   InstanceType instance_type() const { return instance_type_; }
   Flags flags() const { return flags_; }
+  ElementsKind elements_kind() const { return elements_kind_; }
 
   bool is_callable() const { return flags_ & kCallable; }
   bool is_undetectable() const { return flags_ & kUndetectable; }
 
  private:
   InstanceType const instance_type_;
+  ElementsKind const elements_kind_;
   OddballType const oddball_type_;
   HoleType const hole_type_;
   Flags const flags_;
@@ -476,9 +531,9 @@ class HeapObjectRef : public ObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(HeapObject, ObjectRef)
 
-  Handle<HeapObject> object() const;
+  IndirectHandle<HeapObject> object() const;
 
-  MapRef map(JSHeapBroker* broker) const;
+  V8_EXPORT_PRIVATE MapRef map(JSHeapBroker* broker) const;
 
   // Only for use in special situations where we need to read the object's
   // current map (instead of returning the cached map). Use with care.
@@ -492,7 +547,7 @@ class PropertyCellRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(PropertyCell, HeapObjectRef)
 
-  Handle<PropertyCell> object() const;
+  IndirectHandle<PropertyCell> object() const;
 
   V8_WARN_UNUSED_RESULT bool Cache(JSHeapBroker* broker) const;
   void CacheAsProtector(JSHeapBroker* broker) const {
@@ -510,21 +565,32 @@ class JSReceiverRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(JSReceiver, HeapObjectRef)
 
-  Handle<JSReceiver> object() const;
+  IndirectHandle<JSReceiver> object() const;
+};
+
+class JSProxyRef : public JSReceiverRef {
+ public:
+  DEFINE_REF_CONSTRUCTOR(JSProxy, JSReceiverRef)
+
+  IndirectHandle<JSProxy> object() const;
+
+  bool is_revocable() const;
+
+  OptionalObjectRef GetTarget(JSHeapBroker* broker) const;
+  OptionalObjectRef GetHandler(JSHeapBroker* broker) const;
 };
 
 class JSObjectRef : public JSReceiverRef {
  public:
   DEFINE_REF_CONSTRUCTOR(JSObject, JSReceiverRef)
 
-  Handle<JSObject> object() const;
+  IndirectHandle<JSObject> object() const;
 
   OptionalObjectRef raw_properties_or_hash(JSHeapBroker* broker) const;
 
   // Usable only for in-object properties. Only use this if the underlying
-  // value can be an uninitialized-sentinel, or if HeapNumber construction must
-  // be avoided for some reason. Otherwise, use the higher-level
-  // GetOwnFastDataProperty.
+  // value can be an uninitialized-sentinel. Otherwise, use the higher-level
+  // GetOwnFastConstantDataProperty/GetOwnFastConstantDoubleProperty.
   OptionalObjectRef RawInobjectPropertyAt(JSHeapBroker* broker,
                                           FieldIndex index) const;
 
@@ -538,19 +604,31 @@ class JSObjectRef : public JSReceiverRef {
   // The direct-read implementation of the above, extracted into a helper since
   // it's also called from compilation-dependency validation. This helper is
   // guaranteed to not create new Ref instances.
-  base::Optional<Tagged<Object>> GetOwnConstantElementFromHeap(
+  std::optional<Tagged<Object>> GetOwnConstantElementFromHeap(
       JSHeapBroker* broker, Tagged<FixedArrayBase> elements,
       ElementsKind elements_kind, uint32_t index) const;
 
   // Return the value of the property identified by the field {index}
-  // if {index} is known to be an own data property of the object.
-  // If {dependencies} is non-null, and a property was successfully read,
-  // then the function will take a dependency to check the value of the
-  // property at code finalization time.
-  OptionalObjectRef GetOwnFastDataProperty(
+  // if {index} is known to be an own data property of the object and the field
+  // is constant.
+  // If a property was successfully read, then the function will take a
+  // dependency to check the value of the property at code finalization time.
+  //
+  // This is *not* allowed to be a double representation field. Those should use
+  // GetOwnFastDoubleProperty, to avoid unnecessary HeapNumber allocation.
+  OptionalObjectRef GetOwnFastConstantDataProperty(
       JSHeapBroker* broker, Representation field_representation,
       FieldIndex index, CompilationDependencies* dependencies) const;
 
+  // Return the value of the double property identified by the field {index}
+  // if {index} is known to be an own data property of the object and the field
+  // is constant.
+  // If a property was successfully read, then the function will take a
+  // dependency to check the value of the property at code finalization time.
+  std::optional<Float64> GetOwnFastConstantDoubleProperty(
+      JSHeapBroker* broker, FieldIndex index,
+      CompilationDependencies* dependencies) const;
+
   // Return the value of the dictionary property at {index} in the dictionary
   // if {index} is known to be an own data property of the object.
   OptionalObjectRef GetOwnDictionaryProperty(
@@ -571,7 +649,7 @@ class JSDataViewRef : public JSObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(JSDataView, JSObjectRef)
 
-  Handle<JSDataView> object() const;
+  IndirectHandle<JSDataView> object() const;
 
   size_t byte_length() const;
 };
@@ -580,7 +658,7 @@ class JSBoundFunctionRef : public JSObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(JSBoundFunction, JSObjectRef)
 
-  Handle<JSBoundFunction> object() const;
+  IndirectHandle<JSBoundFunction> object() const;
 
   JSReceiverRef bound_target_function(JSHeapBroker* broker) const;
   ObjectRef bound_this(JSHeapBroker* broker) const;
@@ -591,7 +669,7 @@ class V8_EXPORT_PRIVATE JSFunctionRef : public JSObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(JSFunction, JSObjectRef)
 
-  Handle<JSFunction> object() const;
+  IndirectHandle<JSFunction> object() const;
 
   // Returns true, iff the serialized JSFunctionData contents are consistent
   // with the state of the underlying JSFunction object. Must be called from
@@ -611,15 +689,18 @@ class V8_EXPORT_PRIVATE JSFunctionRef : public JSObjectRef {
   int InitialMapInstanceSizeWithMinSlack(JSHeapBroker* broker) const;
   FeedbackCellRef raw_feedback_cell(JSHeapBroker* broker) const;
   OptionalFeedbackVectorRef feedback_vector(JSHeapBroker* broker) const;
+#ifdef V8_ENABLE_LEAPTIERING
+  JSDispatchHandle dispatch_handle() const;
+#endif
 };
 
 class RegExpBoilerplateDescriptionRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(RegExpBoilerplateDescription, HeapObjectRef)
 
-  Handle<RegExpBoilerplateDescription> object() const;
+  IndirectHandle<RegExpBoilerplateDescription> object() const;
 
-  FixedArrayRef data(JSHeapBroker* broker) const;
+  HeapObjectRef data(JSHeapBroker* broker) const;
   StringRef source(JSHeapBroker* broker) const;
   int flags() const;
 };
@@ -633,17 +714,27 @@ class HeapNumberRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(HeapNumber, HeapObjectRef)
 
-  Handle<HeapNumber> object() const;
+  IndirectHandle<HeapNumber> object() const;
 
   double value() const;
   uint64_t value_as_bits() const;
 };
 
+class ContextCellRef : public HeapObjectRef {
+ public:
+  DEFINE_REF_CONSTRUCTOR(ContextCell, HeapObjectRef)
+
+  IndirectHandle<ContextCell> object() const;
+
+  ContextCell::State state() const;
+  OptionalObjectRef tagged_value(JSHeapBroker* broker) const;
+};
+
 class ContextRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(Context, HeapObjectRef)
 
-  Handle<Context> object() const;
+  IndirectHandle<Context> object() const;
 
   // {previous} decrements {depth} by 1 for each previous link successfully
   // followed. If {depth} != 0 on function return, then it only got partway to
@@ -672,6 +763,7 @@ class ContextRef : public HeapObjectRef {
   V(JSGlobalObject, global_object)               \
   V(JSGlobalProxy, global_proxy_object)          \
   V(JSObject, initial_array_prototype)           \
+  V(JSObject, initial_object_prototype)          \
   V(JSObject, promise_prototype)                 \
   V(Map, async_function_object_map)              \
   V(Map, block_context_map)                      \
@@ -693,6 +785,7 @@ class ContextRef : public HeapObjectRef {
   V(Map, map_key_iterator_map)                   \
   V(Map, map_key_value_iterator_map)             \
   V(Map, map_value_iterator_map)                 \
+  V(Map, meta_map)                               \
   V(Map, set_key_value_iterator_map)             \
   V(Map, set_value_iterator_map)                 \
   V(Map, sloppy_arguments_map)                   \
@@ -705,7 +798,7 @@ class NativeContextRef : public ContextRef {
  public:
   DEFINE_REF_CONSTRUCTOR(NativeContext, ContextRef)
 
-  Handle<NativeContext> object() const;
+  IndirectHandle<NativeContext> object() const;
 
 #define DECL_ACCESSOR(type, name) type##Ref name(JSHeapBroker* broker) const;
   BROKER_NATIVE_CONTEXT_FIELDS(DECL_ACCESSOR)
@@ -722,7 +815,7 @@ class NameRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(Name, HeapObjectRef)
 
-  Handle<Name> object() const;
+  IndirectHandle<Name> object() const;
 
   bool IsUniqueName() const;
 };
@@ -731,7 +824,7 @@ class DescriptorArrayRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(DescriptorArray, HeapObjectRef)
 
-  Handle<DescriptorArray> object() const;
+  IndirectHandle<DescriptorArray> object() const;
 
   PropertyDetails GetPropertyDetails(InternalIndex descriptor_index) const;
   NameRef GetPropertyKey(JSHeapBroker* broker,
@@ -744,7 +837,7 @@ class FeedbackCellRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(FeedbackCell, HeapObjectRef)
 
-  Handle<FeedbackCell> object() const;
+  IndirectHandle<FeedbackCell> object() const;
 
   ObjectRef value(JSHeapBroker* broker) const;
 
@@ -752,41 +845,36 @@ class FeedbackCellRef : public HeapObjectRef {
   OptionalFeedbackVectorRef feedback_vector(JSHeapBroker* broker) const;
   OptionalSharedFunctionInfoRef shared_function_info(
       JSHeapBroker* broker) const;
+#ifdef V8_ENABLE_LEAPTIERING
+  JSDispatchHandle dispatch_handle() const;
+#endif
 };
 
 class FeedbackVectorRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(FeedbackVector, HeapObjectRef)
 
-  Handle<FeedbackVector> object() const;
+  IndirectHandle<FeedbackVector> object() const;
 
   SharedFunctionInfoRef shared_function_info(JSHeapBroker* broker) const;
 
   FeedbackCellRef GetClosureFeedbackCell(JSHeapBroker* broker, int index) const;
-};
 
-class CallHandlerInfoRef : public HeapObjectRef {
- public:
-  DEFINE_REF_CONSTRUCTOR(CallHandlerInfo, HeapObjectRef)
-
-  Handle<CallHandlerInfo> object() const;
-
-  Address callback(JSHeapBroker* broker) const;
-  ObjectRef data(JSHeapBroker* broker) const;
+  bool was_once_deoptimized() const;
 };
 
 class AccessorInfoRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(AccessorInfo, HeapObjectRef)
 
-  Handle<AccessorInfo> object() const;
+  IndirectHandle<AccessorInfo> object() const;
 };
 
 class AllocationSiteRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(AllocationSite, HeapObjectRef)
 
-  Handle<AllocationSite> object() const;
+  IndirectHandle<AllocationSite> object() const;
 
   bool PointsToLiteral() const;
   AllocationType GetAllocationType() const;
@@ -801,7 +889,7 @@ class BigIntRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(BigInt, HeapObjectRef)
 
-  Handle<BigInt> object() const;
+  IndirectHandle<BigInt> object() const;
 
   uint64_t AsUint64() const;
   int64_t AsInt64(bool* lossless) const;
@@ -811,7 +899,7 @@ class V8_EXPORT_PRIVATE MapRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(Map, HeapObjectRef)
 
-  Handle<Map> object() const;
+  IndirectHandle<Map> object() const;
 
   int instance_size() const;
   InstanceType instance_type() const;
@@ -837,11 +925,18 @@ class V8_EXPORT_PRIVATE MapRef : public HeapObjectRef {
   bool is_undetectable() const;
   bool is_callable() const;
   bool has_indexed_interceptor() const;
+  bool has_named_interceptor() const;
   int construction_counter() const;
   bool is_migration_target() const;
+  bool is_extensible() const;
   bool supports_fast_array_iteration(JSHeapBroker* broker) const;
   bool supports_fast_array_resize(JSHeapBroker* broker) const;
   bool is_abandoned_prototype_map() const;
+  bool IsOneByteStringMap() const;
+  bool IsTwoByteStringMap() const;
+  bool IsSeqStringMap() const;
+  bool IsThinStringMap() const;
+  bool IsStringWrapperMap() const;
 
   OddballType oddball_type(JSHeapBroker* broker) const;
 
@@ -855,6 +950,10 @@ class V8_EXPORT_PRIVATE MapRef : public HeapObjectRef {
   INSTANCE_TYPE_CHECKERS(DEF_TESTER)
 #undef DEF_TESTER
 
+  bool IsBooleanMap(JSHeapBroker* broker) const;
+  bool IsNullMap(JSHeapBroker* broker) const;
+  bool IsUndefinedMap(JSHeapBroker* broker) const;
+
   HeapObjectRef GetBackPointer(JSHeapBroker* broker) const;
 
   HeapObjectRef prototype(JSHeapBroker* broker) const;
@@ -881,7 +980,7 @@ class V8_EXPORT_PRIVATE MapRef : public HeapObjectRef {
 struct HolderLookupResult {
   HolderLookupResult(CallOptimization::HolderLookup lookup_ =
                          CallOptimization::kHolderNotFound,
-                     OptionalJSObjectRef holder_ = base::nullopt)
+                     OptionalJSObjectRef holder_ = std::nullopt)
       : lookup(lookup_), holder(holder_) {}
   CallOptimization::HolderLookup lookup;
   OptionalJSObjectRef holder;
@@ -891,14 +990,18 @@ class FunctionTemplateInfoRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(FunctionTemplateInfo, HeapObjectRef)
 
-  Handle<FunctionTemplateInfo> object() const;
+  IndirectHandle<FunctionTemplateInfo> object() const;
 
   bool is_signature_undefined(JSHeapBroker* broker) const;
   bool accept_any_receiver() const;
   int16_t allowed_receiver_instance_type_range_start() const;
   int16_t allowed_receiver_instance_type_range_end() const;
 
-  OptionalCallHandlerInfoRef call_code(JSHeapBroker* broker) const;
+  // Function pointer and a data value that should be passed to the callback.
+  // The |callback_data| must be read before the |callback|.
+  Address callback(JSHeapBroker* broker) const;
+  OptionalObjectRef callback_data(JSHeapBroker* broker) const;
+
   ZoneVector<Address> c_functions(JSHeapBroker* broker) const;
   ZoneVector<const CFunctionInfo*> c_signatures(JSHeapBroker* broker) const;
   HolderLookupResult LookupHolderOfExpectedType(JSHeapBroker* broker,
@@ -909,15 +1012,15 @@ class FixedArrayBaseRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(FixedArrayBase, HeapObjectRef)
 
-  Handle<FixedArrayBase> object() const;
+  IndirectHandle<FixedArrayBase> object() const;
 
-  int length() const;
+  uint32_t length() const;
 };
 
 class ArrayBoilerplateDescriptionRef : public HeapObjectRef {
  public:
   using HeapObjectRef::HeapObjectRef;
-  Handle<ArrayBoilerplateDescription> object() const;
+  IndirectHandle<ArrayBoilerplateDescription> object() const;
 
   int constants_elements_length() const;
 };
@@ -926,7 +1029,7 @@ class FixedArrayRef : public FixedArrayBaseRef {
  public:
   DEFINE_REF_CONSTRUCTOR(FixedArray, FixedArrayBaseRef)
 
-  Handle<FixedArray> object() const;
+  IndirectHandle<FixedArray> object() const;
 
   OptionalObjectRef TryGet(JSHeapBroker* broker, int i) const;
 };
@@ -935,7 +1038,7 @@ class FixedDoubleArrayRef : public FixedArrayBaseRef {
  public:
   DEFINE_REF_CONSTRUCTOR(FixedDoubleArray, FixedArrayBaseRef)
 
-  Handle<FixedDoubleArray> object() const;
+  IndirectHandle<FixedDoubleArray> object() const;
 
   // Due to 64-bit unaligned reads, only usable for
   // immutable-after-initialization FixedDoubleArrays protected by
@@ -943,48 +1046,53 @@ class FixedDoubleArrayRef : public FixedArrayBaseRef {
   Float64 GetFromImmutableFixedDoubleArray(int i) const;
 };
 
-class BytecodeArrayRef : public FixedArrayBaseRef {
+class BytecodeArrayRef : public HeapObjectRef {
  public:
-  DEFINE_REF_CONSTRUCTOR(BytecodeArray, FixedArrayBaseRef)
+  DEFINE_REF_CONSTRUCTOR(BytecodeArray, HeapObjectRef)
 
-  Handle<BytecodeArray> object() const;
+  IndirectHandle<BytecodeArray> object() const;
 
   // NOTE: Concurrent reads of the actual bytecodes as well as the constant pool
   // (both immutable) do not go through BytecodeArrayRef but are performed
   // directly through the handle by BytecodeArrayIterator.
 
+  int length() const;
+
   int register_count() const;
-  int parameter_count() const;
+  uint16_t parameter_count() const;
+  uint16_t parameter_count_without_receiver() const;
+  uint16_t max_arguments() const;
   interpreter::Register incoming_new_target_or_generator_register() const;
 
-  Handle<ByteArray> SourcePositionTable(JSHeapBroker* broker) const;
+  IndirectHandle<TrustedByteArray> SourcePositionTable(
+      JSHeapBroker* broker) const;
 
   // Exception handler table.
   Address handler_table_address() const;
   int handler_table_size() const;
 };
 
-class ScriptContextTableRef : public FixedArrayRef {
+class ScriptContextTableRef : public FixedArrayBaseRef {
  public:
-  DEFINE_REF_CONSTRUCTOR(ScriptContextTable, FixedArrayRef)
+  DEFINE_REF_CONSTRUCTOR(ScriptContextTable, FixedArrayBaseRef)
 
-  Handle<ScriptContextTable> object() const;
+  IndirectHandle<ScriptContextTable> object() const;
 };
 
 class ObjectBoilerplateDescriptionRef : public FixedArrayRef {
  public:
   DEFINE_REF_CONSTRUCTOR(ObjectBoilerplateDescription, FixedArrayRef)
 
-  Handle<ObjectBoilerplateDescription> object() const;
+  IndirectHandle<ObjectBoilerplateDescription> object() const;
 
-  int size() const;
+  int boilerplate_properties_count() const;
 };
 
 class JSArrayRef : public JSObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(JSArray, JSObjectRef)
 
-  Handle<JSArray> object() const;
+  IndirectHandle<JSArray> object() const;
 
   // The `length` property of boilerplate JSArray objects. Boilerplates are
   // immutable after initialization. Must not be used for non-boilerplate
@@ -1010,42 +1118,40 @@ class ScopeInfoRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(ScopeInfo, HeapObjectRef)
 
-  Handle<ScopeInfo> object() const;
+  IndirectHandle<ScopeInfo> object() const;
 
   int ContextLength() const;
+  bool HasContext() const;
   bool HasOuterScopeInfo() const;
   bool HasContextExtensionSlot() const;
+  bool SomeContextHasExtension() const;
   bool ClassScopeHasPrivateBrand() const;
+  bool SloppyEvalCanExtendVars() const;
+  ScopeType scope_type() const;
 
   ScopeInfoRef OuterScopeInfo(JSHeapBroker* broker) const;
 };
 
-#define BROKER_SFI_FIELDS(V)                                    \
-  V(int, internal_formal_parameter_count_with_receiver)         \
-  V(int, internal_formal_parameter_count_without_receiver)      \
-  V(bool, IsDontAdaptArguments)                                 \
-  V(bool, has_simple_parameters)                                \
-  V(bool, has_duplicate_parameters)                             \
-  V(int, function_map_index)                                    \
-  V(FunctionKind, kind)                                         \
-  V(LanguageMode, language_mode)                                \
-  V(bool, native)                                               \
-  V(bool, HasBuiltinId)                                         \
-  V(bool, construct_as_builtin)                                 \
-  V(bool, HasBytecodeArray)                                     \
-  V(int, StartPosition)                                         \
-  V(bool, is_compiled)                                          \
-  V(bool, IsUserJavaScript)                                     \
-  V(bool, requires_instance_members_initializer)                \
-  IF_WASM(V, const wasm::WasmModule*, wasm_module)              \
-  IF_WASM(V, const wasm::FunctionSig*, wasm_function_signature) \
-  IF_WASM(V, int, wasm_function_index)
+#define BROKER_SFI_FIELDS(V)                               \
+  V(bool, has_simple_parameters)                           \
+  V(bool, has_duplicate_parameters)                        \
+  V(int, function_map_index)                               \
+  V(FunctionKind, kind)                                    \
+  V(LanguageMode, language_mode)                           \
+  V(bool, native)                                          \
+  V(bool, HasBuiltinId)                                    \
+  V(bool, construct_as_builtin)                            \
+  V(bool, HasBytecodeArray)                                \
+  V(int, StartPosition)                                    \
+  V(bool, is_compiled)                                     \
+  V(bool, IsUserJavaScript)                                \
+  V(bool, requires_instance_members_initializer)
 
 class V8_EXPORT_PRIVATE SharedFunctionInfoRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(SharedFunctionInfo, HeapObjectRef)
 
-  Handle<SharedFunctionInfo> object() const;
+  IndirectHandle<SharedFunctionInfo> object() const;
 
   Builtin builtin_id() const;
   int context_header_size() const;
@@ -1053,17 +1159,25 @@ class V8_EXPORT_PRIVATE SharedFunctionInfoRef : public HeapObjectRef {
   BytecodeArrayRef GetBytecodeArray(JSHeapBroker* broker) const;
   bool HasBreakInfo(JSHeapBroker* broker) const;
   SharedFunctionInfo::Inlineability GetInlineability(
-      JSHeapBroker* broker) const;
+      CodeKind code_kind, JSHeapBroker* broker) const;
   OptionalFunctionTemplateInfoRef function_template_info(
       JSHeapBroker* broker) const;
   ScopeInfoRef scope_info(JSHeapBroker* broker) const;
 
+  // TODO(370343328): The compiler should not rely on the parameter count
+  // stored on the SFI but instead use the parameter count from the
+  // BytecodeArray or JSDispatchTable. Once remaining uses of the field are
+  // gone, these accessors should probably be removed.
+  int internal_formal_parameter_count_with_receiver_deprecated() const;
+  int internal_formal_parameter_count_without_receiver_deprecated() const;
+
 #define DECL_ACCESSOR(type, name) type name() const;
   BROKER_SFI_FIELDS(DECL_ACCESSOR)
 #undef DECL_ACCESSOR
 
-  bool IsInlineable(JSHeapBroker* broker) const {
-    return GetInlineability(broker) == SharedFunctionInfo::kIsInlineable;
+  bool IsInlineable(CodeKind code_kind, JSHeapBroker* broker) const {
+    return GetInlineability(code_kind, broker) ==
+           SharedFunctionInfo::kIsInlineable;
   }
 };
 
@@ -1071,28 +1185,30 @@ class StringRef : public NameRef {
  public:
   DEFINE_REF_CONSTRUCTOR(String, NameRef)
 
-  Handle<String> object() const;
+  IndirectHandle<String> object() const;
 
-  // With concurrent inlining on, we return base::nullopt due to not being able
+  // With concurrent inlining on, we return std::nullopt due to not being able
   // to use LookupIterator in a thread-safe way.
   OptionalObjectRef GetCharAsStringOrUndefined(JSHeapBroker* broker,
                                                uint32_t index) const;
+  // Returns ThinString::actual() if the current (uncached) map is a ThinString
+  // map, a self reference for all other strings.
+  StringRef UnpackIfThin(JSHeapBroker* broker);
 
   // When concurrently accessing non-read-only non-supported strings, we return
-  // base::nullopt for these methods.
-  base::Optional<Handle<String>> ObjectIfContentAccessible(
-      JSHeapBroker* broker);
-  int length() const;
-  base::Optional<uint16_t> GetFirstChar(JSHeapBroker* broker) const;
-  base::Optional<uint16_t> GetChar(JSHeapBroker* broker, int index) const;
-  base::Optional<double> ToNumber(JSHeapBroker* broker);
-  base::Optional<double> ToInt(JSHeapBroker* broker, int radix);
+  // std::nullopt for these methods.
+  std::optional<Handle<String>> ObjectIfContentAccessible(JSHeapBroker* broker);
+  uint32_t length() const;
+  std::optional<uint16_t> GetFirstChar(JSHeapBroker* broker) const;
+  std::optional<uint16_t> GetChar(JSHeapBroker* broker, uint32_t index) const;
+  std::optional<double> ToNumber(JSHeapBroker* broker);
+  std::optional<double> ToInt(JSHeapBroker* broker, int radix);
 
-  bool IsSeqString() const;
+  V8_EXPORT_PRIVATE bool IsSeqString() const;
   bool IsExternalString() const;
 
   bool IsContentAccessible() const;
-  bool IsOneByteRepresentation() const;
+  V8_EXPORT_PRIVATE bool IsOneByteRepresentation() const;
 
  private:
   // With concurrent inlining on, we currently support reading directly
@@ -1105,26 +1221,42 @@ class SymbolRef : public NameRef {
  public:
   DEFINE_REF_CONSTRUCTOR(Symbol, NameRef)
 
-  Handle<Symbol> object() const;
+  IndirectHandle<Symbol> object() const;
 };
 
 class JSTypedArrayRef : public JSObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(JSTypedArray, JSObjectRef)
 
-  Handle<JSTypedArray> object() const;
+  IndirectHandle<JSTypedArray> object() const;
 
   bool is_on_heap() const;
-  size_t length() const;
+  size_t length(JSHeapBroker* broker) const;
+  size_t byte_length() const;
+  ElementsKind elements_kind(JSHeapBroker* broker) const;
   void* data_ptr() const;
   HeapObjectRef buffer(JSHeapBroker* broker) const;
+
+  bool is_off_heap_non_rab_gsab(JSHeapBroker* broker) const {
+    return !is_on_heap() &&
+           !IsRabGsabTypedArrayElementsKind(elements_kind(broker));
+  }
+};
+
+class JSPrimitiveWrapperRef : public JSObjectRef {
+ public:
+  DEFINE_REF_CONSTRUCTOR(JSPrimitiveWrapper, JSObjectRef)
+
+  bool IsStringWrapper(JSHeapBroker* broker) const;
+
+  IndirectHandle<JSPrimitiveWrapper> object() const;
 };
 
 class SourceTextModuleRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(SourceTextModule, HeapObjectRef)
 
-  Handle<SourceTextModule> object() const;
+  IndirectHandle<SourceTextModule> object() const;
 
   OptionalCellRef GetCell(JSHeapBroker* broker, int cell_index) const;
   OptionalObjectRef import_meta(JSHeapBroker* broker) const;
@@ -1134,21 +1266,21 @@ class TemplateObjectDescriptionRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(TemplateObjectDescription, HeapObjectRef)
 
-  Handle<TemplateObjectDescription> object() const;
+  IndirectHandle<TemplateObjectDescription> object() const;
 };
 
 class CellRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(Cell, HeapObjectRef)
 
-  Handle<Cell> object() const;
+  IndirectHandle<Cell> object() const;
 };
 
 class JSGlobalObjectRef : public JSObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(JSGlobalObject, JSObjectRef)
 
-  Handle<JSGlobalObject> object() const;
+  IndirectHandle<JSGlobalObject> object() const;
 
   bool IsDetachedFrom(JSGlobalProxyRef proxy) const;
 
@@ -1161,14 +1293,14 @@ class JSGlobalProxyRef : public JSObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(JSGlobalProxy, JSObjectRef)
 
-  Handle<JSGlobalProxy> object() const;
+  IndirectHandle<JSGlobalProxy> object() const;
 };
 
 class CodeRef : public HeapObjectRef {
  public:
   DEFINE_REF_CONSTRUCTOR(Code, HeapObjectRef)
 
-  Handle<Code> object() const;
+  IndirectHandle<Code> object() const;
 
   unsigned GetInlinedBytecodeSize() const;
 };
@@ -1177,7 +1309,7 @@ class InternalizedStringRef : public StringRef {
  public:
   DEFINE_REF_CONSTRUCTOR(InternalizedString, StringRef)
 
-  Handle<InternalizedString> object() const;
+  IndirectHandle<InternalizedString> object() const;
 };
 
 #undef DEFINE_REF_CONSTRUCTOR
@@ -1192,8 +1324,7 @@ class InternalizedStringRef : public StringRef {
   static_assert(std::is_trivially_destructible_v<Name##Ref>);     \
   static_assert(std::is_trivially_destructible_v<OptionalName##Ref>);
 
-V(Object)
-HEAP_BROKER_OBJECT_LIST(V)
+V(Object) HEAP_BROKER_OBJECT_LIST(V)
 #undef V
 
 }  // namespace compiler
@@ -1216,6 +1347,16 @@ namespace compiler {
 template <typename T>
 using ZoneRefSet = ZoneCompactSet<typename ref_traits<T>::ref_type>;
 
+inline bool AnyMapIsHeapNumber(const ZoneRefSet<Map>& maps) {
+  return std::any_of(maps.begin(), maps.end(),
+                     [](MapRef map) { return map.IsHeapNumberMap(); });
+}
+
+inline bool AnyMapIsHeapNumber(const base::Vector<const MapRef>& maps) {
+  return std::any_of(maps.begin(), maps.end(),
+                     [](MapRef map) { return map.IsHeapNumberMap(); });
+}
+
 }  // namespace compiler
 
 }  // namespace internal
