@@ -8,6 +8,7 @@
 #include <iterator>
 #include <map>
 #include <memory>
+#include <optional>
 #include <queue>
 #include <string>
 #include <unordered_map>
@@ -22,6 +23,7 @@
 #include "src/base/platform/time.h"
 #include "src/base/platform/wrappers.h"
 #include "src/d8/async-hooks-wrapper.h"
+#include "src/handles/global-handles.h"
 #include "src/heap/parked-scope.h"
 
 namespace v8 {
@@ -32,7 +34,7 @@ class D8Console;
 class Message;
 class TryCatch;
 
-enum class ModuleType { kJavaScript, kJSON, kInvalid };
+enum class ModuleType { kJavaScript, kJSON, kWebAssembly, kInvalid };
 
 namespace internal {
 class CancelableTaskManager;
@@ -82,6 +84,7 @@ class CounterCollection {
 };
 
 using CounterMap = std::unordered_map<std::string, Counter*>;
+using PerformanceMarkMap = std::unordered_map<std::string, double>;
 
 class SourceGroup {
  public:
@@ -150,7 +153,7 @@ class SerializationData {
   const std::vector<CompiledWasmModule>& compiled_wasm_modules() {
     return compiled_wasm_modules_;
   }
-  const base::Optional<v8::SharedValueConveyor>& shared_value_conveyor() {
+  const std::optional<v8::SharedValueConveyor>& shared_value_conveyor() {
     return shared_value_conveyor_;
   }
 
@@ -164,7 +167,7 @@ class SerializationData {
   std::vector<std::shared_ptr<v8::BackingStore>> backing_stores_;
   std::vector<std::shared_ptr<v8::BackingStore>> sab_backing_stores_;
   std::vector<CompiledWasmModule> compiled_wasm_modules_;
-  base::Optional<v8::SharedValueConveyor> shared_value_conveyor_;
+  std::optional<v8::SharedValueConveyor> shared_value_conveyor_;
 
  private:
   friend class Serializer;
@@ -184,7 +187,10 @@ class SerializationDataQueue {
 
 class Worker : public std::enable_shared_from_this<Worker> {
  public:
-  explicit Worker(const char* script);
+  static constexpr i::ExternalPointerTag kManagedTag = i::kD8WorkerTag;
+
+  explicit Worker(Isolate* parent_isolate, const char* script,
+                  bool flush_denormals);
   ~Worker();
 
   // Post a message to the worker. The worker will take ownership of the
@@ -197,6 +203,11 @@ class Worker : public std::enable_shared_from_this<Worker> {
   // return nullptr.
   // This function should only be called by the thread that created the Worker.
   std::unique_ptr<SerializationData> GetMessage(Isolate* requester);
+  // Synchronously retrieve messages from the worker's outgoing message queue.
+  // If there is no message in the queue, or the worker is no longer running,
+  // return nullptr.
+  // This function should only be called by the thread that created the Worker.
+  std::unique_ptr<SerializationData> TryGetMessage();
   // Terminate the worker's event loop. Messages from the worker that have been
   // queued can still be read via GetMessage().
   // This function can be called by any thread.
@@ -212,6 +223,7 @@ class Worker : public std::enable_shared_from_this<Worker> {
 
   // Enters State::kTerminated for the Worker and resets the task runner.
   void EnterTerminatedState();
+  bool IsTerminated() const { return state_ == State::kTerminated; }
 
   // Returns the Worker instance for this thread.
   static Worker* GetCurrentWorker();
@@ -247,13 +259,17 @@ class Worker : public std::enable_shared_from_this<Worker> {
 
   void ExecuteInThread();
   static void PostMessageOut(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void ImportScripts(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void Close(const v8::FunctionCallbackInfo<v8::Value>& info);
 
   static void SetCurrentWorker(Worker* worker);
 
   i::ParkingSemaphore out_semaphore_{0};
   SerializationDataQueue out_queue_;
+
   base::Thread* thread_ = nullptr;
   char* script_;
+  bool flush_denormals_;
   std::atomic<State> state_;
   bool is_joined_ = false;
   // For signalling that the worker has started.
@@ -270,9 +286,10 @@ class Worker : public std::enable_shared_from_this<Worker> {
   // The isolate should only be accessed by the worker itself, or when holding
   // the worker_mutex_ and after checking the worker state.
   Isolate* isolate_ = nullptr;
+  Isolate* parent_isolate_;
 
   // Only accessed by the worker thread.
-  v8::Persistent<v8::Context> context_;
+  Global<Context> context_;
 };
 
 class PerIsolateData {
@@ -287,7 +304,7 @@ class PerIsolateData {
 
   class V8_NODISCARD RealmScope {
    public:
-    explicit RealmScope(PerIsolateData* data);
+    explicit RealmScope(Isolate* isolate, const Global<Context>& context);
     ~RealmScope();
 
    private:
@@ -310,10 +327,6 @@ class PerIsolateData {
     int previous_index_;
   };
 
-  inline void SetTimeout(Local<Function> callback, Local<Context> context);
-  inline MaybeLocal<Function> GetTimeoutCallback();
-  inline MaybeLocal<Context> GetTimeoutContext();
-
   AsyncHooks* GetAsyncHooks() { return async_hooks_wrapper_; }
 
   void RemoveUnhandledPromise(Local<Promise> promise);
@@ -332,6 +345,15 @@ class PerIsolateData {
   Local<FunctionTemplate> GetDomNodeCtor() const;
   void SetDomNodeCtor(Local<FunctionTemplate> ctor);
 
+  bool HasRunningSubscribedWorkers();
+  void RegisterWorker(std::shared_ptr<Worker> worker);
+  void SubscribeWorkerOnMessage(const std::shared_ptr<Worker>& worker,
+                                Local<Context> context,
+                                Local<Function> callback);
+  std::pair<Local<Context>, Local<Function>> GetWorkerOnMessage(
+      const std::shared_ptr<Worker>& worker) const;
+  void UnregisterWorker(const std::shared_ptr<Worker>& worker);
+
  private:
   friend class Shell;
   friend class RealmScope;
@@ -341,8 +363,6 @@ class PerIsolateData {
   int realm_switch_;
   Global<Context>* realms_;
   Global<Value> realm_shared_;
-  std::queue<Global<Function>> set_timeout_callbacks_;
-  std::queue<Global<Context>> set_timeout_contexts_;
   bool ignore_unhandled_promises_;
   std::vector<std::tuple<Global<Promise>, Global<Message>, Global<Value>>>
       unhandled_promises_;
@@ -352,6 +372,16 @@ class PerIsolateData {
 #endif
   Global<FunctionTemplate> test_api_object_ctor_;
   Global<FunctionTemplate> dom_node_ctor_;
+  // Track workers and their callbacks separately, so that we know both which
+  // workers are still registered, and which of them have callbacks. We can't
+  // rely on Shell::running_workers_ or worker.IsTerminated(), because these are
+  // set concurrently and may race with callback subscription.
+  std::set<std::shared_ptr<Worker>> registered_workers_;
+  std::map<std::shared_ptr<Worker>,
+           std::pair<Global<Context>, Global<Function>>>
+      worker_message_callbacks_;
+
+  PerformanceMarkMap performance_mark_map_;
 
   int RealmIndexOrThrow(const v8::FunctionCallbackInfo<v8::Value>& info,
                         int arg_offset);
@@ -374,8 +404,7 @@ class ShellOptions {
   // repeated flags for identical boolean values. We allow exceptions for flags
   // with enum-like arguments since their conflicts can also be specified
   // completely.
-  template <class T,
-            bool kAllowIdenticalAssignment = std::is_same<T, bool>::value>
+  template <class T, bool kAllowIdenticalAssignment = std::is_same_v<T, bool>>
   class DisallowReassignment {
    public:
     DisallowReassignment(const char* name, T value)
@@ -406,7 +435,7 @@ class ShellOptions {
     T value_;
     bool specified_ = false;
   };
-
+  DisallowReassignment<bool> can_block = {"can_block", true};
   DisallowReassignment<const char*> d8_path = {"d8-path", ""};
   DisallowReassignment<bool> fuzzilli_coverage_statistics = {
       "fuzzilli-coverage-statistics", false};
@@ -435,10 +464,8 @@ class ShellOptions {
       "mock-arraybuffer-allocator", false};
   DisallowReassignment<size_t> mock_arraybuffer_allocator_limit = {
       "mock-arraybuffer-allocator-limit", 0};
-#if MULTI_MAPPED_ALLOCATOR_AVAILABLE
   DisallowReassignment<bool> multi_mapped_mock_allocator = {
       "multi-mapped-mock-allocator", false};
-#endif
   DisallowReassignment<bool> enable_inspector = {"enable-inspector", false};
   int num_isolates = 1;
   DisallowReassignment<v8::ScriptCompiler::CompileOptions, true>
@@ -455,6 +482,14 @@ class ShellOptions {
   DisallowReassignment<const char*> trace_path = {"trace-path", nullptr};
   DisallowReassignment<const char*> trace_config = {"trace-config", nullptr};
   DisallowReassignment<const char*> lcov_file = {"lcov", nullptr};
+#ifdef V8_OS_LINUX
+  // Allow linux perf to be started and stopped by performance.mark and
+  // performance.measure, respectively.
+  DisallowReassignment<bool> scope_linux_perf_to_mark_measure = {
+      "scope-linux-perf-to-mark-measure", false};
+  DisallowReassignment<int> perf_ctl_fd = {"perf-ctl-fd", -1};
+  DisallowReassignment<int> perf_ack_fd = {"perf-ack-fd", -1};
+#endif
   DisallowReassignment<bool> disable_in_process_stack_traces = {
       "disable-in-process-stack-traces", false};
   DisallowReassignment<int> read_from_tcp_port = {"read-from-tcp-port", -1};
@@ -481,10 +516,7 @@ class ShellOptions {
   DisallowReassignment<bool> wasm_trap_handler = {"wasm-trap-handler", true};
 #endif  // V8_ENABLE_WEBASSEMBLY
   DisallowReassignment<bool> expose_fast_api = {"expose-fast-api", false};
-#if V8_ENABLE_SANDBOX
-  DisallowReassignment<bool> enable_sandbox_crash_filter = {
-      "enable-sandbox-crash-filter", false};
-#endif  // V8_ENABLE_SANDBOX
+  DisallowReassignment<bool> flush_denormals = {"flush-denormals", false};
   DisallowReassignment<size_t> max_serializer_memory = {"max-serializer-memory",
                                                         1 * i::MB};
 };
@@ -502,33 +534,38 @@ class Shell : public i::AllStatic {
   };
   enum class CodeType { kFileName, kString, kFunction, kInvalid, kNone };
 
+  // Boolean return values (for any method below) typically denote "success".
+  // We return `false` on uncaught exceptions, except for termination
+  // exceptions.
   static bool ExecuteString(Isolate* isolate, Local<String> source,
-                            Local<String> name, PrintResult print_result,
+                            Local<String> name,
                             ReportExceptions report_exceptions,
-                            ProcessMessageQueue process_message_queue);
+                            Global<Value>* out_result = nullptr);
   static bool ExecuteModule(Isolate* isolate, const char* file_name);
   static bool LoadJSON(Isolate* isolate, const char* file_name);
   static void ReportException(Isolate* isolate, Local<Message> message,
                               Local<Value> exception);
-  static void ReportException(Isolate* isolate, TryCatch* try_catch);
+  static void ReportException(Isolate* isolate, const TryCatch& try_catch);
   static MaybeLocal<String> ReadFile(Isolate* isolate, const char* name,
                                      bool should_throw = true);
+  static std::unique_ptr<base::OS::MemoryMappedFile> ReadFileData(
+      Isolate* isolate, const char* name, bool should_throw = true);
   static Local<String> WasmLoadSourceMapCallback(Isolate* isolate,
                                                  const char* name);
   static MaybeLocal<Context> CreateEvaluationContext(Isolate* isolate);
   static int RunMain(Isolate* isolate, bool last_run);
   static int Main(int argc, char* argv[]);
+  static void InitializeDefaultCounters(Isolate* isolate);
   static void Exit(int exit_code);
   static void OnExit(Isolate* isolate, bool dispose);
+  static void DumpCounters();
   static void CollectGarbage(Isolate* isolate);
   static bool EmptyMessageQueues(Isolate* isolate);
   static bool CompleteMessageLoop(Isolate* isolate);
+  static bool FinishExecuting(Isolate* isolate, const Global<Context>& context);
 
   static bool HandleUnhandledPromiseRejections(Isolate* isolate);
 
-  static void PostForegroundTask(Isolate* isolate, std::unique_ptr<Task> task);
-  static void PostBlockingBackgroundTask(std::unique_ptr<Task> task);
-
   static std::unique_ptr<SerializationData> SerializeValue(
       Isolate* isolate, Local<Value> value, Local<Value> transfer);
   static MaybeLocal<Value> DeserializeValue(
@@ -545,10 +582,13 @@ class Shell : public i::AllStatic {
 
   static void PerformanceNow(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void PerformanceMark(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static bool LookupPerformanceMark(Isolate* isolate, v8::Local<String> name,
+                                    double* result);
   static void PerformanceMeasure(
       const v8::FunctionCallbackInfo<v8::Value>& info);
   static void PerformanceMeasureMemory(
       const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void PerformanceStub(const v8::FunctionCallbackInfo<v8::Value>& info);
 
   static void RealmCurrent(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void RealmOwner(const v8::FunctionCallbackInfo<v8::Value>& info);
@@ -562,9 +602,9 @@ class Shell : public i::AllStatic {
   static void RealmDispose(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void RealmSwitch(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void RealmEval(const v8::FunctionCallbackInfo<v8::Value>& info);
-  static void RealmSharedGet(Local<String> property,
+  static void RealmSharedGet(Local<Name> property,
                              const PropertyCallbackInfo<Value>& info);
-  static void RealmSharedSet(Local<String> property, Local<Value> value,
+  static void RealmSharedSet(Local<Name> property, Local<Value> value,
                              const PropertyCallbackInfo<void>& info);
 
   static void LogGetAndStop(const v8::FunctionCallbackInfo<v8::Value>& info);
@@ -573,6 +613,9 @@ class Shell : public i::AllStatic {
 
   static void InstallConditionalFeatures(
       const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void EnableJSPI(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void SetFlushDenormals(
+      const v8::FunctionCallbackInfo<v8::Value>& info);
 
   static void AsyncHooksCreateHook(
       const v8::FunctionCallbackInfo<v8::Value>& info);
@@ -610,10 +653,14 @@ class Shell : public i::AllStatic {
   static void NotifyDone(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void QuitOnce(v8::FunctionCallbackInfo<v8::Value>* info);
   static void Quit(const v8::FunctionCallbackInfo<v8::Value>& info);
-  static void Terminate(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void TerminateNow(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void ScheduleTermination(
+      const v8::FunctionCallbackInfo<v8::Value>& info);
   static void Version(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void WriteFile(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void ReadFile(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void CreateWasmMemoryMapDescriptor(
+      const v8::FunctionCallbackInfo<v8::Value>& info);
   static char* ReadChars(const char* name, int* size_out);
   static MaybeLocal<PrimitiveArray> ReadLines(Isolate* isolate,
                                               const char* name);
@@ -637,6 +684,10 @@ class Shell : public i::AllStatic {
   static void WorkerPostMessage(
       const v8::FunctionCallbackInfo<v8::Value>& info);
   static void WorkerGetMessage(const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void WorkerOnMessageGetter(
+      const v8::FunctionCallbackInfo<v8::Value>& info);
+  static void WorkerOnMessageSetter(
+      const v8::FunctionCallbackInfo<v8::Value>& info);
   static void WorkerTerminate(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void WorkerTerminateAndWait(
       const v8::FunctionCallbackInfo<v8::Value>& info);
@@ -673,10 +724,23 @@ class Shell : public i::AllStatic {
   // the "mkdir -p" command.
   static void MakeDirectory(const v8::FunctionCallbackInfo<v8::Value>& info);
   static void RemoveDirectory(const v8::FunctionCallbackInfo<v8::Value>& info);
+
+#ifdef V8_ENABLE_CONTINUATION_PRESERVED_EMBEDDER_DATA
+  static void GetContinuationPreservedEmbedderData(
+      const v8::FunctionCallbackInfo<v8::Value>& info);
+#endif  // V8_ENABLE_CONTINUATION_PRESERVER_EMBEDDER_DATA
+
+  static void GetExtrasBindingObject(
+      const v8::FunctionCallbackInfo<v8::Value>& info);
+
   static MaybeLocal<Promise> HostImportModuleDynamically(
       Local<Context> context, Local<Data> host_defined_options,
       Local<Value> resource_name, Local<String> specifier,
-      Local<FixedArray> import_assertions);
+      Local<FixedArray> import_attributes);
+  static MaybeLocal<Promise> HostImportModuleWithPhaseDynamically(
+      Local<Context> context, Local<Data> host_defined_options,
+      Local<Value> resource_name, Local<String> specifier,
+      ModuleImportPhase phase, Local<FixedArray> import_attributes);
 
   static void ModuleResolutionSuccessCallback(
       const v8::FunctionCallbackInfo<v8::Value>& info);
@@ -704,8 +768,6 @@ class Shell : public i::AllStatic {
   static ArrayBuffer::Allocator* array_buffer_allocator;
 
   static void SetWaitUntilDone(Isolate* isolate, bool value);
-  static void NotifyStartStreamingTask(Isolate* isolate);
-  static void NotifyFinishStreamingTask(Isolate* isolate);
 
   static char* ReadCharsFromTcpPort(const char* name, int* size_out);
 
@@ -726,6 +788,7 @@ class Shell : public i::AllStatic {
 
   static void Initialize(Isolate* isolate, D8Console* console,
                          bool isOnMainThread = true);
+  static void InitializeMainThreadCounters(Isolate* isolate);
 
   static void PromiseRejectCallback(v8::PromiseRejectMessage reject_message);
 
@@ -744,7 +807,7 @@ class Shell : public i::AllStatic {
 
   static const char* stringify_source_;
   static CounterMap* counter_map_;
-  static base::SharedMutex counter_mutex_;
+  static base::Mutex counter_mutex_;
   // We statically allocate a set of local counters to be used if we
   // don't want to store the stats in a memory-mapped file
   static CounterCollection local_counters_;
@@ -779,7 +842,6 @@ class Shell : public i::AllStatic {
   static Local<ObjectTemplate> CreateOSTemplate(Isolate* isolate);
   static Local<FunctionTemplate> CreateWorkerTemplate(Isolate* isolate);
   static Local<ObjectTemplate> CreateAsyncHookTemplate(Isolate* isolate);
-  static Local<ObjectTemplate> CreateTestRunnerTemplate(Isolate* isolate);
   static Local<ObjectTemplate> CreatePerformanceTemplate(Isolate* isolate);
   static Local<ObjectTemplate> CreateRealmTemplate(Isolate* isolate);
   static Local<ObjectTemplate> CreateD8Template(Isolate* isolate);
@@ -792,6 +854,10 @@ class Shell : public i::AllStatic {
       v8::MaybeLocal<Value> global_object);
   static void DisposeRealm(const v8::FunctionCallbackInfo<v8::Value>& info,
                            int index);
+
+  static MaybeLocal<Object> FetchModuleSource(
+      v8::Local<v8::Module> origin_module, v8::Local<v8::Context> context,
+      const std::string& file_name, ModuleType module_type);
   static MaybeLocal<Module> FetchModuleTree(v8::Local<v8::Module> origin_module,
                                             v8::Local<v8::Context> context,
                                             const std::string& file_name,
