@@ -6,12 +6,17 @@
 
 #include <unordered_set>
 
+#include "src/codegen/external-reference-encoder.h"
 #include "src/common/assert-scope.h"
 #include "src/execution/isolate.h"
-#include "src/heap/combined-heap.h"
+#include "src/heap/heap-layout-inl.h"
 #include "src/heap/heap.h"
+#include "src/heap/read-only-spaces.h"
+#include "src/heap/visit-object.h"
 #include "src/objects/heap-object-inl.h"
+#include "src/objects/objects-inl.h"
 #include "src/sandbox/external-pointer-table.h"
+#include "src/utils/ostreams.h"
 
 namespace v8 {
 namespace internal {
@@ -22,6 +27,8 @@ using HeapObjectSet = std::unordered_set<Tagged<HeapObject>, Object::Hasher,
                                          Object::KeyEqualSafe>;
 using HeapObjectMap = std::unordered_map<Tagged<HeapObject>, Tagged<HeapObject>,
                                          Object::Hasher, Object::KeyEqualSafe>;
+using HeapObjectList = std::vector<Tagged<HeapObject>>;
+
 bool Contains(const HeapObjectSet& s, Tagged<HeapObject> o) {
   return s.count(o) != 0;
 }
@@ -29,27 +36,58 @@ bool Contains(const HeapObjectMap& s, Tagged<HeapObject> o) {
   return s.count(o) != 0;
 }
 
+enum class PromoRecommendation {
+  // Do not promote this object.
+  kReject,
+  // Do promote this object.
+  kPromote,
+
+  // Promote if it's referenced by an object that must be promoted, otherwise
+  // do not promote.
+  // This is a rare case when by looking at an object itself it's not clear
+  // whether it should be promoted or not. The decision can be made later
+  // only taking into account where it's used.
+  // One of such examples is ArrayList which shouldn't be promoted unless
+  // it's a value of TemplateInfoWithProperties::property_list or
+  // TemplateInfoWithProperties::property_accessors of a template that
+  // should be promoted to RO space.
+  kMatchHost,
+};
+using enum PromoRecommendation;
+
 class Committee final {
  public:
-  static std::vector<Tagged<HeapObject>> DeterminePromotees(
+  static HeapObjectList DeterminePromotees(
       Isolate* isolate, const DisallowGarbageCollection& no_gc,
       const SafepointScope& safepoint_scope) {
     return Committee(isolate).DeterminePromotees(safepoint_scope);
   }
 
  private:
-  explicit Committee(Isolate* isolate) : isolate_(isolate) {}
+  explicit Committee(Isolate* isolate)
+      : isolate_(isolate), ref_encoder_(isolate) {}
 
-  std::vector<Tagged<HeapObject>> DeterminePromotees(
-      const SafepointScope& safepoint_scope) {
+  const ExternalReferenceEncoder& ref_encoder() const { return ref_encoder_; }
+
+  HeapObjectList DeterminePromotees(const SafepointScope& safepoint_scope) {
     DCHECK(promo_accepted_.empty());
     DCHECK(promo_rejected_.empty());
 
+    // List of promotees as discovered in insertion order. Using the
+    // `HeapObjectIterator` will visit pages in insertion order from a GC
+    // perspective which is deterministic independent of the absolute pages
+    // provided by the OS allocators in case of 32-bit builds (that don't use
+    // cages).
+    //
+    // We keep a separate HeapObjectList as there's no standard-equivalent for a
+    // hash set that can maintain insertion order.
+    HeapObjectList promo_accepted_list;
+
     // We assume that a full and precise GC has reclaimed all dead objects
     // and therefore that no filtering of unreachable objects is required here.
     HeapObjectIterator it(isolate_->heap(), safepoint_scope);
     for (Tagged<HeapObject> o = it.Next(); !o.is_null(); o = it.Next()) {
-      DCHECK(!o.InReadOnlySpace());
+      DCHECK(!HeapLayout::InReadOnlySpace(o));
 
       // Note that cycles prevent us from promoting/rejecting each subgraph as
       // we visit it, since locally we cannot determine whether the deferred
@@ -57,23 +95,43 @@ class Committee final {
       // could be solved if necessary (with more complex code), but for now
       // there are no performance issues.
       HeapObjectSet accepted_subgraph;  // Either all are accepted or none.
+      HeapObjectList accepted_subgraph_list;
       HeapObjectSet visited;            // Cycle detection.
-      if (!EvaluateSubgraph(o, &accepted_subgraph, &visited)) continue;
-      if (accepted_subgraph.empty()) continue;
+      if (!EvaluateSubgraph(std::nullopt, o, &accepted_subgraph, &visited,
+                            &accepted_subgraph_list)) {
+        continue;
+      }
+      if (accepted_subgraph.empty()) {
+        continue;
+      }
 
       if (V8_UNLIKELY(v8_flags.trace_read_only_promotion)) {
         LogAcceptedPromotionSet(accepted_subgraph);
       }
       promo_accepted_.insert(accepted_subgraph.begin(),
                              accepted_subgraph.end());
+      promo_accepted_list.insert(promo_accepted_list.end(),
+                                 accepted_subgraph_list.begin(),
+                                 accepted_subgraph_list.end());
+    }
+    if (V8_UNLIKELY(v8_flags.trace_read_only_promotion)) {
+      for (auto o : promo_deferred_) {
+        LogRejectedPromotionForNoMatchingHost(o);
+      }
     }
 
-    // Return promotees as a sorted list. Note that sorting uses object
-    // addresses; the list order is deterministic only if heap layout
-    // itself is deterministic (see v8_flags.predictable).
-    std::vector<Tagged<HeapObject>> promotees{promo_accepted_.begin(),
-                                              promo_accepted_.end()};
-    std::sort(promotees.begin(), promotees.end(), Object::Comparer());
+    // Remove duplicates from the promo_accepted_list. Note we have to jump
+    // through these hoops in order to preserve deterministic ordering
+    // (otherwise simply using the promo_accepted_ set would be sufficient).
+    HeapObjectSet seen_promotees;
+    HeapObjectList promotees;
+    promotees.reserve(promo_accepted_list.size());
+    for (Tagged<HeapObject> o : promo_accepted_list) {
+      if (Contains(seen_promotees, o)) continue;
+      seen_promotees.insert(o);
+      promotees.push_back(o);
+    }
+    CHECK_EQ(promotees.size(), promo_accepted_.size());
 
     return promotees;
   }
@@ -81,41 +139,79 @@ class Committee final {
   // Returns `false` if the subgraph rooted at `o` is rejected.
   // Returns `true` if it is accepted, or if we've reached a cycle and `o`
   // will be processed further up the callchain.
-  bool EvaluateSubgraph(Tagged<HeapObject> o, HeapObjectSet* accepted_subgraph,
-                        HeapObjectSet* visited) {
-    if (o.InReadOnlySpace()) return true;
+  bool EvaluateSubgraph(std::optional<Tagged<HeapObject>> maybe_host,
+                        Tagged<HeapObject> o, HeapObjectSet* accepted_subgraph,
+                        HeapObjectSet* visited, HeapObjectList* promotees) {
+    if (HeapLayout::InReadOnlySpace(o)) return true;
     if (Contains(promo_rejected_, o)) return false;
     if (Contains(promo_accepted_, o)) return true;
     if (Contains(*visited, o)) return true;
     visited->insert(o);
-    if (!IsPromoCandidate(isolate_, o)) {
-      const auto& [it, inserted] = promo_rejected_.insert(o);
-      if (V8_UNLIKELY(v8_flags.trace_read_only_promotion) && inserted) {
-        LogRejectedPromotionForFailedPredicate(o);
+    PromoRecommendation recommendation =
+        GetPromoRecommendation(this, isolate_, o);
+    switch (recommendation) {
+      case kPromote:
+        break;
+      case kReject: {
+        const auto& [it, inserted] = promo_rejected_.insert(o);
+        if (V8_UNLIKELY(v8_flags.trace_read_only_promotion) && inserted) {
+          LogRejectedPromotionForFailedPredicate(o);
+        }
+        return false;
       }
-      return false;
+      case kMatchHost:
+        if (!maybe_host.has_value()) {
+          // We've met this object during regular heap iteration and don't know
+          // yet how it's used. Postpone the decision for later.
+          DCHECK(!Contains(promo_deferred_, o));
+          promo_deferred_.insert(o);
+          visited->erase(o);
+          return false;
+        }
+        // This object is referenced by another promotion candidate, so
+        // let this one be promoted too.
+        promo_deferred_.erase(o);
+        break;
     }
     // Recurse into outgoing pointers.
-    CandidateVisitor v(this, accepted_subgraph, visited);
-    o->Iterate(isolate_, &v);
+    CandidateVisitor v(this, accepted_subgraph, visited, promotees, maybe_host);
+    VisitObject(isolate_, o, &v);
     if (!v.all_slots_are_promo_candidates()) {
       const auto& [it, inserted] = promo_rejected_.insert(o);
       if (V8_UNLIKELY(v8_flags.trace_read_only_promotion) && inserted) {
         LogRejectedPromotionForInvalidSubgraph(o,
                                                v.first_rejected_slot_offset());
       }
+      if (Tagged<TemplateInfo> info; TryCast<TemplateInfo>(o, &info)) {
+        CHECK_WITH_MSG(!info->should_promote_to_read_only(),
+                       "v8::Template was asked to be promoted to "
+                       "read only space but it wasn't possible. "
+                       "Use --trace-read-only-promotion for debugging.");
+      }
       return false;
     }
 
     accepted_subgraph->insert(o);
+    promotees->push_back(o);
+
+    // Some sanity checks for general-purpose objects (they should be included
+    // only as a part of expected hosts).
+    DCHECK_IMPLIES(IsArrayList(o),
+                   maybe_host.has_value() &&
+                       IsTemplateInfoWithProperties(maybe_host.value()));
     return true;
   }
 
 #define PROMO_CANDIDATE_TYPE_LIST(V) \
   V(AccessCheckInfo)                 \
   V(AccessorInfo)                    \
-  V(CallHandlerInfo)                 \
+  V(ArrayList)                       \
   V(Code)                            \
+  V(CodeWrapper)                     \
+  V(JSExternalObject)                \
+  V(ObjectTemplateInfo)              \
+  V(FunctionTemplateInfo)            \
+  V(FunctionTemplateRareData)        \
   V(InterceptorInfo)                 \
   V(ScopeInfo)                       \
   V(SharedFunctionInfo)              \
@@ -123,39 +219,74 @@ class Committee final {
   // TODO(jgruber): Don't forget to extend ReadOnlyPromotionImpl::Verify when
   // adding new object types here.
 
-  static bool IsPromoCandidate(Isolate* isolate, Tagged<HeapObject> o) {
+  static PromoRecommendation GetPromoRecommendation(Committee* committee,
+                                                    Isolate* isolate,
+                                                    Tagged<HeapObject> o) {
     const InstanceType itype = o->map(isolate)->instance_type();
-#define V(TYPE)                                            \
-  if (InstanceTypeChecker::Is##TYPE(itype)) {              \
-    return IsPromoCandidate##TYPE(isolate, TYPE::cast(o)); \
-    /* NOLINTNEXTLINE(readability/braces) */               \
+#define V(TYPE)                                                             \
+  if (InstanceTypeChecker::Is##TYPE(itype)) {                               \
+    return GetPromoRecommendation##TYPE(committee, isolate, Cast<TYPE>(o)); \
+    /* NOLINTNEXTLINE(readability/braces) */                                \
   } else
     PROMO_CANDIDATE_TYPE_LIST(V)
     /* if { ... } else */ {
-      return false;
+      return kReject;
     }
 #undef V
     UNREACHABLE();
   }
 #undef PROMO_CANDIDATE_TYPE_LIST
 
-#define DEF_PROMO_CANDIDATE(Type)                                        \
-  static bool IsPromoCandidate##Type(Isolate* isolate, Tagged<Type> o) { \
-    return true;                                                         \
+#define DEF_PROMO_CANDIDATE(Type)                               \
+  static PromoRecommendation GetPromoRecommendation##Type(      \
+      Committee* committee, Isolate* isolate, Tagged<Type> o) { \
+    return kPromote;                                            \
+  }
+#define DEF_MATCH_HOST_CANDIDATE(Type)                          \
+  static PromoRecommendation GetPromoRecommendation##Type(      \
+      Committee* committee, Isolate* isolate, Tagged<Type> o) { \
+    return kMatchHost;                                          \
   }
 
   DEF_PROMO_CANDIDATE(AccessCheckInfo)
   DEF_PROMO_CANDIDATE(AccessorInfo)
-  DEF_PROMO_CANDIDATE(CallHandlerInfo)
-  static bool IsPromoCandidateCode(Isolate* isolate, Tagged<Code> o) {
-    return Builtins::kCodeObjectsAreInROSpace && o->is_builtin();
+  DEF_MATCH_HOST_CANDIDATE(ArrayList)
+  // Promote v8::External only if it's used by a v8::Template that was
+  // explicitly requested to be promoted to RO space.
+  DEF_MATCH_HOST_CANDIDATE(JSExternalObject)
+
+  static PromoRecommendation GetPromoRecommendationFunctionTemplateInfo(
+      Committee* committee, Isolate* isolate, Tagged<FunctionTemplateInfo> o) {
+    // This flag is set by the embedder explicitly by calling
+    // v8::FunctionTemplate::SealAndPrepareForPromotionToReadOnly(..).
+    return o->should_promote_to_read_only() ? kPromote : kReject;
+  }
+  DEF_PROMO_CANDIDATE(FunctionTemplateRareData)
+
+  static PromoRecommendation GetPromoRecommendationCode(Committee* committee,
+                                                        Isolate* isolate,
+                                                        Tagged<Code> o) {
+    return Builtins::kCodeObjectsAreInROSpace && o->is_builtin() ? kPromote
+                                                                 : kReject;
+  }
+  static PromoRecommendation GetPromoRecommendationCodeWrapper(
+      Committee* committee, Isolate* isolate, Tagged<CodeWrapper> o) {
+    return GetPromoRecommendationCode(committee, isolate, o->code(isolate));
   }
   DEF_PROMO_CANDIDATE(InterceptorInfo)
+
+  static PromoRecommendation GetPromoRecommendationObjectTemplateInfo(
+      Committee* committee, Isolate* isolate, Tagged<ObjectTemplateInfo> o) {
+    // This flag is set by the embedder explicitly by calling
+    // v8::ObjectTemplate::SealAndPrepareForPromotionToReadOnly(..).
+    return o->should_promote_to_read_only() ? kPromote : kReject;
+  }
+
   DEF_PROMO_CANDIDATE(ScopeInfo)
-  static bool IsPromoCandidateSharedFunctionInfo(Isolate* isolate,
-                                                 Tagged<SharedFunctionInfo> o) {
+  static PromoRecommendation GetPromoRecommendationSharedFunctionInfo(
+      Committee* committee, Isolate* isolate, Tagged<SharedFunctionInfo> o) {
     // Only internal SFIs are guaranteed to remain immutable.
-    if (o->has_script(kAcquireLoad)) return false;
+    if (o->has_script(kAcquireLoad)) return kReject;
     // kIllegal is used for js_global_object_function, which is created during
     // bootstrapping but never rooted. We currently assumed that all objects in
     // the snapshot are live. But RO space is 1) not GC'd and 2) serialized
@@ -164,7 +295,12 @@ class Committee final {
     // TODO(jgruber): A better solution. Remove the liveness assumption (see
     // test-heap-profiler.cc)? Overwrite dead RO objects with fillers
     // pre-serialization? Implement a RO GC pass pre-serialization?
-    return o->HasBuiltinId() && o->builtin_id() != Builtin::kIllegal;
+    if (o->HasBuiltinId() && o->builtin_id() != Builtin::kIllegal) {
+      return kPromote;
+    }
+    // Api functions are good candidates for promotion.
+    if (o->IsApiFunction()) return kPromote;
+    return kReject;
   }
   DEF_PROMO_CANDIDATE(Symbol)
 
@@ -175,10 +311,17 @@ class Committee final {
   class CandidateVisitor : public ObjectVisitor {
    public:
     CandidateVisitor(Committee* committee, HeapObjectSet* accepted_subgraph,
-                     HeapObjectSet* visited)
+                     HeapObjectSet* visited, HeapObjectList* promotees,
+                     std::optional<Tagged<HeapObject>> maybe_host)
         : committee_(committee),
           accepted_subgraph_(accepted_subgraph),
-          visited_(visited) {}
+          visited_(visited),
+          promotees_(promotees),
+          maybe_host_(maybe_host) {}
+
+    // Promotion decisions for some objects might depend on the properties
+    // of the "outer" object.
+    std::optional<Tagged<HeapObject>> maybe_host() const { return maybe_host_; }
 
     int first_rejected_slot_offset() const {
       return first_rejected_slot_offset_;
@@ -191,11 +334,12 @@ class Committee final {
                        MaybeObjectSlot end) final {
       if (!all_slots_are_promo_candidates()) return;
       for (MaybeObjectSlot slot = start; slot < end; slot++) {
-        MaybeObject maybe_object = slot.load(committee_->isolate_);
+        Tagged<MaybeObject> maybe_object = slot.load(committee_->isolate_);
         Tagged<HeapObject> heap_object;
         if (!maybe_object.GetHeapObject(&heap_object)) continue;
-        if (!committee_->EvaluateSubgraph(heap_object, accepted_subgraph_,
-                                          visited_)) {
+        if (!committee_->EvaluateSubgraph({host}, heap_object,
+                                          accepted_subgraph_, visited_,
+                                          promotees_)) {
           first_rejected_slot_offset_ =
               static_cast<int>(slot.address() - host.address());
           DCHECK_GE(first_rejected_slot_offset_, 0);
@@ -220,6 +364,8 @@ class Committee final {
     Committee* const committee_;
     HeapObjectSet* const accepted_subgraph_;
     HeapObjectSet* const visited_;
+    HeapObjectList* const promotees_;
+    std::optional<Tagged<HeapObject>> maybe_host_;
     int first_rejected_slot_offset_ = -1;
   };
 
@@ -238,6 +384,14 @@ class Committee final {
               << "\n";
   }
 
+  void LogRejectedPromotionForNoMatchingHost(Tagged<HeapObject> o) {
+    std::cout << "ro-promotion: rejected because it wasn't included into any "
+                 "accepted set "
+              << reinterpret_cast<void*>(o.ptr()) << " ("
+              << o->map()->instance_type() << ")"
+              << "\n";
+  }
+
   void LogRejectedPromotionForInvalidSubgraph(Tagged<HeapObject> o,
                                               int first_rejected_slot_offset) {
     std::cout << "ro-promotion: rejected due to rejected subgraph "
@@ -246,7 +400,7 @@ class Committee final {
               << " at slot offset " << first_rejected_slot_offset << " ";
 
     MaybeObjectSlot slot = o->RawMaybeWeakField(first_rejected_slot_offset);
-    MaybeObject maybe_object = slot.load(isolate_);
+    Tagged<MaybeObject> maybe_object = slot.load(isolate_);
     Tagged<HeapObject> heap_object;
     if (maybe_object.GetHeapObject(&heap_object)) {
       std::cout << reinterpret_cast<void*>(heap_object.ptr()) << " ("
@@ -258,8 +412,10 @@ class Committee final {
   }
 
   Isolate* const isolate_;
+  ExternalReferenceEncoder ref_encoder_;
   HeapObjectSet promo_accepted_;
   HeapObjectSet promo_rejected_;
+  HeapObjectSet promo_deferred_;
 };
 
 class ReadOnlyPromotionImpl final : public AllStatic {
@@ -274,6 +430,10 @@ class ReadOnlyPromotionImpl final : public AllStatic {
           rospace->AllocateRaw(size, kTaggedAligned).ToObjectChecked();
       Heap::CopyBlock(dst.address(), src.address(), size);
       moves->emplace(src, dst);
+
+      if (V8_UNLIKELY(v8_flags.trace_read_only_promotion_verbose)) {
+        LogPromotedObject(src, dst);
+      }
     }
   }
 
@@ -289,7 +449,7 @@ class ReadOnlyPromotionImpl final : public AllStatic {
 
     // Iterate all roots.
     EmbedderStackStateScope stack_scope(
-        isolate->heap(), EmbedderStackStateScope::kExplicitInvocation,
+        isolate->heap(), EmbedderStackStateOrigin::kExplicitInvocation,
         StackState::kNoHeapPointers);
     heap->IterateRoots(&v, base::EnumSet<SkipRoot>{});
 
@@ -298,12 +458,45 @@ class ReadOnlyPromotionImpl final : public AllStatic {
     // and therefore that no filtering of unreachable objects is required here.
     HeapObjectIterator it(heap, safepoint_scope);
     for (Tagged<HeapObject> o = it.Next(); !o.is_null(); o = it.Next()) {
-      o->Iterate(isolate, &v);
+      VisitObject(isolate, o, &v);
     }
 
     // Iterate all objects we just copied into RO space.
     for (auto [src, dst] : moves) {
-      dst->Iterate(isolate, &v);
+      VisitObject(isolate, dst, &v);
+    }
+
+#ifdef V8_ENABLE_LEAPTIERING
+    // Iterate all entries in the JSDispatchTable as they could contain
+    // pointers to promoted Code objects.
+    JSDispatchTable* const jdt = IsolateGroup::current()->js_dispatch_table();
+    jdt->IterateActiveEntriesIn(heap->js_dispatch_table_space(),
+                                [&](JSDispatchHandle handle) {
+                                  Tagged<Code> old_code = jdt->GetCode(handle);
+                                  auto it = moves.find(old_code);
+                                  if (it == moves.end()) return;
+                                  Tagged<HeapObject> new_code = it->second;
+                                  CHECK(IsCode(new_code));
+                                  // TODO(saelo): is it worth logging something
+                                  // in this case?
+                                  jdt->SetCodeNoWriteBarrier(
+                                      handle, Cast<Code>(new_code));
+                                });
+#endif  // V8_ENABLE_LEAPTIERING
+  }
+
+  static void DeleteDeadObjects(Isolate* isolate,
+                                const SafepointScope& safepoint_scope,
+                                const HeapObjectMap& moves) {
+    // After moving a source object to a new destination, overwrite the source
+    // memory with a filler. This is needed for moved objects that are verified
+    // by the heap verifier to have a 1-1 relation with some other object (e.g.
+    // objects related to trusted space). The verifier won't compute liveness
+    // and instead just iterates linearly over pages. Without this change the
+    // verifier would fail on this now-dead object.
+    for (auto [src, dst] : moves) {
+      CHECK(!HeapLayout::InReadOnlySpace(src));
+      isolate->heap()->CreateFillerObjectAt(src.address(), src->Size(isolate));
     }
   }
 
@@ -313,7 +506,11 @@ class ReadOnlyPromotionImpl final : public AllStatic {
     //
     // Known objects.
     Heap* heap = isolate->heap();
-    CHECK(heap->promise_all_resolve_element_shared_fun().InReadOnlySpace());
+    CHECK(HeapLayout::InReadOnlySpace(
+        heap->promise_all_resolve_element_closure_shared_fun()));
+    CHECK(HeapLayout::InReadOnlySpace(heap->error_stack_getter_fun_template()));
+    CHECK(HeapLayout::InReadOnlySpace(heap->error_stack_setter_fun_template()));
+
     // TODO(jgruber): Extend here with more objects as they are added to
     // the promotion algorithm.
 
@@ -321,7 +518,8 @@ class ReadOnlyPromotionImpl final : public AllStatic {
     if (Builtins::kCodeObjectsAreInROSpace) {
       Builtins* builtins = isolate->builtins();
       for (int i = 0; i < Builtins::kBuiltinCount; i++) {
-        CHECK(builtins->code(static_cast<Builtin>(i)).InReadOnlySpace());
+        CHECK(HeapLayout::InReadOnlySpace(
+            builtins->code(static_cast<Builtin>(i))));
       }
     }
 #endif  // DEBUG
@@ -332,9 +530,14 @@ class ReadOnlyPromotionImpl final : public AllStatic {
    public:
     UpdatePointersVisitor(Isolate* isolate, const HeapObjectMap* moves)
         : isolate_(isolate), moves_(moves) {
-      for (auto [src, dst] : *moves_) {
-        moves_reverse_lookup_.emplace(dst, src);
+#ifdef V8_ENABLE_SANDBOX
+      for (auto [_src, dst] : *moves_) {
+        promoted_objects_.emplace(dst);
+        if (IsCode(dst)) {
+          PromoteCodePointerEntryFor(Cast<Code>(dst));
+        }
       }
+#endif  // V8_ENABLE_SANDBOX
     }
 
     // The RootVisitor interface.
@@ -366,8 +569,7 @@ class ReadOnlyPromotionImpl final : public AllStatic {
     void VisitExternalPointer(Tagged<HeapObject> host,
                               ExternalPointerSlot slot) final {
 #ifdef V8_ENABLE_SANDBOX
-      auto it = moves_reverse_lookup_.find(host);
-      if (it == moves_reverse_lookup_.end()) return;
+      if (promoted_objects_.find(host) == promoted_objects_.end()) return;
 
       // If we reach here, `host` is a moved object with external pointer slots
       // located in RO space. To preserve the 1:1 relation between slots and
@@ -375,7 +577,8 @@ class ReadOnlyPromotionImpl final : public AllStatic {
       // read_only_external_pointer_space) now.
       RecordProcessedSlotIfDebug(slot.address());
       Address slot_value = slot.load(isolate_);
-      slot.init(isolate_, slot_value);
+      DCHECK(slot.ExactTagIsKnown());
+      slot.init(isolate_, host, slot_value, slot.exact_tag());
 
       if (V8_UNLIKELY(v8_flags.trace_read_only_promotion_verbose)) {
         LogUpdatedExternalPointerTableEntry(host, slot, slot_value);
@@ -383,68 +586,42 @@ class ReadOnlyPromotionImpl final : public AllStatic {
 #endif  // V8_ENABLE_SANDBOX
     }
     void VisitIndirectPointer(Tagged<HeapObject> host, IndirectPointerSlot slot,
-                              IndirectPointerMode mode) final {}
-    void VisitIndirectPointerTableEntry(Tagged<HeapObject> host,
-                                        IndirectPointerSlot slot) final {
+                              IndirectPointerMode mode) final {
 #ifdef V8_ENABLE_SANDBOX
-      // When an object owning an indirect pointer table entry is relocated, it
-      // needs to update the entry to point to its new location. Currently, only
-      // Code objects are referenced through indirect pointers, and they use the
-      // code pointer table.
-      CHECK(IsCode(host));
-      CHECK_EQ(slot.tag(), kCodeIndirectPointerTag);
-
-      // Due to the way we handle baseline code during serialization (we
-      // manually skip over them), we may encounter such live Code objects in
-      // mutable space during iteration. Do a lookup to make sure we only
-      // update CPT entries for moved objects.
-      auto it = moves_reverse_lookup_.find(host);
-      if (it == moves_reverse_lookup_.end()) return;
-
-      // If we reach here, `host` is a moved Code object located in RO space.
-      CHECK(host.InReadOnlySpace());
-      RecordProcessedSlotIfDebug(slot.address());
-
-      Tagged<Code> dead_code = Code::cast(it->second);
-      CHECK(IsCode(dead_code));
-      CHECK(!dead_code.InReadOnlySpace());
-
-      IndirectPointerHandle handle = slot.Relaxed_LoadHandle();
-      CodePointerTable* cpt = GetProcessWideCodePointerTable();
-      CHECK_EQ(dead_code, Tagged<Object>(cpt->GetCodeObject(handle)));
-
-      // The old Code object (in mutable space) is dead. To preserve the 1:1
-      // relation between Code objects and CPT entries, overwrite it immediately
-      // with the filler object.
-      isolate_->heap()->CreateFillerObjectAt(dead_code.address(), Code::kSize);
-
-      // Update the CPT entry to point at the moved RO Code object.
-      cpt->SetCodeObject(handle, host.ptr());
-
-      if (V8_UNLIKELY(v8_flags.trace_read_only_promotion_verbose)) {
-        LogUpdatedCodePointerTableEntry(host, slot, dead_code);
+      if (slot.tag() == kCodeIndirectPointerTag) {
+        VisitCodePointer(host, slot);
       }
-#else
-      UNREACHABLE();
-#endif
+#endif  // V8_ENABLE_SANDBOX
+    }
+    void VisitTrustedPointerTableEntry(Tagged<HeapObject> host,
+                                       IndirectPointerSlot slot) final {
+#ifdef V8_ENABLE_SANDBOX
+      if (slot.tag() == kCodeIndirectPointerTag) {
+        VisitCodePointer(host, slot);
+      }
+#endif  // V8_ENABLE_SANDBOX
     }
     void VisitRootPointers(Root root, const char* description,
                            OffHeapObjectSlot start,
                            OffHeapObjectSlot end) override {
-      // We shouldn't have moved any string table contents (which is what
-      // OffHeapObjectSlot currently refers to).
+      // We shouldn't have moved any string table contents or SharedStructType
+      // registry contents (which is what OffHeapObjectSlot currently refers
+      // to).
       for (OffHeapObjectSlot slot = start; slot < end; slot++) {
         Tagged<Object> o = slot.load(isolate_);
         if (!IsHeapObject(o)) continue;
-        CHECK(!Contains(*moves_, HeapObject::cast(o)));
+        CHECK(!Contains(*moves_, Cast<HeapObject>(o)));
       }
     }
 
    private:
     void ProcessSlot(Root root, FullObjectSlot slot) {
       Tagged<Object> old_slot_value_obj = slot.load(isolate_);
+#ifdef V8_ENABLE_DIRECT_HANDLE
+      if (old_slot_value_obj.ptr() == kTaggedNullAddress) return;
+#endif
       if (!IsHeapObject(old_slot_value_obj)) return;
-      Tagged<HeapObject> old_slot_value = HeapObject::cast(old_slot_value_obj);
+      Tagged<HeapObject> old_slot_value = Cast<HeapObject>(old_slot_value_obj);
       auto it = moves_->find(old_slot_value);
       if (it == moves_->end()) return;
       Tagged<HeapObject> new_slot_value = it->second;
@@ -459,12 +636,60 @@ class ReadOnlyPromotionImpl final : public AllStatic {
       auto it = moves_->find(old_slot_value);
       if (it == moves_->end()) return;
       Tagged<HeapObject> new_slot_value = it->second;
-      slot.store(MaybeObject::FromObject(new_slot_value));
+      slot.store(new_slot_value);
       if (V8_UNLIKELY(v8_flags.trace_read_only_promotion_verbose)) {
         LogUpdatedPointer(host, slot, old_slot_value, new_slot_value);
       }
     }
 
+#ifdef V8_ENABLE_SANDBOX
+    void VisitCodePointer(Tagged<HeapObject> host, IndirectPointerSlot slot) {
+      CHECK_EQ(kCodeIndirectPointerTag, slot.tag());
+      IndirectPointerHandle old_handle = slot.Relaxed_LoadHandle();
+      auto it = code_pointer_moves_.find(old_handle);
+      if (it == code_pointer_moves_.end()) return;
+
+      // If we reach here, `host` is a moved object with a code pointer slot
+      // located in RO space. To preserve the 1:1 relation between slots and
+      // table entries, we need to use the relocated code pointer table entry.
+      RecordProcessedSlotIfDebug(slot.address());
+      IndirectPointerHandle new_handle = it->second;
+      slot.Relaxed_StoreHandle(new_handle);
+
+      if (V8_UNLIKELY(v8_flags.trace_read_only_promotion_verbose)) {
+        LogUpdatedCodePointerTableEntry(host, slot, old_handle, new_handle);
+      }
+    }
+
+    void PromoteCodePointerEntryFor(Tagged<Code> code) {
+      // If we reach here, `code` is a moved Code object located in RO space.
+      CHECK(HeapLayout::InReadOnlySpace(code));
+
+      IndirectPointerSlot slot = code->RawIndirectPointerField(
+          Code::kSelfIndirectPointerOffset, kCodeIndirectPointerTag);
+      CodeEntrypointTag entrypoint_tag = code->entrypoint_tag();
+
+      IndirectPointerHandle old_handle = slot.Relaxed_LoadHandle();
+      CodePointerTable* cpt = IsolateGroup::current()->code_pointer_table();
+
+      // To preserve the 1:1 relation between slots and code table entries,
+      // allocate a new entry (in the code_pointer_space of the RO heap) now.
+      // The slot will be updated later, when the Code object is visited.
+      CodePointerTable::Space* space =
+          IsolateForSandbox(isolate_).GetCodePointerTableSpaceFor(
+              slot.address());
+      IndirectPointerHandle new_handle = cpt->AllocateAndInitializeEntry(
+          space, code.address(), cpt->GetEntrypoint(old_handle, entrypoint_tag),
+          entrypoint_tag);
+
+      code_pointer_moves_.emplace(old_handle, new_handle);
+
+      if (V8_UNLIKELY(v8_flags.trace_read_only_promotion_verbose)) {
+        LogPromotedCodePointerTableEntry(code, old_handle, new_handle);
+      }
+    }
+#endif  // V8_ENABLE_SANDBOX
+
     void LogUpdatedPointer(Root root, FullObjectSlot slot,
                            Tagged<HeapObject> old_slot_value,
                            Tagged<HeapObject> new_slot_value) {
@@ -493,11 +718,13 @@ class ReadOnlyPromotionImpl final : public AllStatic {
     }
     void LogUpdatedCodePointerTableEntry(Tagged<HeapObject> host,
                                          IndirectPointerSlot slot,
-                                         Tagged<Code> old_code_object) {
+                                         IndirectPointerHandle old_handle,
+                                         IndirectPointerHandle new_handle) {
       std::cout << "ro-promotion: updated code pointer table entry {host "
                 << reinterpret_cast<void*>(host.address()) << " slot "
                 << reinterpret_cast<void*>(slot.address()) << " from "
-                << reinterpret_cast<void*>(old_code_object.ptr()) << "}\n";
+                << AsHex(old_handle, 8, true) << " to "
+                << AsHex(new_handle, 8, true) << "}\n";
     }
 
 #ifdef DEBUG
@@ -513,8 +740,36 @@ class ReadOnlyPromotionImpl final : public AllStatic {
 
     Isolate* const isolate_;
     const HeapObjectMap* moves_;
-    HeapObjectMap moves_reverse_lookup_;
+
+#ifdef V8_ENABLE_SANDBOX
+    HeapObjectSet promoted_objects_;
+
+    // When an object owning an pointer table entry is relocated to the RO
+    // space, it cannot just update the entry to point to its new location
+    // (see b/330450848). A new pointer table entry must be allocated for the
+    // relocated object, in a RO segment of the table.
+
+    using IndirectPointerHandleMap =
+        std::unordered_map<IndirectPointerHandle, IndirectPointerHandle>;
+    IndirectPointerHandleMap code_pointer_moves_;
+#endif  // V8_ENABLE_SANDBOX
   };
+
+  static void LogPromotedObject(Tagged<HeapObject> src,
+                                Tagged<HeapObject> dst) {
+    std::cout << "ro-promotion: promoted object {from "
+              << reinterpret_cast<void*>(src.ptr()) << " to "
+              << reinterpret_cast<void*>(dst.ptr()) << "}\n";
+  }
+
+  static void LogPromotedCodePointerTableEntry(
+      Tagged<Code> code, IndirectPointerHandle old_handle,
+      IndirectPointerHandle new_handle) {
+    std::cout << "ro-promotion: promoted code pointer table entry {code "
+              << reinterpret_cast<void*>(code.ptr()) << " slot "
+              << AsHex(old_handle, 8, true) << " to "
+              << AsHex(new_handle, 8, true) << "}\n";
+  }
 };
 
 }  // namespace
@@ -533,6 +788,7 @@ void ReadOnlyPromotion::Promote(Isolate* isolate,
   // Update all references to moved objects to point at their new location in
   // RO space.
   ReadOnlyPromotionImpl::UpdatePointers(isolate, safepoint_scope, moves);
+  ReadOnlyPromotionImpl::DeleteDeadObjects(isolate, safepoint_scope, moves);
   ReadOnlyPromotionImpl::Verify(isolate, safepoint_scope);
 }
 
