@@ -20,7 +20,7 @@
 #include "src/maglev/maglev-graph-labeller.h"
 #include "src/maglev/maglev-graph-processor.h"
 #include "src/maglev/maglev-graph.h"
-#include "src/maglev/maglev-ir.h"
+#include "src/maglev/maglev-ir-inl.h"
 #include "src/objects/script-inl.h"
 #include "src/objects/shared-function-info-inl.h"
 #include "src/utils/utils.h"
@@ -31,24 +31,25 @@ namespace maglev {
 
 namespace {
 
-int IntWidth(int val) { return std::ceil(std::log10(val + 1)); }
+int IntWidth(int val) {
+  if (val == -1) return 2;
+  return std::ceil(std::log10(val + 1));
+}
 
-int MaxIdWidth(MaglevGraphLabeller* graph_labeller, NodeIdT max_node_id,
-               int padding_adjustement = 0) {
-  int max_width = IntWidth(graph_labeller->max_node_id());
+int MaxIdWidth(NodeIdT max_node_id, int padding_adjustement = 0) {
+  int max_width = IntWidth(GetCurrentGraphLabeller()->max_node_id());
   if (max_node_id != kInvalidNodeId) {
     max_width += IntWidth(max_node_id) + 1;
   }
   return max_width + 2 + padding_adjustement;
 }
 
-void PrintPaddedId(std::ostream& os, MaglevGraphLabeller* graph_labeller,
-                   NodeIdT max_node_id, NodeBase* node,
+void PrintPaddedId(std::ostream& os, NodeIdT max_node_id, NodeBase* node,
                    std::string padding = " ", int padding_adjustement = 0) {
-  int id = graph_labeller->NodeId(node);
+  int id = GetCurrentGraphLabeller()->NodeId(node);
   int id_width = IntWidth(id);
   int other_id_width = node->has_id() ? 1 + IntWidth(node->id()) : 0;
-  int max_width = MaxIdWidth(graph_labeller, max_node_id, padding_adjustement);
+  int max_width = MaxIdWidth(max_node_id, padding_adjustement);
   int padding_width = std::max(0, max_width - id_width - other_id_width);
 
   for (int i = 0; i < padding_width; ++i) {
@@ -58,17 +59,16 @@ void PrintPaddedId(std::ostream& os, MaglevGraphLabeller* graph_labeller,
   if (node->has_id()) {
     os << node->id() << "/";
   }
-  os << graph_labeller->NodeId(node) << ": ";
+  os << GetCurrentGraphLabeller()->NodeId(node) << ": ";
 }
 
 void PrintPadding(std::ostream& os, int size) {
   os << std::setfill(' ') << std::setw(size) << "";
 }
 
-void PrintPadding(std::ostream& os, MaglevGraphLabeller* graph_labeller,
-                  NodeIdT max_node_id, int padding_adjustement) {
-  PrintPadding(os,
-               MaxIdWidth(graph_labeller, max_node_id, padding_adjustement));
+void PrintPadding(std::ostream& os, NodeIdT max_node_id,
+                  int padding_adjustement) {
+  PrintPadding(os, MaxIdWidth(max_node_id, padding_adjustement));
 }
 
 enum ConnectionLocation {
@@ -251,10 +251,8 @@ int MaglevPrintingVisitorOstream::overflow(int c) {
 
 }  // namespace
 
-MaglevPrintingVisitor::MaglevPrintingVisitor(
-    MaglevGraphLabeller* graph_labeller, std::ostream& os)
-    : graph_labeller_(graph_labeller),
-      os_(os),
+MaglevPrintingVisitor::MaglevPrintingVisitor(std::ostream& os)
+    : os_(os),
       os_for_additional_info_(
           new MaglevPrintingVisitorOstream(os_, &targets_)) {}
 
@@ -313,7 +311,8 @@ void MaglevPrintingVisitor::PreProcessGraph(Graph* graph) {
                      [](BasicBlock* block) { return block == nullptr; }));
 }
 
-void MaglevPrintingVisitor::PreProcessBasicBlock(BasicBlock* block) {
+BlockProcessResult MaglevPrintingVisitor::PreProcessBasicBlock(
+    BasicBlock* block) {
   size_t loop_position = static_cast<size_t>(-1);
   if (loop_headers_.erase(block) > 0) {
     loop_position = AddTarget(targets_, block);
@@ -362,21 +361,61 @@ void MaglevPrintingVisitor::PreProcessBasicBlock(BasicBlock* block) {
     if (v8_flags.log_colour) os_ << "\033[0m";
   }
 
-  int block_id = graph_labeller_->BlockId(block);
-  os_ << "Block b" << block_id;
+  os_ << "Block b" << block->id();
+  if (block->has_state() && block->state()->is_resumable_loop()) {
+    os_ << " (resumable)";
+  }
   if (block->is_exception_handler_block()) {
     os_ << " (exception handler)";
   }
+  if (block->is_loop() && block->has_state()) {
+    if (block->state()->is_loop_with_peeled_iteration()) {
+      os_ << " peeled";
+    }
+    if (const LoopEffects* loop_effects = block->state()->loop_effects()) {
+      os_ << " (effects:";
+      if (loop_effects->unstable_aspects_cleared) {
+        if (loop_effects->unstable_aspects_cleared) {
+          os_ << " ua";
+        }
+        if (loop_effects->context_slot_written.size()) {
+          os_ << " c" << loop_effects->context_slot_written.size();
+        }
+        if (loop_effects->objects_written.size()) {
+          os_ << " o" << loop_effects->objects_written.size();
+        }
+        if (loop_effects->keys_cleared.size()) {
+          os_ << " k" << loop_effects->keys_cleared.size();
+        }
+      }
+      os_ << ")";
+    }
+  }
   os_ << "\n";
 
   MaglevPrintingVisitorOstream::cast(os_for_additional_info_)->set_padding(1);
+  return BlockProcessResult::kContinue;
 }
 
 namespace {
 
+void PrintInputLocationAndAdvance(std::ostream& os, ValueNode* node,
+                                  InputLocation*& input_location) {
+  if (InlinedAllocation* allocation = node->TryCast<InlinedAllocation>()) {
+    if (allocation->HasBeenAnalysed() && allocation->HasBeenElided()) {
+      os << "(elided)";
+      return;
+    }
+  }
+  if (input_location) {
+    os << input_location->operand();
+    input_location++;
+  }
+}
+
 void PrintSingleDeoptFrame(
-    std::ostream& os, MaglevGraphLabeller* graph_labeller,
-    const DeoptFrame& frame, InputLocation*& current_input_location,
+    std::ostream& os, const DeoptFrame& frame,
+    InputLocation*& current_input_location,
     LazyDeoptInfo* lazy_deopt_info_if_top_frame = nullptr) {
   switch (frame.type()) {
     case DeoptFrame::FrameType::kInterpretedFrame: {
@@ -390,23 +429,20 @@ void PrintSingleDeoptFrame(
         return;
       }
       os << " : {";
-      bool first = true;
+      os << "<closure>:" << PrintNodeLabel(frame.as_interpreted().closure())
+         << ":";
+      PrintInputLocationAndAdvance(os, frame.as_interpreted().closure(),
+                                   current_input_location);
       frame.as_interpreted().frame_state()->ForEachValue(
           frame.as_interpreted().unit(),
           [&](ValueNode* node, interpreter::Register reg) {
-            if (first) {
-              first = false;
-            } else {
-              os << ", ";
-            }
-            os << reg.ToString() << ":";
+            os << ", " << reg.ToString() << ":";
             if (lazy_deopt_info_if_top_frame &&
                 lazy_deopt_info_if_top_frame->IsResultRegister(reg)) {
               os << "<result>";
             } else {
-              os << PrintNodeLabel(graph_labeller, node) << ":"
-                 << current_input_location->operand();
-              current_input_location++;
+              os << PrintNodeLabel(node) << ":";
+              PrintInputLocationAndAdvance(os, node, current_input_location);
             }
           });
       os << "}";
@@ -416,14 +452,14 @@ void PrintSingleDeoptFrame(
       os << "@ConstructInvokeStub";
       if (!v8_flags.print_maglev_deopt_verbose) return;
       os << " : {";
-      os << "<this>:"
-         << PrintNodeLabel(graph_labeller, frame.as_construct_stub().receiver())
-         << ":" << current_input_location->operand();
-      current_input_location++;
+      os << "<this>:" << PrintNodeLabel(frame.as_construct_stub().receiver())
+         << ":";
+      PrintInputLocationAndAdvance(os, frame.as_construct_stub().receiver(),
+                                   current_input_location);
       os << ", <context>:"
-         << PrintNodeLabel(graph_labeller, frame.as_construct_stub().context())
-         << ":" << current_input_location->operand();
-      current_input_location++;
+         << PrintNodeLabel(frame.as_construct_stub().context()) << ":";
+      PrintInputLocationAndAdvance(os, frame.as_construct_stub().context(),
+                                   current_input_location);
       os << "}";
       break;
     }
@@ -433,17 +469,14 @@ void PrintSingleDeoptFrame(
       os << " : {";
       auto arguments = frame.as_inlined_arguments().arguments();
       DCHECK_GT(arguments.size(), 0);
-      os << "<this>:" << PrintNodeLabel(graph_labeller, arguments[0]) << ":"
-         << current_input_location->operand();
-      current_input_location++;
+      os << "<this>:" << PrintNodeLabel(arguments[0]) << ":";
+      PrintInputLocationAndAdvance(os, arguments[0], current_input_location);
       if (arguments.size() > 1) {
         os << ", ";
       }
       for (size_t i = 1; i < arguments.size(); i++) {
-        os << "a" << (i - 1) << ":"
-           << PrintNodeLabel(graph_labeller, arguments[i]) << ":"
-           << current_input_location->operand();
-        current_input_location++;
+        os << "a" << (i - 1) << ":" << PrintNodeLabel(arguments[i]) << ":";
+        PrintInputLocationAndAdvance(os, arguments[i], current_input_location);
         os << ", ";
       }
       os << "}";
@@ -455,110 +488,144 @@ void PrintSingleDeoptFrame(
       os << " : {";
       int arg_index = 0;
       for (ValueNode* node : frame.as_builtin_continuation().parameters()) {
-        os << "a" << arg_index << ":" << PrintNodeLabel(graph_labeller, node)
-           << ":" << current_input_location->operand();
+        os << "a" << arg_index << ":" << PrintNodeLabel(node) << ":";
+        PrintInputLocationAndAdvance(os, node, current_input_location);
         arg_index++;
-        current_input_location++;
         os << ", ";
       }
       os << "<context>:"
-         << PrintNodeLabel(graph_labeller,
-                           frame.as_builtin_continuation().context())
-         << ":" << current_input_location->operand();
-      current_input_location++;
+         << PrintNodeLabel(frame.as_builtin_continuation().context()) << ":";
+      PrintInputLocationAndAdvance(os,
+                                   frame.as_builtin_continuation().context(),
+                                   current_input_location);
       os << "}";
       break;
     }
   }
 }
 
+void PrintVirtualObjects(std::ostream& os, std::vector<BasicBlock*> targets,
+                         const DeoptFrame& frame, int max_node_id) {
+  if (!v8_flags.trace_deopt_verbose) return;
+  PrintVerticalArrows(os, targets);
+  PrintPadding(os, max_node_id, 0);
+  os << "  │       VOs : { ";
+  const VirtualObjectList& virtual_objects = frame.GetVirtualObjects();
+  for (auto vo : virtual_objects) {
+    os << PrintNodeLabel(vo) << "; ";
+  }
+  os << "}\n";
+}
+
+void PrintDeoptInfoInputLocation(std::ostream& os,
+                                 std::vector<BasicBlock*> targets,
+                                 DeoptInfo* deopt_info,
+                                 int max_node_id) {
+#ifdef DEBUG
+  if (!v8_flags.print_maglev_deopt_verbose) return;
+  PrintVerticalArrows(os, targets);
+  PrintPadding(os, max_node_id, 0);
+  if (deopt_info->has_input_locations()) {
+    os << "  input locations: " << deopt_info->input_locations() << " ("
+       << deopt_info->input_location_count() << " slots)";
+  }
+  os << "\n";
+#endif  // DEBUG
+}
+
 void RecursivePrintEagerDeopt(std::ostream& os,
                               std::vector<BasicBlock*> targets,
                               const DeoptFrame& frame,
-                              MaglevGraphLabeller* graph_labeller,
                               int max_node_id,
                               InputLocation*& current_input_location) {
   if (frame.parent()) {
-    RecursivePrintEagerDeopt(os, targets, *frame.parent(), graph_labeller,
-                             max_node_id, current_input_location);
+    RecursivePrintEagerDeopt(os, targets, *frame.parent(), max_node_id,
+                             current_input_location);
   }
 
   PrintVerticalArrows(os, targets);
-  PrintPadding(os, graph_labeller, max_node_id, 0);
+  PrintPadding(os, max_node_id, 0);
   if (!frame.parent()) {
     os << "  ↱ eager ";
   } else {
     os << "  │       ";
   }
-  PrintSingleDeoptFrame(os, graph_labeller, frame, current_input_location);
+  PrintSingleDeoptFrame(os, frame, current_input_location);
   os << "\n";
+  PrintVirtualObjects(os, targets, frame, max_node_id);
 }
 
 void PrintEagerDeopt(std::ostream& os, std::vector<BasicBlock*> targets,
-                     NodeBase* node, MaglevGraphLabeller* graph_labeller,
-                     int max_node_id) {
+                     NodeBase* node, int max_node_id) {
   EagerDeoptInfo* deopt_info = node->eager_deopt_info();
-  InputLocation* current_input_location = deopt_info->input_locations();
-  RecursivePrintEagerDeopt(os, targets, deopt_info->top_frame(), graph_labeller,
-                           max_node_id, current_input_location);
+  InputLocation* current_input_location = nullptr;
+  if (deopt_info->has_input_locations()) {
+    current_input_location = deopt_info->input_locations();
+  }
+  PrintDeoptInfoInputLocation(os, targets, deopt_info, max_node_id);
+  RecursivePrintEagerDeopt(os, targets, deopt_info->top_frame(), max_node_id,
+                           current_input_location);
 }
 
 void MaybePrintEagerDeopt(std::ostream& os, std::vector<BasicBlock*> targets,
-                          NodeBase* node, MaglevGraphLabeller* graph_labeller,
-                          int max_node_id) {
+                          NodeBase* node, int max_node_id) {
   if (node->properties().can_eager_deopt()) {
-    PrintEagerDeopt(os, targets, node, graph_labeller, max_node_id);
+    PrintEagerDeopt(os, targets, node, max_node_id);
   }
 }
 
 void RecursivePrintLazyDeopt(std::ostream& os, std::vector<BasicBlock*> targets,
                              const DeoptFrame& frame,
-                             MaglevGraphLabeller* graph_labeller,
                              int max_node_id,
                              InputLocation*& current_input_location) {
   if (frame.parent()) {
-    RecursivePrintLazyDeopt(os, targets, *frame.parent(), graph_labeller,
-                            max_node_id, current_input_location);
+    RecursivePrintLazyDeopt(os, targets, *frame.parent(), max_node_id,
+                            current_input_location);
   }
 
   PrintVerticalArrows(os, targets);
-  PrintPadding(os, graph_labeller, max_node_id, 0);
+  PrintPadding(os, max_node_id, 0);
   os << "  │      ";
-  PrintSingleDeoptFrame(os, graph_labeller, frame, current_input_location);
+  PrintSingleDeoptFrame(os, frame, current_input_location);
   os << "\n";
+  PrintVirtualObjects(os, targets, frame, max_node_id);
 }
 
 template <typename NodeT>
 void PrintLazyDeopt(std::ostream& os, std::vector<BasicBlock*> targets,
-                    NodeT* node, MaglevGraphLabeller* graph_labeller,
-                    int max_node_id) {
+                    NodeT* node, int max_node_id) {
   LazyDeoptInfo* deopt_info = node->lazy_deopt_info();
-  InputLocation* current_input_location = deopt_info->input_locations();
+  InputLocation* current_input_location = nullptr;
+  if (deopt_info->has_input_locations()) {
+    current_input_location = deopt_info->input_locations();
+  }
+
+  PrintDeoptInfoInputLocation(os, targets, deopt_info, max_node_id);
+
   const DeoptFrame& top_frame = deopt_info->top_frame();
   if (top_frame.parent()) {
-    RecursivePrintLazyDeopt(os, targets, *top_frame.parent(), graph_labeller,
-                            max_node_id, current_input_location);
+    RecursivePrintLazyDeopt(os, targets, *top_frame.parent(), max_node_id,
+                            current_input_location);
   }
 
   PrintVerticalArrows(os, targets);
-  PrintPadding(os, graph_labeller, max_node_id, 0);
+  PrintPadding(os, max_node_id, 0);
 
   os << "  ↳ lazy ";
-  PrintSingleDeoptFrame(os, graph_labeller, top_frame, current_input_location,
-                        deopt_info);
+  PrintSingleDeoptFrame(os, top_frame, current_input_location, deopt_info);
   os << "\n";
+  PrintVirtualObjects(os, targets, top_frame, max_node_id);
 }
 
 template <typename NodeT>
 void PrintExceptionHandlerPoint(std::ostream& os,
                                 std::vector<BasicBlock*> targets, NodeT* node,
-                                MaglevGraphLabeller* graph_labeller,
                                 int max_node_id) {
   // If no handler info, then we cannot throw.
   ExceptionHandlerInfo* info = node->exception_handler_info();
-  if (!info->HasExceptionHandler()) return;
+  if (!info->HasExceptionHandler() || info->ShouldLazyDeopt()) return;
 
-  BasicBlock* block = info->catch_block.block_ptr();
+  BasicBlock* block = info->catch_block();
   DCHECK(block->is_exception_handler_block());
 
   if (!block->has_phi()) {
@@ -586,7 +653,7 @@ void PrintExceptionHandlerPoint(std::ostream& os,
   }
 
   PrintVerticalArrows(os, targets);
-  PrintPadding(os, graph_labeller, max_node_id, 0);
+  PrintPadding(os, max_node_id, 0);
 
   os << "  ↳ throw @" << handler_offset << " : {";
   bool first = true;
@@ -602,7 +669,7 @@ void PrintExceptionHandlerPoint(std::ostream& os,
         } else {
           os << ", ";
         }
-        os << reg.ToString() << ":" << PrintNodeLabel(graph_labeller, node);
+        os << reg.ToString() << ":" << PrintNodeLabel(node);
       });
   os << "}\n";
 }
@@ -610,19 +677,17 @@ void PrintExceptionHandlerPoint(std::ostream& os,
 void MaybePrintLazyDeoptOrExceptionHandler(std::ostream& os,
                                            std::vector<BasicBlock*> targets,
                                            NodeBase* node,
-                                           MaglevGraphLabeller* graph_labeller,
                                            int max_node_id) {
   switch (node->opcode()) {
-#define CASE(Name)                                                          \
-  case Opcode::k##Name:                                                     \
-    if constexpr (Name::kProperties.can_lazy_deopt()) {                     \
-      PrintLazyDeopt<Name>(os, targets, node->Cast<Name>(), graph_labeller, \
-                           max_node_id);                                    \
-    }                                                                       \
-    if constexpr (Name::kProperties.can_throw()) {                          \
-      PrintExceptionHandlerPoint<Name>(os, targets, node->Cast<Name>(),     \
-                                       graph_labeller, max_node_id);        \
-    }                                                                       \
+#define CASE(Name)                                                        \
+  case Opcode::k##Name:                                                   \
+    if constexpr (Name::kProperties.can_lazy_deopt()) {                   \
+      PrintLazyDeopt<Name>(os, targets, node->Cast<Name>(), max_node_id); \
+    }                                                                     \
+    if constexpr (Name::kProperties.can_throw()) {                        \
+      PrintExceptionHandlerPoint<Name>(os, targets, node->Cast<Name>(),   \
+                                       max_node_id);                      \
+    }                                                                     \
     break;
     NODE_BASE_LIST(CASE)
 #undef CASE
@@ -645,7 +710,7 @@ void MaybePrintProvenance(std::ostream& os, std::vector<BasicBlock*> targets,
   if (provenance.position.IsKnown() &&
       (provenance.position != existing_provenance.position ||
        provenance.unit != existing_provenance.unit)) {
-    script = Script::cast(
+    script = Cast<Script>(
         provenance.unit->shared_function_info().object()->script());
     has_position_info = script->GetPositionInfo(
         provenance.position.ScriptOffset(), &position_info,
@@ -656,7 +721,7 @@ void MaybePrintProvenance(std::ostream& os, std::vector<BasicBlock*> targets,
   // Do the actual function + position print.
   if (needs_function_print) {
     if (script.is_null()) {
-      script = Script::cast(
+      script = Cast<Script>(
           provenance.unit->shared_function_info().object()->script());
     }
     PrintVerticalArrows(os, targets);
@@ -703,7 +768,7 @@ void MaybePrintProvenance(std::ostream& os, std::vector<BasicBlock*> targets,
 ProcessResult MaglevPrintingVisitor::Process(Phi* phi,
                                              const ProcessingState& state) {
   PrintVerticalArrows(os_, targets_);
-  PrintPaddedId(os_, graph_labeller_, max_node_id_, phi);
+  PrintPaddedId(os_, max_node_id_, phi);
   os_ << "φ";
   switch (phi->value_representation()) {
     case ValueRepresentation::kTagged:
@@ -721,19 +786,23 @@ ProcessResult MaglevPrintingVisitor::Process(Phi* phi,
     case ValueRepresentation::kHoleyFloat64:
       os_ << "ʰᶠ";
       break;
-    case ValueRepresentation::kWord64:
+    case ValueRepresentation::kIntPtr:
       UNREACHABLE();
   }
+  if (phi->uses_require_31_bit_value()) {
+    os_ << "ⁱ";
+  }
   if (phi->input_count() == 0) {
-    os_ << "ₑ " << phi->owner().ToString();
+    os_ << "ₑ " << (phi->owner().is_valid() ? phi->owner().ToString() : "VO");
   } else {
-    os_ << " " << phi->owner().ToString() << " (";
+    os_ << " " << (phi->owner().is_valid() ? phi->owner().ToString() : "VO")
+        << " (";
     // Manually walk Phi inputs to print just the node labels, without
     // input locations (which are shown in the predecessor block's gap
     // moves).
     for (int i = 0; i < phi->input_count(); ++i) {
       if (i > 0) os_ << ", ";
-      os_ << PrintNodeLabel(graph_labeller_, phi->input(i).node());
+      os_ << PrintNodeLabel(phi->input(i).node());
     }
     os_ << ")";
   }
@@ -745,6 +814,10 @@ ProcessResult MaglevPrintingVisitor::Process(Phi* phi,
     }
   }
   os_ << " → " << phi->result().operand();
+  if (phi->result().operand().IsAllocated() && phi->is_spilled() &&
+      phi->spill_slot() != phi->result().operand()) {
+    os_ << " (spilled: " << phi->spill_slot() << ")";
+  }
   if (phi->has_valid_live_range()) {
     os_ << ", live range: [" << phi->live_range().start << "-"
         << phi->live_range().end << "]";
@@ -755,47 +828,45 @@ ProcessResult MaglevPrintingVisitor::Process(Phi* phi,
   os_ << "\n";
 
   MaglevPrintingVisitorOstream::cast(os_for_additional_info_)
-      ->set_padding(MaxIdWidth(graph_labeller_, max_node_id_, 2));
+      ->set_padding(MaxIdWidth(max_node_id_, 2));
   return ProcessResult::kContinue;
 }
 
 ProcessResult MaglevPrintingVisitor::Process(Node* node,
                                              const ProcessingState& state) {
   MaglevGraphLabeller::Provenance provenance =
-      graph_labeller_->GetNodeProvenance(node);
+      GetCurrentGraphLabeller()->GetNodeProvenance(node);
   if (provenance.unit != nullptr) {
     MaybePrintProvenance(os_, targets_, provenance, existing_provenance_);
     existing_provenance_ = provenance;
   }
 
-  MaybePrintEagerDeopt(os_, targets_, node, graph_labeller_, max_node_id_);
+  MaybePrintEagerDeopt(os_, targets_, node, max_node_id_);
 
   PrintVerticalArrows(os_, targets_);
-  PrintPaddedId(os_, graph_labeller_, max_node_id_, node);
+  PrintPaddedId(os_, max_node_id_, node);
   if (node->properties().is_call()) {
     os_ << "🐢 ";
   }
-  os_ << PrintNode(graph_labeller_, node) << "\n";
+  os_ << PrintNode(node) << "\n";
 
   MaglevPrintingVisitorOstream::cast(os_for_additional_info_)
-      ->set_padding(MaxIdWidth(graph_labeller_, max_node_id_, 2));
+      ->set_padding(MaxIdWidth(max_node_id_, 2));
 
-  MaybePrintLazyDeoptOrExceptionHandler(os_, targets_, node, graph_labeller_,
-                                        max_node_id_);
+  MaybePrintLazyDeoptOrExceptionHandler(os_, targets_, node, max_node_id_);
   return ProcessResult::kContinue;
 }
 
 ProcessResult MaglevPrintingVisitor::Process(ControlNode* control_node,
                                              const ProcessingState& state) {
   MaglevGraphLabeller::Provenance provenance =
-      graph_labeller_->GetNodeProvenance(control_node);
+      GetCurrentGraphLabeller()->GetNodeProvenance(control_node);
   if (provenance.unit != nullptr) {
     MaybePrintProvenance(os_, targets_, provenance, existing_provenance_);
     existing_provenance_ = provenance;
   }
 
-  MaybePrintEagerDeopt(os_, targets_, control_node, graph_labeller_,
-                       max_node_id_);
+  MaybePrintEagerDeopt(os_, targets_, control_node, max_node_id_);
 
   bool has_fallthrough = false;
 
@@ -804,7 +875,7 @@ ProcessResult MaglevPrintingVisitor::Process(ControlNode* control_node,
 
     PrintVerticalArrows(os_, targets_, {}, {target}, true);
     os_ << "◄─";
-    PrintPaddedId(os_, graph_labeller_, max_node_id_, control_node, "─", -2);
+    PrintPaddedId(os_, max_node_id_, control_node, "─", -2);
     std::replace(targets_.begin(), targets_.end(), target,
                  static_cast<BasicBlock*>(nullptr));
 
@@ -816,8 +887,7 @@ ProcessResult MaglevPrintingVisitor::Process(ControlNode* control_node,
     has_fallthrough |= !AddTargetIfNotNext(targets_, target, state.next_block(),
                                            &arrows_starting_here);
     PrintVerticalArrows(os_, targets_, arrows_starting_here);
-    PrintPaddedId(os_, graph_labeller_, max_node_id_, control_node,
-                  has_fallthrough ? " " : "─");
+    PrintPaddedId(os_, max_node_id_, control_node, has_fallthrough ? " " : "─");
 
   } else if (control_node->Is<BranchControlNode>()) {
     BasicBlock* true_target =
@@ -831,7 +901,7 @@ ProcessResult MaglevPrintingVisitor::Process(ControlNode* control_node,
     has_fallthrough |= !AddTargetIfNotNext(
         targets_, true_target, state.next_block(), &arrows_starting_here);
     PrintVerticalArrows(os_, targets_, arrows_starting_here);
-    PrintPaddedId(os_, graph_labeller_, max_node_id_, control_node, "─");
+    PrintPaddedId(os_, max_node_id_, control_node, "─");
   } else if (control_node->Is<Switch>()) {
     std::set<size_t> arrows_starting_here;
     for (int i = 0; i < control_node->Cast<Switch>()->size(); i++) {
@@ -850,14 +920,14 @@ ProcessResult MaglevPrintingVisitor::Process(ControlNode* control_node,
     }
 
     PrintVerticalArrows(os_, targets_, arrows_starting_here);
-    PrintPaddedId(os_, graph_labeller_, max_node_id_, control_node, "─");
+    PrintPaddedId(os_, max_node_id_, control_node, "─");
 
   } else {
     PrintVerticalArrows(os_, targets_);
-    PrintPaddedId(os_, graph_labeller_, max_node_id_, control_node);
+    PrintPaddedId(os_, max_node_id_, control_node);
   }
 
-  os_ << PrintNode(graph_labeller_, control_node) << "\n";
+  os_ << PrintNode(control_node) << "\n";
 
   bool printed_phis = false;
   if (control_node->Is<UnconditionalControlNode>()) {
@@ -866,17 +936,17 @@ ProcessResult MaglevPrintingVisitor::Process(ControlNode* control_node,
     if (target->has_phi()) {
       printed_phis = true;
       PrintVerticalArrows(os_, targets_);
-      PrintPadding(os_, graph_labeller_, max_node_id_, -1);
+      PrintPadding(os_, max_node_id_, -1);
       os_ << (has_fallthrough ? "│" : " ");
       os_ << "  with gap moves:\n";
       int pid = state.block()->predecessor_id();
       for (Phi* phi : *target->phis()) {
         PrintVerticalArrows(os_, targets_);
-        PrintPadding(os_, graph_labeller_, max_node_id_, -1);
+        PrintPadding(os_, max_node_id_, -1);
         os_ << (has_fallthrough ? "│" : " ");
         os_ << "    - ";
-        graph_labeller_->PrintInput(os_, phi->input(pid));
-        os_ << " → " << graph_labeller_->NodeId(phi) << ": φ";
+        GetCurrentGraphLabeller()->PrintInput(os_, phi->input(pid));
+        os_ << " → " << GetCurrentGraphLabeller()->NodeId(phi) << ": φ";
         switch (phi->value_representation()) {
           case ValueRepresentation::kTagged:
             os_ << "ᵀ";
@@ -893,15 +963,19 @@ ProcessResult MaglevPrintingVisitor::Process(ControlNode* control_node,
           case ValueRepresentation::kHoleyFloat64:
             os_ << "ʰᶠ";
             break;
-          case ValueRepresentation::kWord64:
+          case ValueRepresentation::kIntPtr:
             UNREACHABLE();
         }
-        os_ << " " << phi->owner().ToString() << " " << phi->result().operand()
-            << "\n";
+        if (phi->uses_require_31_bit_value()) {
+          os_ << "ⁱ";
+        }
+        os_ << " " << (phi->owner().is_valid() ? phi->owner().ToString() : "VO")
+            << " " << phi->result().operand() << "\n";
       }
+#ifdef V8_ENABLE_MAGLEV
       if (target->state()->register_state().is_initialized()) {
         PrintVerticalArrows(os_, targets_);
-        PrintPadding(os_, graph_labeller_, max_node_id_, -1);
+        PrintPadding(os_, max_node_id_, -1);
         os_ << (has_fallthrough ? "│" : " ");
         os_ << "  with register merges:\n";
         auto print_register_merges = [&](auto reg, RegisterState& state) {
@@ -910,7 +984,7 @@ ProcessResult MaglevPrintingVisitor::Process(ControlNode* control_node,
           if (LoadMergeState(state, &node, &merge)) {
             compiler::InstructionOperand source = merge->operand(pid);
             PrintVerticalArrows(os_, targets_);
-            PrintPadding(os_, graph_labeller_, max_node_id_, -1);
+            PrintPadding(os_, max_node_id_, -1);
             os_ << (has_fallthrough ? "│" : " ");
             os_ << "    - " << source << " → " << reg << "\n";
           }
@@ -920,12 +994,13 @@ ProcessResult MaglevPrintingVisitor::Process(ControlNode* control_node,
         target->state()->register_state().ForEachDoubleRegister(
             print_register_merges);
       }
+#endif
     }
   }
 
   PrintVerticalArrows(os_, targets_);
   if (has_fallthrough) {
-    PrintPadding(os_, graph_labeller_, max_node_id_, -1);
+    PrintPadding(os_, max_node_id_, -1);
     if (printed_phis) {
       os_ << "▼";
     } else {
@@ -937,24 +1012,48 @@ ProcessResult MaglevPrintingVisitor::Process(ControlNode* control_node,
   // TODO(leszeks): Allow MaglevPrintingVisitorOstream to print the arrowhead
   // so that it overlaps the fallthrough arrow.
   MaglevPrintingVisitorOstream::cast(os_for_additional_info_)
-      ->set_padding(MaxIdWidth(graph_labeller_, max_node_id_, 2));
+      ->set_padding(MaxIdWidth(max_node_id_, 2));
 
   return ProcessResult::kContinue;
 }
 
-void PrintGraph(std::ostream& os, MaglevCompilationInfo* compilation_info,
-                Graph* const graph) {
+void PrintGraph(std::ostream& os, Graph* const graph) {
   GraphProcessor<MaglevPrintingVisitor, /*visit_identity_nodes*/ true> printer(
-      compilation_info->graph_labeller(), os);
+      os);
   printer.ProcessGraph(graph);
 }
 
 void PrintNode::Print(std::ostream& os) const {
-  node_->Print(os, graph_labeller_, skip_targets_);
+  node_->Print(os, skip_targets_);
 }
 
 void PrintNodeLabel::Print(std::ostream& os) const {
-  graph_labeller_->PrintNodeLabel(os, node_);
+  GetCurrentGraphLabeller()->PrintNodeLabel(os, node_);
+}
+
+// For GDB: Print any basic block with `print bb->Print()`.
+void BasicBlock::Print() const {
+  std::cout << "Block";
+  if (is_loop()) {
+    if (state()->is_loop_with_peeled_iteration()) {
+      std::cout << " (peeled loop)";
+    } else if (has_state() && state()->is_resumable_loop()) {
+      std::cout << " (resumable loop)";
+    } else {
+      std::cout << " (loop header)";
+    }
+  } else if (is_exception_handler_block()) {
+    std::cout << " (exception handler)";
+  }
+  std::cout << "\n";
+  for (auto node : nodes_) {
+    node->Print();
+  }
+  if (control_node_) {
+    control_node_->Print();
+  } else {
+    std::cout << " (missing control node)\n";
+  }
 }
 
 }  // namespace maglev
