@@ -31,8 +31,12 @@
 #include "src/codegen/arm/register-arm.h"
 #elif V8_TARGET_ARCH_ARM64
 #include "src/codegen/arm64/register-arm64.h"
+#elif V8_TARGET_ARCH_RISCV64
+#include "src/codegen/riscv/register-riscv.h"
 #elif V8_TARGET_ARCH_X64
 #include "src/codegen/x64/register-x64.h"
+#elif V8_TARGET_ARCH_S390X
+#include "src/codegen/s390/register-s390.h"
 #else
 #error "Maglev does not supported this architecture."
 #endif
@@ -68,9 +72,15 @@ ControlNode* NearestPostDominatingHole(ControlNode* node) {
   // If the node is a Jump, it may be a hole, but only if it is not a
   // fallthrough (jump to the immediately next block). Otherwise, it will point
   // to the nearest post-dominating hole in its own "next" field.
-  if (Jump* jump = node->TryCast<Jump>()) {
-    if (IsTargetOfNodeFallthrough(jump, jump->target())) {
-      return jump->next_post_dominating_hole();
+  if (node->Is<Jump>() || node->Is<CheckpointedJump>()) {
+    BasicBlock* target;
+    if (auto jmp = node->TryCast<Jump>()) {
+      target = jmp->target();
+    } else {
+      target = node->Cast<CheckpointedJump>()->target();
+    }
+    if (IsTargetOfNodeFallthrough(node, target)) {
+      return node->next_post_dominating_hole();
     }
   }
 
@@ -147,7 +157,7 @@ bool IsLiveAtTarget(ValueNode* node, ControlNode* source, BasicBlock* target) {
   }
 
   // Drop all values on resumable loop headers.
-  if (target->has_state() && target->state()->is_resumable_loop()) return false;
+  if (target->is_loop() && target->state()->is_resumable_loop()) return false;
 
   // TODO(verwaest): This should be true but isn't because we don't yet
   // eliminate dead code.
@@ -157,40 +167,45 @@ bool IsLiveAtTarget(ValueNode* node, ControlNode* source, BasicBlock* target) {
   return node->live_range().end >= target->first_id();
 }
 
-// TODO(dmercadier): this function should never clear any registers, since dead
-// registers should always have been cleared:
-//  - Nodes without uses have their output registers cleared right after their
-//    allocation by `FreeRegistersUsedBy(node)`.
-//  - Once the last use of a Node has been processed, its register is freed (by
-//    UpdateUse, called from Assigned***Input, called by AssignInputs).
-// Thus, this function should DCHECK that all of the registers are live at
-// target, rather than clearing the ones that aren't.
-template <typename RegisterT>
-void ClearDeadFallthroughRegisters(RegisterFrameState<RegisterT>& registers,
-                                   ConditionalControlNode* control_node,
-                                   BasicBlock* target) {
-  RegListBase<RegisterT> list = registers.used();
-  while (list != registers.empty()) {
-    RegisterT reg = list.PopFirst();
-    ValueNode* node = registers.GetValue(reg);
-    if (!IsLiveAtTarget(node, control_node, target)) {
-      registers.FreeRegistersUsedBy(node);
-      // Update the registers we're visiting to avoid revisiting this node.
-      list.clear(registers.free());
-    }
-  }
-}
-
 bool IsDeadNodeToSkip(Node* node) {
-  return node->Is<ValueNode>() && node->Cast<ValueNode>()->has_no_more_uses() &&
-         !node->properties().is_required_when_unused();
+  if (!node->Is<ValueNode>()) return false;
+  ValueNode* value = node->Cast<ValueNode>();
+  return value->has_no_more_uses() &&
+         !value->properties().is_required_when_unused();
 }
 
 }  // namespace
 
+void StraightForwardRegisterAllocator::ApplyPatches(BasicBlock* block) {
+  // TODO(verwaest): Perhaps don't actually merge these in but let the code
+  // generator pick up the gap moves from a separate list.
+  size_t diff = patches_.size();
+  if (diff == 0) return;
+  block->nodes().resize(block->nodes().size() + diff);
+  auto patches_it = patches_.end() - 1;
+  for (auto node_it = block->nodes().end() - 1 - diff;
+       node_it >= block->nodes().begin(); --node_it) {
+    *(node_it + diff) = *node_it;
+    for (; patches_it->diff == (node_it - block->nodes().begin());
+         --patches_it) {
+      --diff;
+      *(node_it + diff) = patches_it->new_node;
+      if (diff == 0) {
+        patches_.resize(0);
+        return;
+      }
+    }
+  }
+  UNREACHABLE();
+}
+
 StraightForwardRegisterAllocator::StraightForwardRegisterAllocator(
-    MaglevCompilationInfo* compilation_info, Graph* graph)
-    : compilation_info_(compilation_info), graph_(graph) {
+    MaglevCompilationInfo* compilation_info, Graph* graph,
+    RegallocInfo* regalloc_info)
+    : compilation_info_(compilation_info),
+      graph_(graph),
+      patches_(compilation_info_->zone()),
+      regalloc_info_(regalloc_info) {
   ComputePostDominatingHoles();
   AllocateRegisters();
   uint32_t tagged_stack_slots = tagged_.top;
@@ -293,37 +308,39 @@ void StraightForwardRegisterAllocator::ComputePostDominatingHoles() {
   // the block. Such a list of jumps terminates in return or jumploop.
   for (BasicBlock* block : base::Reversed(*graph_)) {
     ControlNode* control = block->control_node();
-    if (auto node = control->TryCast<UnconditionalControlNode>()) {
+    if (auto unconditional_control =
+            control->TryCast<UnconditionalControlNode>()) {
       // If the current control node is a jump, prepend it to the list of jumps
       // at the target.
-      control->set_next_post_dominating_hole(
-          NearestPostDominatingHole(node->target()->control_node()));
-    } else if (auto node = control->TryCast<BranchControlNode>()) {
+      control->set_next_post_dominating_hole(NearestPostDominatingHole(
+          unconditional_control->target()->control_node()));
+    } else if (auto branch = control->TryCast<BranchControlNode>()) {
       ControlNode* first =
-          NearestPostDominatingHole(node->if_true()->control_node());
+          NearestPostDominatingHole(branch->if_true()->control_node());
       ControlNode* second =
-          NearestPostDominatingHole(node->if_false()->control_node());
+          NearestPostDominatingHole(branch->if_false()->control_node());
       control->set_next_post_dominating_hole(
           HighestPostDominatingHole(first, second));
-    } else if (auto node = control->TryCast<Switch>()) {
-      int num_targets = node->size() + (node->has_fallthrough() ? 1 : 0);
+    } else if (auto switch_node = control->TryCast<Switch>()) {
+      int num_targets =
+          switch_node->size() + (switch_node->has_fallthrough() ? 1 : 0);
       if (num_targets == 1) {
         // If we have a single target, the next post dominating hole
         // is the same one as the target.
-        DCHECK(!node->has_fallthrough());
+        DCHECK(!switch_node->has_fallthrough());
         control->set_next_post_dominating_hole(NearestPostDominatingHole(
-            node->targets()[0].block_ptr()->control_node()));
+            switch_node->targets()[0].block_ptr()->control_node()));
         continue;
       }
       // Calculate the post dominating hole for each target.
       base::SmallVector<ControlNode*, 16> holes(num_targets);
-      for (int i = 0; i < node->size(); i++) {
+      for (int i = 0; i < switch_node->size(); i++) {
         holes[i] = NearestPostDominatingHole(
-            node->targets()[i].block_ptr()->control_node());
+            switch_node->targets()[i].block_ptr()->control_node());
       }
-      if (node->has_fallthrough()) {
-        holes[node->size()] =
-            NearestPostDominatingHole(node->fallthrough()->control_node());
+      if (switch_node->has_fallthrough()) {
+        holes[switch_node->size()] = NearestPostDominatingHole(
+            switch_node->fallthrough()->control_node());
       }
       control->set_next_post_dominating_hole(HighestPostDominatingHole(holes));
     }
@@ -346,8 +363,7 @@ void StraightForwardRegisterAllocator::PrintLiveRegs() const {
 
 void StraightForwardRegisterAllocator::AllocateRegisters() {
   if (v8_flags.trace_maglev_regalloc) {
-    printing_visitor_.reset(new MaglevPrintingVisitor(
-        compilation_info_->graph_labeller(), std::cout));
+    printing_visitor_.reset(new MaglevPrintingVisitor(std::cout));
     printing_visitor_->PreProcessGraph(graph_);
   }
 
@@ -371,31 +387,35 @@ void StraightForwardRegisterAllocator::AllocateRegisters() {
     constant->SetConstantLocation();
     USE(value);
   }
+  for (const auto& [value, constant] : graph_->uint32()) {
+    constant->SetConstantLocation();
+    USE(value);
+  }
+  for (const auto& [value, constant] : graph_->intptr()) {
+    constant->SetConstantLocation();
+    USE(value);
+  }
   for (const auto& [value, constant] : graph_->float64()) {
     constant->SetConstantLocation();
     USE(value);
   }
-  for (const auto& [address, constant] : graph_->external_references()) {
+  for (const auto& [ref, constant] : graph_->trusted_constants()) {
     constant->SetConstantLocation();
-    USE(address);
+    USE(ref);
   }
 
   for (block_it_ = graph_->begin(); block_it_ != graph_->end(); ++block_it_) {
     BasicBlock* block = *block_it_;
+    DCHECK(!block->is_dead());
     current_node_ = nullptr;
 
     // Restore mergepoint state.
     if (block->has_state()) {
-      if (block->state()->is_exception_handler()) {
-        // Exceptions start from a blank state of register values.
-        ClearRegisterValues();
-      } else if (block->state()->is_resumable_loop() &&
-                 block->state()->predecessor_count() <= 1) {
-        // Loops that are only reachable through JumpLoop start from a blank
-        // state of register values.
-        // This should actually only support predecessor_count == 1, but we
-        // currently don't eliminate resumable loop headers (and subsequent code
-        // until the next resume) that end up being unreachable from JumpLoop.
+      if (block->state()->is_exception_handler() ||
+          block->state()->IsUnreachableByForwardEdge()) {
+        // Exceptions and loops only reachable from a JumpLoop (i.e., resumable
+        // loops with no fall-through edge) start with a blank state of register
+        // values.
         ClearRegisterValues();
       } else {
         InitializeRegisterValues(block->state()->register_state());
@@ -487,7 +507,7 @@ void StraightForwardRegisterAllocator::AllocateRegisters() {
           } else if (phi->owner().is_parameter() &&
                      phi->owner().is_receiver()) {
             // The receiver is a special case for a fairly silly reason:
-            // OptimizedFrame::Summarize requires the receiver (and the
+            // OptimizedJSFrame::Summarize requires the receiver (and the
             // function) to be in a stack slot, since its value must be
             // available even though we're not deoptimizing (and thus register
             // states are not available).
@@ -562,18 +582,19 @@ void StraightForwardRegisterAllocator::AllocateRegisters() {
       general_registers_.clear_blocked();
       double_registers_.clear_blocked();
     }
+    DCHECK(AllUsedRegistersLiveAt(block));
     VerifyRegisterState();
 
     node_it_ = block->nodes().begin();
-    for (; node_it_ != block->nodes().end();) {
+    for (; node_it_ != block->nodes().end(); ++node_it_) {
       Node* node = *node_it_;
+      if (node == nullptr) continue;
 
       if (IsDeadNodeToSkip(node)) {
         // We remove unused pure nodes.
         if (v8_flags.trace_maglev_regalloc) {
           printing_visitor_->os()
-              << "Removing unused node "
-              << PrintNodeLabel(graph_labeller(), node) << "\n";
+              << "Removing unused node " << PrintNodeLabel(node) << "\n";
         }
 
         if (!node->Is<Identity>()) {
@@ -588,14 +609,14 @@ void StraightForwardRegisterAllocator::AllocateRegisters() {
               });
         }
 
-        node_it_ = block->nodes().RemoveAt(node_it_);
+        *node_it_ = nullptr;
         continue;
       }
 
       AllocateNode(node);
-      ++node_it_;
     }
     AllocateControlNode(block->control_node(), block);
+    ApplyPatches(block);
   }
 }
 
@@ -610,8 +631,7 @@ void StraightForwardRegisterAllocator::FreeRegistersUsedBy(ValueNode* node) {
 void StraightForwardRegisterAllocator::UpdateUse(
     ValueNode* node, InputLocation* input_location) {
   if (v8_flags.trace_maglev_regalloc) {
-    printing_visitor_->os()
-        << "Using " << PrintNodeLabel(graph_labeller(), node) << "...\n";
+    printing_visitor_->os() << "Using " << PrintNodeLabel(node) << "...\n";
   }
 
   DCHECK(!node->has_no_more_uses());
@@ -622,8 +642,7 @@ void StraightForwardRegisterAllocator::UpdateUse(
   if (!node->has_no_more_uses()) return;
 
   if (v8_flags.trace_maglev_regalloc) {
-    printing_visitor_->os()
-        << "  freeing " << PrintNodeLabel(graph_labeller(), node) << "\n";
+    printing_visitor_->os() << "  freeing " << PrintNodeLabel(node) << "\n";
   }
 
   // If a value is dead, make sure it's cleared.
@@ -649,27 +668,31 @@ void StraightForwardRegisterAllocator::UpdateUse(
 
 void StraightForwardRegisterAllocator::AllocateEagerDeopt(
     const EagerDeoptInfo& deopt_info) {
-  detail::DeepForEachInput(
-      &deopt_info, [&](ValueNode* node, InputLocation* input) {
-        // We might have dropped this node without spilling it. Spill it now.
-        if (!node->has_register() && !node->is_loadable()) {
-          Spill(node);
-        }
-        input->InjectLocation(node->allocation());
-        UpdateUse(node, input);
-      });
+  InputLocation* input = deopt_info.input_locations();
+  deopt_info.ForEachInput([&](ValueNode* node) {
+    DCHECK(!node->Is<Identity>());
+    // We might have dropped this node without spilling it. Spill it now.
+    if (!node->has_register() && !node->is_loadable()) {
+      Spill(node);
+    }
+    input->InjectLocation(node->allocation());
+    UpdateUse(node, input);
+    input++;
+  });
 }
 
 void StraightForwardRegisterAllocator::AllocateLazyDeopt(
     const LazyDeoptInfo& deopt_info) {
-  detail::DeepForEachInput(&deopt_info,
-                           [&](ValueNode* node, InputLocation* input) {
-                             // Lazy deopts always need spilling, and should
-                             // always be loaded from their loadable slot.
-                             Spill(node);
-                             input->InjectLocation(node->loadable_slot());
-                             UpdateUse(node, input);
-                           });
+  InputLocation* input = deopt_info.input_locations();
+  deopt_info.ForEachInput([&](ValueNode* node) {
+    DCHECK(!node->Is<Identity>());
+    // Lazy deopts always need spilling, and should
+    // always be loaded from their loadable slot.
+    Spill(node);
+    input->InjectLocation(node->loadable_slot());
+    UpdateUse(node, input);
+    input++;
+  });
 }
 
 #ifdef DEBUG
@@ -698,8 +721,7 @@ void StraightForwardRegisterAllocator::AllocateNode(Node* node) {
   current_node_ = node;
   if (v8_flags.trace_maglev_regalloc) {
     printing_visitor_->os()
-        << "Allocating " << PrintNodeLabel(graph_labeller(), node)
-        << " inputs...\n";
+        << "Allocating " << PrintNodeLabel(node) << " inputs...\n";
   }
   AssignInputs(node);
   VerifyInputs(node);
@@ -732,8 +754,9 @@ void StraightForwardRegisterAllocator::AllocateNode(Node* node) {
     // spilled so they can properly be merged after the catch block.
     if (node->properties().can_throw()) {
       ExceptionHandlerInfo* info = node->exception_handler_info();
-      if (info->HasExceptionHandler() && !node->properties().is_call()) {
-        BasicBlock* block = info->catch_block.block_ptr();
+      if (info->HasExceptionHandler() && !info->ShouldLazyDeopt() &&
+          !node->properties().is_call()) {
+        BasicBlock* block = info->catch_block();
         auto spill = [&](auto reg, ValueNode* node) {
           if (node->live_range().end < block->first_id()) return;
           Spill(node);
@@ -745,6 +768,7 @@ void StraightForwardRegisterAllocator::AllocateNode(Node* node) {
     AllocateLazyDeopt(*node->lazy_deopt_info());
   }
 
+  // Make sure to save snapshot after allocate eager deopt registers.
   if (node->properties().needs_register_snapshot()) SaveRegisterSnapshot(node);
 
   if (v8_flags.trace_maglev_regalloc) {
@@ -883,8 +907,8 @@ void StraightForwardRegisterAllocator::DropRegisterValue(
   ValueNode* node = registers.GetValue(reg);
 
   if (v8_flags.trace_maglev_regalloc) {
-    printing_visitor_->os() << "  dropping " << reg << " value "
-                            << PrintNodeLabel(graph_labeller(), node) << "\n";
+    printing_visitor_->os()
+        << "  dropping " << reg << " value " << PrintNodeLabel(node) << "\n";
   }
 
   MachineRepresentation mach_repr = node->GetMachineRepresentation();
@@ -951,6 +975,38 @@ void StraightForwardRegisterAllocator::InitializeBranchTargetPhis(
   }
 }
 
+#ifdef DEBUG
+
+bool StraightForwardRegisterAllocator::AllUsedRegistersLiveAt(
+    ConditionalControlNode* control_node, BasicBlock* target) {
+  auto ForAllRegisters = [&](const auto& registers) {
+    for (auto reg : registers.used()) {
+      if (!IsLiveAtTarget(registers.GetValue(reg), control_node, target)) {
+        return false;
+      }
+    }
+    return true;
+  };
+  return ForAllRegisters(general_registers_) &&
+         ForAllRegisters(double_registers_);
+}
+
+bool StraightForwardRegisterAllocator::AllUsedRegistersLiveAt(
+    BasicBlock* target) {
+  auto ForAllRegisters = [&](const auto& registers) {
+    for (auto reg : registers.used()) {
+      if (registers.GetValue(reg)->live_range().end < target->first_id()) {
+        return false;
+      }
+    }
+    return true;
+  };
+  return ForAllRegisters(general_registers_) &&
+         ForAllRegisters(double_registers_);
+}
+
+#endif  // DEBUG
+
 void StraightForwardRegisterAllocator::InitializeConditionalBranchTarget(
     ConditionalControlNode* control_node, BasicBlock* target) {
   DCHECK(!target->has_phi());
@@ -963,12 +1019,8 @@ void StraightForwardRegisterAllocator::InitializeConditionalBranchTarget(
     return InitializeEmptyBlockRegisterValues(control_node, target);
   }
 
-  // Clear dead fall-through registers.
   DCHECK_EQ(control_node->id() + 1, target->first_id());
-  ClearDeadFallthroughRegisters<Register>(general_registers_, control_node,
-                                          target);
-  ClearDeadFallthroughRegisters<DoubleRegister>(double_registers_, control_node,
-                                                target);
+  DCHECK(AllUsedRegistersLiveAt(control_node, target));
 }
 
 void StraightForwardRegisterAllocator::AllocateControlNode(ControlNode* node,
@@ -1140,35 +1192,34 @@ void StraightForwardRegisterAllocator::AddMoveBeforeCurrentNode(
   if (source.IsConstant()) {
     DCHECK(IsConstantNode(node->opcode()));
     if (v8_flags.trace_maglev_regalloc) {
-      printing_visitor_->os()
-          << "  constant gap move: " << target << " ← "
-          << PrintNodeLabel(graph_labeller(), node) << std::endl;
+      printing_visitor_->os() << "  constant gap move: " << target << " ← "
+                              << PrintNodeLabel(node) << std::endl;
     }
     gap_move =
-        Node::New<ConstantGapMove>(compilation_info_->zone(), {}, node, target);
+        Node::New<ConstantGapMove>(compilation_info_->zone(), 0, node, target);
   } else {
     if (v8_flags.trace_maglev_regalloc) {
-      printing_visitor_->os() << "  gap move: " << target << " ← "
-                              << PrintNodeLabel(graph_labeller(), node) << ":"
-                              << source << std::endl;
+      printing_visitor_->os()
+          << "  gap move: " << target << " ← " << PrintNodeLabel(node) << ":"
+          << source << std::endl;
     }
     gap_move =
-        Node::New<GapMove>(compilation_info_->zone(), {},
+        Node::New<GapMove>(compilation_info_->zone(), 0,
                            compiler::AllocatedOperand::cast(source), target);
   }
+  gap_move->InitTemporaries();
   if (compilation_info_->has_graph_labeller()) {
     graph_labeller()->RegisterNode(gap_move);
   }
-  if (*node_it_ == nullptr) {
+  BasicBlock* block = *block_it_;
+  if (node_it_ == block->nodes().end()) {
     DCHECK(current_node_->Is<ControlNode>());
     // We're at the control node, so append instead.
-    (*block_it_)->nodes().Add(gap_move);
-    node_it_ = (*block_it_)->nodes().end();
+    block->nodes().push_back(gap_move);
+    node_it_ = block->nodes().end();
   } else {
-    DCHECK_NE(node_it_, (*block_it_)->nodes().end());
     // We should not add any gap move before a GetSecondReturnedValue.
-    DCHECK_NE(node_it_->opcode(), Opcode::kGetSecondReturnedValue);
-    node_it_.InsertBefore(gap_move);
+    patches_.emplace_back(node_it_ - block->nodes().begin(), gap_move);
   }
 }
 
@@ -1176,9 +1227,8 @@ void StraightForwardRegisterAllocator::Spill(ValueNode* node) {
   if (node->is_loadable()) return;
   AllocateSpillSlot(node);
   if (v8_flags.trace_maglev_regalloc) {
-    printing_visitor_->os()
-        << "  spill: " << node->spill_slot() << " ← "
-        << PrintNodeLabel(graph_labeller(), node) << std::endl;
+    printing_visitor_->os() << "  spill: " << node->spill_slot() << " ← "
+                            << PrintNodeLabel(node) << std::endl;
   }
 }
 
@@ -1192,18 +1242,16 @@ void StraightForwardRegisterAllocator::AssignFixedInput(Input& input) {
     case compiler::UnallocatedOperand::MUST_HAVE_REGISTER:
       // Allocated in AssignArbitraryRegisterInput.
       if (v8_flags.trace_maglev_regalloc) {
-        printing_visitor_->os()
-            << "- " << PrintNodeLabel(graph_labeller(), input.node())
-            << " has arbitrary register\n";
+        printing_visitor_->os() << "- " << PrintNodeLabel(input.node())
+                                << " has arbitrary register\n";
       }
       return;
 
     case compiler::UnallocatedOperand::REGISTER_OR_SLOT_OR_CONSTANT:
       // Allocated in AssignAnyInput.
       if (v8_flags.trace_maglev_regalloc) {
-        printing_visitor_->os()
-            << "- " << PrintNodeLabel(graph_labeller(), input.node())
-            << " has arbitrary location\n";
+        printing_visitor_->os() << "- " << PrintNodeLabel(input.node())
+                                << " has arbitrary location\n";
       }
       return;
 
@@ -1227,9 +1275,8 @@ void StraightForwardRegisterAllocator::AssignFixedInput(Input& input) {
       UNREACHABLE();
   }
   if (v8_flags.trace_maglev_regalloc) {
-    printing_visitor_->os()
-        << "- " << PrintNodeLabel(graph_labeller(), input.node())
-        << " in forced " << input.operand() << "\n";
+    printing_visitor_->os() << "- " << PrintNodeLabel(input.node())
+                            << " in forced " << input.operand() << "\n";
   }
 
   compiler::AllocatedOperand allocated =
@@ -1339,10 +1386,9 @@ void StraightForwardRegisterAllocator::AssignArbitraryRegisterInput(
     // Reuse an existing register if possible.
     if (existing_register_location.IsAnyLocationOperand()) {
       if (v8_flags.trace_maglev_regalloc) {
-        printing_visitor_->os()
-            << "- " << PrintNodeLabel(graph_labeller(), input.node()) << " in "
-            << (is_clobbered ? "clobbered " : "") << existing_register_location
-            << "\n";
+        printing_visitor_->os() << "- " << PrintNodeLabel(input.node())
+                                << " in " << (is_clobbered ? "clobbered " : "")
+                                << existing_register_location << "\n";
       }
       return compiler::AllocatedOperand::cast(existing_register_location);
     }
@@ -1355,7 +1401,7 @@ void StraightForwardRegisterAllocator::AssignArbitraryRegisterInput(
 
     if (v8_flags.trace_maglev_regalloc) {
       printing_visitor_->os()
-          << "- " << PrintNodeLabel(graph_labeller(), input.node()) << " in "
+          << "- " << PrintNodeLabel(input.node()) << " in "
           << (is_clobbered ? "clobbered " : "") << allocation << " ← "
           << node->allocation() << "\n";
     }
@@ -1398,9 +1444,8 @@ void StraightForwardRegisterAllocator::AssignAnyInput(Input& input) {
     }
   }
   if (v8_flags.trace_maglev_regalloc) {
-    printing_visitor_->os()
-        << "- " << PrintNodeLabel(graph_labeller(), input.node())
-        << " in original " << location << "\n";
+    printing_visitor_->os() << "- " << PrintNodeLabel(input.node())
+                            << " in original " << location << "\n";
   }
   UpdateUse(&input);
 }
@@ -1460,8 +1505,8 @@ void StraightForwardRegisterAllocator::VerifyRegisterState() {
 
   auto NodeNameForFatal = [&](ValueNode* node) {
     std::stringstream ss;
-    if (compilation_info_->has_graph_labeller()) {
-      ss << PrintNodeLabel(compilation_info_->graph_labeller(), node);
+    if (has_graph_labeller()) {
+      ss << PrintNodeLabel(node);
     } else {
       ss << "<" << node << ">";
     }
@@ -1516,6 +1561,7 @@ void StraightForwardRegisterAllocator::VerifyRegisterState() {
       }
     }
     for (Node* node : block->nodes()) {
+      if (node == nullptr) continue;
       if (ValueNode* value_node = node->TryCast<ValueNode>()) {
         if (node->Is<Identity>()) continue;
         ValidateValueNode(value_node);
@@ -1539,8 +1585,8 @@ void StraightForwardRegisterAllocator::SpillAndClearRegisters(
     RegisterT reg = registers.used().first();
     ValueNode* node = registers.GetValue(reg);
     if (v8_flags.trace_maglev_regalloc) {
-      printing_visitor_->os() << "  clearing registers with "
-                              << PrintNodeLabel(graph_labeller(), node) << "\n";
+      printing_visitor_->os()
+          << "  clearing registers with " << PrintNodeLabel(node) << "\n";
     }
     Spill(node);
     registers.FreeRegistersUsedBy(node);
@@ -1574,6 +1620,27 @@ void StraightForwardRegisterAllocator::SaveRegisterSnapshot(NodeBase* node) {
       snapshot.live_tagged_registers.clear(reg);
     }
   }
+  if (node->properties().can_eager_deopt()) {
+    // If we eagerly deopt after a deferred call, the registers saved by the
+    // runtime call might not include the inputs into the eager deopt. Here, we
+    // make sure that all the eager deopt registers are included in the
+    // snapshot.
+    InputLocation* input = node->eager_deopt_info()->input_locations();
+    node->eager_deopt_info()->ForEachInput([&](ValueNode* node) {
+      if (input->IsAnyRegister()) {
+        if (input->IsDoubleRegister()) {
+          snapshot.live_double_registers.set(input->AssignedDoubleRegister());
+        } else {
+          snapshot.live_registers.set(input->AssignedGeneralRegister());
+          if (node->is_tagged()) {
+            snapshot.live_tagged_registers.set(
+                input->AssignedGeneralRegister());
+          }
+        }
+      }
+      input++;
+    });
+  }
   node->set_register_snapshot(snapshot);
 }
 
@@ -1755,8 +1822,7 @@ compiler::AllocatedOperand StraightForwardRegisterAllocator::ForceAllocate(
   DCHECK(!registers.is_blocked(reg));
   if (v8_flags.trace_maglev_regalloc) {
     printing_visitor_->os()
-        << "  forcing " << reg << " to "
-        << PrintNodeLabel(graph_labeller(), node) << "...\n";
+        << "  forcing " << reg << " to " << PrintNodeLabel(node) << "...\n";
   }
   if (registers.free().has(reg)) {
     // If it's already free, remove it from the free list.
@@ -1770,9 +1836,7 @@ compiler::AllocatedOperand StraightForwardRegisterAllocator::ForceAllocate(
     DCHECK(!registers.is_blocked(reg));
     DropRegisterValue(registers, reg);
   }
-#ifdef DEBUG
   DCHECK(!registers.free().has(reg));
-#endif
   registers.unblock(reg);
   registers.SetValue(reg, node);
   return compiler::AllocatedOperand(compiler::LocationOperand::REGISTER,
@@ -1976,18 +2040,6 @@ void StraightForwardRegisterAllocator::AssignArbitraryTemporaries(
   AssignArbitraryTemporaries(double_registers_, node);
 }
 
-namespace {
-template <typename RegisterT>
-void ClearRegisterState(RegisterFrameState<RegisterT>& registers) {
-  while (!registers.used().is_empty()) {
-    RegisterT reg = registers.used().first();
-    ValueNode* node = registers.GetValue(reg);
-    registers.FreeRegistersUsedBy(node);
-    DCHECK(!registers.used().has(reg));
-  }
-}
-}  // namespace
-
 template <typename Function>
 void StraightForwardRegisterAllocator::ForEachMergePointRegisterState(
     MergePointRegisterState& merge_point_state, Function&& f) {
@@ -2002,6 +2054,15 @@ void StraightForwardRegisterAllocator::ForEachMergePointRegisterState(
 }
 
 void StraightForwardRegisterAllocator::ClearRegisterValues() {
+  auto ClearRegisterState = [&](auto& registers) {
+    while (!registers.used().is_empty()) {
+      auto reg = registers.used().first();
+      ValueNode* node = registers.GetValue(reg);
+      registers.FreeRegistersUsedBy(node);
+      DCHECK(!registers.used().has(reg));
+    }
+  };
+
   ClearRegisterState(general_registers_);
   ClearRegisterState(double_registers_);
 
@@ -2094,7 +2155,9 @@ bool StraightForwardRegisterAllocator::IsForwardReachable(
 template <typename RegisterT>
 void StraightForwardRegisterAllocator::HoistLoopReloads(
     BasicBlock* target, RegisterFrameState<RegisterT>& registers) {
-  for (ValueNode* node : target->reload_hints()) {
+  auto info = regalloc_info_->loop_info_.find(target->id());
+  if (info == regalloc_info_->loop_info_.end()) return;
+  for (ValueNode* node : info->second.reload_hints_) {
     DCHECK(general_registers_.blocked().is_empty());
     if (registers.free().is_empty()) break;
     if (node->has_register()) continue;
@@ -2109,12 +2172,12 @@ void StraightForwardRegisterAllocator::HoistLoopReloads(
     if (!registers.free().has(target_reg)) {
       target_reg = registers.free().first();
     }
-    compiler::AllocatedOperand target(compiler::LocationOperand::REGISTER,
-                                      node->GetMachineRepresentation(),
-                                      target_reg.code());
+    compiler::AllocatedOperand target_operand(
+        compiler::LocationOperand::REGISTER, node->GetMachineRepresentation(),
+        target_reg.code());
     registers.RemoveFromFree(target_reg);
     registers.SetValueWithoutBlocking(target_reg, node);
-    AddMoveBeforeCurrentNode(node, node->loadable_slot(), target);
+    AddMoveBeforeCurrentNode(node, node->loadable_slot(), target_operand);
   }
 }
 
@@ -2122,7 +2185,9 @@ void StraightForwardRegisterAllocator::HoistLoopReloads(
 // first call and after the last call of the loop, keep it spilled in the merge
 // state to avoid an unnecessary reload + spill on every iteration.
 void StraightForwardRegisterAllocator::HoistLoopSpills(BasicBlock* target) {
-  for (ValueNode* node : target->spill_hints()) {
+  auto info = regalloc_info_->loop_info_.find(target->id());
+  if (info == regalloc_info_->loop_info_.end()) return;
+  for (ValueNode* node : info->second.spill_hints_) {
     if (!node->has_register()) continue;
     // Do not move to a different register, the goal is to keep the value
     // spilled on the back-edge.
@@ -2217,9 +2282,9 @@ void StraightForwardRegisterAllocator::MergeRegisterValues(ControlNode* control,
       incoming = registers.GetValue(reg);
       if (!IsLiveAtTarget(incoming, control, target)) {
         if (v8_flags.trace_maglev_regalloc) {
-          printing_visitor_->os() << "  " << reg << " - incoming node "
-                                  << PrintNodeLabel(graph_labeller(), incoming)
-                                  << " dead at target\n";
+          printing_visitor_->os()
+              << "  " << reg << " - incoming node " << PrintNodeLabel(incoming)
+              << " dead at target\n";
         }
         incoming = nullptr;
       }
@@ -2231,8 +2296,9 @@ void StraightForwardRegisterAllocator::MergeRegisterValues(ControlNode* control,
       if (v8_flags.trace_maglev_regalloc) {
         if (node) {
           printing_visitor_->os()
-              << "  " << reg << " - incoming node same as node: "
-              << PrintNodeLabel(graph_labeller(), node) << "\n";
+              << "  " << reg
+              << " - incoming node same as node: " << PrintNodeLabel(node)
+              << "\n";
         }
       }
       if (merge) merge->operand(predecessor_id) = register_info;
@@ -2250,9 +2316,9 @@ void StraightForwardRegisterAllocator::MergeRegisterValues(ControlNode* control,
       // TODO(v8:7700): Overeager DCHECK.
       // DCHECK(node->properties().is_conversion());
       if (v8_flags.trace_maglev_regalloc) {
-        printing_visitor_->os() << "  " << reg << " - can't load "
-                                << PrintNodeLabel(graph_labeller(), node)
-                                << ", dropping the merge\n";
+        printing_visitor_->os()
+            << "  " << reg << " - can't load " << PrintNodeLabel(node)
+            << ", dropping the merge\n";
       }
       // We always need to be able to restore values on JumpLoop since the value
       // is definitely live at the loop header.
@@ -2266,9 +2332,9 @@ void StraightForwardRegisterAllocator::MergeRegisterValues(ControlNode* control,
       // where that node is allocated on the incoming branch.
       merge->operand(predecessor_id) = node->allocation();
       if (v8_flags.trace_maglev_regalloc) {
-        printing_visitor_->os() << "  " << reg << " - merge: loading "
-                                << PrintNodeLabel(graph_labeller(), node)
-                                << " from " << node->allocation() << " \n";
+        printing_visitor_->os()
+            << "  " << reg << " - merge: loading " << PrintNodeLabel(node)
+            << " from " << node->allocation() << " \n";
       }
 
       if (incoming != nullptr) {
@@ -2299,7 +2365,7 @@ void StraightForwardRegisterAllocator::MergeRegisterValues(ControlNode* control,
       if (v8_flags.trace_maglev_regalloc) {
         printing_visitor_->os()
             << "  " << reg << " - can't load incoming "
-            << PrintNodeLabel(graph_labeller(), incoming) << ", bailing out\n";
+            << PrintNodeLabel(incoming) << ", bailing out\n";
       }
       return;
     }
@@ -2328,16 +2394,16 @@ void StraightForwardRegisterAllocator::MergeRegisterValues(ControlNode* control,
     if (node == nullptr) {
       merge->operand(predecessor_id) = register_info;
       if (v8_flags.trace_maglev_regalloc) {
-        printing_visitor_->os() << "  " << reg << " - new merge: loading new "
-                                << PrintNodeLabel(graph_labeller(), incoming)
-                                << " from " << register_info << " \n";
+        printing_visitor_->os()
+            << "  " << reg << " - new merge: loading new "
+            << PrintNodeLabel(incoming) << " from " << register_info << " \n";
       }
     } else {
       merge->operand(predecessor_id) = node->allocation();
       if (v8_flags.trace_maglev_regalloc) {
-        printing_visitor_->os() << "  " << reg << " - new merge: loading "
-                                << PrintNodeLabel(graph_labeller(), node)
-                                << " from " << node->allocation() << " \n";
+        printing_visitor_->os()
+            << "  " << reg << " - new merge: loading " << PrintNodeLabel(node)
+            << " from " << node->allocation() << " \n";
       }
     }
     state = {merge, initialized_merge};
