@@ -13,70 +13,56 @@ namespace internal {
 
 namespace {
 
-const int kLiteralEntryLength = 2;
-const int kLiteralInitialLength = 2;
-const int kLiteralContextOffset = 0;
-const int kLiteralLiteralsOffset = 1;
-
-int SearchLiteralsMapEntry(Tagged<CompilationCacheTable> cache,
-                           InternalIndex cache_entry,
-                           Tagged<Context> native_context) {
+Tagged<JSFunction> SearchLiteralsMap(Tagged<CompilationCacheTable> cache,
+                                     InternalIndex cache_entry,
+                                     Tagged<Context> native_context) {
   DisallowGarbageCollection no_gc;
   DCHECK(IsNativeContext(native_context));
-  Tagged<Object> obj = cache->EvalFeedbackValueAt(cache_entry);
+  Tagged<Object> obj = cache->EvalJSFunctionsValueAt(cache_entry);
 
   // Check that there's no confusion between FixedArray and WeakFixedArray (the
   // object used to be a FixedArray here).
   DCHECK(!IsFixedArray(obj));
   if (IsWeakFixedArray(obj)) {
-    Tagged<WeakFixedArray> literals_map = WeakFixedArray::cast(obj);
+    Tagged<WeakFixedArray> literals_map = Cast<WeakFixedArray>(obj);
     int length = literals_map->length();
-    for (int i = 0; i < length; i += kLiteralEntryLength) {
-      DCHECK(literals_map->Get(i + kLiteralContextOffset)->IsWeakOrCleared());
-      if (literals_map->Get(i + kLiteralContextOffset) ==
-          HeapObjectReference::Weak(native_context)) {
-        return i;
+    for (int i = 0; i < length; i++) {
+      DCHECK(literals_map->get(i).IsWeakOrCleared());
+      Tagged<JSFunction> js_function;
+      if (literals_map->get(i).GetHeapObjectIfWeak(&js_function) &&
+          js_function->GetCreationContext().value() == native_context) {
+        return js_function;
       }
     }
   }
-  return -1;
+  return {};
 }
 
-void AddToFeedbackCellsMap(Handle<CompilationCacheTable> cache,
-                           InternalIndex cache_entry,
-                           Handle<Context> native_context,
-                           Handle<FeedbackCell> feedback_cell) {
-  Isolate* isolate = native_context->GetIsolate();
-  DCHECK(IsNativeContext(*native_context));
-  static_assert(kLiteralEntryLength == 2);
-  Handle<WeakFixedArray> new_literals_map;
-  int entry;
+void AddToJSFunctionMap(DirectHandle<CompilationCacheTable> cache,
+                        InternalIndex cache_entry,
+                        DirectHandle<JSFunction> js_function) {
+  Isolate* isolate = Isolate::Current();
+  DirectHandle<WeakFixedArray> new_js_functions;
+  int entry = -1;
 
-  Tagged<Object> obj = cache->EvalFeedbackValueAt(cache_entry);
+  Tagged<Object> obj = cache->EvalJSFunctionsValueAt(cache_entry);
 
   // Check that there's no confusion between FixedArray and WeakFixedArray (the
   // object used to be a FixedArray here).
   DCHECK(!IsFixedArray(obj));
-  if (!IsWeakFixedArray(obj) || WeakFixedArray::cast(obj)->length() == 0) {
-    new_literals_map = isolate->factory()->NewWeakFixedArray(
-        kLiteralInitialLength, AllocationType::kOld);
+  if (!IsWeakFixedArray(obj) || Cast<WeakFixedArray>(obj)->length() == 0) {
+    new_js_functions =
+        isolate->factory()->NewWeakFixedArray(1, AllocationType::kOld);
     entry = 0;
   } else {
-    Handle<WeakFixedArray> old_literals_map(WeakFixedArray::cast(obj), isolate);
-    entry = SearchLiteralsMapEntry(*cache, cache_entry, *native_context);
-    if (entry >= 0) {
-      // Just set the code of the entry.
-      old_literals_map->Set(entry + kLiteralLiteralsOffset,
-                            HeapObjectReference::Weak(*feedback_cell));
-      return;
-    }
-
+    DirectHandle<WeakFixedArray> old_js_functions(Cast<WeakFixedArray>(obj),
+                                                  isolate);
     // Can we reuse an entry?
     DCHECK_LT(entry, 0);
-    int length = old_literals_map->length();
-    for (int i = 0; i < length; i += kLiteralEntryLength) {
-      if (old_literals_map->Get(i + kLiteralContextOffset)->IsCleared()) {
-        new_literals_map = old_literals_map;
+    int length = old_js_functions->length();
+    for (int i = 0; i < length; i++) {
+      if (old_js_functions->get(i).IsCleared()) {
+        new_js_functions = old_js_functions;
         entry = i;
         break;
       }
@@ -84,53 +70,27 @@ void AddToFeedbackCellsMap(Handle<CompilationCacheTable> cache,
 
     if (entry < 0) {
       // Copy old optimized code map and append one new entry.
-      new_literals_map = isolate->factory()->CopyWeakFixedArrayAndGrow(
-          old_literals_map, kLiteralEntryLength);
-      entry = old_literals_map->length();
+      new_js_functions =
+          isolate->factory()->CopyWeakFixedArrayAndGrow(old_js_functions, 1);
+      entry = old_js_functions->length();
     }
   }
 
-  new_literals_map->Set(entry + kLiteralContextOffset,
-                        HeapObjectReference::Weak(*native_context));
-  new_literals_map->Set(entry + kLiteralLiteralsOffset,
-                        HeapObjectReference::Weak(*feedback_cell));
+  new_js_functions->set(entry, MakeWeak(*js_function));
 
 #ifdef DEBUG
-  for (int i = 0; i < new_literals_map->length(); i += kLiteralEntryLength) {
-    MaybeObject object = new_literals_map->Get(i + kLiteralContextOffset);
-    DCHECK(object->IsCleared() ||
-           IsNativeContext(object.GetHeapObjectAssumeWeak()));
-    object = new_literals_map->Get(i + kLiteralLiteralsOffset);
-    DCHECK(object->IsCleared() ||
-           IsFeedbackCell(object.GetHeapObjectAssumeWeak()));
+  for (int i = 0; i < new_js_functions->length(); i++) {
+    Tagged<MaybeObject> object = new_js_functions->get(i);
+    DCHECK_IMPLIES(!object.IsCleared(),
+                   IsJSFunction(object.GetHeapObjectAssumeWeak()));
   }
 #endif
-
-  Tagged<Object> old_literals_map = cache->EvalFeedbackValueAt(cache_entry);
-  if (old_literals_map != *new_literals_map) {
-    cache->SetEvalFeedbackValueAt(cache_entry, *new_literals_map);
+  Tagged<Object> old_js_functions = cache->EvalJSFunctionsValueAt(cache_entry);
+  if (old_js_functions != *new_js_functions) {
+    cache->SetEvalJSFunctionsValueAt(cache_entry, *new_js_functions);
   }
 }
 
-Tagged<FeedbackCell> SearchLiteralsMap(Tagged<CompilationCacheTable> cache,
-                                       InternalIndex cache_entry,
-                                       Tagged<Context> native_context) {
-  Tagged<FeedbackCell> result;
-  int entry = SearchLiteralsMapEntry(cache, cache_entry, native_context);
-  if (entry >= 0) {
-    Tagged<WeakFixedArray> literals_map =
-        WeakFixedArray::cast(cache->EvalFeedbackValueAt(cache_entry));
-    DCHECK_LE(entry + kLiteralEntryLength, literals_map->length());
-    MaybeObject object = literals_map->Get(entry + kLiteralLiteralsOffset);
-
-    if (!object->IsCleared()) {
-      result = FeedbackCell::cast(object.GetHeapObjectAssumeWeak());
-    }
-  }
-  DCHECK(result.is_null() || IsFeedbackCell(result));
-  return result;
-}
-
 // EvalCacheKeys are used as keys in the eval cache.
 class EvalCacheKey : public HashTableKey {
  public:
@@ -144,7 +104,8 @@ class EvalCacheKey : public HashTableKey {
   // * When positive, position is the position in the source where eval is
   //   called. When negative, position is the negation of the position in the
   //   dynamic function's effective source where the ')' ends the parameters.
-  EvalCacheKey(Handle<String> source, Handle<SharedFunctionInfo> shared,
+  EvalCacheKey(DirectHandle<String> source,
+               DirectHandle<SharedFunctionInfo> shared,
                LanguageMode language_mode, int position)
       : HashTableKey(CompilationCacheShape::EvalHash(*source, *shared,
                                                      language_mode, position)),
@@ -155,12 +116,7 @@ class EvalCacheKey : public HashTableKey {
 
   bool IsMatch(Tagged<Object> other) override {
     DisallowGarbageCollection no_gc;
-    if (!IsFixedArray(other)) {
-      DCHECK(IsNumber(other));
-      uint32_t other_hash = static_cast<uint32_t>(Object::Number(other));
-      return Hash() == other_hash;
-    }
-    Tagged<FixedArray> other_array = FixedArray::cast(other);
+    Tagged<FixedArray> other_array = Cast<FixedArray>(other);
     DCHECK(IsSharedFunctionInfo(other_array->get(0)));
     if (*shared_ != other_array->get(0)) return false;
     int language_unchecked = Smi::ToInt(other_array->get(2));
@@ -169,23 +125,23 @@ class EvalCacheKey : public HashTableKey {
     if (language_mode != language_mode_) return false;
     int position = Smi::ToInt(other_array->get(3));
     if (position != position_) return false;
-    Tagged<String> source = String::cast(other_array->get(1));
+    Tagged<String> source = Cast<String>(other_array->get(1));
     return source->Equals(*source_);
   }
 
-  Handle<Object> AsHandle(Isolate* isolate) {
-    Handle<FixedArray> array = isolate->factory()->NewFixedArray(4);
+  DirectHandle<Object> AsHandle(Isolate* isolate) {
+    DirectHandle<FixedArray> array = isolate->factory()->NewFixedArray(4);
     array->set(0, *shared_);
     array->set(1, *source_);
     array->set(2, Smi::FromEnum(language_mode_));
     array->set(3, Smi::FromInt(position_));
-    array->set_map(ReadOnlyRoots(isolate).fixed_cow_array_map());
+    array->set_map(isolate, ReadOnlyRoots(isolate).fixed_cow_array_map());
     return array;
   }
 
  private:
-  Handle<String> source_;
-  Handle<SharedFunctionInfo> shared_;
+  DirectHandle<String> source_;
+  DirectHandle<SharedFunctionInfo> shared_;
   LanguageMode language_mode_;
   int position_;
 };
@@ -193,24 +149,28 @@ class EvalCacheKey : public HashTableKey {
 // RegExpKey carries the source and flags of a regular expression as key.
 class RegExpKey : public HashTableKey {
  public:
-  RegExpKey(Handle<String> string, JSRegExp::Flags flags)
+  RegExpKey(Isolate* isolate, DirectHandle<String> string,
+            JSRegExp::Flags flags)
       : HashTableKey(
             CompilationCacheShape::RegExpHash(*string, Smi::FromInt(flags))),
+        isolate_(isolate),
         string_(string),
-        flags_(Smi::FromInt(flags)) {}
+        flags_(flags) {}
 
   // Rather than storing the key in the hash table, a pointer to the
   // stored value is stored where the key should be.  IsMatch then
   // compares the search key to the found object, rather than comparing
   // a key to a key.
+  // TODO(pthier): Loading the data via TrustedPointerTable on every key check
+  // is not great.
   bool IsMatch(Tagged<Object> obj) override {
-    Tagged<FixedArray> val = FixedArray::cast(obj);
-    return string_->Equals(String::cast(val->get(JSRegExp::kSourceIndex))) &&
-           (flags_ == val->get(JSRegExp::kFlagsIndex));
+    Tagged<RegExpData> val = Cast<RegExpDataWrapper>(obj)->data(isolate_);
+    return string_->Equals(val->source()) && (flags_ == val->flags());
   }
 
-  Handle<String> string_;
-  Tagged<Smi> flags_;
+  Isolate* isolate_;
+  DirectHandle<String> string_;
+  JSRegExp::Flags flags_;
 };
 
 // CodeKey carries the SharedFunctionInfo key associated with a
@@ -225,16 +185,17 @@ class CodeKey : public HashTableKey {
   Handle<SharedFunctionInfo> key_;
 };
 
-Tagged<Smi> ScriptHash(Tagged<String> source, MaybeHandle<Object> maybe_name,
-                       int line_offset, int column_offset,
+Tagged<Smi> ScriptHash(Tagged<String> source,
+                       MaybeDirectHandle<Object> maybe_name, int line_offset,
+                       int column_offset,
                        v8::ScriptOriginOptions origin_options,
                        Isolate* isolate) {
   DisallowGarbageCollection no_gc;
   size_t hash = base::hash_combine(source->EnsureHash());
-  if (Handle<Object> name;
+  if (DirectHandle<Object> name;
       maybe_name.ToHandle(&name) && IsString(*name, isolate)) {
     hash =
-        base::hash_combine(hash, String::cast(*name)->EnsureHash(), line_offset,
+        base::hash_combine(hash, Cast<String>(*name)->EnsureHash(), line_offset,
                            column_offset, origin_options.Flags());
   }
   // The upper bits of the hash are discarded so that the value fits in a Smi.
@@ -243,10 +204,10 @@ Tagged<Smi> ScriptHash(Tagged<String> source, MaybeHandle<Object> maybe_name,
 
 }  // namespace
 
-// We only re-use a cached function for some script source code if the
+// We only reuse a cached function for some script source code if the
 // script originates from the same place. This is to avoid issues
 // when reporting errors, etc.
-bool ScriptCacheKey::MatchesOrigin(Tagged<Script> script) {
+bool ScriptCacheKey::MatchesScript(Tagged<Script> script) {
   DisallowGarbageCollection no_gc;
 
   // If the script name isn't set, the boilerplate script should have
@@ -266,19 +227,50 @@ bool ScriptCacheKey::MatchesOrigin(Tagged<Script> script) {
     return false;
   }
   // Compare the two name strings for equality.
-  if (!String::cast(*name)->Equals(String::cast(script->name()))) {
+  if (!Cast<String>(*name)->Equals(Cast<String>(script->name()))) {
     return false;
   }
 
+  Handle<FixedArray> wrapped_arguments_handle;
+  if (wrapped_arguments_.ToHandle(&wrapped_arguments_handle)) {
+    if (!script->is_wrapped()) {
+      return false;
+    }
+    Tagged<FixedArray> wrapped_arguments = *wrapped_arguments_handle;
+    Tagged<FixedArray> other_wrapped_arguments = script->wrapped_arguments();
+    int length = wrapped_arguments->length();
+    if (length != other_wrapped_arguments->length()) {
+      return false;
+    }
+    for (int i = 0; i < length; i++) {
+      Tagged<Object> arg = wrapped_arguments->get(i);
+      Tagged<Object> other_arg = other_wrapped_arguments->get(i);
+      DCHECK(IsString(arg));
+      DCHECK(IsString(other_arg));
+      if (!Cast<String>(arg)->Equals(Cast<String>(other_arg))) {
+        return false;
+      }
+    }
+  } else if (script->is_wrapped()) {
+    return false;
+  }
+
+  // Don't compare host options if the script was deserialized because we didn't
+  // serialize host options (see CodeSerializer::SerializeObjectImpl())
+  if (script->deserialized() &&
+      script->host_defined_options() ==
+          ReadOnlyRoots(isolate_).empty_fixed_array()) {
+    return true;
+  }
   // TODO(cbruni, chromium:1244145): Remove once migrated to the context
   Handle<Object> maybe_host_defined_options;
   if (!host_defined_options_.ToHandle(&maybe_host_defined_options)) {
     maybe_host_defined_options = isolate_->factory()->empty_fixed_array();
   }
   Tagged<FixedArray> host_defined_options =
-      FixedArray::cast(*maybe_host_defined_options);
+      Cast<FixedArray>(*maybe_host_defined_options);
   Tagged<FixedArray> script_options =
-      FixedArray::cast(script->host_defined_options());
+      Cast<FixedArray>(script->host_defined_options());
   int length = host_defined_options->length();
   if (length != script_options->length()) return false;
 
@@ -300,12 +292,14 @@ ScriptCacheKey::ScriptCacheKey(Handle<String> source,
     : ScriptCacheKey(source, script_details->name_obj,
                      script_details->line_offset, script_details->column_offset,
                      script_details->origin_options,
-                     script_details->host_defined_options, isolate) {}
+                     script_details->host_defined_options,
+                     script_details->wrapped_arguments, isolate) {}
 
 ScriptCacheKey::ScriptCacheKey(Handle<String> source, MaybeHandle<Object> name,
                                int line_offset, int column_offset,
                                v8::ScriptOriginOptions origin_options,
                                MaybeHandle<Object> host_defined_options,
+                               MaybeHandle<FixedArray> maybe_wrapped_arguments,
                                Isolate* isolate)
     : HashTableKey(static_cast<uint32_t>(ScriptHash(*source, name, line_offset,
                                                     column_offset,
@@ -317,42 +311,53 @@ ScriptCacheKey::ScriptCacheKey(Handle<String> source, MaybeHandle<Object> name,
       column_offset_(column_offset),
       origin_options_(origin_options),
       host_defined_options_(host_defined_options),
+      wrapped_arguments_(maybe_wrapped_arguments),
       isolate_(isolate) {
   DCHECK(Smi::IsValid(static_cast<int>(Hash())));
+#ifdef DEBUG
+  DirectHandle<FixedArray> wrapped_arguments;
+  if (maybe_wrapped_arguments.ToHandle(&wrapped_arguments)) {
+    int length = wrapped_arguments->length();
+    for (int i = 0; i < length; i++) {
+      Tagged<Object> arg = wrapped_arguments->get(i);
+      DCHECK(IsString(arg));
+    }
+  }
+#endif
 }
 
 bool ScriptCacheKey::IsMatch(Tagged<Object> other) {
   DisallowGarbageCollection no_gc;
   DCHECK(IsWeakFixedArray(other));
-  Tagged<WeakFixedArray> other_array = WeakFixedArray::cast(other);
+  Tagged<WeakFixedArray> other_array = Cast<WeakFixedArray>(other);
   DCHECK_EQ(other_array->length(), kEnd);
 
   // A hash check can quickly reject many non-matches, even though this step
   // isn't strictly necessary.
   uint32_t other_hash =
-      static_cast<uint32_t>(other_array->Get(kHash).ToSmi().value());
+      static_cast<uint32_t>(other_array->get(kHash).ToSmi().value());
   if (other_hash != Hash()) return false;
 
   Tagged<HeapObject> other_script_object;
-  if (!other_array->Get(kWeakScript)
+  if (!other_array->get(kWeakScript)
            .GetHeapObjectIfWeak(&other_script_object)) {
     return false;
   }
-  Tagged<Script> other_script = Script::cast(other_script_object);
-  Tagged<String> other_source = String::cast(other_script->source());
-  return other_source->Equals(*source_) && MatchesOrigin(other_script);
+  Tagged<Script> other_script = Cast<Script>(other_script_object);
+  Tagged<String> other_source = Cast<String>(other_script->source());
+
+  return other_source->Equals(*source_) && MatchesScript(other_script);
 }
 
-Handle<Object> ScriptCacheKey::AsHandle(Isolate* isolate,
-                                        Handle<SharedFunctionInfo> shared) {
-  Handle<WeakFixedArray> array = isolate->factory()->NewWeakFixedArray(kEnd);
+DirectHandle<Object> ScriptCacheKey::AsHandle(
+    Isolate* isolate, DirectHandle<SharedFunctionInfo> shared) {
+  DirectHandle<WeakFixedArray> array =
+      isolate->factory()->NewWeakFixedArray(kEnd);
   // Any SharedFunctionInfo being stored in the script cache should have a
   // Script.
   DCHECK(IsScript(shared->script()));
-  array->Set(kHash,
-             MaybeObject::FromObject(Smi::FromInt(static_cast<int>(Hash()))));
-  array->Set(kWeakScript,
-             MaybeObject::MakeWeak(MaybeObject::FromObject(shared->script())));
+  array->set(kHash, Smi::FromInt(static_cast<int>(Hash())));
+  array->set(kWeakScript, MakeWeak(shared->script()));
   return array;
 }
 
@@ -386,7 +391,7 @@ CompilationCacheScriptLookupResult::FromRawObjects(
 }
 
 CompilationCacheScriptLookupResult CompilationCacheTable::LookupScript(
-    Handle<CompilationCacheTable> table, Handle<String> src,
+    DirectHandle<CompilationCacheTable> table, Handle<String> src,
     const ScriptDetails& script_details, Isolate* isolate) {
   src = String::Flatten(isolate, src);
   ScriptCacheKey key(src, &script_details, isolate);
@@ -395,14 +400,14 @@ CompilationCacheScriptLookupResult CompilationCacheTable::LookupScript(
 
   DisallowGarbageCollection no_gc;
   Tagged<Object> key_in_table = table->KeyAt(entry);
-  Tagged<Script> script = Script::cast(WeakFixedArray::cast(key_in_table)
-                                           ->Get(ScriptCacheKey::kWeakScript)
+  Tagged<Script> script = Cast<Script>(Cast<WeakFixedArray>(key_in_table)
+                                           ->get(ScriptCacheKey::kWeakScript)
                                            .GetHeapObjectAssumeWeak());
 
   Tagged<Object> obj = table->PrimaryValueAt(entry);
   Tagged<SharedFunctionInfo> toplevel_sfi;
   if (!IsUndefined(obj, isolate)) {
-    toplevel_sfi = SharedFunctionInfo::cast(obj);
+    toplevel_sfi = Cast<SharedFunctionInfo>(obj);
     DCHECK_EQ(toplevel_sfi->script(), script);
   }
 
@@ -411,11 +416,12 @@ CompilationCacheScriptLookupResult CompilationCacheTable::LookupScript(
 }
 
 InfoCellPair CompilationCacheTable::LookupEval(
-    Handle<CompilationCacheTable> table, Handle<String> src,
-    Handle<SharedFunctionInfo> outer_info, Handle<Context> native_context,
-    LanguageMode language_mode, int position) {
+    DirectHandle<CompilationCacheTable> table, DirectHandle<String> src,
+    DirectHandle<SharedFunctionInfo> outer_info,
+    DirectHandle<NativeContext> native_context, LanguageMode language_mode,
+    int position) {
   InfoCellPair empty_result;
-  Isolate* isolate = native_context->GetIsolate();
+  Isolate* isolate = Isolate::Current();
   src = String::Flatten(isolate, src);
 
   EvalCacheKey key(src, outer_info, language_mode, position);
@@ -427,19 +433,19 @@ InfoCellPair CompilationCacheTable::LookupEval(
   if (!IsSharedFunctionInfo(obj)) return empty_result;
 
   static_assert(CompilationCacheShape::kEntrySize == 3);
-  Tagged<FeedbackCell> feedback_cell =
+  Tagged<JSFunction> js_function =
       SearchLiteralsMap(*table, entry, *native_context);
-  return InfoCellPair(isolate, SharedFunctionInfo::cast(obj), feedback_cell);
+  return InfoCellPair(isolate, Cast<SharedFunctionInfo>(obj), js_function);
 }
 
-Handle<Object> CompilationCacheTable::LookupRegExp(Handle<String> src,
-                                                   JSRegExp::Flags flags) {
-  Isolate* isolate = GetIsolate();
+DirectHandle<Object> CompilationCacheTable::LookupRegExp(
+    DirectHandle<String> src, JSRegExp::Flags flags) {
+  Isolate* isolate = Isolate::Current();
   DisallowGarbageCollection no_gc;
-  RegExpKey key(src, flags);
+  RegExpKey key(isolate, src, flags);
   InternalIndex entry = FindEntry(isolate, &key);
   if (entry.is_not_found()) return isolate->factory()->undefined_value();
-  return Handle<Object>(PrimaryValueAt(entry), isolate);
+  return DirectHandle<Object>(PrimaryValueAt(entry), isolate);
 }
 
 Handle<CompilationCacheTable> CompilationCacheTable::EnsureScriptTableCapacity(
@@ -453,8 +459,8 @@ Handle<CompilationCacheTable> CompilationCacheTable::EnsureScriptTableCapacity(
     for (InternalIndex entry : cache->IterateEntries()) {
       Tagged<Object> key;
       if (!cache->ToKey(isolate, entry, &key)) continue;
-      if (WeakFixedArray::cast(key)
-              ->Get(ScriptCacheKey::kWeakScript)
+      if (Cast<WeakFixedArray>(key)
+              ->get(ScriptCacheKey::kWeakScript)
               .IsCleared()) {
         DCHECK(IsUndefined(cache->PrimaryValueAt(entry)));
         cache->RemoveEntry(entry);
@@ -465,11 +471,12 @@ Handle<CompilationCacheTable> CompilationCacheTable::EnsureScriptTableCapacity(
   return EnsureCapacity(isolate, cache);
 }
 
-Handle<CompilationCacheTable> CompilationCacheTable::PutScript(
+DirectHandle<CompilationCacheTable> CompilationCacheTable::PutScript(
     Handle<CompilationCacheTable> cache, Handle<String> src,
-    Handle<SharedFunctionInfo> value, Isolate* isolate) {
+    MaybeHandle<FixedArray> maybe_wrapped_arguments,
+    DirectHandle<SharedFunctionInfo> value, Isolate* isolate) {
   src = String::Flatten(isolate, src);
-  Handle<Script> script = handle(Script::cast(value->script()), isolate);
+  DirectHandle<Script> script(Cast<Script>(value->script()), isolate);
   MaybeHandle<Object> script_name;
   if (IsString(script->name(), isolate)) {
     script_name = handle(script->name(), isolate);
@@ -478,8 +485,8 @@ Handle<CompilationCacheTable> CompilationCacheTable::PutScript(
                                           isolate);
   ScriptCacheKey key(src, script_name, script->line_offset(),
                      script->column_offset(), script->origin_options(),
-                     host_defined_options, isolate);
-  Handle<Object> k = key.AsHandle(isolate, value);
+                     host_defined_options, maybe_wrapped_arguments, isolate);
+  DirectHandle<Object> k = key.AsHandle(isolate, value);
 
   // Check whether there is already a matching entry. If so, we must overwrite
   // it. This allows an entry whose value is undefined to upgrade to contain a
@@ -496,7 +503,7 @@ Handle<CompilationCacheTable> CompilationCacheTable::PutScript(
   // fixing. Consider the following unlikely sequence of events:
   // 1. BackgroundMergeTask::SetUpOnMainThread finds a script S1 in the cache.
   // 2. DevTools is attached and clears the cache.
-  // 3. DevTools is detached; the cache is reenabled.
+  // 3. DevTools is detached; the cache is re-enabled.
   // 4. A new instance of the script, S2, is compiled and placed into the cache.
   // 5. The merge from step 1 finishes on the main thread, still using S1, and
   //    places S1 into the cache, replacing S2.
@@ -508,53 +515,60 @@ Handle<CompilationCacheTable> CompilationCacheTable::PutScript(
   return cache;
 }
 
-Handle<CompilationCacheTable> CompilationCacheTable::PutEval(
-    Handle<CompilationCacheTable> cache, Handle<String> src,
-    Handle<SharedFunctionInfo> outer_info, Handle<SharedFunctionInfo> value,
-    Handle<Context> native_context, Handle<FeedbackCell> feedback_cell,
+void CompilationCacheTable::UpdateEval(
+    DirectHandle<CompilationCacheTable> table, DirectHandle<String> src,
+    DirectHandle<SharedFunctionInfo> outer_info,
+    DirectHandle<JSFunction> js_function, LanguageMode language_mode,
     int position) {
-  Isolate* isolate = native_context->GetIsolate();
+  Isolate* isolate = Isolate::Current();
   src = String::Flatten(isolate, src);
-  EvalCacheKey key(src, outer_info, value->language_mode(), position);
 
-  // This block handles 'real' insertions, i.e. the initial dummy insert
-  // (below) has already happened earlier.
-  {
-    Handle<Object> k = key.AsHandle(isolate);
-    InternalIndex entry = cache->FindEntry(isolate, &key);
-    if (entry.is_found()) {
-      cache->SetKeyAt(entry, *k);
-      cache->SetPrimaryValueAt(entry, *value);
-      // AddToFeedbackCellsMap may allocate a new sub-array to live in the
-      // entry, but it won't change the cache array. Therefore EntryToIndex
-      // and entry remains correct.
-      AddToFeedbackCellsMap(cache, entry, native_context, feedback_cell);
-      // Add hash again even on cache hit to avoid unnecessary cache delay in
-      // case of hash collisions.
-    }
-  }
+  EvalCacheKey key(src, outer_info, language_mode, position);
+  InternalIndex entry = table->FindEntry(isolate, &key);
+  if (entry.is_not_found()) return;
+
+  if (!IsFixedArray(table->KeyAt(entry))) return;
+  Tagged<Object> obj = table->PrimaryValueAt(entry);
+  if (!IsSharedFunctionInfo(obj)) return;
+  AddToJSFunctionMap(table, entry, js_function);
+}
+
+DirectHandle<CompilationCacheTable> CompilationCacheTable::PutEval(
+    DirectHandle<CompilationCacheTable> cache, DirectHandle<String> src,
+    DirectHandle<SharedFunctionInfo> outer_info,
+    DirectHandle<JSFunction> js_function, int position) {
+  Isolate* isolate = Isolate::Current();
+  src = String::Flatten(isolate, src);
+  EvalCacheKey key(src, outer_info, js_function->shared()->language_mode(),
+                   position);
 
   // Create a dummy entry to mark that this key has already been inserted once.
   cache = EnsureCapacity(isolate, cache);
   InternalIndex entry = cache->FindInsertionEntry(isolate, key.Hash());
-  Handle<Object> k =
-      isolate->factory()->NewNumber(static_cast<double>(key.Hash()));
+  DirectHandle<Object> k = key.AsHandle(isolate);
   cache->SetKeyAt(entry, *k);
-  cache->SetPrimaryValueAt(entry, Smi::FromInt(kHashGenerations));
+  cache->SetPrimaryValueAt(entry, js_function->shared());
+  cache->SetEvalJSFunctionsValueAt(entry,
+                                   ReadOnlyRoots(isolate).the_hole_value());
+  // AddToFeedbackCellsMap may allocate a new sub-array to live in the
+  // entry, but it won't change the cache array. Therefore EntryToIndex
+  // and entry remains correct.
+  AddToJSFunctionMap(cache, entry, js_function);
   cache->ElementAdded();
   return cache;
 }
 
-Handle<CompilationCacheTable> CompilationCacheTable::PutRegExp(
-    Isolate* isolate, Handle<CompilationCacheTable> cache, Handle<String> src,
-    JSRegExp::Flags flags, Handle<FixedArray> value) {
-  RegExpKey key(src, flags);
+DirectHandle<CompilationCacheTable> CompilationCacheTable::PutRegExp(
+    Isolate* isolate, DirectHandle<CompilationCacheTable> cache,
+    DirectHandle<String> src, JSRegExp::Flags flags,
+    DirectHandle<RegExpData> value) {
+  RegExpKey key(isolate, src, flags);
   cache = EnsureCapacity(isolate, cache);
   InternalIndex entry = cache->FindInsertionEntry(isolate, key.Hash());
   // We store the value in the key slot, and compare the search key
   // to the stored value with a custom IsMatch function during lookups.
-  cache->SetKeyAt(entry, *value);
-  cache->SetPrimaryValueAt(entry, *value);
+  cache->SetKeyAt(entry, value->wrapper());
+  cache->SetPrimaryValueAt(entry, value->wrapper());
   cache->ElementAdded();
   return cache;
 }
@@ -572,7 +586,7 @@ void CompilationCacheTable::RemoveEntry(InternalIndex entry) {
   int entry_index = EntryToIndex(entry);
   Tagged<Object> the_hole_value = GetReadOnlyRoots().the_hole_value();
   for (int i = 0; i < kEntrySize; i++) {
-    NoWriteBarrierSet(*this, entry_index + i, the_hole_value);
+    this->set(entry_index + i, the_hole_value, SKIP_WRITE_BARRIER);
   }
   ElementRemoved();
 
