@@ -8,9 +8,11 @@
 #include "include/v8config.h"
 #include "src/base/atomicops.h"
 #include "src/base/memory.h"
-#include "src/base/platform/mutex.h"
 #include "src/common/globals.h"
-#include "src/sandbox/external-entity-table.h"
+#include "src/sandbox/check.h"
+#include "src/sandbox/compactible-external-entity-table.h"
+#include "src/sandbox/tagged-payload.h"
+#include "src/utils/allocation.h"
 
 #ifdef V8_COMPRESS_POINTERS
 
@@ -36,22 +38,25 @@ class ReadOnlyArtifacts;
  *    algorithm overview for more details about these entries.
  */
 struct ExternalPointerTableEntry {
+  enum class EvacuateMarkMode { kTransferMark, kLeaveUnmarked, kClearMark };
+
   // Make this entry an external pointer entry containing the given pointer
   // tagged with the given tag.
-  inline void MakeExternalPointerEntry(Address value, ExternalPointerTag tag);
+  inline void MakeExternalPointerEntry(Address value, ExternalPointerTag tag,
+                                       bool mark_as_alive);
 
   // Load and untag the external pointer stored in this entry.
   // This entry must be an external pointer entry.
   // If the specified tag doesn't match the actual tag of this entry, the
   // resulting pointer will be invalid and cannot be dereferenced.
-  inline Address GetExternalPointer(ExternalPointerTag tag) const;
+  inline Address GetExternalPointer(ExternalPointerTagRange tag_range) const;
 
   // Tag and store the given external pointer in this entry.
   // This entry must be an external pointer entry.
   inline void SetExternalPointer(Address value, ExternalPointerTag tag);
 
   // Returns true if this entry contains an external pointer with the given tag.
-  inline bool HasExternalPointer(ExternalPointerTag tag) const;
+  inline bool HasExternalPointer(ExternalPointerTagRange tag_range) const;
 
   // Exchanges the external pointer stored in this entry with the provided one.
   // Returns the old external pointer. This entry must be an external pointer
@@ -59,6 +64,18 @@ struct ExternalPointerTableEntry {
   // returned pointer will be invalid.
   inline Address ExchangeExternalPointer(Address value, ExternalPointerTag tag);
 
+  // Load the tag of the external pointer stored in this entry.
+  // This entry must be an external pointer entry.
+  inline ExternalPointerTag GetExternalPointerTag() const;
+
+  // Returns the address of the managed resource contained in this entry or
+  // nullptr if this entry does not reference a managed resource.
+  inline Address ExtractManagedResourceOrNull() const;
+
+  // Invalidate the entry. Any access to a zapped entry will result in an
+  // invalid pointer that will crash upon dereference.
+  inline void MakeZappedEntry();
+
   // Make this entry a freelist entry, containing the index of the next entry
   // on the freelist.
   inline void MakeFreelistEntry(uint32_t next_entry_index);
@@ -73,36 +90,63 @@ struct ExternalPointerTableEntry {
   // the entry being evacuated.
   inline void MakeEvacuationEntry(Address handle_location);
 
-  // Move the content of this entry into the provided entry while also clearing
-  // the marking bit. Used during table compaction. This invalidates the entry.
-  inline void UnmarkAndMigrateInto(ExternalPointerTableEntry& other);
+  // Returns true if this entry contains an evacuation entry.
+  inline bool HasEvacuationEntry() const;
+
+  // Move the content of this entry into the provided entry, possibly clearing
+  // the marking bit. Used during table compaction and during promotion.
+  // Invalidates the source entry.
+  inline void Evacuate(ExternalPointerTableEntry& dest, EvacuateMarkMode mode);
 
   // Mark this entry as alive during table garbage collection.
   inline void Mark();
 
+  static constexpr bool IsWriteProtected = false;
+
  private:
   friend class ExternalPointerTable;
 
-  // ExternalPointerTable entries consist of a single pointer-sized word
-  // containing a tag and marking bit together with the actual content (e.g. an
-  // external pointer).
+  // TODO(saelo): generalize this payload struct and reuse it for all other
+  // pointer table that use type tags. For that, we probably will have to make
+  // this more flexible, allowing shifts and masks to be applied to both the
+  // tag- and payload bits since the tables store the tag bits differently.
   struct Payload {
     Payload(Address pointer, ExternalPointerTag tag)
         : encoded_word_(Tag(pointer, tag)) {}
 
-    Address Untag(ExternalPointerTag tag) const { return encoded_word_ & ~tag; }
-
     static Address Tag(Address pointer, ExternalPointerTag tag) {
-      return pointer | tag;
+      DCHECK_LE(tag, kLastExternalPointerTag);
+      return pointer | (static_cast<Address>(tag) << kExternalPointerTagShift);
+    }
+
+    static bool CheckTag(Address content, ExternalPointerTagRange tag_range) {
+      // TODO(saelo): use well-known null entries per type tag instead of a
+      // generic null entry. Then this check can be removed.
+      if (ExternalPointerCanBeEmpty(tag_range) && !content) {
+        return true;
+      }
+
+      ExternalPointerTag tag = static_cast<ExternalPointerTag>(
+          (content & kExternalPointerTagMask) >> kExternalPointerTagShift);
+      return tag_range.Contains(tag);
+    }
+
+    Address Untag(ExternalPointerTagRange tag_range) const {
+      Address content = encoded_word_;
+      SBXCHECK(CheckTag(content, tag_range));
+      return content & kExternalPointerPayloadMask;
+    }
+
+    Address Untag(ExternalPointerTag tag) const {
+      return Untag(ExternalPointerTagRange(tag, tag));
+    }
+
+    bool IsTaggedWithTagIn(ExternalPointerTagRange tag_range) const {
+      return CheckTag(encoded_word_, tag_range);
     }
 
     bool IsTaggedWith(ExternalPointerTag tag) const {
-      // We have to explicitly ignore the marking bit (which is part of the
-      // tag) since an unmarked entry with tag kXyzTag is still considered to
-      // be tagged with kXyzTag.
-      uint64_t expected = tag & ~kExternalPointerMarkBit;
-      uint64_t actual = encoded_word_ & kExternalPointerTagMaskWithoutMarkBit;
-      return expected == actual;
+      return IsTaggedWithTagIn(ExternalPointerTagRange(tag));
     }
 
     void SetMarkBit() { encoded_word_ |= kExternalPointerMarkBit; }
@@ -110,17 +154,23 @@ struct ExternalPointerTableEntry {
     void ClearMarkBit() { encoded_word_ &= ~kExternalPointerMarkBit; }
 
     bool HasMarkBitSet() const {
-      return (encoded_word_ & kExternalPointerMarkBit) != 0;
-    }
-
-    bool ContainsFreelistLink() const {
-      return IsTaggedWith(kExternalPointerFreeEntryTag);
+      return encoded_word_ & kExternalPointerMarkBit;
     }
 
     uint32_t ExtractFreelistLink() const {
       return static_cast<uint32_t>(encoded_word_);
     }
 
+    ExternalPointerTag ExtractTag() const {
+      return static_cast<ExternalPointerTag>(
+          (encoded_word_ & kExternalPointerTagMask) >>
+          kExternalPointerTagShift);
+    }
+
+    bool ContainsFreelistLink() const {
+      return IsTaggedWith(kExternalPointerFreeEntryTag);
+    }
+
     bool ContainsEvacuationEntry() const {
       return IsTaggedWith(kExternalPointerEvacuationEntryTag);
     }
@@ -129,13 +179,14 @@ struct ExternalPointerTableEntry {
       return Untag(kExternalPointerEvacuationEntryTag);
     }
 
-    bool ContainsExternalPointer() const {
+    bool ContainsPointer() const {
       return !ContainsFreelistLink() && !ContainsEvacuationEntry();
     }
 
     bool operator==(Payload other) const {
       return encoded_word_ == other.encoded_word_;
     }
+
     bool operator!=(Payload other) const {
       return encoded_word_ != other.encoded_word_;
     }
@@ -157,6 +208,9 @@ struct ExternalPointerTableEntry {
 #endif  // LEAK_SANITIZER
   }
 
+  // ExternalPointerTable entries consist of a single pointer-sized word
+  // containing a tag and marking bit together with the actual content (e.g. an
+  // external pointer).
   std::atomic<Payload> payload_;
 
 #if defined(LEAK_SANITIZER)
@@ -217,11 +271,7 @@ static_assert(sizeof(ExternalPointerTableEntry) == 8);
  *
  * Table memory management:
  * ------------------------
- * For the purpose of memory management, the table is partitioned into Segments
- * (for example 64kb memory chunks) that are grouped together in "Spaces". All
- * segments in a space share a freelist, and so entry allocation and garbage
- * collection happen on the level of spaces. The garbage collection algorithm
- * then works as follows:
+ * The garbage collection algorithm works as follows:
  *  - One bit of every entry is reserved for the marking bit.
  *  - Every store to an entry automatically sets the marking bit when ORing
  *    with the tag. This avoids the need for write barriers.
@@ -232,127 +282,76 @@ static_assert(sizeof(ExternalPointerTableEntry) == 8);
  *    marking bit using an atomic CAS operation.
  *  - When marking is finished, SweepAndCompact() iterates over a Space once
  *    while the mutator is stopped and builds a freelist from all dead entries
- *    while also removing the marking bit from any live entry.
+ *    while also possibly clearing the marking bit from any live entry.
+ *
+ * Generational collection for tables:
+ * -----------------------------------
+ * Young-generation objects with external pointer slots allocate their
+ * ExternalPointerTable entries in a spatially partitioned young external
+ * pointer space.  There are two different mechanisms:
+ *  - When using the semi-space nursery, promoting an object evacuates its EPT
+ *    entries to the old external pointer space.
+ *  - For the in-place MinorMS nursery, possibly-concurrent marking populates
+ *    the SURVIVOR_TO_EXTERNAL_POINTER remembered sets.  In the pause, promoted
+ *    objects use this remembered set to evacuate their EPT entries to the old
+ *    external pointer space.  Survivors have their EPT entries are left in
+ *    place.
+ * In a full collection, segments from the young EPT space are eagerly promoted
+ * during the pause, leaving the young generation empty.
  *
  * Table compaction:
  * -----------------
- * The table's spaces are to some degree self-compacting: since the freelists
- * are sorted in ascending order (see SweepAndCompact()), segments at the start
- * of the table will usually be fairly well utilized, while later segments
- * might become completely free, in which case they will be deallocated.
- * However, as a single live entry may keep an entire segment alive, the
- * following simple algorithm is used to compact a space if that is deemed
- * necessary:
- *  - At the start of the GC marking phase, determine if a space needs to be
- *    compacted. This decisiont is mostly based on the absolute and relative
- *    size of the freelist.
- *  - If compaction is needed, this algorithm determines by how many segments
- *    it would like to shrink the space (N). It will then attempts to move all
- *    live entries out of these segments so that they can be deallocated
- *    afterwards during sweeping.
- *  - The algorithm then simply selects the last N segments for evacuation, and
- *    it "marks" them for evacuation simply by remembering the start of the
- *    first selected segment. Everything after this threshold value then
- *    becomes the evacuation area. In this way, it becomes very cheap to test
- *    if an entry or segment should be evacuated: only a single integer
- *    comparison against the threshold is required. It also establishes a
- *    simple compaction invariant that can be verified with a few DCHECKs:
- *    compaction always moves an entry at or above the threshold to a new
- *    position before the threshold.
- *  - During marking, whenever a live entry inside the evacuation area is
- *    found, a new "evacuation entry" is allocated from the freelist (which is
- *    assumed to have enough free slots) and the address of the handle in the
- *    object owning the table entry is written into it.
- *  - During sweeping, these evacuation entries are resolved: the content of
- *    the old entry is copied into the new entry and the handle in the object
- *    is updated to point to the new entry.
- *
- * When compacting, it is expected that the evacuation area contains few live
- * entries and that the freelist will be able to serve all evacuation entry
- * allocations. In that case, compaction is essentially free (very little
- * marking overhead, no memory overhead). However, it can happen that the
- * application allocates a large number of table entries during marking, in
- * which case we might end up allocating new entries inside the evacuation area
- * or even allocate entire new segments for the space that's being compacted.
- * If that situation is detected, compaction is aborted during marking.
- *
- * This algorithm assumes that table entries (except for the null entry) are
- * never shared between multiple objects. Otherwise, the following could
- * happen: object A initially has handle H1 and is scanned during incremental
- * marking. Next, object B with handle H2 is scanned and marked for
- * evacuation. Afterwards, object A copies the handle H2 from object B.
- * During sweeping, only object B's handle will be updated to point to the
- * new entry while object A's handle is now dangling. If shared entries ever
- * become necessary, setting external pointer handles would have to be
- * guarded by write barriers to avoid this scenario.
+ * Additionally, the external pointer table supports compaction.
+ * For details about the compaction algorithm see the
+ * CompactibleExternalEntityTable class.
  */
 class V8_EXPORT_PRIVATE ExternalPointerTable
-    : public ExternalEntityTable<ExternalPointerTableEntry,
-                                 kExternalPointerTableReservationSize> {
+    : public CompactibleExternalEntityTable<
+          ExternalPointerTableEntry, kExternalPointerTableReservationSize> {
+  using Base =
+      CompactibleExternalEntityTable<ExternalPointerTableEntry,
+                                     kExternalPointerTableReservationSize>;
+
 #if defined(LEAK_SANITIZER)
   //  When LSan is active, we use "fat" entries, see above.
   static_assert(kMaxExternalPointers == kMaxCapacity * 2);
 #else
   static_assert(kMaxExternalPointers == kMaxCapacity);
 #endif
+  static_assert(kSupportsCompaction);
 
  public:
-  // Size of an ExternalPointerTable, for layout computation in IsolateData.
-  static int constexpr kSize = 2 * kSystemPointerSize;
+  using EvacuateMarkMode = ExternalPointerTableEntry::EvacuateMarkMode;
 
   ExternalPointerTable() = default;
   ExternalPointerTable(const ExternalPointerTable&) = delete;
   ExternalPointerTable& operator=(const ExternalPointerTable&) = delete;
 
-  // The Spaces used by an ExternalPointerTable also contain the state related
-  // to compaction.
-  struct Space : public ExternalEntityTable<
-                     ExternalPointerTableEntry,
-                     kExternalPointerTableReservationSize>::Space {
+  // The Spaces used by an ExternalPointerTable.
+  struct Space : public Base::Space {
    public:
-    Space() : start_of_evacuation_area_(kNotCompactingMarker) {}
+    // During table compaction, we may record the addresses of fields
+    // containing external pointer handles (if they are evacuation candidates).
+    // As such, if such a field is invalidated (for example because the host
+    // object is converted to another object type), we need to be notified of
+    // that. Note that we do not need to care about "re-validated" fields here:
+    // if an external pointer field is first converted to different kind of
+    // field, then again converted to a external pointer field, then it will be
+    // re-initialized, at which point it will obtain a new entry in the
+    // external pointer table which cannot be a candidate for evacuation.
+    inline void NotifyExternalPointerFieldInvalidated(
+        Address field_address, ExternalPointerTagRange tag_range);
 
-    // Determine if compaction is needed and if so start the compaction.
-    // This is expected to be called at the start of the GC marking phase.
-    void StartCompactingIfNeeded();
+    // Not atomic.  Mutators and concurrent marking must be paused.
+    void AssertEmpty() { CHECK(segments_.empty()); }
+
+    bool allocate_black() { return allocate_black_; }
+    void set_allocate_black(bool allocate_black) {
+      allocate_black_ = allocate_black;
+    }
 
    private:
-    friend class ExternalPointerTable;
-
-    // Routines for compaction. See the comment about table compaction above.
-    inline bool IsCompacting();
-    inline void StartCompacting(uint32_t start_of_evacuation_area);
-    inline void StopCompacting();
-    inline void AbortCompacting(uint32_t start_of_evacuation_area);
-    inline bool CompactingWasAborted();
-
-    // This value indicates that this space is not currently being compacted. It
-    // is set to uint32_t max so that determining whether an entry should be
-    // evacuated becomes a single comparison:
-    // `bool should_be_evacuated = index >= start_of_evacuation_area`.
-    static constexpr uint32_t kNotCompactingMarker =
-        std::numeric_limits<uint32_t>::max();
-
-    // This value may be ORed into the start of evacuation area threshold
-    // during the GC marking phase to indicate that compaction has been
-    // aborted because the freelist grew to short and so evacuation entry
-    // allocation is no longer possible. This will prevent any further
-    // evacuation attempts as entries will be evacuated if their index is at or
-    // above the start of the evacuation area, which is now a huge value.
-    static constexpr uint32_t kCompactionAbortedMarker = 0xf0000000;
-
-    // When compacting this space, this field contains the index of the first
-    // entry in the evacuation area. The evacuation area then consists of all
-    // segments above this threshold, and the goal of compaction is to move all
-    // live entries out of these segments so that they can be deallocated after
-    // sweeping. The field can have the following values:
-    // - kNotCompactingMarker: compaction is not currently running.
-    // - A kEntriesPerSegment aligned value within: compaction is running and
-    //   all entries after this value should be evacuated.
-    // - A value that has kCompactionAbortedMarker in its top bits:
-    //   compaction has been aborted during marking. The original start of the
-    //   evacuation area is still contained in the lower bits.
-    std::atomic<uint32_t> start_of_evacuation_area_;
+    bool allocate_black_ = false;
   };
 
   // Initializes all slots in the RO space from pre-existing artifacts.
@@ -363,7 +362,7 @@ class V8_EXPORT_PRIVATE ExternalPointerTable
   //
   // This method is atomic and can be called from background threads.
   inline Address Get(ExternalPointerHandle handle,
-                     ExternalPointerTag tag) const;
+                     ExternalPointerTagRange tag_range) const;
 
   // Sets the entry referenced by the given handle.
   //
@@ -379,6 +378,14 @@ class V8_EXPORT_PRIVATE ExternalPointerTable
   inline Address Exchange(ExternalPointerHandle handle, Address value,
                           ExternalPointerTag tag);
 
+  // Retrieves the tag used for the entry referenced by the given handle.
+  //
+  // This method is atomic and can be called from background threads.
+  inline ExternalPointerTag GetTag(ExternalPointerHandle handle) const;
+
+  // Invalidates the entry referenced by the given handle.
+  inline void Zap(ExternalPointerHandle handle);
+
   // Allocates a new entry in the given space. The caller must provide the
   // initial value and tag for the entry.
   //
@@ -397,54 +404,83 @@ class V8_EXPORT_PRIVATE ExternalPointerTable
   inline void Mark(Space* space, ExternalPointerHandle handle,
                    Address handle_location);
 
-  // Frees unmarked entries and finishes space compaction (if running).
+  // Evacuate the specified entry from one space to another, updating the handle
+  // location in place.
+  //
+  // This method is not atomic and can be called only when the mutator is
+  // paused.
+  inline void Evacuate(Space* from_space, Space* to_space,
+                       ExternalPointerHandle handle, Address handle_location,
+                       EvacuateMarkMode mode);
+
+  // Evacuate all segments from from_space to to_space, leaving from_space empty
+  // with an empty free list.  Then free unmarked entries, finishing compaction
+  // if it was running, and collecting freed entries onto to_space's free list.
+  //
+  // The from_space will be left empty with an empty free list.
   //
   // This method must only be called while mutator threads are stopped as it is
   // not safe to allocate table entries while the table is being swept.
   //
+  // SweepAndCompact is the same as EvacuateAndSweepAndCompact, except without
+  // the evacuation phase.
+  //
+  // Sweep is the same as SweepAndCompact, but assumes that compaction was not
+  // running.
+  //
   // Returns the number of live entries after sweeping.
+  uint32_t EvacuateAndSweepAndCompact(Space* to_space, Space* from_space,
+                                      Counters* counters);
   uint32_t SweepAndCompact(Space* space, Counters* counters);
+  uint32_t Sweep(Space* space, Counters* counters);
+
+  inline bool Contains(Space* space, ExternalPointerHandle handle) const;
+
+  // A resource outside of the V8 heap whose lifetime is tied to something
+  // inside the V8 heap. This class makes that relationship explicit.
+  //
+  // Knowing about such objects is important for the sandbox to guarantee
+  // memory safety. In particular, it is necessary to prevent issues where the
+  // external resource is destroyed before the entry in the
+  // ExternalPointerTable (EPT) that references it is freed. In that case, the
+  // EPT entry would then contain a dangling pointer which could be abused by
+  // an attacker to cause a use-after-free outside of the sandbox.
+  //
+  // Currently, this is solved by remembering the EPT entry in the external
+  // object and zapping/invalidating it when the resource is destroyed. An
+  // alternative approach that might be preferable in the future would be to
+  // destroy the external resource only when the EPT entry is freed. This would
+  // avoid the need to manually keep track of the entry, for example.
+  class ManagedResource : public Malloced {
+   public:
+    // This method must be called before destroying the external resource.
+    // When the sandbox is enabled, it will take care of zapping its EPT entry.
+    inline void ZapExternalPointerTableEntry();
+
+   private:
+    friend class ExternalPointerTable;
+    // Currently required for snapshot stress mode, see deserializer.cc.
+    template <typename IsolateT>
+    friend class Deserializer;
+
+    ExternalPointerTable* owning_table_ = nullptr;
+    ExternalPointerHandle ept_entry_ = kNullExternalPointerHandle;
+  };
 
  private:
-  inline bool IsValidHandle(ExternalPointerHandle handle) const;
-  inline uint32_t HandleToIndex(ExternalPointerHandle handle) const;
-  inline ExternalPointerHandle IndexToHandle(uint32_t index) const;
+  static inline bool IsValidHandle(ExternalPointerHandle handle);
+  static inline uint32_t HandleToIndex(ExternalPointerHandle handle);
+  static inline ExternalPointerHandle IndexToHandle(uint32_t index);
 
-  inline void MaybeCreateEvacuationEntry(Space* space, uint32_t index,
-                                         Address handle_location);
+  inline void TakeOwnershipOfManagedResourceIfNecessary(
+      Address value, ExternalPointerHandle handle, ExternalPointerTag tag);
+  inline void FreeManagedResourceIfPresent(uint32_t entry_index);
 
   void ResolveEvacuationEntryDuringSweeping(
       uint32_t index, ExternalPointerHandle* handle_location,
       uint32_t start_of_evacuation_area);
-
-#ifdef DEBUG
-  // In debug builds during GC marking, this value is ORed into
-  // ExternalPointerHandles whose entries are marked for evacuation. During
-  // sweeping, the Handles for evacuated entries are checked to have this
-  // marker value. This allows detecting re-initialized entries, which are
-  // problematic for compaction. This is only possible for entries marked for
-  // evacuation as the location of the Handle is only known for those.
-  static constexpr uint32_t kVisitedHandleMarker = 0x1;
-  static_assert(kExternalPointerIndexShift >= 1);
-
-  inline bool HandleWasVisitedDuringMarking(ExternalPointerHandle handle) {
-    return (handle & kVisitedHandleMarker) == kVisitedHandleMarker;
-  }
-#endif  // DEBUG
-
-  // Outcome of external pointer table compaction to use for the
-  // ExternalPointerTableCompactionOutcome histogram.
-  enum class TableCompactionOutcome {
-    // Table compaction was successful.
-    kSuccess = 0,
-    // Outcome 1, partial success, is no longer supported.
-    // Table compaction was aborted because the freelist grew to short.
-    kAborted = 2,
-  };
 };
 
-static_assert(sizeof(ExternalPointerTable) == ExternalPointerTable::kSize);
-
 }  // namespace internal
 }  // namespace v8
 
