@@ -4,8 +4,10 @@
 
 #include "src/snapshot/read-only-serializer.h"
 
+#include "src/common/globals.h"
 #include "src/heap/heap-inl.h"
 #include "src/heap/read-only-heap.h"
+#include "src/heap/visit-object.h"
 #include "src/objects/objects-inl.h"
 #include "src/objects/slots.h"
 #include "src/snapshot/read-only-serializer-deserializer.h"
@@ -23,14 +25,16 @@ class ObjectPreProcessor final {
 
 #define PRE_PROCESS_TYPE_LIST(V) \
   V(AccessorInfo)                \
-  V(CallHandlerInfo)             \
+  V(InterceptorInfo)             \
+  V(JSExternalObject)            \
+  V(FunctionTemplateInfo)        \
   V(Code)
 
   void PreProcessIfNeeded(Tagged<HeapObject> o) {
     const InstanceType itype = o->map(isolate_)->instance_type();
 #define V(TYPE)                               \
   if (InstanceTypeChecker::Is##TYPE(itype)) { \
-    return PreProcess##TYPE(TYPE::cast(o));   \
+    return PreProcess##TYPE(Cast<TYPE>(o));   \
   }
     PRE_PROCESS_TYPE_LIST(V)
 #undef V
@@ -68,15 +72,60 @@ class ObjectPreProcessor final {
     EncodeExternalPointerSlot(o->RawExternalPointerField(
         AccessorInfo::kSetterOffset, kAccessorInfoSetterTag));
   }
-  void PreProcessCallHandlerInfo(Tagged<CallHandlerInfo> o) {
+  void PreProcessInterceptorInfo(Tagged<InterceptorInfo> o) {
+    const bool is_named = o->is_named();
+
+#define PROCESS_FIELD(Name, name)                       \
+  EncodeExternalPointerSlot(o->RawExternalPointerField( \
+      InterceptorInfo::k##Name##Offset,                 \
+      is_named ? kApiNamedProperty##Name##CallbackTag   \
+               : kApiIndexedProperty##Name##CallbackTag));
+
+    INTERCEPTOR_INFO_CALLBACK_LIST(PROCESS_FIELD)
+#undef PROCESS_FIELD
+  }
+  void PreProcessJSExternalObject(Tagged<JSExternalObject> o) {
+    EncodeExternalPointerSlot(
+        o->RawExternalPointerField(JSExternalObject::kValueOffset,
+                                   kExternalObjectValueTag),
+        reinterpret_cast<Address>(o->value(isolate_)));
+  }
+  void PreProcessFunctionTemplateInfo(Tagged<FunctionTemplateInfo> o) {
     EncodeExternalPointerSlot(
         o->RawExternalPointerField(
-            CallHandlerInfo::kMaybeRedirectedCallbackOffset,
-            kCallHandlerInfoCallbackTag),
+            FunctionTemplateInfo::kMaybeRedirectedCallbackOffset,
+            kFunctionTemplateInfoCallbackTag),
         o->callback(isolate_));  // Pass the non-redirected value.
   }
+#if V8_ENABLE_GEARBOX
+  V8_INLINE void ResetGearboxPlaceholderBuiltin(Tagged<Code> code) {
+    // In order to ensure predictable state of placeholder builtins Code
+    // objects after serialization we replace their fields with the contents
+    // of kIllegal builtin.
+    if (code->is_gearbox_placeholder_builtin()) {
+      Builtin variant_builtin_id = code->builtin_id();
+      Builtin placeholder_builtin_id =
+          Builtins::GetGearboxPlaceholderFromVariant(variant_builtin_id);
+      DCHECK_EQ(
+          isolate_->builtins()->code(placeholder_builtin_id)->builtin_id(),
+          code->builtin_id());
+      Tagged<Code> src = isolate_->builtins()->code(Builtin::kIllegal);
+      Code::CopyFieldsWithGearboxForSerialization(code, src, isolate_);
+      // We should use the placeholder id instead of kIllegal.
+      code->set_builtin_id(placeholder_builtin_id);
+    }
+  }
+#endif
   void PreProcessCode(Tagged<Code> o) {
     o->ClearInstructionStartForSerialization(isolate_);
+    CHECK(!o->has_source_position_table_or_bytecode_offset_table());
+    CHECK(!o->has_deoptimization_data_or_interpreter_data());
+#ifdef V8_ENABLE_LEAPTIERING
+    CHECK_EQ(o->js_dispatch_handle(), kNullJSDispatchHandle);
+#endif
+#if V8_ENABLE_GEARBOX
+    ResetGearboxPlaceholderBuiltin(o);
+#endif
   }
 
   Isolate* const isolate_;
@@ -84,7 +133,8 @@ class ObjectPreProcessor final {
 };
 
 struct ReadOnlySegmentForSerialization {
-  ReadOnlySegmentForSerialization(Isolate* isolate, const ReadOnlyPage* page,
+  ReadOnlySegmentForSerialization(Isolate* isolate,
+                                  const ReadOnlyPageMetadata* page,
                                   Address segment_start, size_t segment_size,
                                   ObjectPreProcessor* pre_processor)
       : page(page),
@@ -95,6 +145,9 @@ struct ReadOnlySegmentForSerialization {
         tagged_slots(segment_size / kTaggedSize) {
     // .. because tagged_slots records a bit for each slot:
     DCHECK(IsAligned(segment_size, kTaggedSize));
+    // Ensure incoming pointers to this page are representable.
+    CHECK_LT(isolate->read_only_heap()->read_only_space()->IndexOf(page),
+             1UL << ro::EncodedTagged::kPageIndexBits);
 
     MemCopy(contents.get(), reinterpret_cast<void*>(segment_start),
             segment_size);
@@ -115,13 +168,13 @@ struct ReadOnlySegmentForSerialization {
       size_t o_offset = o.ptr() - segment_start;
       Address o_dst = reinterpret_cast<Address>(contents.get()) + o_offset;
       pre_processor->PreProcessIfNeeded(
-          HeapObject::cast(Tagged<Object>(o_dst)));
+          Cast<HeapObject>(Tagged<Object>(o_dst)));
     }
   }
 
   void EncodeTaggedSlots(Isolate* isolate);
 
-  const ReadOnlyPage* const page;
+  const ReadOnlyPageMetadata* const page;
   const Address segment_start;
   const size_t segment_size;
   const size_t segment_offset;
@@ -135,19 +188,14 @@ struct ReadOnlySegmentForSerialization {
 
 ro::EncodedTagged Encode(Isolate* isolate, Tagged<HeapObject> o) {
   Address o_address = o.address();
-  BasicMemoryChunk* chunk = BasicMemoryChunk::FromAddress(o_address);
+  MemoryChunkMetadata* chunk = MemoryChunkMetadata::FromAddress(o_address);
 
-  ro::EncodedTagged encoded;
   ReadOnlySpace* ro_space = isolate->read_only_heap()->read_only_space();
   int index = static_cast<int>(ro_space->IndexOf(chunk));
-  DCHECK_LT(index, 1UL << ro::EncodedTagged::kPageIndexBits);
-  encoded.page_index = index;
   uint32_t offset = static_cast<int>(chunk->Offset(o_address));
   DCHECK(IsAligned(offset, kTaggedSize));
-  DCHECK_LT(offset / kTaggedSize, 1UL << ro::EncodedTagged::kOffsetBits);
-  encoded.offset = offset / kTaggedSize;
 
-  return encoded;
+  return ro::EncodedTagged(index, offset / kTaggedSize);
 }
 
 // If relocations are needed, this class
@@ -206,7 +254,7 @@ class EncodeRelocationsVisitor final : public ObjectVisitor {
     ExternalPointerSlot slot_in_segment{
         reinterpret_cast<Address>(segment_->contents.get() +
                                   SegmentOffsetOf(slot)),
-        slot.tag()};
+        slot.exact_tag()};
     // Constructing no_gc here is not the intended use pattern (instead we
     // should pass it along the entire callchain); but there's little point of
     // doing that here - all of the code in this file relies on GC being
@@ -225,7 +273,7 @@ class EncodeRelocationsVisitor final : public ObjectVisitor {
 
  private:
   void ProcessSlot(MaybeObjectSlot slot) {
-    MaybeObject o = *slot;
+    Tagged<MaybeObject> o = *slot;
     if (!o.IsStrongOrWeak()) return;  // Smis don't need relocation.
     DCHECK(o.IsStrong());
 
@@ -268,7 +316,7 @@ void ReadOnlySegmentForSerialization::EncodeTaggedSlots(Isolate* isolate) {
                                 SkipFreeSpaceOrFiller::kNo);
   for (Tagged<HeapObject> o = it.Next(); !o.is_null(); o = it.Next()) {
     if (o.address() >= segment_end) break;
-    o->Iterate(cage_base, &v);
+    VisitObject(isolate, o, &v);
   }
 }
 
@@ -297,12 +345,12 @@ class ReadOnlyHeapImageSerializer {
 
     // Allocate all pages first s.t. the deserializer can easily handle forward
     // references (e.g.: an object on page i points at an object on page i+1).
-    for (const ReadOnlyPage* page : ro_space->pages()) {
+    for (const ReadOnlyPageMetadata* page : ro_space->pages()) {
       EmitAllocatePage(page, unmapped_regions);
     }
 
     // Now write the page contents.
-    for (const ReadOnlyPage* page : ro_space->pages()) {
+    for (const ReadOnlyPageMetadata* page : ro_space->pages()) {
       SerializePage(page, unmapped_regions);
     }
 
@@ -310,26 +358,30 @@ class ReadOnlyHeapImageSerializer {
     sink_->Put(Bytecode::kFinalizeReadOnlySpace, "space end");
   }
 
-  uint32_t IndexOf(const ReadOnlyPage* page) {
+  uint32_t IndexOf(const ReadOnlyPageMetadata* page) {
     ReadOnlySpace* ro_space = isolate_->read_only_heap()->read_only_space();
     return static_cast<uint32_t>(ro_space->IndexOf(page));
   }
 
-  void EmitAllocatePage(const ReadOnlyPage* page,
+  void EmitAllocatePage(const ReadOnlyPageMetadata* page,
                         const std::vector<MemoryRegion>& unmapped_regions) {
-    sink_->Put(Bytecode::kAllocatePage, "page begin");
+    if (V8_STATIC_ROOTS_BOOL) {
+      sink_->Put(Bytecode::kAllocatePageAt, "fixed page begin");
+    } else {
+      sink_->Put(Bytecode::kAllocatePage, "page begin");
+    }
     sink_->PutUint30(IndexOf(page), "page index");
     sink_->PutUint30(
         static_cast<uint32_t>(page->HighWaterMark() - page->area_start()),
         "area size in bytes");
     if (V8_STATIC_ROOTS_BOOL) {
-      auto page_addr = reinterpret_cast<Address>(page);
+      auto page_addr = page->ChunkAddress();
       sink_->PutUint32(V8HeapCompressionScheme::CompressAny(page_addr),
                        "page start offset");
     }
   }
 
-  void SerializePage(const ReadOnlyPage* page,
+  void SerializePage(const ReadOnlyPageMetadata* page,
                      const std::vector<MemoryRegion>& unmapped_regions) {
     Address pos = page->area_start();
 
@@ -380,7 +432,7 @@ class ReadOnlyHeapImageSerializer {
       ReadOnlyRoots roots(isolate_);
       for (size_t i = 0; i < ReadOnlyRoots::kEntriesCount; i++) {
         RootIndex rudi = static_cast<RootIndex>(i);
-        Tagged<HeapObject> rudolf = HeapObject::cast(roots.object_at(rudi));
+        Tagged<HeapObject> rudolf = Cast<HeapObject>(roots.object_at(rudi));
         ro::EncodedTagged encoded = Encode(isolate_, rudolf);
         sink_->PutUint32(encoded.ToUint32(), "read only roots entry");
       }
@@ -394,7 +446,7 @@ class ReadOnlyHeapImageSerializer {
 
 std::vector<ReadOnlyHeapImageSerializer::MemoryRegion> GetUnmappedRegions(
     Isolate* isolate) {
-#ifdef V8_STATIC_ROOTS
+#if defined(V8_STATIC_ROOTS) && defined(V8_ENABLE_WEBASSEMBLY)
   // WasmNull's payload is aligned to the OS page and consists of
   // WasmNull::kPayloadSize bytes of unmapped memory. To avoid inflating the
   // snapshot size and accessing uninitialized and/or unmapped memory, the
@@ -404,7 +456,7 @@ std::vector<ReadOnlyHeapImageSerializer::MemoryRegion> GetUnmappedRegions(
   Tagged<HeapObject> wasm_null_padding = ro_roots.wasm_null_padding();
   CHECK(IsFreeSpace(wasm_null_padding));
   Address wasm_null_padding_start =
-      wasm_null_padding.address() + FreeSpace::kHeaderSize;
+      wasm_null_padding.address() + sizeof(FreeSpace);
   std::vector<ReadOnlyHeapImageSerializer::MemoryRegion> unmapped;
   if (wasm_null.address() > wasm_null_padding_start) {
     unmapped.push_back({wasm_null_padding_start,
@@ -414,7 +466,7 @@ std::vector<ReadOnlyHeapImageSerializer::MemoryRegion> GetUnmappedRegions(
   return unmapped;
 #else
   return {};
-#endif  // V8_STATIC_ROOTS
+#endif  // V8_STATIC_ROOTS && V8_ENABLE_WEBASSEMBLY
 }
 
 }  // namespace
