@@ -37,10 +37,6 @@ class TestCode : public HandleAndZoneScope {
     End();
     return pos;
   }
-  void Fallthru() {
-    Start();
-    End();
-  }
   int Branch(int ttarget, int ftarget) {
     Start();
     InstructionOperand ops[] = {UseRpo(ttarget), UseRpo(ftarget)};
@@ -65,6 +61,24 @@ class TestCode : public HandleAndZoneScope {
     Start();
     sequence_.AddInstruction(Instruction::New(main_zone(), kArchNop));
   }
+  template <int... T>
+  int Switch() {
+    Start();
+    int size = sizeof...(T);
+    InstructionOperand ops[] = {Immediate(T)...};
+    sequence_.AddInstruction(Instruction::New(main_zone(), kArchTableSwitch, 0,
+                                              nullptr, size, ops, 0, nullptr));
+    int pos = static_cast<int>(sequence_.instructions().size() - 1);
+    End();
+    return pos;
+  }
+  RpoNumber Case() {
+    CHECK_NULL(current_);
+    Start();
+    CHECK_NOT_NULL(current_);
+    current_->set_table_switch_target(true);
+    return rpo_number_;
+  }
   void RedundantMoves() {
     Start();
     sequence_.AddInstruction(Instruction::New(main_zone(), kArchNop));
@@ -101,7 +115,7 @@ class TestCode : public HandleAndZoneScope {
     Start();
     sequence_.AddInstruction(Instruction::New(main_zone(), 155));
   }
-  void End() {
+  RpoNumber End() {
     Start();
     int end = static_cast<int>(sequence_.instructions().size());
     if (current_->code_start() == end) {  // Empty block.  Insert a nop.
@@ -109,7 +123,9 @@ class TestCode : public HandleAndZoneScope {
     }
     sequence_.EndBlock(current_->rpo_number());
     current_ = nullptr;
+    RpoNumber rpo_number = rpo_number_;
     rpo_number_ = RpoNumber::FromInt(rpo_number_.ToInt() + 1);
+    return rpo_number;
   }
   InstructionOperand UseRpo(int num) {
     return sequence_.AddImmediate(Constant(RpoNumber::FromInt(num)));
@@ -220,7 +236,7 @@ TEST(FwMoves2) {
 
   // B0
   code.RedundantMoves();
-  code.Fallthru();
+  code.Jump(1);
   // B1
   code.End();
 
@@ -235,7 +251,7 @@ TEST(FwMoves2b) {
 
   // B0
   code.NonRedundantMoves();
-  code.Fallthru();
+  code.Jump(1);
   // B1
   code.End();
 
@@ -289,7 +305,7 @@ TEST(FwOther2) {
 
   // B0
   code.Other();
-  code.Fallthru();
+  code.Jump(1);
   // B1
   code.End();
 
@@ -303,7 +319,7 @@ TEST(FwNone2a) {
   TestCode code(kBlockCount);
 
   // B0
-  code.Fallthru();
+  code.Jump(1);
   // B1
   code.End();
 
@@ -343,7 +359,7 @@ TEST(FwLoop2) {
   TestCode code(kBlockCount);
 
   // B0
-  code.Fallthru();
+  code.Jump(1);
   // B1
   code.Jump(0);
 
@@ -357,9 +373,9 @@ TEST(FwLoop3) {
   TestCode code(kBlockCount);
 
   // B0
-  code.Fallthru();
+  code.Jump(1);
   // B1
-  code.Fallthru();
+  code.Jump(2);
   // B2
   code.Jump(0);
 
@@ -373,7 +389,7 @@ TEST(FwLoop1b) {
   TestCode code(kBlockCount);
 
   // B0
-  code.Fallthru();
+  code.Jump(1);
   // B1
   code.Jump(1);
 
@@ -387,9 +403,9 @@ TEST(FwLoop2b) {
   TestCode code(kBlockCount);
 
   // B0
-  code.Fallthru();
+  code.Jump(1);
   // B1
-  code.Fallthru();
+  code.Jump(2);
   // B2
   code.Jump(1);
 
@@ -403,11 +419,11 @@ TEST(FwLoop3b) {
   TestCode code(kBlockCount);
 
   // B0
-  code.Fallthru();
+  code.Jump(1);
   // B1
-  code.Fallthru();
+  code.Jump(2);
   // B2
-  code.Fallthru();
+  code.Jump(3);
   // B3
   code.Jump(1);
 
@@ -421,11 +437,11 @@ TEST(FwLoop2_1a) {
   TestCode code(kBlockCount);
 
   // B0
-  code.Fallthru();
+  code.Jump(1);
   // B1
-  code.Fallthru();
+  code.Jump(2);
   // B2
-  code.Fallthru();
+  code.Jump(3);
   // B3
   code.Jump(1);
   // B4
@@ -441,9 +457,9 @@ TEST(FwLoop2_1b) {
   TestCode code(kBlockCount);
 
   // B0
-  code.Fallthru();
+  code.Jump(1);
   // B1
-  code.Fallthru();
+  code.Jump(2);
   // B2
   code.Jump(4);
   // B3
@@ -461,9 +477,9 @@ TEST(FwLoop2_1c) {
   TestCode code(kBlockCount);
 
   // B0
-  code.Fallthru();
+  code.Jump(1);
   // B1
-  code.Fallthru();
+  code.Jump(2);
   // B2
   code.Jump(4);
   // B3
@@ -481,9 +497,9 @@ TEST(FwLoop2_1d) {
   TestCode code(kBlockCount);
 
   // B0
-  code.Fallthru();
+  code.Jump(1);
   // B1
-  code.Fallthru();
+  code.Jump(2);
   // B2
   code.Jump(1);
   // B3
@@ -501,11 +517,11 @@ TEST(FwLoop3_1a) {
   TestCode code(kBlockCount);
 
   // B0
-  code.Fallthru();
+  code.Jump(1);
   // B1
-  code.Fallthru();
+  code.Jump(2);
   // B2
-  code.Fallthru();
+  code.Jump(3);
   // B3
   code.Jump(2);
   // B4
@@ -779,6 +795,12 @@ void CheckBranch(TestCode* code, int pos, int t1, int t2) {
   CHECK_EQ(t2, code->sequence_.InputRpo(instr, 1).ToInt());
 }
 
+void CheckBlockSwitchTarget(TestCode* code, RpoNumber rpo,
+                            bool expect_switch_target) {
+  CHECK_EQ(code->sequence_.InstructionBlockAt(rpo)->IsTableSwitchTarget(),
+           expect_switch_target);
+}
+
 void CheckAssemblyOrder(TestCode* code, int size, int* expected) {
   int i = 0;
   for (auto const block : code->sequence_.instruction_blocks()) {
@@ -841,7 +863,7 @@ TEST(Rewire2_deferred) {
   int j1 = code.Jump(1);
   // B1
   code.Defer();
-  code.Fallthru();
+  code.Jump(2);
   // B2
   code.Defer();
   int j2 = code.Jump(3);
@@ -864,7 +886,7 @@ TEST(Rewire_deferred_diamond) {
   // B0
   int b1 = code.Branch(1, 2);
   // B1
-  code.Fallthru();  // To B3
+  code.Jump(3);
   // B2
   code.Defer();
   int j1 = code.Jump(3);
@@ -1063,6 +1085,51 @@ TEST(RewireGapJump2) {
   CheckAssemblyOrder(&code, kBlockCount, assembly);
 }
 
+TEST(PropagateSwitchTarget) {
+  constexpr size_t kBlockCount = 6;
+  TestCode code(kBlockCount);
+
+  // B0
+  code.Switch<1, 2, 3, 4>();
+  // B1
+  auto b1 = code.Case();
+  int j1 = code.Jump(5);
+  // B2
+  auto b2 = code.Case();
+  int j2 = code.Jump(5);
+  // B3
+  auto b3 = code.Case();
+  int j3 = code.Jump(5);
+  // B4
+  auto b4 = code.Case();
+  int j4 = code.Jump(5);
+  // B5
+  auto b5 = code.End();
+
+  // All cases should forward to the end B5 block and be turned into nops.
+
+  int forward[] = {0, 5, 5, 5, 5, 5};
+  VerifyForwarding(&code, kBlockCount, forward);
+  ApplyForwarding(&code, kBlockCount, forward);
+
+  CheckNop(&code, j1);
+  CheckNop(&code, j2);
+  CheckNop(&code, j3);
+  CheckNop(&code, j4);
+
+  static int assembly[] = {0, 1, 1, 1, 1, 1};
+  CheckAssemblyOrder(&code, kBlockCount, assembly);
+
+  // If jump-threading elides a block, it should propagate the switch target
+  // property.
+  CheckBlockSwitchTarget(&code, b1, false);
+  CheckBlockSwitchTarget(&code, b2, false);
+  CheckBlockSwitchTarget(&code, b3, false);
+  CheckBlockSwitchTarget(&code, b4, false);
+
+  CheckBlockSwitchTarget(&code, b5, true);
+}
+
 }  // namespace compiler
 }  // namespace internal
 }  // namespace v8
