@@ -5,8 +5,6 @@
 #include "src/codegen/tick-counter.h"
 #include "src/compiler/access-builder.h"
 #include "src/compiler/common-operator.h"
-#include "src/compiler/graph-visualizer.h"
-#include "src/compiler/graph.h"
 #include "src/compiler/js-graph.h"
 #include "src/compiler/js-operator.h"
 #include "src/compiler/loop-analysis.h"
@@ -16,6 +14,8 @@
 #include "src/compiler/schedule.h"
 #include "src/compiler/scheduler.h"
 #include "src/compiler/simplified-operator.h"
+#include "src/compiler/turbofan-graph-visualizer.h"
+#include "src/compiler/turbofan-graph.h"
 #include "src/compiler/verifier.h"
 #include "test/cctest/cctest.h"
 
@@ -36,8 +36,7 @@ static const int kNumLeafs = 4;
 class LoopFinderTester : HandleAndZoneScope {
  public:
   LoopFinderTester()
-      : HandleAndZoneScope(kCompressGraphZone),
-        isolate(main_isolate()),
+      : isolate(main_isolate()),
         common(main_zone()),
         graph(main_zone()),
         jsgraph(main_isolate(), &graph, &common, nullptr, nullptr, nullptr),
@@ -46,7 +45,7 @@ class LoopFinderTester : HandleAndZoneScope {
         p0(graph.NewNode(common.Parameter(0), start)),
         zero(jsgraph.Int32Constant(0)),
         one(jsgraph.OneConstant()),
-        half(jsgraph.Constant(0.5)),
+        half(jsgraph.ConstantNoHole(0.5)),
         self(graph.NewNode(common.Int32Constant(0xAABBCCDD))),
         dead(graph.NewNode(common.Dead())),
         loop_tree(nullptr) {
@@ -61,7 +60,7 @@ class LoopFinderTester : HandleAndZoneScope {
   Isolate* isolate;
   TickCounter tick_counter;
   CommonOperatorBuilder common;
-  Graph graph;
+  TFGraph graph;
   JSGraph jsgraph;
   Node* start;
   Node* end;
