@@ -3,8 +3,11 @@
 // found in the LICENSE file.
 
 #include "src/api/api-inl.h"
+#include "src/base/logging.h"
 #include "src/base/strings.h"
+#include "src/common/globals.h"
 #include "src/objects/js-array-buffer-inl.h"
+#include "src/sandbox/sandbox.h"
 #include "test/cctest/heap/heap-utils.h"
 #include "test/cctest/test-api.h"
 #include "test/common/flag-utils.h"
@@ -12,10 +15,18 @@
 using ::v8::Array;
 using ::v8::Context;
 using ::v8::Local;
+using ::v8::Maybe;
 using ::v8::Value;
 
 namespace {
 
+// Too large ArrayBuffer.
+#if V8_TARGET_ARCH_64_BIT
+size_t kUnreasonableSize = size_t{1} << 53;
+#else
+size_t kUnreasonableSize = size_t{1} << 31;
+#endif
+
 void CheckDataViewIsDetached(v8::Local<v8::DataView> dv) {
   CHECK_EQ(0, static_cast<int>(dv->ByteLength()));
   CHECK_EQ(0, static_cast<int>(dv->ByteOffset()));
@@ -62,7 +73,7 @@ std::shared_ptr<v8::BackingStore> Externalize(Local<v8::SharedArrayBuffer> ab) {
 
 THREADED_TEST(ArrayBuffer_ApiInternalToExternal) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   Local<v8::ArrayBuffer> ab = v8::ArrayBuffer::New(isolate, 1024);
@@ -94,9 +105,243 @@ THREADED_TEST(ArrayBuffer_ApiInternalToExternal) {
   CHECK_EQ(0xDD, result->Int32Value(env.local()).FromJust());
 }
 
+THREADED_TEST(ArrayBuffer_ApiMaybeNew) {
+  LocalContext env;
+  v8::Isolate* isolate = env.isolate();
+  v8::HandleScope handle_scope(isolate);
+
+  // Reasonable-sized ArrayBuffer.
+  v8::MaybeLocal<v8::ArrayBuffer> maybe_ab =
+      v8::ArrayBuffer::MaybeNew(isolate, 1024);
+  CHECK(!maybe_ab.IsEmpty());
+  auto ab = v8::Local<v8::ArrayBuffer>::Cast(maybe_ab.ToLocalChecked());
+  CheckInternalFieldsAreZero(ab);
+  CHECK_EQ(1024, ab->ByteLength());
+  i::heap::InvokeMajorGC(CcTest::heap());
+
+  std::shared_ptr<v8::BackingStore> backing_store = Externalize(ab);
+  CHECK_EQ(1024, backing_store->ByteLength());
+
+  uint8_t* data = static_cast<uint8_t*>(backing_store->Data());
+  CHECK_NOT_NULL(data);
+  CHECK(env->Global()->Set(env.local(), v8_str("ab"), ab).FromJust());
+
+  v8::Local<v8::Value> result = CompileRun("ab.byteLength");
+  CHECK_EQ(1024, result->Int32Value(env.local()).FromJust());
+
+  v8::MaybeLocal<v8::ArrayBuffer> maybe_ab_2 =
+      v8::ArrayBuffer::MaybeNew(isolate, kUnreasonableSize);
+  CHECK(maybe_ab_2.IsEmpty());
+}
+
+// Limits the max size of each allocation, but doesn't limit the
+// total allocation size.
+class RestrictedAllocator final : public v8::ArrayBuffer::Allocator {
+ public:
+  explicit RestrictedAllocator(size_t limit) : limit_(limit) {}
+  ~RestrictedAllocator() { delete underlying_allocator_; }
+
+  void* Allocate(size_t length) override {
+    return underlying_allocator_->Allocate(length);
+  }
+
+  void* AllocateUninitialized(size_t length) override {
+    return underlying_allocator_->AllocateUninitialized(length);
+  }
+
+  void Free(void* data, size_t length) override {
+    underlying_allocator_->Free(data, length);
+  }
+
+  size_t MaxAllocationSize() const override { return limit_; }
+
+ private:
+  size_t limit_;
+  v8::ArrayBuffer::Allocator* underlying_allocator_ =
+      v8::ArrayBuffer::Allocator::NewDefaultAllocator();
+};
+
+TEST(ArrayBuffer_MaxSize) {
+  RestrictedAllocator allocator(32768);
+  v8::Isolate::CreateParams create_params;
+  create_params.array_buffer_allocator = &allocator;
+
+  v8::Isolate* isolate = v8::Isolate::New(create_params);
+  isolate->Enter();
+
+  {
+    i::Heap* heap = reinterpret_cast<i::Isolate*>(isolate)->heap();
+    int gc_count = heap->gc_count();
+    v8::HandleScope scope(isolate);
+    LocalContext context(isolate);
+
+    // OK size.
+    v8::MaybeLocal<v8::ArrayBuffer> maybe_ab =
+        v8::ArrayBuffer::MaybeNew(isolate, 32768);
+    CHECK(!maybe_ab.IsEmpty());
+
+    // OK size.
+    v8::MaybeLocal<v8::ArrayBuffer> maybe_ab2 =
+        v8::ArrayBuffer::MaybeNew(isolate, 32768);
+    CHECK(!maybe_ab2.IsEmpty());
+
+    // Too big.
+    v8::MaybeLocal<v8::ArrayBuffer> maybe_ab3 =
+        v8::ArrayBuffer::MaybeNew(isolate, 32769);
+    CHECK(maybe_ab3.IsEmpty());
+
+    // The large allocations should be rejected without running a GC first.
+    CHECK_EQ(gc_count, heap->gc_count());
+  }
+
+  isolate->Exit();
+  isolate->Dispose();
+}
+
+// Limits the total allocated size of array buffer backings.
+class TotalAllocator final : public v8::ArrayBuffer::Allocator {
+ public:
+  explicit TotalAllocator(size_t limit) : limit_(limit), remaining_(limit) {}
+  ~TotalAllocator() {
+    CHECK(remaining_.load() == limit_);
+    delete underlying_allocator_;
+  }
+
+  void* Allocate(size_t length) override {
+    if (length > remaining_.load()) return nullptr;
+    void* result = underlying_allocator_->Allocate(length);
+    Update(0 - length);
+    return result;
+  }
+  void* AllocateUninitialized(size_t length) override {
+    if (length > remaining_.load()) return nullptr;
+    void* result = underlying_allocator_->AllocateUninitialized(length);
+    Update(0 - length);
+    return result;
+  }
+  void Free(void* data, size_t length) override {
+    Update(length);
+    underlying_allocator_->Free(data, length);
+  }
+
+  size_t MaxAllocationSize() const override { return limit_; }
+
+ private:
+  void Update(intptr_t difference) {
+    while (true) {
+      size_t previous = remaining_.load();
+      size_t next = previous + difference;
+      bool success = remaining_.compare_exchange_weak(previous, next);
+      if (success) return;
+    }
+  }
+
+  size_t limit_;
+  std::atomic<size_t> remaining_;
+  v8::ArrayBuffer::Allocator* underlying_allocator_ =
+      v8::ArrayBuffer::Allocator::NewDefaultAllocator();
+};
+
+TEST(ArrayBuffer_TotalSize) {
+  TotalAllocator allocator(32768);
+  v8::Isolate::CreateParams create_params;
+  create_params.array_buffer_allocator = &allocator;
+
+  v8::Isolate* isolate = v8::Isolate::New(create_params);
+  isolate->Enter();
+
+  {
+    i::Heap* heap = reinterpret_cast<i::Isolate*>(isolate)->heap();
+    int gc_count = heap->gc_count();
+    v8::HandleScope scope(isolate);
+    LocalContext context(isolate);
+    // When some allocation below fails, we need to invoke GC without stack,
+    // otherwise some objects may not be reclaimed because of conservative
+    // stack scanning (CSS). Objects that are to be retained from this GC must
+    // be placed in the array of globals, as CSS is disabled.
+    i::DisableConservativeStackScanningScopeForTesting no_stack_scanning(heap);
+    std::vector<v8::Global<v8::ArrayBuffer>> globals;
+
+    // Too big.
+    {
+      v8::MaybeLocal<v8::ArrayBuffer> maybe_ab1 =
+          v8::ArrayBuffer::MaybeNew(isolate, 32769);
+      CHECK(maybe_ab1.IsEmpty());
+    }
+
+    // Too big shared.
+    v8::MaybeLocal<v8::SharedArrayBuffer> maybe_sab =
+        v8::SharedArrayBuffer::MaybeNew(isolate, 32769);
+    CHECK(maybe_sab.IsEmpty());
+
+    // Too big backing store.
+    std::unique_ptr<v8::BackingStore> backing_store =
+        v8::ArrayBuffer::NewBackingStore(
+            isolate, 32769,
+            v8::BackingStoreInitializationMode::kZeroInitialized,
+            v8::BackingStoreOnFailureMode::kReturnNull);
+    CHECK(!backing_store);
+
+    // Too big shared backing store.
+    std::unique_ptr<v8::BackingStore> shared_backing_store =
+        v8::SharedArrayBuffer::NewBackingStore(
+            isolate, 32769,
+            v8::BackingStoreInitializationMode::kZeroInitialized,
+            v8::BackingStoreOnFailureMode::kReturnNull);
+    CHECK(!shared_backing_store);
+
+    // Take half size.
+    {
+      v8::MaybeLocal<v8::ArrayBuffer> maybe_ab2 =
+          v8::ArrayBuffer::MaybeNew(isolate, 16384);
+      CHECK(!maybe_ab2.IsEmpty());
+      globals.emplace_back(isolate, maybe_ab2.ToLocalChecked());
+    }
+
+    // The large allocations should be rejected without running a GC first.
+    CHECK_EQ(gc_count, heap->gc_count());
+
+    {
+      v8::HandleScope inner_scope(isolate);
+
+      // Take second half size.
+      v8::MaybeLocal<v8::ArrayBuffer> maybe_ab3 =
+          v8::ArrayBuffer::MaybeNew(isolate, 16384);
+      CHECK(!maybe_ab3.IsEmpty());
+
+      // Fails because of total size
+      v8::MaybeLocal<v8::ArrayBuffer> maybe_ab4 =
+          v8::ArrayBuffer::MaybeNew(isolate, 16384);
+      CHECK(maybe_ab4.IsEmpty());
+    }
+
+    // We exit the inner scope and the last array buffer can be GCed.
+    // Take second half size again.
+    {
+      v8::MaybeLocal<v8::ArrayBuffer> maybe_ab5 =
+          v8::ArrayBuffer::MaybeNew(isolate, 16384);
+      CHECK(!maybe_ab5.IsEmpty());
+      globals.emplace_back(isolate, maybe_ab5.ToLocalChecked());
+    }
+
+    // The last allocation can't be done without a GC.
+    CHECK_LT(gc_count, heap->gc_count());
+
+    // Fails because of total size
+    {
+      v8::MaybeLocal<v8::ArrayBuffer> maybe_ab6 =
+          v8::ArrayBuffer::MaybeNew(isolate, 16384);
+      CHECK(maybe_ab6.IsEmpty());
+    }
+  }
+
+  isolate->Exit();
+  isolate->Dispose();
+}
+
 THREADED_TEST(ArrayBuffer_JSInternalToExternal) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   v8::Local<v8::Value> result = CompileRun(
@@ -135,13 +380,13 @@ THREADED_TEST(ArrayBuffer_JSInternalToExternal) {
 
 THREADED_TEST(ArrayBuffer_DisableDetach) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   Local<v8::ArrayBuffer> ab = v8::ArrayBuffer::New(isolate, 100);
   CHECK(ab->IsDetachable());
 
-  i::Handle<i::JSArrayBuffer> buf = v8::Utils::OpenHandle(*ab);
+  i::DirectHandle<i::JSArrayBuffer> buf = v8::Utils::OpenDirectHandle(*ab);
   buf->set_is_detachable(false);
 
   CHECK(!ab->IsDetachable());
@@ -149,7 +394,7 @@ THREADED_TEST(ArrayBuffer_DisableDetach) {
 
 THREADED_TEST(ArrayBuffer_DetachingApi) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   v8::Local<v8::ArrayBuffer> buffer = v8::ArrayBuffer::New(isolate, 1024);
@@ -198,7 +443,7 @@ THREADED_TEST(ArrayBuffer_DetachingApi) {
 
 THREADED_TEST(ArrayBuffer_DetachingScript) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   CompileRun(
@@ -249,7 +494,7 @@ THREADED_TEST(ArrayBuffer_DetachingScript) {
 
 THREADED_TEST(ArrayBuffer_WasDetached) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   Local<v8::ArrayBuffer> ab = v8::ArrayBuffer::New(isolate, 0);
@@ -261,7 +506,7 @@ THREADED_TEST(ArrayBuffer_WasDetached) {
 
 THREADED_TEST(ArrayBuffer_NonDetachableWasDetached) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   CompileRun(R"JS(
@@ -280,7 +525,7 @@ THREADED_TEST(ArrayBuffer_NonDetachableWasDetached) {
 
 THREADED_TEST(ArrayBuffer_ExternalizeEmpty) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   Local<v8::ArrayBuffer> ab = v8::ArrayBuffer::New(isolate, 2);
@@ -301,7 +546,7 @@ THREADED_TEST(ArrayBuffer_ExternalizeEmpty) {
 
 THREADED_TEST(SharedArrayBuffer_ApiInternalToExternal) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   Local<v8::SharedArrayBuffer> ab = v8::SharedArrayBuffer::New(isolate, 1024);
@@ -335,7 +580,7 @@ THREADED_TEST(SharedArrayBuffer_ApiInternalToExternal) {
 
 THREADED_TEST(SharedArrayBuffer_JSInternalToExternal) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   v8::Local<v8::Value> result = CompileRun(
@@ -375,7 +620,7 @@ THREADED_TEST(SharedArrayBuffer_JSInternalToExternal) {
 
 THREADED_TEST(SkipArrayBufferBackingStoreDuringGC) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   void* buffer = CcTest::array_buffer_allocator()->Allocate(100);
@@ -403,13 +648,13 @@ THREADED_TEST(SkipArrayBufferBackingStoreDuringGC) {
 
 THREADED_TEST(SkipArrayBufferDuringScavenge) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   // Make sure the pointer looks like a heap object
   Local<v8::Object> tmp = v8::Object::New(isolate);
   uint8_t* store_ptr =
-      reinterpret_cast<uint8_t*>(*reinterpret_cast<uintptr_t*>(*tmp));
+      reinterpret_cast<uint8_t*>(i::ValueHelper::ValueAsAddress(*tmp));
   auto backing_store = v8::ArrayBuffer::NewBackingStore(
       store_ptr, 8, [](void*, size_t, void*) {}, nullptr);
 
@@ -429,7 +674,7 @@ THREADED_TEST(SkipArrayBufferDuringScavenge) {
 
 THREADED_TEST(Regress1006600) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   Local<v8::Value> ab = CompileRunChecked(isolate, "new ArrayBuffer()");
@@ -440,7 +685,7 @@ THREADED_TEST(Regress1006600) {
 
 THREADED_TEST(ArrayBuffer_NewBackingStore) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
   std::shared_ptr<v8::BackingStore> backing_store =
       v8::ArrayBuffer::NewBackingStore(isolate, 100);
@@ -452,10 +697,8 @@ THREADED_TEST(ArrayBuffer_NewBackingStore) {
 }
 
 THREADED_TEST(ArrayBuffer_NewResizableBackingStore) {
-  FLAG_SCOPE(harmony_rab_gsab);
-
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
   std::shared_ptr<v8::BackingStore> backing_store =
       v8::ArrayBuffer::NewResizableBackingStore(32, 1024);
@@ -470,7 +713,7 @@ THREADED_TEST(ArrayBuffer_NewResizableBackingStore) {
 
 THREADED_TEST(SharedArrayBuffer_NewBackingStore) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
   std::shared_ptr<v8::BackingStore> backing_store =
       v8::SharedArrayBuffer::NewBackingStore(isolate, 100);
@@ -482,6 +725,30 @@ THREADED_TEST(SharedArrayBuffer_NewBackingStore) {
   CHECK_EQ(backing_store->Data(), ab->Data());
 }
 
+THREADED_TEST(SharedArrayBuffer_NewBackingStore_Unreasonable) {
+  LocalContext env;
+  v8::Isolate* isolate = env.isolate();
+  v8::HandleScope handle_scope(isolate);
+  std::unique_ptr<v8::BackingStore> backing_store =
+      v8::SharedArrayBuffer::NewBackingStore(
+          isolate, kUnreasonableSize,
+          v8::BackingStoreInitializationMode::kZeroInitialized,
+          v8::BackingStoreOnFailureMode::kReturnNull);
+  CHECK(!backing_store);
+}
+
+THREADED_TEST(ArrayBuffer_NewBackingStore_Unreasonable) {
+  LocalContext env;
+  v8::Isolate* isolate = env.isolate();
+  v8::HandleScope handle_scope(isolate);
+  std::unique_ptr<v8::BackingStore> backing_store =
+      v8::ArrayBuffer::NewBackingStore(
+          isolate, kUnreasonableSize,
+          v8::BackingStoreInitializationMode::kZeroInitialized,
+          v8::BackingStoreOnFailureMode::kReturnNull);
+  CHECK(!backing_store);
+}
+
 static void* backing_store_custom_data = nullptr;
 static size_t backing_store_custom_length = 0;
 static bool backing_store_custom_called = false;
@@ -528,18 +795,18 @@ TEST(SharedArrayBuffer_NewBackingStore_CustomDeleter) {
 
 TEST(ArrayBuffer_NewBackingStore_EmptyDeleter) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
   size_t size = 100;
   void* buffer = CcTest::array_buffer_allocator()->Allocate(size);
   std::unique_ptr<v8::BackingStore> backing_store =
       v8::ArrayBuffer::NewBackingStore(buffer, size,
                                        v8::BackingStore::EmptyDeleter, nullptr);
-  uint64_t external_memory_before =
-      isolate->AdjustAmountOfExternalAllocatedMemory(0);
+  uint64_t external_memory_before = v8::ExternalMemoryAccounter::
+      GetTotalAmountOfExternalAllocatedMemoryForTesting(isolate);
   v8::ArrayBuffer::New(isolate, std::move(backing_store));
-  uint64_t external_memory_after =
-      isolate->AdjustAmountOfExternalAllocatedMemory(0);
+  uint64_t external_memory_after = v8::ExternalMemoryAccounter::
+      GetTotalAmountOfExternalAllocatedMemoryForTesting(isolate);
   // The ArrayBuffer constructor does not increase the external memory counter.
   // The counter may decrease however if the allocation triggers GC.
   CHECK_GE(external_memory_before, external_memory_after);
@@ -548,18 +815,18 @@ TEST(ArrayBuffer_NewBackingStore_EmptyDeleter) {
 
 TEST(SharedArrayBuffer_NewBackingStore_EmptyDeleter) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
   size_t size = 100;
   void* buffer = CcTest::array_buffer_allocator()->Allocate(size);
   std::unique_ptr<v8::BackingStore> backing_store =
       v8::SharedArrayBuffer::NewBackingStore(
           buffer, size, v8::BackingStore::EmptyDeleter, nullptr);
-  uint64_t external_memory_before =
-      isolate->AdjustAmountOfExternalAllocatedMemory(0);
+  uint64_t external_memory_before = v8::ExternalMemoryAccounter::
+      GetTotalAmountOfExternalAllocatedMemoryForTesting(isolate);
   v8::SharedArrayBuffer::New(isolate, std::move(backing_store));
-  uint64_t external_memory_after =
-      isolate->AdjustAmountOfExternalAllocatedMemory(0);
+  uint64_t external_memory_after = v8::ExternalMemoryAccounter::
+      GetTotalAmountOfExternalAllocatedMemoryForTesting(isolate);
   // The SharedArrayBuffer constructor does not increase the external memory
   // counter. The counter may decrease however if the allocation triggers GC.
   CHECK_GE(external_memory_before, external_memory_after);
@@ -568,7 +835,7 @@ TEST(SharedArrayBuffer_NewBackingStore_EmptyDeleter) {
 
 THREADED_TEST(BackingStore_NotShared) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
   Local<v8::ArrayBuffer> ab = v8::ArrayBuffer::New(isolate, 8);
   CHECK(!ab->GetBackingStore()->IsShared());
@@ -585,7 +852,7 @@ THREADED_TEST(BackingStore_NotShared) {
 
 THREADED_TEST(BackingStore_Shared) {
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
   Local<v8::SharedArrayBuffer> ab = v8::SharedArrayBuffer::New(isolate, 8);
   CHECK(ab->GetBackingStore()->IsShared());
@@ -606,7 +873,7 @@ THREADED_TEST(ArrayBuffer_NewBackingStore_NullData) {
   // into that TypedArray at index 0 is expected to be a no-op, reading from
   // that TypedArray at index 0 should result in the default value '0'.
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
   std::unique_ptr<v8::BackingStore> backing_store =
       v8::ArrayBuffer::NewBackingStore(nullptr, 0,
@@ -756,82 +1023,9 @@ TEST(BackingStore_ReleaseAllocator_NullptrBackingStore) {
   CHECK(allocator_weak.expired());
 }
 
-TEST(BackingStore_ReallocateExpand) {
-  LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
-  std::unique_ptr<v8::BackingStore> backing_store =
-      v8::ArrayBuffer::NewBackingStore(isolate, 10);
-  {
-    uint8_t* data = reinterpret_cast<uint8_t*>(
-        reinterpret_cast<uintptr_t>(backing_store->Data()));
-    for (uint8_t i = 0; i < 10; i++) {
-      data[i] = i;
-    }
-  }
-  std::unique_ptr<v8::BackingStore> new_backing_store =
-      v8::BackingStore::Reallocate(isolate, std::move(backing_store), 20);
-  CHECK_EQ(new_backing_store->ByteLength(), 20);
-  CHECK(!new_backing_store->IsShared());
-  {
-    uint8_t* data = reinterpret_cast<uint8_t*>(
-        reinterpret_cast<uintptr_t>(new_backing_store->Data()));
-    for (uint8_t i = 0; i < 10; i++) {
-      CHECK_EQ(data[i], i);
-    }
-    for (uint8_t i = 10; i < 20; i++) {
-      CHECK_EQ(data[i], 0);
-    }
-  }
-}
-
-TEST(BackingStore_ReallocateShrink) {
-  LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
-  std::unique_ptr<v8::BackingStore> backing_store =
-      v8::ArrayBuffer::NewBackingStore(isolate, 20);
-  {
-    uint8_t* data = reinterpret_cast<uint8_t*>(backing_store->Data());
-    for (uint8_t i = 0; i < 20; i++) {
-      data[i] = i;
-    }
-  }
-  std::unique_ptr<v8::BackingStore> new_backing_store =
-      v8::BackingStore::Reallocate(isolate, std::move(backing_store), 10);
-  CHECK_EQ(new_backing_store->ByteLength(), 10);
-  CHECK(!new_backing_store->IsShared());
-  {
-    uint8_t* data = reinterpret_cast<uint8_t*>(new_backing_store->Data());
-    for (uint8_t i = 0; i < 10; i++) {
-      CHECK_EQ(data[i], i);
-    }
-  }
-}
-
-TEST(BackingStore_ReallocateNotShared) {
-  LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
-  std::unique_ptr<v8::BackingStore> backing_store =
-      v8::ArrayBuffer::NewBackingStore(isolate, 20);
-  std::unique_ptr<v8::BackingStore> new_backing_store =
-      v8::BackingStore::Reallocate(isolate, std::move(backing_store), 10);
-  CHECK(!new_backing_store->IsShared());
-}
-
-TEST(BackingStore_ReallocateShared) {
-  LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
-  std::unique_ptr<v8::BackingStore> backing_store =
-      v8::SharedArrayBuffer::NewBackingStore(isolate, 20);
-  std::unique_ptr<v8::BackingStore> new_backing_store =
-      v8::BackingStore::Reallocate(isolate, std::move(backing_store), 10);
-  CHECK(new_backing_store->IsShared());
-}
-
 TEST(ArrayBuffer_Resizable) {
-  FLAG_SCOPE(harmony_rab_gsab);
-
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   const char rab_source[] = "new ArrayBuffer(32, { maxByteLength: 1024 });";
@@ -851,10 +1045,8 @@ TEST(ArrayBuffer_Resizable) {
 }
 
 TEST(ArrayBuffer_FixedLength) {
-  FLAG_SCOPE(harmony_rab_gsab);
-
   LocalContext env;
-  v8::Isolate* isolate = env->GetIsolate();
+  v8::Isolate* isolate = env.isolate();
   v8::HandleScope handle_scope(isolate);
 
   // Fixed-length ArrayBuffers' byte length are equal to their max byte length.
@@ -871,3 +1063,214 @@ TEST(ArrayBuffer_FixedLength) {
   CHECK_EQ(32, sab->MaxByteLength());
   CHECK_EQ(sab->MaxByteLength(), sab->GetBackingStore()->MaxByteLength());
 }
+
+THREADED_TEST(ArrayBuffer_DataApiWithEmptyExternal) {
+  LocalContext env;
+  v8::Isolate* isolate = env.isolate();
+  v8::HandleScope handle_scope(isolate);
+
+  Local<v8::ArrayBuffer> ab = v8::ArrayBuffer::New(isolate, 0);
+  void* empty_buffer_ptr = v8::internal::EmptyBackingStoreBuffer();
+  CHECK_EQ(empty_buffer_ptr, ab->Data());
+  CHECK_EQ(0, ab->ByteLength());
+  CHECK_EQ(empty_buffer_ptr, ab->GetBackingStore()->Data());
+  // Repeat test to make sure that accessing the backing store buffer hasn't
+  // changed what the AB's Data method returns.
+  CHECK_EQ(empty_buffer_ptr, ab->Data());
+  CHECK_EQ(0, ab->ByteLength());
+
+  void* buffer = CcTest::array_buffer_allocator()->Allocate(1);
+  std::unique_ptr<v8::BackingStore> backing_store =
+      v8::ArrayBuffer::NewBackingStore(buffer, 0,
+                                       v8::BackingStore::EmptyDeleter, nullptr);
+  Local<v8::ArrayBuffer> ab2 =
+      v8::ArrayBuffer::New(isolate, std::move(backing_store));
+  CHECK_EQ(buffer, ab2->Data());
+  CHECK_EQ(0, ab->ByteLength());
+}
+
+namespace {
+void TestArrayBufferViewGetContent(const char* source, void* expected) {
+  LocalContext env;
+  v8::Isolate* isolate = env.isolate();
+  v8::HandleScope handle_scope(isolate);
+
+  auto view = v8::Local<v8::ArrayBufferView>::Cast(CompileRun(source));
+  uint8_t buffer[i::JSTypedArray::kMaxSizeInHeap];
+  v8::MemorySpan<uint8_t> storage(buffer);
+  storage = view->GetContents(storage);
+  CHECK_EQ(view->ByteLength(), storage.size());
+  if (expected) {
+    CHECK_EQ(0, memcmp(storage.data(), expected, view->ByteLength()));
+  } else {
+    CHECK_EQ(0, storage.size());
+  }
+}
+}  // namespace
+
+TEST(ArrayBufferView_GetContentsSmallUint8) {
+  const char* source = "new Uint8Array([1, 2, 3, 4, 5, 6, 7, 8, 9])";
+  uint8_t expected[]{1, 2, 3, 4, 5, 6, 7, 8, 9};
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsLargeUint8) {
+  const char* source =
+      "let array = new Uint8Array(100);"
+      "for (let i = 0; i < 100; ++i) {"
+      "  array[i] = i;"
+      "}"
+      "array";
+  uint8_t expected[100];
+  for (uint8_t i = 0; i < 100; ++i) {
+    expected[i] = i;
+  }
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsUint8View) {
+  const char* source =
+      "let array = new Uint8Array(100);"
+      "for (let i = 0; i < 100; ++i) {"
+      "  array[i] = i;"
+      "}"
+      "new Uint8Array(array.buffer, 70, 9)";
+  uint8_t expected[]{70, 71, 72, 73, 74, 75, 76, 77, 78, 79};
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsSmallUint32) {
+  const char* source = "new Uint16Array([1, 2, 3, 4, 5, 6, 7, 8, 9])";
+  uint16_t expected[]{1, 2, 3, 4, 5, 6, 7, 8, 9};
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsLargeUint16) {
+  const char* source =
+      "let array = new Uint16Array(100);"
+      "for (let i = 0; i < 100; ++i) {"
+      "  array[i] = i;"
+      "}"
+      "array";
+  uint16_t expected[100];
+  for (uint16_t i = 0; i < 100; ++i) {
+    expected[i] = i;
+  }
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsUint16View) {
+  const char* source =
+      "let array = new Uint16Array(100);"
+      "for (let i = 0; i < 100; ++i) {"
+      "  array[i] = i;"
+      "}"
+      "new Uint16Array(array.buffer, 140, 9)";
+  uint16_t expected[]{70, 71, 72, 73, 74, 75, 76, 77, 78, 79};
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsSmallDataView) {
+  const char* source =
+      "let array = new Uint8Array([1, 2, 3, 4, 5, 6, 7, 8, 9]);"
+      "new DataView(array.buffer)";
+  uint8_t expected[]{1, 2, 3, 4, 5, 6, 7, 8, 9};
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsLargeDataView) {
+  const char* source =
+      "let array = new Uint8Array(100);"
+      "for (let i = 0; i < 100; ++i) {"
+      "  array[i] = i;"
+      "}"
+      "new DataView(array.buffer)";
+  uint8_t expected[100];
+  for (uint8_t i = 0; i < 100; ++i) {
+    expected[i] = i;
+  }
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsDataViewWithOffset) {
+  const char* source =
+      "let array = new Uint8Array(100);"
+      "for (let i = 0; i < 100; ++i) {"
+      "  array[i] = i;"
+      "}"
+      "new DataView(array.buffer, 70, 9)";
+  uint8_t expected[]{70, 71, 72, 73, 74, 75, 76, 77, 78, 79};
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsSmallResizableDataView) {
+  const char* source =
+      "let rsab = new ArrayBuffer(10, {maxByteLength: 20});"
+      "let array = new Uint8Array(rsab);"
+      "for (let i = 0; i < 10; ++i) {"
+      "  array[i] = i;"
+      "}"
+      "new DataView(rsab)";
+  uint8_t expected[]{0, 1, 2, 3, 4, 5, 6, 7, 8, 9};
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsResizableTypedArray) {
+  const char* source =
+      "let rsab = new ArrayBuffer(8, {maxByteLength: 8});"
+      "let array = new Uint8Array(rsab);"
+      "for (let i = 0; i < 8; ++i) {"
+      "  array[i] = i;"
+      "};"
+      "array";
+  uint8_t expected[]{0, 1, 2, 3, 4, 5, 6, 7};
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsLargeResizableDataView) {
+  const char* source =
+      "let rsab = new ArrayBuffer(100, {maxByteLength: 200});"
+      "let array = new Uint8Array(rsab);"
+      "for (let i = 0; i < 100; ++i) {"
+      "  array[i] = i;"
+      "}"
+      "new DataView(rsab)";
+  uint8_t expected[100];
+  for (uint8_t i = 0; i < 100; ++i) {
+    expected[i] = i;
+  }
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsResizableDataViewWithOffset) {
+  const char* source =
+      "let rsab = new ArrayBuffer(100, {maxByteLength: 200});"
+      "let array = new Uint8Array(rsab);"
+      "for (let i = 0; i < 100; ++i) {"
+      "  array[i] = i;"
+      "}"
+      "new DataView(rsab, 70, 9)";
+  uint8_t expected[]{70, 71, 72, 73, 74, 75, 76, 77, 78, 79};
+  TestArrayBufferViewGetContent(source, expected);
+}
+
+TEST(ArrayBufferView_GetContentsDetached) {
+  const char* source =
+      "let array = new Uint8Array(100);"
+      "for (let i = 0; i < 100; ++i) {"
+      "  array[i] = i;"
+      "}"
+      "const data_view = new DataView(array.buffer);"
+      "let buffer = array.buffer.transfer();"
+      "data_view";
+  TestArrayBufferViewGetContent(source, nullptr);
+}
+
+TEST(ArrayBufferView_GetContentsOutOfBounds) {
+  const char* source =
+      "let ab = new ArrayBuffer(100, {maxByteLength: 200}); "
+      "let ta = new Uint8Array(ab, 2); "
+      "ab.resize(1); "
+      "ta";
+  TestArrayBufferViewGetContent(source, nullptr);
+}
