@@ -26,7 +26,7 @@
 // OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 // Check that we can traverse very deep stacks of ConsStrings using
-// StringCharacterStram.  Check that Get(int) works on very deep stacks
+// StringCharacterStream.  Check that Get(int) works on very deep stacks
 // of ConsStrings.  These operations may not be very fast, but they
 // should be possible without getting errors due to too deep recursion.
 
@@ -56,7 +56,7 @@ class MyRandomNumberGenerator {
     Q[0] = seed;
     Q[1] = seed + phi;
     Q[2] = seed + phi + phi;
-    for (unsigned j = 3; j < kQSize; j++) {
+    for (uint32_t j = 3; j < kQSize; j++) {
       Q[j] = Q[j - 3] ^ Q[j - 2] ^ phi ^ j;
     }
   }
@@ -133,10 +133,10 @@ static void InitializeBuildingBlocks(Handle<String>* building_blocks,
   Isolate* isolate = CcTest::i_isolate();
   Factory* factory = isolate->factory();
   for (int i = 0; i < bb_length; i++) {
-    int len = rng->next(16);
-    int slice_head_chars = 0;
-    int slice_tail_chars = 0;
-    int slice_depth = 0;
+    uint32_t len = rng->next(16);
+    uint32_t slice_head_chars = 0;
+    uint32_t slice_tail_chars = 0;
+    uint32_t slice_depth = 0;
     for (int j = 0; j < 3; j++) {
       if (rng->next(0.35)) slice_depth++;
     }
@@ -154,12 +154,12 @@ static void InitializeBuildingBlocks(Handle<String>* building_blocks,
     }
     // Don't slice 0 length strings.
     if (len == 0) slice_depth = 0;
-    int slice_length = slice_depth * (slice_head_chars + slice_tail_chars);
+    uint32_t slice_length = slice_depth * (slice_head_chars + slice_tail_chars);
     len += slice_length;
     switch (rng->next(4)) {
       case 0: {
         base::uc16 buf[2000];
-        for (int j = 0; j < len; j++) {
+        for (uint32_t j = 0; j < len; j++) {
           buf[j] = rng->next(0x10000);
         }
         building_blocks[i] =
@@ -167,48 +167,48 @@ static void InitializeBuildingBlocks(Handle<String>* building_blocks,
                 ->NewStringFromTwoByte(
                     v8::base::Vector<const base::uc16>(buf, len))
                 .ToHandleChecked();
-        for (int j = 0; j < len; j++) {
+        for (uint32_t j = 0; j < len; j++) {
           CHECK_EQ(buf[j], building_blocks[i]->Get(j));
         }
         break;
       }
       case 1: {
         char buf[2000];
-        for (int j = 0; j < len; j++) {
+        for (uint32_t j = 0; j < len; j++) {
           buf[j] = rng->next(0x80);
         }
         building_blocks[i] =
             factory->NewStringFromOneByte(v8::base::OneByteVector(buf, len))
                 .ToHandleChecked();
-        for (int j = 0; j < len; j++) {
+        for (uint32_t j = 0; j < len; j++) {
           CHECK_EQ(buf[j], building_blocks[i]->Get(j));
         }
         break;
       }
       case 2: {
         base::uc16* buf = NewArray<base::uc16>(len);
-        for (int j = 0; j < len; j++) {
+        for (uint32_t j = 0; j < len; j++) {
           buf[j] = rng->next(0x10000);
         }
         Resource* resource = new Resource(buf, len);
         building_blocks[i] = v8::Utils::OpenHandle(
             *v8::String::NewExternalTwoByte(CcTest::isolate(), resource)
                  .ToLocalChecked());
-        for (int j = 0; j < len; j++) {
+        for (uint32_t j = 0; j < len; j++) {
           CHECK_EQ(buf[j], building_blocks[i]->Get(j));
         }
         break;
       }
       case 3: {
         char* buf = NewArray<char>(len);
-        for (int j = 0; j < len; j++) {
+        for (uint32_t j = 0; j < len; j++) {
           buf[j] = rng->next(0x80);
         }
         OneByteResource* resource = new OneByteResource(buf, len);
         building_blocks[i] = v8::Utils::OpenHandle(
             *v8::String::NewExternalOneByte(CcTest::isolate(), resource)
                  .ToLocalChecked());
-        for (int j = 0; j < len; j++) {
+        for (uint32_t j = 0; j < len; j++) {
           CHECK_EQ(buf[j], building_blocks[i]->Get(j));
         }
         break;
@@ -308,14 +308,14 @@ void ConsStringGenerationData::Reset() {
 }
 
 void AccumulateStats(Tagged<ConsString> cons_string, ConsStringStats* stats) {
-  int left_length = cons_string->first()->length();
-  int right_length = cons_string->second()->length();
+  uint32_t left_length = cons_string->first()->length();
+  uint32_t right_length = cons_string->second()->length();
   CHECK(cons_string->length() == left_length + right_length);
   // Check left side.
   bool left_is_cons = IsConsString(cons_string->first());
   if (left_is_cons) {
     stats->left_traversals_++;
-    AccumulateStats(ConsString::cast(cons_string->first()), stats);
+    AccumulateStats(Cast<ConsString>(cons_string->first()), stats);
   } else {
     CHECK_NE(left_length, 0);
     stats->leaves_++;
@@ -324,7 +324,7 @@ void AccumulateStats(Tagged<ConsString> cons_string, ConsStringStats* stats) {
   // Check right side.
   if (IsConsString(cons_string->second())) {
     stats->right_traversals_++;
-    AccumulateStats(ConsString::cast(cons_string->second()), stats);
+    AccumulateStats(Cast<ConsString>(cons_string->second()), stats);
   } else {
     if (right_length == 0) {
       stats->empty_leaves_++;
@@ -335,10 +335,10 @@ void AccumulateStats(Tagged<ConsString> cons_string, ConsStringStats* stats) {
   }
 }
 
-void AccumulateStats(Handle<String> cons_string, ConsStringStats* stats) {
+void AccumulateStats(DirectHandle<String> cons_string, ConsStringStats* stats) {
   DisallowGarbageCollection no_gc;
   if (IsConsString(*cons_string)) {
-    return AccumulateStats(ConsString::cast(*cons_string), stats);
+    return AccumulateStats(Cast<ConsString>(*cons_string), stats);
   }
   // This string got flattened by gc.
   stats->chars_ += cons_string->length();
@@ -357,17 +357,18 @@ void AccumulateStatsWithOperator(Tagged<ConsString> cons_string,
   }
 }
 
-void VerifyConsString(Handle<String> root, ConsStringGenerationData* data) {
+void VerifyConsString(DirectHandle<String> root,
+                      ConsStringGenerationData* data) {
   // Verify basic data.
   CHECK(IsConsString(*root));
   CHECK_EQ(root->length(), data->stats_.chars_);
   // Recursive verify.
   ConsStringStats stats;
-  AccumulateStats(ConsString::cast(*root), &stats);
+  AccumulateStats(Cast<ConsString>(*root), &stats);
   stats.VerifyEqual(data->stats_);
   // Iteratively verify.
   stats.Reset();
-  AccumulateStatsWithOperator(ConsString::cast(*root), &stats);
+  AccumulateStatsWithOperator(Cast<ConsString>(*root), &stats);
   // Don't see these. Must copy over.
   stats.empty_leaves_ = data->stats_.empty_leaves_;
   stats.left_traversals_ = data->stats_.left_traversals_;
@@ -502,7 +503,7 @@ static Handle<String> ConstructBalanced(ConsStringGenerationData* data,
   return string;
 }
 
-static void Traverse(Handle<String> s1, Handle<String> s2) {
+static void Traverse(DirectHandle<String> s1, DirectHandle<String> s2) {
   int i = 0;
   StringCharacterStream character_stream_1(*s1);
   StringCharacterStream character_stream_2(*s2);
@@ -518,7 +519,8 @@ static void Traverse(Handle<String> s1, Handle<String> s2) {
   CHECK_EQ(s2->length(), i);
 }
 
-static void TraverseFirst(Handle<String> s1, Handle<String> s2, int chars) {
+static void TraverseFirst(DirectHandle<String> s1, DirectHandle<String> s2,
+                          int chars) {
   int i = 0;
   StringCharacterStream character_stream_1(*s1);
   StringCharacterStream character_stream_2(*s2);
@@ -551,7 +553,7 @@ TEST(Traverse) {
   Traverse(flat, right_asymmetric);
   printf("4\n");
   Handle<String> left_deep_asymmetric = ConstructLeft(&data, SUPER_DEEP_DEPTH);
-  Handle<String> right_deep_asymmetric =
+  DirectHandle<String> right_deep_asymmetric =
       ConstructRight(&data, SUPER_DEEP_DEPTH);
   printf("5\n");
   TraverseFirst(left_asymmetric, left_deep_asymmetric, 1050);
@@ -588,25 +590,25 @@ TEST(ConsStringWithEmptyFirstFlatten) {
                                  ->NewConsString(initial_fst, initial_snd)
                                  .ToHandleChecked();
   CHECK(IsConsString(*str));
-  auto cons = i::Handle<i::ConsString>::cast(str);
+  auto cons = i::Cast<i::ConsString>(str);
 
   const int initial_length = cons->length();
 
   // set_first / set_second does not update the length (which the heap verifier
   // checks), so we need to ensure the length stays the same.
 
-  i::Handle<i::String> new_fst = isolate->factory()->empty_string();
-  i::Handle<i::String> new_snd =
+  i::DirectHandle<i::String> new_fst = isolate->factory()->empty_string();
+  i::DirectHandle<i::String> new_snd =
       isolate->factory()->NewStringFromAsciiChecked("snd012345012345678");
   cons->set_first(*new_fst);
   cons->set_second(*new_snd);
-  CHECK(!cons->IsFlat(GetPtrComprCageBase(*cons)));
+  CHECK(!cons->IsFlat());
   CHECK_EQ(initial_length, new_fst->length() + new_snd->length());
   CHECK_EQ(initial_length, cons->length());
 
   // Make sure Flatten doesn't alloc a new string.
   DisallowGarbageCollection no_alloc;
-  i::Handle<i::String> flat = i::String::Flatten(isolate, cons);
+  i::DirectHandle<i::String> flat = i::String::Flatten(isolate, cons);
   CHECK(flat->IsFlat());
   CHECK_EQ(initial_length, flat->length());
 }
@@ -661,13 +663,13 @@ void TestStringCharacterStream(BuildString build, int test_cases) {
     HandleScope inner_scope(isolate);
     AlwaysAllocateScopeForTesting always_allocate(isolate->heap());
     // Build flat version of cons string.
-    Handle<String> flat_string = build(i, &data);
+    DirectHandle<String> flat_string = build(i, &data);
     ConsStringStats flat_string_stats;
     AccumulateStats(flat_string, &flat_string_stats);
     // Flatten string.
     String::Flatten(isolate, flat_string);
     // Build unflattened version of cons string to test.
-    Handle<String> cons_string = build(i, &data);
+    DirectHandle<String> cons_string = build(i, &data);
     ConsStringStats cons_string_stats;
     AccumulateStats(cons_string, &cons_string_stats);
     DisallowGarbageCollection no_gc;
@@ -680,7 +682,7 @@ void TestStringCharacterStream(BuildString build, int test_cases) {
     static_assert(kTaggedCanConvertToRawObjects);
     Tagged<String> flat_string_ptr =
         IsConsString(*flat_string)
-            ? Tagged(Tagged<ConsString>::cast(*flat_string)->first())
+            ? Tagged(Cast<ConsString>(*flat_string)->first())
             : *flat_string;
     VerifyCharacterStream(flat_string_ptr, *cons_string);
   }
@@ -688,8 +690,8 @@ void TestStringCharacterStream(BuildString build, int test_cases) {
 
 static const int kCharacterStreamNonRandomCases = 8;
 
-static Handle<String> BuildEdgeCaseConsString(int test_case,
-                                              ConsStringGenerationData* data) {
+static DirectHandle<String> BuildEdgeCaseConsString(
+    int test_case, ConsStringGenerationData* data) {
   Isolate* isolate = CcTest::i_isolate();
   Factory* factory = isolate->factory();
   data->Reset();
@@ -833,8 +835,8 @@ static void InitializeGenerationData(int test_case,
       0.03 * static_cast<double>(test_case % kEmptyLeaves);
 }
 
-static Handle<String> BuildRandomConsString(int test_case,
-                                            ConsStringGenerationData* data) {
+static DirectHandle<String> BuildRandomConsString(
+    int test_case, ConsStringGenerationData* data) {
   InitializeGenerationData(test_case, data);
   return ConstructRandomString(data, 200);
 }
@@ -880,11 +882,12 @@ TEST(Utf8Conversion) {
   v8::HandleScope handle_scope(CcTest::isolate());
   // A simple one-byte string
   const char* one_byte_string = "abcdef12345";
-  int len = v8::String::NewFromUtf8(CcTest::isolate(), one_byte_string,
-                                    v8::NewStringType::kNormal,
-                                    static_cast<int>(strlen(one_byte_string)))
-                .ToLocalChecked()
-                ->Utf8Length(CcTest::isolate());
+  size_t len =
+      v8::String::NewFromUtf8(CcTest::isolate(), one_byte_string,
+                              v8::NewStringType::kNormal,
+                              static_cast<int>(strlen(one_byte_string)))
+          .ToLocalChecked()
+          ->Utf8LengthV2(CcTest::isolate());
   CHECK_EQ(strlen(one_byte_string), len);
   // A mixed one-byte and two-byte string
   // U+02E4 -> CB A4
@@ -894,32 +897,28 @@ TEST(Utf8Conversion) {
   // U+3045 -> E3 81 85
   const uint16_t mixed_string[] = {0x02E4, 0x0064, 0x12E4, 0x0030, 0x3045};
   // The characters we expect to be output
-  const unsigned char as_utf8[11] = {0xCB, 0xA4, 0x64, 0xE1, 0x8B, 0xA4,
-                                     0x30, 0xE3, 0x81, 0x85, 0x00};
+  const unsigned char as_utf8[10] = {0xCB, 0xA4, 0x64, 0xE1, 0x8B,
+                                     0xA4, 0x30, 0xE3, 0x81, 0x85};
   // The number of bytes expected to be written for each length
-  const int lengths[12] = {0, 0, 2, 3, 3, 3, 6, 7, 7, 7, 10, 11};
-  const int char_lengths[12] = {0, 0, 1, 2, 2, 2, 3, 4, 4, 4, 5, 5};
+  const uint32_t lengths[11] = {0, 0, 2, 3, 3, 3, 6, 7, 7, 7, 10};
   v8::Local<v8::String> mixed =
       v8::String::NewFromTwoByte(CcTest::isolate(), mixed_string,
                                  v8::NewStringType::kNormal, 5)
           .ToLocalChecked();
-  CHECK_EQ(10, mixed->Utf8Length(CcTest::isolate()));
+  CHECK_EQ(10, mixed->Utf8LengthV2(CcTest::isolate()));
   // Try encoding the string with all capacities
   char buffer[11];
   const char kNoChar = static_cast<char>(-1);
-  for (int i = 0; i <= 11; i++) {
+  for (int i = 0; i <= 10; i++) {
     // Clear the buffer before reusing it
     for (int j = 0; j < 11; j++) buffer[j] = kNoChar;
-    int chars_written;
-    int written =
-        mixed->WriteUtf8(CcTest::isolate(), buffer, i, &chars_written);
+    size_t written = mixed->WriteUtf8V2(CcTest::isolate(), buffer, i);
     CHECK_EQ(lengths[i], written);
-    CHECK_EQ(char_lengths[i], chars_written);
     // Check that the contents are correct
-    for (int j = 0; j < lengths[i]; j++)
+    for (uint32_t j = 0; j < lengths[i]; j++)
       CHECK_EQ(as_utf8[j], static_cast<unsigned char>(buffer[j]));
     // Check that the rest of the buffer hasn't been touched
-    for (int j = lengths[i]; j < 11; j++) CHECK_EQ(kNoChar, buffer[j]);
+    for (uint32_t j = lengths[i]; j < 11; j++) CHECK_EQ(kNoChar, buffer[j]);
   }
 }
 
@@ -940,21 +939,21 @@ TEST(Utf8ConversionPerf) {
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    ascii_string->WriteUtf8(CcTest::isolate(), buffer, size, nullptr);
+    ascii_string->WriteUtf8V2(CcTest::isolate(), buffer, size);
     printf("ascii string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    ascii_string->WriteUtf8(CcTest::isolate(), buffer, size, nullptr);
+    ascii_string->WriteUtf8V2(CcTest::isolate(), buffer, size);
     printf("ascii string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    ascii_string->WriteUtf8(CcTest::isolate(), buffer, 4 * size, nullptr);
+    ascii_string->WriteUtf8V2(CcTest::isolate(), buffer, 4 * size);
     printf("ascii string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
@@ -962,21 +961,21 @@ TEST(Utf8ConversionPerf) {
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    one_byte_string->WriteUtf8(CcTest::isolate(), buffer, size, nullptr);
+    one_byte_string->WriteUtf8V2(CcTest::isolate(), buffer, size);
     printf("one byte string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    one_byte_string->WriteUtf8(CcTest::isolate(), buffer, size, nullptr);
+    one_byte_string->WriteUtf8V2(CcTest::isolate(), buffer, size);
     printf("one byte string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    one_byte_string->WriteUtf8(CcTest::isolate(), buffer, 4 * size, nullptr);
+    one_byte_string->WriteUtf8V2(CcTest::isolate(), buffer, 4 * size);
     printf("one byte string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
@@ -984,21 +983,21 @@ TEST(Utf8ConversionPerf) {
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    two_byte_string->WriteUtf8(CcTest::isolate(), buffer, size, nullptr);
+    two_byte_string->WriteUtf8V2(CcTest::isolate(), buffer, size);
     printf("two byte string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    two_byte_string->WriteUtf8(CcTest::isolate(), buffer, size, nullptr);
+    two_byte_string->WriteUtf8V2(CcTest::isolate(), buffer, size);
     printf("two byte string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    two_byte_string->WriteUtf8(CcTest::isolate(), buffer, 4 * size, nullptr);
+    two_byte_string->WriteUtf8V2(CcTest::isolate(), buffer, 4 * size);
     printf("two byte string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
@@ -1006,21 +1005,21 @@ TEST(Utf8ConversionPerf) {
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    surrogate_string->WriteUtf8(CcTest::isolate(), buffer, size, nullptr);
+    surrogate_string->WriteUtf8V2(CcTest::isolate(), buffer, size);
     printf("surrogate string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    surrogate_string->WriteUtf8(CcTest::isolate(), buffer, size, nullptr);
+    surrogate_string->WriteUtf8V2(CcTest::isolate(), buffer, size);
     printf("surrogate string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
   {
     v8::base::ElapsedTimer timer;
     timer.Start();
-    surrogate_string->WriteUtf8(CcTest::isolate(), buffer, 4 * size, nullptr);
+    surrogate_string->WriteUtf8V2(CcTest::isolate(), buffer, 4 * size);
     printf("surrogate string %0.3f\n", timer.Elapsed().InMillisecondsF());
     timer.Stop();
   }
@@ -1142,19 +1141,16 @@ TEST(ReplaceInvalidUtf8) {
   v8::Local<v8::String> string = CompileRun("'ab\\ud800cd'").As<v8::String>();
   char buffer[7];
   memset(buffer, 0, 7);
-  int chars_written = 0;
-  int size = string->WriteUtf8(CcTest::isolate(), buffer, 7, &chars_written,
-                               v8::String::REPLACE_INVALID_UTF8);
+  size_t size =
+      string->WriteUtf8V2(CcTest::isolate(), buffer, 7,
+                          v8::String::WriteFlags::kReplaceInvalidUtf8);
   CHECK_EQ(7, size);
-  CHECK_EQ(5, chars_written);
   CHECK_EQ(0, memcmp("\x61\x62\xef\xbf\xbd\x63\x64", buffer, 7));
 
   memset(buffer, 0, 7);
-  chars_written = 0;
-  size = string->WriteUtf8(CcTest::isolate(), buffer, 6, &chars_written,
-                           v8::String::REPLACE_INVALID_UTF8);
+  size = string->WriteUtf8V2(CcTest::isolate(), buffer, 6,
+                             v8::String::WriteFlags::kReplaceInvalidUtf8);
   CHECK_EQ(6, size);
-  CHECK_EQ(4, chars_written);
   CHECK_EQ(0, memcmp("\x61\x62\xef\xbf\xbd\x63", buffer, 6));
 }
 
@@ -1309,17 +1305,17 @@ TEST(SliceFromCons) {
       factory->NewConsString(string, string).ToHandleChecked();
   CHECK(IsConsString(*parent));
   CHECK(!parent->IsFlat());
-  Handle<String> slice = factory->NewSubString(parent, 1, 25);
+  DirectHandle<String> slice = factory->NewSubString(parent, 1, 25);
   // After slicing, the original string becomes a flat cons.
   CHECK(parent->IsFlat());
   CHECK(IsSlicedString(*slice));
   // TODO(leszeks): Remove Tagged cast when .first() returns a Tagged.
   static_assert(kTaggedCanConvertToRawObjects);
-  CHECK_EQ(SlicedString::cast(*slice)->parent(),
+  CHECK_EQ(Cast<SlicedString>(*slice)->parent(),
            // Parent could have been short-circuited.
-           IsConsString(*parent) ? Tagged(ConsString::cast(*parent)->first())
+           IsConsString(*parent) ? Tagged(Cast<ConsString>(*parent)->first())
                                  : *parent);
-  CHECK(IsSeqString(SlicedString::cast(*slice)->parent()));
+  CHECK(IsSeqString(Cast<SlicedString>(*slice)->parent()));
   CHECK(slice->IsFlat());
 }
 
@@ -1337,6 +1333,7 @@ class OneByteVectorResource : public v8::String::ExternalOneByteStringResource {
 
 TEST(InternalizeExternal) {
   v8_flags.stress_incremental_marking = false;
+  ManualGCScope manual_gc_scope;
   CcTest::InitializeVM();
   i::Isolate* isolate = CcTest::i_isolate();
   Factory* factory = isolate->factory();
@@ -1351,14 +1348,14 @@ TEST(InternalizeExternal) {
     Handle<String> string = v8::Utils::OpenHandle(*ext_string);
     CHECK(IsExternalString(*string));
     CHECK(!IsInternalizedString(*string));
-    CHECK(!i::Heap::InYoungGeneration(*string));
+    CHECK(!i::HeapLayout::InYoungGeneration(*string));
     CHECK_EQ(isolate->string_table()->TryStringToIndexOrLookupExisting(
                  isolate, string->ptr()),
              Smi::FromInt(ResultSentinel::kNotFound).ptr());
     factory->InternalizeName(string);
     CHECK(IsExternalString(*string));
     CHECK(IsInternalizedString(*string));
-    CHECK(!i::Heap::InYoungGeneration(*string));
+    CHECK(!i::HeapLayout::InYoungGeneration(*string));
   }
   i::heap::InvokeMajorGC(CcTest::heap());
   i::heap::InvokeMajorGC(CcTest::heap());
@@ -1387,7 +1384,7 @@ TEST(Regress1402187) {
     v8::Local<v8::String> ext_string =
         Utils::ToLocal(factory->NewStringFromAsciiChecked(
             ext_string_content, AllocationType::kOld));
-    CHECK(ext_string->MakeExternal(resource));
+    CHECK(ext_string->MakeExternal(CcTest::isolate(), resource));
     Handle<String> string = v8::Utils::OpenHandle(*ext_string);
     string->set_raw_hash_field(fake_hash);
     CHECK(IsExternalString(*string));
@@ -1418,15 +1415,15 @@ TEST(SliceFromExternal) {
   Handle<String> string =
       factory->NewExternalStringFromOneByte(&resource).ToHandleChecked();
   CHECK(IsExternalString(*string));
-  Handle<String> slice = factory->NewSubString(string, 1, 25);
+  DirectHandle<String> slice = factory->NewSubString(string, 1, 25);
   CHECK(IsSlicedString(*slice));
   CHECK(IsExternalString(*string));
-  CHECK_EQ(SlicedString::cast(*slice)->parent(), *string);
-  CHECK(IsExternalString(SlicedString::cast(*slice)->parent()));
+  CHECK_EQ(Cast<SlicedString>(*slice)->parent(), *string);
+  CHECK(IsExternalString(Cast<SlicedString>(*slice)->parent()));
   CHECK(slice->IsFlat());
   // This avoids the GC from trying to free stack allocated resources.
-  i::Handle<i::ExternalOneByteString>::cast(string)->SetResource(
-      CcTest::i_isolate(), nullptr);
+  i::Cast<i::ExternalOneByteString>(string)->SetResource(CcTest::i_isolate(),
+                                                         nullptr);
 }
 
 static void ExternalizeDuringJsonStringifyCallback(
@@ -1434,11 +1431,11 @@ static void ExternalizeDuringJsonStringifyCallback(
   v8::Local<v8::Value> key = v8_compile("p")
                                  ->Run(CcTest::isolate()->GetCurrentContext())
                                  .ToLocalChecked();
-  const static char ext_string_content[] = "prop-1234567890asdf";
+  static const char ext_string_content[] = "prop-1234567890asdf";
   OneByteVectorResource* resource =
       new OneByteVectorResource(v8::base::Vector<const char>(
           ext_string_content, strlen(ext_string_content)));
-  CHECK(v8::String::Cast(*key)->MakeExternal(resource));
+  CHECK(v8::String::Cast(*key)->MakeExternal(CcTest::isolate(), resource));
 }
 
 TEST(ExternalizeDuringJsonStringify) {
@@ -1503,7 +1500,7 @@ TEST(SliceFromSlice) {
   CcTest::InitializeVM();
   v8::HandleScope scope(CcTest::isolate());
   v8::Local<v8::Value> result;
-  Handle<String> string;
+  DirectHandle<String> string;
   const char* init = "var str = 'abcdefghijklmnopqrstuvwxyz';";
   const char* slice = "var slice = ''; slice = str.slice(1,-1); slice";
   const char* slice_from_slice = "slice.slice(1,-1);";
@@ -1511,16 +1508,16 @@ TEST(SliceFromSlice) {
   CompileRun(init);
   result = CompileRun(slice);
   CHECK(result->IsString());
-  string = v8::Utils::OpenHandle(v8::String::Cast(*result));
+  string = v8::Utils::OpenDirectHandle(v8::String::Cast(*result));
   CHECK(IsSlicedString(*string));
-  CHECK(IsSeqString(SlicedString::cast(*string)->parent()));
+  CHECK(IsSeqString(Cast<SlicedString>(*string)->parent()));
   CHECK_EQ(0, strcmp("bcdefghijklmnopqrstuvwxy", string->ToCString().get()));
 
   result = CompileRun(slice_from_slice);
   CHECK(result->IsString());
-  string = v8::Utils::OpenHandle(v8::String::Cast(*result));
+  string = v8::Utils::OpenDirectHandle(v8::String::Cast(*result));
   CHECK(IsSlicedString(*string));
-  CHECK(IsSeqString(SlicedString::cast(*string)->parent()));
+  CHECK(IsSeqString(Cast<SlicedString>(*string)->parent()));
   CHECK_EQ(0, strcmp("cdefghijklmnopqrstuvwx", string->ToCString().get()));
 }
 
@@ -1602,7 +1599,8 @@ TEST(StringReplaceAtomTwoByteResult) {
       "var replace = '\x80';            "
       "subject.replace(/~/g, replace);  ");
   CHECK(result->IsString());
-  Handle<String> string = v8::Utils::OpenHandle(v8::String::Cast(*result));
+  DirectHandle<String> string =
+      v8::Utils::OpenDirectHandle(v8::String::Cast(*result));
   CHECK(string->IsTwoByteRepresentation());
 
   v8::Local<v8::String> expected = v8_str("one_byte\x80only\x80string\x80");
@@ -1634,11 +1632,25 @@ static void CheckCanonicalEquivalence(uint16_t c, uint16_t test) {
   CHECK_EQ(expect, test);
 }
 
+static inline uint16_t TryConvertToLatin1(uint16_t c) {
+  switch (c) {
+    // This are equivalent characters in unicode.
+    case 0x39c:
+    case 0x3bc:
+      return 0xb5;
+    // This is an uppercase of a Latin-1 character
+    // outside of Latin-1.
+    case 0x178:
+      return 0xff;
+  }
+  return c;
+}
+
 TEST(Latin1IgnoreCase) {
   for (uint16_t c = unibrow::Latin1::kMaxChar + 1; c != 0; c++) {
     uint16_t lower = ConvertLatin1<unibrow::ToLowercase, false>(c);
     uint16_t upper = ConvertLatin1<unibrow::ToUppercase, false>(c);
-    uint16_t test = unibrow::Latin1::TryConvertToLatin1(c);
+    uint16_t test = TryConvertToLatin1(c);
     // Filter out all character whose upper is not their lower or vice versa.
     if (lower == 0 && upper == 0) {
       CheckCanonicalEquivalence(c, test);
@@ -1689,16 +1701,16 @@ TEST(InvalidExternalString) {
     HandleScope scope(isolate);
     DummyOneByteResource r;
     CHECK(isolate->factory()->NewExternalStringFromOneByte(&r).is_null());
-    CHECK(isolate->has_pending_exception());
-    isolate->clear_pending_exception();
+    CHECK(isolate->has_exception());
+    isolate->clear_exception();
   }
 
   {
     HandleScope scope(isolate);
     DummyResource r;
     CHECK(isolate->factory()->NewExternalStringFromTwoByte(&r).is_null());
-    CHECK(isolate->has_pending_exception());
-    isolate->clear_pending_exception();
+    CHECK(isolate->has_exception());
+    isolate->clear_exception();
   }
 }
 
@@ -1712,12 +1724,10 @@ TEST(InvalidExternalString) {
     HandleScope scope(isolate);                                          \
     v8::base::Vector<TYPE> dummy = v8::base::Vector<TYPE>::New(invalid); \
     memset(dummy.begin(), 0x0, dummy.length() * sizeof(TYPE));           \
-    CHECK(isolate->factory()                                             \
-              ->FUN(v8::base::Vector<const TYPE>::cast(dummy))           \
-              .is_null());                                               \
+    CHECK(isolate->factory()->FUN(dummy).is_null());                     \
     memset(dummy.begin(), 0x20, dummy.length() * sizeof(TYPE));          \
-    CHECK(isolate->has_pending_exception());                             \
-    isolate->clear_pending_exception();                                  \
+    CHECK(isolate->has_exception());                                     \
+    isolate->clear_exception();                                          \
     dummy.Dispose();                                                     \
   }
 
@@ -1731,14 +1741,18 @@ TEST(FormatMessage) {
   LocalContext context;
   Isolate* isolate = CcTest::i_isolate();
   HandleScope scope(isolate);
-  Handle<String> arg0 = isolate->factory()->NewStringFromAsciiChecked("arg0");
-  Handle<String> arg1 = isolate->factory()->NewStringFromAsciiChecked("arg1");
-  Handle<String> arg2 = isolate->factory()->NewStringFromAsciiChecked("arg2");
-  Handle<String> result =
-      MessageFormatter::TryFormat(
-          isolate, MessageTemplate::kPropertyNotFunction, arg0, arg1, arg2)
+  DirectHandle<String> arg0 =
+      isolate->factory()->NewStringFromAsciiChecked("arg0");
+  DirectHandle<String> arg1 =
+      isolate->factory()->NewStringFromAsciiChecked("arg1");
+  DirectHandle<String> arg2 =
+      isolate->factory()->NewStringFromAsciiChecked("arg2");
+  DirectHandle<String> result =
+      MessageFormatter::TryFormat(isolate,
+                                  MessageTemplate::kPropertyNotFunction,
+                                  base::VectorOf({arg0, arg1, arg2}))
           .ToHandleChecked();
-  Handle<String> expected = isolate->factory()->NewStringFromAsciiChecked(
+  DirectHandle<String> expected = isolate->factory()->NewStringFromAsciiChecked(
       "'arg0' returned for property 'arg1' of object 'arg2' is not a function");
   CHECK(String::Equals(isolate, result, expected));
 }
@@ -1753,7 +1767,7 @@ TEST(Regress609831) {
         "String.fromCharCode(32, 32, 32, 32, 32, "
         "32, 32, 32, 32, 32, 32, 32, 32, 32, 32, "
         "32, 32, 32, 32, 32, 32, 32, 32, 32, 32)");
-    CHECK(IsSeqOneByteString(*v8::Utils::OpenHandle(*result)));
+    CHECK(IsSeqOneByteString(*v8::Utils::OpenDirectHandle(*result)));
   }
   {
     HandleScope scope(isolate);
@@ -1761,7 +1775,7 @@ TEST(Regress609831) {
         "String.fromCharCode(432, 432, 432, 432, 432, "
         "432, 432, 432, 432, 432, 432, 432, 432, 432, "
         "432, 432, 432, 432, 432, 432, 432, 432, 432)");
-    CHECK(IsSeqTwoByteString(*v8::Utils::OpenHandle(*result)));
+    CHECK(IsSeqTwoByteString(*v8::Utils::OpenDirectHandle(*result)));
   }
 }
 
@@ -1811,7 +1825,8 @@ struct IndexData {
 };
 
 void TestString(i::Isolate* isolate, const IndexData& data) {
-  Handle<String> s = isolate->factory()->NewStringFromAsciiChecked(data.string);
+  DirectHandle<String> s =
+      isolate->factory()->NewStringFromAsciiChecked(data.string);
   if (data.is_array_index) {
     uint32_t index;
     CHECK(s->AsArrayIndex(&index));
@@ -1916,16 +1931,16 @@ TEST(InternalizeExternalString) {
   const char* raw_string = "external";
   OneByteResource* resource =
       new OneByteResource(i::StrDup(raw_string), strlen(raw_string));
-  Handle<String> string =
+  DirectHandle<String> string =
       factory->NewExternalStringFromOneByte(resource).ToHandleChecked();
   CHECK(IsExternalString(*string));
 
   // Check it is not uncached.
-  Handle<ExternalString> external = Handle<ExternalString>::cast(string);
+  DirectHandle<ExternalString> external = Cast<ExternalString>(string);
   CHECK(!external->is_uncached());
 
-  // Internalize succesfully, without a copy.
-  Handle<String> internal = factory->InternalizeString(external);
+  // Internalize successfully, without a copy.
+  DirectHandle<String> internal = factory->InternalizeString(external);
   CHECK(IsInternalizedString(*string));
   CHECK(string.equals(internal));
 }
@@ -1941,16 +1956,16 @@ TEST(InternalizeExternalStringTwoByte) {
   const char* raw_string = "external";
   Resource* resource =
       new Resource(AsciiToTwoByteString(raw_string), strlen(raw_string));
-  Handle<String> string =
+  DirectHandle<String> string =
       factory->NewExternalStringFromTwoByte(resource).ToHandleChecked();
   CHECK(IsExternalString(*string));
 
   // Check it is not uncached.
-  Handle<ExternalString> external = Handle<ExternalString>::cast(string);
+  DirectHandle<ExternalString> external = Cast<ExternalString>(string);
   CHECK(!external->is_uncached());
 
-  // Internalize succesfully, without a copy.
-  Handle<String> internal = factory->InternalizeString(external);
+  // Internalize successfully, without a copy.
+  DirectHandle<String> internal = factory->InternalizeString(external);
   CHECK(IsInternalizedString(*string));
   CHECK(string.equals(internal));
 }
@@ -1987,11 +2002,11 @@ TEST(InternalizeExternalStringUncachedWithCopy) {
   CHECK(IsExternalString(*string));
 
   // Check it is uncached.
-  Handle<ExternalString> external = Handle<ExternalString>::cast(string);
+  Handle<ExternalString> external = Cast<ExternalString>(string);
   CHECK(external->is_uncached());
 
-  // Internalize succesfully, with a copy.
-  Handle<String> internal = factory->InternalizeString(external);
+  // Internalize successfully, with a copy.
+  DirectHandle<String> internal = factory->InternalizeString(external);
   CHECK(!IsInternalizedString(*external));
   CHECK(IsInternalizedString(*internal));
 }
@@ -2030,12 +2045,12 @@ TEST(InternalizeExternalStringUncachedWithCopyTwoByte) {
   CHECK(IsExternalString(*string));
 
   // Check it is uncached.
-  Handle<ExternalString> external = Handle<ExternalString>::cast(string);
+  Handle<ExternalString> external = Cast<ExternalString>(string);
   CHECK(external->is_uncached());
 
-  // Internalize succesfully, with a copy.
+  // Internalize successfully, with a copy.
   CHECK(!IsInternalizedString(*external));
-  Handle<String> internal = factory->InternalizeString(external);
+  DirectHandle<String> internal = factory->InternalizeString(external);
   CHECK(!IsInternalizedString(*external));
   CHECK(IsInternalizedString(*internal));
 }
@@ -2057,21 +2072,21 @@ TEST(CheckCachedDataInternalExternalUncachedString) {
   const char* raw_small = "smalls";
 #endif  // V8_COMPRESS_POINTERS
 
-  Handle<String> string =
+  DirectHandle<String> string =
       factory->InternalizeString(factory->NewStringFromAsciiChecked(raw_small));
   OneByteResource* resource =
       new OneByteResource(i::StrDup(raw_small), strlen(raw_small));
 
   // Check it is external, internalized, and uncached with a cacheable resource.
-  string->MakeExternal(resource);
+  string->MakeExternal(CcTest::i_isolate(), resource);
   CHECK(string->IsOneByteRepresentation());
   CHECK(IsExternalString(*string));
   CHECK(IsInternalizedString(*string));
 
   // Check that the external string is uncached, its resource is cacheable, and
   // that we indeed cached it.
-  Handle<ExternalOneByteString> external_string =
-      Handle<ExternalOneByteString>::cast(string);
+  DirectHandle<ExternalOneByteString> external_string =
+      Cast<ExternalOneByteString>(string);
   // If the sandbox is enabled, string objects will always be cacheable because
   // they are smaller.
   CHECK(V8_ENABLE_SANDBOX_BOOL || external_string->is_uncached());
@@ -2102,21 +2117,21 @@ TEST(CheckCachedDataInternalExternalUncachedStringTwoByte) {
 
   size_t len;
   const uint16_t* two_byte = AsciiToTwoByteString(raw_small, &len);
-  Handle<String> string = factory->InternalizeString(
+  DirectHandle<String> string = factory->InternalizeString(
       factory->NewStringFromTwoByte(base::VectorOf(two_byte, len))
           .ToHandleChecked());
   Resource* resource = new Resource(two_byte, len);
 
   // Check it is external, internalized, and uncached with a cacheable resource.
-  string->MakeExternal(resource);
+  string->MakeExternal(CcTest::i_isolate(), resource);
   CHECK(string->IsTwoByteRepresentation());
   CHECK(IsExternalString(*string));
   CHECK(IsInternalizedString(*string));
 
   // Check that the external string is uncached, its resource is cacheable, and
   // that we indeed cached it.
-  Handle<ExternalTwoByteString> external_string =
-      Handle<ExternalTwoByteString>::cast(string);
+  DirectHandle<ExternalTwoByteString> external_string =
+      Cast<ExternalTwoByteString>(string);
   // If the sandbox is enabled, string objects will always be cacheable because
   // they are smaller.
   CHECK(V8_ENABLE_SANDBOX_BOOL || external_string->is_uncached());
@@ -2128,6 +2143,85 @@ TEST(CheckCachedDataInternalExternalUncachedStringTwoByte) {
   }
 }
 
+TEST(CheckIntlSegmentIteratorTerminateExecutionInterrupt) {
+#if V8_INTL_SUPPORT
+  class WorkerThread : public v8::base::Thread {
+   public:
+    WorkerThread(v8::base::Mutex& m, v8::base::ConditionVariable& cv)
+        : Thread(v8::base::Thread::Options("WorkerThread")), m_(m), cv_(cv) {}
+    void Run() override {
+      v8::Isolate::CreateParams create_params;
+      create_params.array_buffer_allocator = CcTest::array_buffer_allocator();
+      isolate = v8::Isolate::New(create_params);
+      {
+        v8::Isolate::Scope isolate_scope(isolate);
+        v8::HandleScope handle_scope(isolate);
+        v8::Local<v8::ObjectTemplate> global = ObjectTemplate::New(isolate);
+        v8::Local<v8::Value> wrapper = v8::External::New(isolate, this);
+        global->Set(isolate, "notifyCV",
+                    v8::FunctionTemplate::New(
+                        isolate, (v8::FunctionCallback)&NotifyCallback, wrapper,
+                        Local<v8::Signature>(), 0, ConstructorBehavior::kThrow,
+                        SideEffectType::kHasNoSideEffect));
+        LocalContext context(isolate, nullptr, global);
+        v8::TryCatch try_catch(isolate);
+        auto result = CompileRun(
+            context.local(),
+            "const kSize = 4 * 1024 * 1024;\n"
+            "const baseText = 'Super big, super bad, itty bitty teeny weeny "
+            "mellow yellow stringy wingy. ';\n"
+            "let text = baseText.repeat(((kSize / baseText.length) + 1) | 0);\n"
+            "let iter = new Intl.Segmenter('en', { granularity: 'word' "
+            "}).segment(text);\n"
+            "notifyCV(); /* Signal CV that we are ready for interrupt */\n"
+            "Array.from(iter)\n");
+        CHECK(result.IsEmpty());
+        CHECK(try_catch.HasTerminated());
+      }
+      isolate->Dispose();
+    }
+    void TerminateExecution() { isolate->TerminateExecution(); }
+    inline void NotifyCV() {
+      v8::base::MutexGuard guard(&m_);
+      cv_.NotifyOne();
+    }
+
+    bool DidEnterLoop() const { return did_enter_loop_; }
+
+   private:
+    static WorkerThread* Unwrap(Local<Value> value) {
+      CHECK(value->IsExternal());
+      return reinterpret_cast<WorkerThread*>(value.As<External>()->Value());
+    }
+    static void NotifyCallback(
+        const v8::FunctionCallbackInfo<v8::Value>& args) {
+      auto self = Unwrap(args.Data());
+      {
+        v8::base::MutexGuard guard(self->m_);
+        self->did_enter_loop_ = true;
+      }
+      self->NotifyCV();
+    }
+    bool did_enter_loop_{false};
+    v8::Isolate* isolate{nullptr};
+    v8::base::Mutex& m_;
+    v8::base::ConditionVariable& cv_;
+  };
+  v8::base::Mutex m;
+  v8::base::ConditionVariable cv;
+  WorkerThread worker_thread(m, cv);
+  CHECK(worker_thread.Start());
+  {
+    v8::base::MutexGuard guard(&m);
+    if (!worker_thread.DidEnterLoop()) {
+      cv.Wait(&m);
+    }
+  }
+  worker_thread.TerminateExecution();
+  worker_thread.Join();
+#endif
+}
+
 }  // namespace test_strings
 }  // namespace internal
 }  // namespace v8
