@@ -8,8 +8,7 @@
 #include "src/base/utils/random-number-generator.h"
 #include "src/builtins/builtins-promise-gen.h"
 #include "src/builtins/builtins-string-gen.h"
-#include "src/codegen/code-factory.h"
-#include "src/codegen/code-stub-assembler.h"
+#include "src/codegen/code-stub-assembler-inl.h"
 #include "src/compiler/node.h"
 #include "src/debug/debug.h"
 #include "src/execution/isolate.h"
@@ -19,12 +18,15 @@
 #include "src/objects/torque-defined-classes-inl.h"
 #include "src/strings/char-predicates.h"
 #include "test/cctest/compiler/function-tester.h"
+#include "test/cctest/heap/heap-utils.h"
 #include "test/common/code-assembler-tester.h"
 
 namespace v8 {
 namespace internal {
 namespace compiler {
 
+#include "src/codegen/define-code-stub-assembler-macros.inc"
+
 namespace {
 
 class TestTorqueAssembler : public CodeStubAssembler {
@@ -163,12 +165,12 @@ TEST(TestTernaryOperator) {
     m.Return(m.TestTernaryOperator(arg));
   }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
-  Handle<Object> result1 =
+  DirectHandle<Object> result1 =
       ft.Call(Handle<Smi>(Smi::FromInt(-5), isolate)).ToHandleChecked();
-  CHECK_EQ(-15, Smi::cast(*result1).value());
-  Handle<Object> result2 =
+  CHECK_EQ(-15, Cast<Smi>(*result1).value());
+  DirectHandle<Object> result2 =
       ft.Call(Handle<Smi>(Smi::FromInt(3), isolate)).ToHandleChecked();
-  CHECK_EQ(103, Smi::cast(*result2).value());
+  CHECK_EQ(103, Cast<Smi>(*result2).value());
 }
 
 TEST(TestFunctionPointerToGeneric) {
@@ -261,7 +263,7 @@ TEST(TestTypeswitch) {
   CodeAssemblerTester asm_tester(isolate, JSParameterCount(kNumParams));
   TestTorqueAssembler m(asm_tester.state());
   {
-    m.TestTypeswitch(m.UncheckedCast<Context>(m.HeapConstant(context)));
+    m.TestTypeswitch(m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
@@ -278,7 +280,8 @@ TEST(TestGenericOverload) {
   CodeAssemblerTester asm_tester(isolate, JSParameterCount(kNumParams));
   TestTorqueAssembler m(asm_tester.state());
   {
-    m.TestGenericOverload(m.UncheckedCast<Context>(m.HeapConstant(context)));
+    m.TestGenericOverload(
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
@@ -295,7 +298,7 @@ TEST(TestEquality) {
   CodeAssemblerTester asm_tester(isolate, JSParameterCount(kNumParams));
   TestTorqueAssembler m(asm_tester.state());
   {
-    m.TestEquality(m.UncheckedCast<Context>(m.HeapConstant(context)));
+    m.TestEquality(m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
@@ -342,7 +345,7 @@ TEST(TestCatch1) {
   TestTorqueAssembler m(asm_tester.state());
   {
     TNode<Smi> result =
-        m.TestCatch1(m.UncheckedCast<Context>(m.HeapConstant(context)));
+        m.TestCatch1(m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     USE(result);
     CSA_DCHECK(&m, m.TaggedEqual(result, m.SmiConstant(1)));
     m.Return(m.UndefinedConstant());
@@ -362,7 +365,7 @@ TEST(TestCatch2) {
   TestTorqueAssembler m(asm_tester.state());
   {
     TNode<Smi> result =
-        m.TestCatch2(m.UncheckedCast<Context>(m.HeapConstant(context)));
+        m.TestCatch2(m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     USE(result);
     CSA_DCHECK(&m, m.TaggedEqual(result, m.SmiConstant(2)));
     m.Return(m.UndefinedConstant());
@@ -382,7 +385,7 @@ TEST(TestCatch3) {
   TestTorqueAssembler m(asm_tester.state());
   {
     TNode<Smi> result =
-        m.TestCatch3(m.UncheckedCast<Context>(m.HeapConstant(context)));
+        m.TestCatch3(m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     USE(result);
     CSA_DCHECK(&m, m.TaggedEqual(result, m.SmiConstant(2)));
     m.Return(m.UndefinedConstant());
@@ -401,7 +404,8 @@ TEST(TestLookup) {
   CodeAssemblerTester asm_tester(isolate, JSParameterCount(kNumParams));
   TestTorqueAssembler m(asm_tester.state());
   {
-    m.TestQualifiedAccess(m.UncheckedCast<Context>(m.HeapConstant(context)));
+    m.TestQualifiedAccess(
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
@@ -417,7 +421,7 @@ TEST(TestFrame1) {
   CodeAssemblerTester asm_tester(isolate);
   TestTorqueAssembler m(asm_tester.state());
   {
-    m.TestFrame1(m.UncheckedCast<Context>(m.HeapConstant(context)));
+    m.TestFrame1(m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), 0);
@@ -433,7 +437,7 @@ TEST(TestNew) {
   CodeAssemblerTester asm_tester(isolate);
   TestTorqueAssembler m(asm_tester.state());
   {
-    m.TestNew(m.UncheckedCast<Context>(m.HeapConstant(context)));
+    m.TestNew(m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), 0);
@@ -449,7 +453,8 @@ TEST(TestStructConstructor) {
   CodeAssemblerTester asm_tester(isolate);
   TestTorqueAssembler m(asm_tester.state());
   {
-    m.TestStructConstructor(m.UncheckedCast<Context>(m.HeapConstant(context)));
+    m.TestStructConstructor(
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), 0);
@@ -465,7 +470,8 @@ TEST(TestInternalClass) {
   CodeAssemblerTester asm_tester(isolate);
   TestTorqueAssembler m(asm_tester.state());
   {
-    m.TestInternalClass(m.UncheckedCast<Context>(m.HeapConstant(context)));
+    m.TestInternalClass(
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), 0);
@@ -482,7 +488,7 @@ TEST(TestNewFixedArrayFromSpread) {
   TestTorqueAssembler m(asm_tester.state());
   {
     m.TestNewFixedArrayFromSpread(
-        m.UncheckedCast<Context>(m.HeapConstant(context)));
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), 0);
@@ -526,7 +532,8 @@ TEST(TestSliceEnumeration) {
   CodeAssemblerTester asm_tester(isolate);
   TestTorqueAssembler m(asm_tester.state());
   {
-    m.TestSliceEnumeration(m.UncheckedCast<Context>(m.HeapConstant(context)));
+    m.TestSliceEnumeration(
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), 0);
@@ -557,7 +564,7 @@ TEST(TestLoadEliminationFixed) {
   TestTorqueAssembler m(asm_tester.state());
   {
     m.TestLoadEliminationFixed(
-        m.UncheckedCast<Context>(m.HeapConstant(context)));
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   asm_tester.GenerateCode();
@@ -573,7 +580,7 @@ TEST(TestLoadEliminationVariable) {
   TestTorqueAssembler m(asm_tester.state());
   {
     m.TestLoadEliminationVariable(
-        m.UncheckedCast<Context>(m.HeapConstant(context)));
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   asm_tester.GenerateCode();
@@ -589,7 +596,7 @@ TEST(TestRedundantArrayElementCheck) {
   TestTorqueAssembler m(asm_tester.state());
   {
     m.Return(m.TestRedundantArrayElementCheck(
-        m.UncheckedCast<Context>(m.HeapConstant(context))));
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context))));
   }
   asm_tester.GenerateCode();
 }
@@ -604,7 +611,7 @@ TEST(TestRedundantSmiCheck) {
   TestTorqueAssembler m(asm_tester.state());
   {
     m.Return(m.TestRedundantSmiCheck(
-        m.UncheckedCast<Context>(m.HeapConstant(context))));
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context))));
   }
   asm_tester.GenerateCode();
 }
@@ -645,7 +652,7 @@ TEST(TestBranchOnBoolOptimization) {
   TestTorqueAssembler m(asm_tester.state());
   {
     m.TestBranchOnBoolOptimization(
-        m.UncheckedCast<Context>(m.HeapConstant(context)),
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)),
         m.UncheckedParameter<Smi>(0));
     m.Return(m.UndefinedConstant());
   }
@@ -809,7 +816,7 @@ TEST(TestTestParentFrameArguments) {
   TestTorqueAssembler m(asm_tester.state());
   {
     m.TestParentFrameArguments(
-        m.UncheckedCast<Context>(m.HeapConstant(context)));
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   asm_tester.GenerateCode();
@@ -824,8 +831,8 @@ TEST(TestFullyGeneratedClassFromCpp) {
   TestTorqueAssembler m(asm_tester.state());
   { m.Return(m.TestFullyGeneratedClassFromCpp()); }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
-  Handle<ExportedSubClass> result =
-      Handle<ExportedSubClass>::cast(ft.Call().ToHandleChecked());
+  DirectHandle<ExportedSubClass> result =
+      Cast<ExportedSubClass>(ft.Call().ToHandleChecked());
   CHECK_EQ(result->c_field(), 7);
   CHECK_EQ(result->d_field(), 8);
   CHECK_EQ(result->e_field(), 9);
@@ -842,7 +849,7 @@ TEST(TestGeneratedCastOperators) {
     Handle<Context> context =
         Utils::OpenHandle(*v8::Isolate::GetCurrent()->GetCurrentContext());
     m.TestGeneratedCastOperators(
-        m.UncheckedCast<Context>(m.HeapConstant(context)));
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
@@ -850,6 +857,7 @@ TEST(TestGeneratedCastOperators) {
 }
 
 TEST(TestNewPretenured) {
+  ManualGCScope manual_gc_scope;
   CcTest::InitializeVM();
   Isolate* isolate(CcTest::i_isolate());
   i::HandleScope scope(isolate);
@@ -859,7 +867,8 @@ TEST(TestNewPretenured) {
   {
     Handle<Context> context =
         Utils::OpenHandle(*v8::Isolate::GetCurrent()->GetCurrentContext());
-    m.TestNewPretenured(m.UncheckedCast<Context>(m.HeapConstant(context)));
+    m.TestNewPretenured(
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
@@ -909,7 +918,7 @@ TEST(TestCallMultiReturnBuiltin) {
     Handle<Context> context =
         Utils::OpenHandle(*v8::Isolate::GetCurrent()->GetCurrentContext());
     m.TestCallMultiReturnBuiltin(
-        m.UncheckedCast<Context>(m.HeapConstant(context)));
+        m.UncheckedCast<Context>(m.HeapConstantNoHole(context)));
     m.Return(m.UndefinedConstant());
   }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
@@ -932,8 +941,8 @@ TEST(TestRunLazyTwice) {
   }
   CHECK_EQ(lazyNumber, 5);
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
-  Handle<Object> result = ft.Call().ToHandleChecked();
-  CHECK_EQ(7, Smi::cast(*result).value());
+  DirectHandle<Object> result = ft.Call().ToHandleChecked();
+  CHECK_EQ(7, Cast<Smi>(*result).value());
 }
 
 TEST(TestCreateLazyNodeFromTorque) {
@@ -965,8 +974,8 @@ TEST(TestReturnNever_NotCalled) {
     m.Return(result);
   }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
-  Handle<Object> result = ft.Call().ToHandleChecked();
-  CHECK_EQ(42, Smi::cast(*result).value());
+  DirectHandle<Object> result = ft.Call().ToHandleChecked();
+  CHECK_EQ(42, Cast<Smi>(*result).value());
 }
 
 // Test calling a builtin that calls a runtime fct with return type {never}.
@@ -984,9 +993,9 @@ TEST(TestReturnNever_Runtime_Called) {
     m.Return(result);
   }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
-  MaybeHandle<Object> result = ft.Call();
+  MaybeDirectHandle<Object> result = ft.Call();
   CHECK(result.is_null());
-  CHECK(isolate->has_pending_exception());
+  CHECK(isolate->has_exception());
 }
 
 // Test calling a builtin that calls another builtin with return type {never}.
@@ -1004,11 +1013,46 @@ TEST(TestReturnNever_Builtin_Called) {
     m.Return(result);
   }
   FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
-  MaybeHandle<Object> result = ft.Call();
+  MaybeDirectHandle<Object> result = ft.Call();
   CHECK(result.is_null());
-  CHECK(isolate->has_pending_exception());
+  CHECK(isolate->has_exception());
 }
 
+int* global_use_counts = nullptr;
+
+void MockUseCounterCallback(v8::Isolate* isolate,
+                            v8::Isolate::UseCounterFeature feature) {
+  ++global_use_counts[feature];
+}
+
+// Test @incrementUseCounter
+TEST(TestIncrementUseCounterInBuiltin) {
+  CcTest::InitializeVM();
+  v8::Isolate* isolate = CcTest::isolate();
+  v8::HandleScope scope(isolate);
+  LocalContext env;
+  int use_counts[v8::Isolate::kUseCounterFeatureCount] = {};
+  global_use_counts = use_counts;
+  CcTest::isolate()->SetUseCounterCallback(MockUseCounterCallback);
+
+  Isolate* i_isolate(CcTest::i_isolate());
+  const int kNumParams = 0;
+  CodeAssemblerTester asm_tester(i_isolate, JSParameterCount(kNumParams));
+  TestTorqueAssembler m(asm_tester.state());
+  {
+    auto context = m.GetJSContextParameter();
+    TNode<Object> result =
+        m.CallBuiltin(Builtin::kTestIncrementArraySpeciesModified, context);
+    m.Return(result);
+  }
+  FunctionTester ft(asm_tester.GenerateCode(), kNumParams);
+  CHECK_EQ(0, use_counts[v8::Isolate::kArraySpeciesModified]);
+  ft.Call();
+  CHECK_EQ(1, use_counts[v8::Isolate::kArraySpeciesModified]);
+}
+
+#include "src/codegen/undef-code-stub-assembler-macros.inc"
+
 }  // namespace compiler
 }  // namespace internal
 }  // namespace v8
