@@ -57,8 +57,8 @@ ManuallyImportedJSFunction CreateJSSelector(FunctionSig* sig, int which) {
   SNPrintF(source, "(function(%s) { return %c; })",
            formals[sig->parameter_count()], param);
 
-  Handle<JSFunction> js_function =
-      Handle<JSFunction>::cast(v8::Utils::OpenHandle(
+  DirectHandle<JSFunction> js_function =
+      Cast<JSFunction>(v8::Utils::OpenDirectHandle(
           *v8::Local<v8::Function>::Cast(CompileRun(source.begin()))));
   ManuallyImportedJSFunction import = {sig, js_function};
 
@@ -103,8 +103,8 @@ WASM_COMPILED_EXEC_TEST(Run_CallJS_Add_jswrapped) {
   TestSignatures sigs;
   HandleScope scope(CcTest::InitIsolateOnce());
   const char* source = "(function(a) { return a + 99; })";
-  Handle<JSFunction> js_function =
-      Handle<JSFunction>::cast(v8::Utils::OpenHandle(
+  DirectHandle<JSFunction> js_function =
+      Cast<JSFunction>(v8::Utils::OpenDirectHandle(
           *v8::Local<v8::Function>::Cast(CompileRun(source))));
   ManuallyImportedJSFunction import = {sigs.i_i(), js_function};
   WasmRunner<int, int> r(execution_tier, kWasmOrigin, &import);
@@ -147,7 +147,7 @@ void RunJSSelectTest(TestExecutionTier tier, int which) {
     }
 
     double expected = inputs.arg_d(which);
-    r.CheckCallApplyViaJS(expected, t.function_index(), nullptr, 0);
+    r.CheckCallApplyViaJS(expected, t.function_index(), {});
   }
 }
 
@@ -205,7 +205,7 @@ void RunWASMSelectTest(TestExecutionTier tier, int which) {
     WasmFunctionCompiler& t = r.NewFunction(&sig);
     t.Build({WASM_LOCAL_GET(which)});
 
-    Handle<Object> args[] = {
+    DirectHandle<Object> args[] = {
         isolate->factory()->NewNumber(inputs.arg_d(0)),
         isolate->factory()->NewNumber(inputs.arg_d(1)),
         isolate->factory()->NewNumber(inputs.arg_d(2)),
@@ -217,7 +217,7 @@ void RunWASMSelectTest(TestExecutionTier tier, int which) {
     };
 
     double expected = inputs.arg_d(which);
-    r.CheckCallApplyViaJS(expected, t.function_index(), args, kMaxParams);
+    r.CheckCallApplyViaJS(expected, t.function_index(), {args, kMaxParams});
   }
 }
 
@@ -277,20 +277,22 @@ void RunWASMSelectAlignTest(TestExecutionTier tier, int num_args,
     WasmFunctionCompiler& t = r.NewFunction(&sig);
     t.Build({WASM_LOCAL_GET(which)});
 
-    Handle<Object> args[] = {isolate->factory()->NewNumber(inputs.arg_d(0)),
-                             isolate->factory()->NewNumber(inputs.arg_d(1)),
-                             isolate->factory()->NewNumber(inputs.arg_d(2)),
-                             isolate->factory()->NewNumber(inputs.arg_d(3)),
-                             isolate->factory()->NewNumber(inputs.arg_d(4)),
-                             isolate->factory()->NewNumber(inputs.arg_d(5)),
-                             isolate->factory()->NewNumber(inputs.arg_d(6)),
-                             isolate->factory()->NewNumber(inputs.arg_d(7)),
-                             isolate->factory()->NewNumber(inputs.arg_d(8)),
-                             isolate->factory()->NewNumber(inputs.arg_d(9))};
+    DirectHandle<Object> args[] = {
+        isolate->factory()->NewNumber(inputs.arg_d(0)),
+        isolate->factory()->NewNumber(inputs.arg_d(1)),
+        isolate->factory()->NewNumber(inputs.arg_d(2)),
+        isolate->factory()->NewNumber(inputs.arg_d(3)),
+        isolate->factory()->NewNumber(inputs.arg_d(4)),
+        isolate->factory()->NewNumber(inputs.arg_d(5)),
+        isolate->factory()->NewNumber(inputs.arg_d(6)),
+        isolate->factory()->NewNumber(inputs.arg_d(7)),
+        isolate->factory()->NewNumber(inputs.arg_d(8)),
+        isolate->factory()->NewNumber(inputs.arg_d(9))};
 
     double nan = std::numeric_limits<double>::quiet_NaN();
     double expected = which < num_args ? inputs.arg_d(which) : nan;
-    r.CheckCallApplyViaJS(expected, t.function_index(), args, num_args);
+    r.CheckCallApplyViaJS(expected, t.function_index(),
+                          {args, static_cast<size_t>(num_args)});
   }
 }
 
@@ -391,7 +393,7 @@ void RunJSSelectAlignTest(TestExecutionTier tier, int num_args,
     WasmFunctionCompiler& t = r.NewFunction(&sig);
     t.Build(base::VectorOf(code.data(), end));
 
-    Handle<Object> args[] = {
+    DirectHandle<Object> args[] = {
         factory->NewNumber(inputs.arg_d(0)),
         factory->NewNumber(inputs.arg_d(1)),
         factory->NewNumber(inputs.arg_d(2)),
@@ -406,7 +408,8 @@ void RunJSSelectAlignTest(TestExecutionTier tier, int num_args,
 
     double nan = std::numeric_limits<double>::quiet_NaN();
     double expected = which < num_args ? inputs.arg_d(which) : nan;
-    r.CheckCallApplyViaJS(expected, t.function_index(), args, num_args);
+    r.CheckCallApplyViaJS(expected, t.function_index(),
+                          {args, static_cast<size_t>(num_args)});
   }
 }
 
@@ -478,14 +481,13 @@ WASM_COMPILED_EXEC_TEST(Run_JSSelectAlign_10) {
 // function (a,b,c){ if(c)return a; return b; }
 
 void RunPickerTest(TestExecutionTier tier, bool indirect) {
-  EXPERIMENTAL_FLAG_SCOPE(return_call);
   Isolate* isolate = CcTest::InitIsolateOnce();
   HandleScope scope(isolate);
   TestSignatures sigs;
 
   const char* source = "(function(a,b,c) { if(c)return a; return b; })";
-  Handle<JSFunction> js_function =
-      Handle<JSFunction>::cast(v8::Utils::OpenHandle(
+  DirectHandle<JSFunction> js_function =
+      Cast<JSFunction>(v8::Utils::OpenDirectHandle(
           *v8::Local<v8::Function>::Cast(CompileRun(source))));
 
   ManuallyImportedJSFunction import = {sigs.i_iii(), js_function};
@@ -499,7 +501,7 @@ void RunPickerTest(TestExecutionTier tier, bool indirect) {
   WasmFunctionCompiler& rc_fn = r.NewFunction(sigs.i_i(), "rc");
 
   if (indirect) {
-    uint8_t sig_index = r.builder().AddSignature(sigs.i_iii());
+    ModuleTypeIndex sig_index = r.builder().AddSignature(sigs.i_iii());
     uint16_t indirect_function_table[] = {static_cast<uint16_t>(js_index)};
 
     r.builder().AddIndirectFunctionTable(indirect_function_table,
@@ -513,11 +515,13 @@ void RunPickerTest(TestExecutionTier tier, bool indirect) {
         js_index, WASM_I32V(left), WASM_I32V(right), WASM_LOCAL_GET(0))});
   }
 
-  Handle<Object> args_left[] = {isolate->factory()->NewNumber(1)};
-  r.CheckCallApplyViaJS(left, rc_fn.function_index(), args_left, 1);
+  DirectHandle<Object> args_left[] = {isolate->factory()->NewNumber(1)};
+  r.CheckCallApplyViaJS(left, rc_fn.function_index(),
+                        base::VectorOf(args_left));
 
-  Handle<Object> args_right[] = {isolate->factory()->NewNumber(0)};
-  r.CheckCallApplyViaJS(right, rc_fn.function_index(), args_right, 1);
+  DirectHandle<Object> args_right[] = {isolate->factory()->NewNumber(0)};
+  r.CheckCallApplyViaJS(right, rc_fn.function_index(),
+                        base::VectorOf(args_right));
 }
 
 WASM_COMPILED_EXEC_TEST(Run_ReturnCallImportedFunction) {
