@@ -1,8 +1,6 @@
 // Copyright 2023 the V8 project authors. All rights reserved.
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
-//
-// Flags: --harmony-array-from-async
 
 async function* asyncGen(n) {
   for (let i = 0; i < n; i++) yield i * 2;
@@ -38,3 +36,25 @@ function* gen() {
   const arr = await Array.fromAsync([1, 2, 3], v => v + 4);
   assertEquals([5, 6, 7], arr);
 })();
+
+(async function TestArrayFromAsyncArrayLike() {
+  const arrayLike = {
+    length: 3,
+    0: 1,
+    1: 2,
+    2: 3
+  };
+  const arr = await Array.fromAsync(arrayLike);
+  assertEquals([1, 2, 3], arr);
+})();
+
+(async function TestArrayFromAsyncArrayLikeWithMapping() {
+  const arrayLike = {
+    length: 3,
+    0: 1,
+    1: 2,
+    2: 3
+  };
+  const arr = await Array.fromAsync(arrayLike, v => v **2);
+  assertEquals([1, 4, 9], arr);
+})();
