@@ -8,7 +8,7 @@ d8.file.execute('test/mjsunit/wasm/wasm-module-builder.js');
 
 const builder = new WasmModuleBuilder();
 builder.addMemory(1, 1, true);
-builder.addGlobal(kWasmI32, 1);
+builder.addGlobal(kWasmI32, true, false);
 builder.addFunction(undefined, kSig_v_i)
   .addLocals(kWasmI32, 5)
   .addBody([
