@@ -2,9 +2,7 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
-// Flags: --experimental-wasm-imported-strings --experimental-wasm-gc
-// For {isOneByteString}:
-// Flags: --expose-externalize-string
+// Flags: --expose-externalize-string --allow-natives-syntax
 
 d8.file.execute("test/mjsunit/wasm/wasm-module-builder.js");
 
@@ -17,6 +15,7 @@ let kSig_e_v = makeSig([], [kRefExtern]);
 let kSig_i_ri = makeSig([kWasmExternRef, kWasmI32], [kWasmI32]);
 let kSig_i_rii = makeSig([kWasmExternRef, kWasmI32, kWasmI32], [kWasmI32]);
 let kSig_i_rr = makeSig([kWasmExternRef, kWasmExternRef], [kWasmI32]);
+let kSig_i_rri = makeSig([kWasmExternRef, kWasmExternRef, kWasmI32], [kWasmI32]);
 let kSig_i_riii = makeSig([kWasmExternRef, kWasmI32, kWasmI32, kWasmI32],
                           [kWasmI32]);
 let kSig_ii_riii = makeSig([kWasmExternRef, kWasmI32, kWasmI32, kWasmI32],
@@ -27,6 +26,12 @@ let kSig_e_rii = makeSig([kWasmExternRef, kWasmI32, kWasmI32],
 let kSig_e_rr = makeSig([kWasmExternRef, kWasmExternRef], [kRefExtern]);
 let kSig_e_r = makeSig([kWasmExternRef], [kRefExtern]);
 
+function MakeExternalString(s) {
+  s = createExternalizableString(s);
+  externalizeString(s);
+  return s;
+}
+
 let interestingStrings = [
   '',
   'ascii',
@@ -40,6 +45,18 @@ let interestingStrings = [
   'ab \ud800',         // Lone lead surrogate at the end.
   'ab \udc00',         // Lone trail surrogate at the end.
   'a \udc00\ud800 b',  // Swapped surrogate pair.
+
+  %ConstructConsString('ascii left and', 'ascii right'),
+  %ConstructConsString('2 \ucccc b', 'must be long enough \ucccc a'),
+
+  %ConstructSlicedString('abcedhfenvejfwqfqw', 5),
+  %ConstructSlicedString('febwvnewkvjabc\uccccdfenvefqw', 8),
+
+  MakeExternalString('external one byte'),
+  MakeExternalString('external two byte \ucccc b'),
+
+  %ConstructThinString('this is thin and must be long enough'),
+  %ConstructThinString('this is two-byte \ucccc thin string'),
 ];
 
 function IsSurrogate(codepoint) {
@@ -53,6 +70,14 @@ function HasIsolatedSurrogate(str) {
   }
   return false;
 }
+function ReplaceIsolatedSurrogates(str, replacement='\ufffd') {
+  let replaced = '';
+  for (let codepoint of str) {
+    replaced +=
+      IsSurrogate(codepoint.codePointAt(0)) ? replacement : codepoint;
+  }
+  return replaced;
+}
 
 function encodeWtf16LE(str) {
   // String iterator coalesces surrogate pairs.
@@ -66,9 +91,9 @@ function encodeWtf16LE(str) {
 }
 
 let kArrayI16;
-let kArrayI8;
+let kStringCast;
+let kStringTest;
 let kStringFromWtf16Array;
-let kStringFromWtf8Array;
 let kStringToWtf16Array;
 let kStringFromCharCode;
 let kStringFromCodePoint;
@@ -79,44 +104,182 @@ let kStringConcat;
 let kStringSubstring;
 let kStringEquals;
 let kStringCompare;
+let kStringIndexOfImported;
+let kStringToLowerCaseImported;
 
 function MakeBuilder() {
   let builder = new WasmModuleBuilder();
   builder.startRecGroup();
   kArrayI16 = builder.addArray(kWasmI16, true, kNoSuperType, true);
   builder.endRecGroup();
-  builder.startRecGroup();
-  kArrayI8 = builder.addArray(kWasmI8, true, kNoSuperType, true);
-  builder.endRecGroup();
   let arrayref = wasmRefNullType(kArrayI16);
-  let array8ref = wasmRefNullType(kArrayI8);
 
+  kStringCast = builder.addImport('wasm:js-string', 'cast', kSig_e_r);
+  kStringTest = builder.addImport('wasm:js-string', 'test', kSig_i_r);
   kStringFromWtf16Array = builder.addImport(
-      'String', 'fromWtf16Array',
+      'wasm:js-string', 'fromCharCodeArray',
       makeSig([arrayref, kWasmI32, kWasmI32], [kRefExtern]));
-  kStringFromWtf8Array = builder.addImport(
-      'String', 'fromWtf8Array',
-      makeSig([array8ref, kWasmI32, kWasmI32], [kRefExtern]));
   kStringToWtf16Array = builder.addImport(
-      'String', 'toWtf16Array',
+      'wasm:js-string', 'intoCharCodeArray',
       makeSig([kWasmExternRef, arrayref, kWasmI32], [kWasmI32]));
-  kStringFromCharCode = builder.addImport('String', 'fromCharCode', kSig_e_i);
-  kStringFromCodePoint = builder.addImport('String', 'fromCodePoint', kSig_e_i);
-  kStringCharCodeAt = builder.addImport('String', 'charCodeAt', kSig_i_ri);
-  kStringCodePointAt = builder.addImport('String', 'codePointAt', kSig_i_ri);
-  kStringLength = builder.addImport('String', 'length', kSig_i_r);
-  kStringConcat = builder.addImport('String', 'concat', kSig_e_rr);
-  kStringSubstring = builder.addImport('String', 'substring', kSig_e_rii);
-  kStringEquals = builder.addImport('String', 'equals', kSig_i_rr);
-  kStringCompare = builder.addImport('String', 'compare', kSig_i_rr);
+  kStringFromCharCode =
+      builder.addImport('wasm:js-string', 'fromCharCode', kSig_e_i);
+  kStringFromCodePoint =
+      builder.addImport('wasm:js-string', 'fromCodePoint', kSig_e_i);
+  kStringCharCodeAt =
+      builder.addImport('wasm:js-string', 'charCodeAt', kSig_i_ri);
+  kStringCodePointAt =
+      builder.addImport('wasm:js-string', 'codePointAt', kSig_i_ri);
+  kStringLength = builder.addImport('wasm:js-string', 'length', kSig_i_r);
+  kStringConcat = builder.addImport('wasm:js-string', 'concat', kSig_e_rr);
+  kStringSubstring =
+      builder.addImport('wasm:js-string', 'substring', kSig_e_rii);
+  kStringEquals = builder.addImport('wasm:js-string', 'equals', kSig_i_rr);
+  kStringCompare = builder.addImport('wasm:js-string', 'compare', kSig_i_rr);
+  kStringIndexOfImported = builder.addImport('m', 'indexOf', kSig_i_rri);
+  kStringToLowerCaseImported = builder.addImport('m', 'toLowerCase', kSig_r_r);
 
   return builder;
 }
 
 let kImports = {
-  String: WebAssembly.String,
   strings: interestingStrings,
+  m: {
+    indexOf: Function.prototype.call.bind(String.prototype.indexOf),
+    toLowerCase: Function.prototype.call.bind(String.prototype.toLowerCase),
+  },
 };
+let kBuiltins = { builtins: ["js-string"] };
+
+(function TestStringCast() {
+  print(arguments.callee.name);
+  let builder = MakeBuilder();
+
+  builder.addFunction("cast", kSig_e_r)
+    .exportFunc()
+    .addBody([
+      kExprLocalGet, 0,
+      kExprCallFunction, kStringCast,
+    ]);
+
+  builder.addFunction("cast_null", kSig_e_v)
+    .exportFunc()
+    .addBody([
+      kExprRefNull, kExternRefCode,
+      kExprCallFunction, kStringCast,
+    ]);
+
+  let instance = builder.instantiate(kImports, kBuiltins);
+
+  assertEquals('foo', instance.exports.cast('foo'));
+  assertThrows(
+      () => instance.exports.cast(123), WebAssembly.RuntimeError,
+      'illegal cast');
+  assertThrows(
+      () => instance.exports.cast(undefined), WebAssembly.RuntimeError,
+      'illegal cast');
+  assertThrows(
+      () => instance.exports.cast(true), WebAssembly.RuntimeError,
+      'illegal cast');
+  assertThrows(
+      () => instance.exports.cast(null), WebAssembly.RuntimeError,
+      'illegal cast');
+  assertThrows(
+      () => instance.exports.cast_null(), WebAssembly.RuntimeError,
+      'illegal cast');
+})();
+
+(function TestStringTest() {
+  print(arguments.callee.name);
+  let builder = MakeBuilder();
+
+  builder.addFunction("test", kSig_i_r)
+    .exportFunc()
+    .addBody([
+      kExprLocalGet, 0,
+      kExprCallFunction, kStringTest,
+    ]);
+
+  builder.addFunction("test_null", kSig_i_v)
+    .exportFunc()
+    .addBody([
+      kExprRefNull, kExternRefCode,
+      kExprCallFunction, kStringTest,
+    ]);
+
+  let instance = builder.instantiate(kImports, kBuiltins);
+
+  assertEquals(1, instance.exports.test("foo"));
+  assertEquals(0, instance.exports.test(123));
+  assertEquals(0, instance.exports.test(undefined));
+  assertEquals(0, instance.exports.test(true));
+  assertEquals(0, instance.exports.test(null));
+  assertEquals(0, instance.exports.test_null());
+})();
+
+(function TestIndexOfImportedStrings() {
+  print(arguments.callee.name);
+  let builder = new MakeBuilder();
+
+  builder.addFunction('indexOf', kSig_i_rri).exportFunc().addBody([
+    kExprLocalGet, 0,
+    kExprCallFunction, kStringCast,
+    kExprLocalGet, 1,
+    kExprCallFunction, kStringCast,
+    kExprLocalGet, 2,
+    kExprCallFunction, kStringIndexOfImported,
+  ]);
+  let instance = builder.instantiate(kImports, kBuiltins);
+
+  assertEquals(2, instance.exports.indexOf('xxfooxx', 'foo', 0));
+  assertEquals(2, instance.exports.indexOf('xxfooxx', 'foo', -2));
+  assertEquals(-1, instance.exports.indexOf('xxfooxx', 'foo', 100));
+  // Make sure we don't lose bits when Smi-tagging of the start position.
+  assertEquals(-1, instance.exports.indexOf('xxfooxx', 'foo', 0x4000_0000));
+  assertEquals(-1, instance.exports.indexOf('xxfooxx', 'foo', 0x2000_0000));
+  assertEquals(
+      2,
+      instance.exports.indexOf(
+          'xxfooxx', 'foo', 0x8000_0000));  // Negative i32.
+
+  // Both first and second args should be non-null strings.
+  assertThrows(
+      () => instance.exports.indexOf('xxnullxx', null, 0),
+      WebAssembly.RuntimeError, 'illegal cast');
+  assertThrows(
+      () => instance.exports.indexOf(12345, 234, 0), WebAssembly.RuntimeError,
+      'illegal cast');
+  assertThrows(
+      () => instance.exports.indexOf(null, 'foo', 0), WebAssembly.RuntimeError,
+      'illegal cast');
+  assertThrows(
+      () => instance.exports.indexOf(null, 'null', 0), WebAssembly.RuntimeError,
+      'illegal cast');
+})();
+
+(function TestStringToLowerCaseImported() {
+  print(arguments.callee.name);
+  let builder = new MakeBuilder();
+
+  builder.addFunction('toLowerCase', kSig_r_r).exportFunc().addBody([
+    kExprLocalGet, 0,
+    kExprCallFunction, kStringCast,
+    kExprCallFunction, kStringToLowerCaseImported,
+  ]);
+  let instance = builder.instantiate(kImports, kBuiltins);
+
+  assertEquals(
+      'make this lowercase!',
+      instance.exports.toLowerCase('MAKE THIS LOWERCASE!'));
+
+  // The argument should be a non-null string.
+  assertThrows(
+      () => instance.exports.toLowerCase(null), WebAssembly.RuntimeError,
+      'illegal cast');
+  assertThrows(
+      () => instance.exports.toLowerCase(123), WebAssembly.RuntimeError,
+      'illegal cast');
+})();
 
 (function TestStringConst() {
   print(arguments.callee.name);
@@ -127,10 +290,10 @@ let kImports = {
       .addBody([kExprGlobalGet, i]);
   }
   for (let i = 0; i < interestingStrings.length; i++) {
-    builder.addGlobal(kRefExtern, false, [kExprGlobalGet, i])
+    builder.addGlobal(kRefExtern, false, false, [kExprGlobalGet, i])
       .exportAs("global" + i);
   }
-  let instance = builder.instantiate(kImports);
+  let instance = builder.instantiate(kImports, kBuiltins);
   for (let [i, str] of interestingStrings.entries()) {
     assertEquals(str, instance.exports["string_const" + i]());
     assertEquals(str, instance.exports["global" + i].value);
@@ -155,7 +318,7 @@ let kImports = {
       kExprCallFunction, kStringLength,
     ]);
 
-  let instance = builder.instantiate(kImports);
+  let instance = builder.instantiate(kImports, kBuiltins);
   for (let str of interestingStrings) {
     assertEquals(str.length, instance.exports.string_length(str));
   }
@@ -194,7 +357,7 @@ let kImports = {
       kExprCallFunction, kStringConcat,
     ]);
 
-  let instance = builder.instantiate(kImports);
+  let instance = builder.instantiate(kImports, kBuiltins);
 
   for (let head of interestingStrings) {
     for (let tail of interestingStrings) {
@@ -248,7 +411,7 @@ let kImports = {
       kExprCallFunction, kStringEquals,
     ]);
 
-  let instance = builder.instantiate(kImports);
+  let instance = builder.instantiate(kImports, kBuiltins);
 
   for (let head of interestingStrings) {
     for (let tail of interestingStrings) {
@@ -319,7 +482,7 @@ let kImports = {
       kExprCallFunction, kStringSubstring,
     ]);
 
-  let instance = builder.instantiate(kImports);
+  let instance = builder.instantiate(kImports, kBuiltins);
   for (let str of interestingStrings) {
     for (let i = 0; i < str.length; i++) {
       assertEquals(str.charCodeAt(i),
@@ -420,7 +583,7 @@ let kImports = {
       kExprCallFunction, kStringCompare,
     ]);
 
-  let instance = builder.instantiate(kImports);
+  let instance = builder.instantiate(kImports, kBuiltins);
   for (let lhs of interestingStrings) {
     for (let rhs of interestingStrings) {
       const expected = lhs < rhs ? -1 : lhs > rhs ? 1 : 0;
@@ -463,7 +626,7 @@ let kImports = {
       kExprCallFunction, kStringCompare,
     ]);
 
-  let instance = builder.instantiate(kImports);
+  let instance = builder.instantiate(kImports, kBuiltins);
   assertThrows(() => instance.exports.compareLhsNullable(null, "abc"),
                WebAssembly.RuntimeError, "dereferencing a null pointer");
   assertThrows(() => instance.exports.compareLhsNullable("abc", null),
@@ -484,7 +647,7 @@ let kImports = {
       kExprCallFunction, kStringFromCodePoint,
     ]);
 
-  let instance = builder.instantiate(kImports);
+  let instance = builder.instantiate(kImports, kBuiltins);
   for (let char of "Az1#\n\ucccc\ud800\udc00") {
     assertEquals(char, instance.exports.asString(char.codePointAt(0)));
   }
@@ -504,7 +667,7 @@ let kImports = {
       kExprCallFunction, kStringFromCharCode,
     ]);
 
-  let instance = builder.instantiate(kImports);
+  let instance = builder.instantiate(kImports, kBuiltins);
   let inputs = "Az1#\n\ucccc\ud800\udc00";
   for (let i = 0; i < inputs.length; i++) {
     assertEquals(inputs.charAt(i),
@@ -512,130 +675,6 @@ let kImports = {
   }
 })();
 
-function encodeWtf8(str) {
-  // String iterator coalesces surrogate pairs.
-  let out = [];
-  for (let codepoint of str) {
-    codepoint = codepoint.codePointAt(0);
-    if (codepoint <= 0x7f) {
-      out.push(codepoint);
-    } else if (codepoint <= 0x7ff) {
-      out.push(0xc0 | (codepoint >> 6));
-      out.push(0x80 | (codepoint & 0x3f));
-    } else if (codepoint <= 0xffff) {
-      out.push(0xe0 | (codepoint >> 12));
-      out.push(0x80 | ((codepoint >> 6) & 0x3f));
-      out.push(0x80 | (codepoint & 0x3f));
-    } else if (codepoint <= 0x10ffff) {
-      out.push(0xf0 | (codepoint >> 18));
-      out.push(0x80 | ((codepoint >> 12) & 0x3f));
-      out.push(0x80 | ((codepoint >> 6) & 0x3f));
-      out.push(0x80 | (codepoint & 0x3f));
-    } else {
-      throw new Error("bad codepoint " + codepoint);
-    }
-  }
-  return out;
-}
-
-function makeWtf8TestDataSegment() {
-  let data = []
-  let valid = {};
-  let invalid = {};
-
-  for (let str of interestingStrings) {
-    let bytes = encodeWtf8(str);
-    valid[str] = { offset: data.length, length: bytes.length };
-    for (let byte of bytes) {
-      data.push(byte);
-    }
-  }
-  for (let bytes of ['trailing high byte \xa9',
-                     'interstitial high \xa9 byte',
-                     'invalid \xc0 byte',
-                     'invalid three-byte \xed\xd0\x80',
-                     'surrogate \xed\xa0\x80\xed\xb0\x80 pair']) {
-    invalid[bytes] = { offset: data.length, length: bytes.length };
-    for (let i = 0; i < bytes.length; i++) {
-      data.push(bytes.charCodeAt(i));
-    }
-  }
-  return { valid, invalid, data: Uint8Array.from(data) };
-};
-
-(function TestStringNewWtf8Array() {
-  print(arguments.callee.name);
-  let builder = MakeBuilder();
-  let data = makeWtf8TestDataSegment();
-  let data_index = builder.addPassiveDataSegment(data.data);
-
-  let ascii_data_index =
-      builder.addPassiveDataSegment(Uint8Array.from(encodeWtf8("ascii")));
-
-  let make_i8_array = builder.addFunction(
-      "make_i8_array", makeSig([], [wasmRefType(kArrayI8)]))
-    .addBody([
-      ...wasmI32Const(0),
-      ...wasmI32Const(data.data.length),
-      kGCPrefix, kExprArrayNewData, kArrayI8, data_index
-    ]).index;
-
-  builder.addFunction("new_wtf8", kSig_e_ii)
-    .exportFunc()
-    .addBody([
-      kExprCallFunction, make_i8_array,
-      kExprLocalGet, 0, kExprLocalGet, 1,
-      kExprCallFunction, kStringFromWtf8Array,
-    ]);
-
-  builder.addFunction("bounds_check", kSig_e_ii)
-    .exportFunc()
-    .addBody([
-      ...wasmI32Const(0),
-      ...wasmI32Const("ascii".length),
-      kGCPrefix, kExprArrayNewData, kArrayI8, ascii_data_index,
-      kExprLocalGet, 0, kExprLocalGet, 1,
-      kExprCallFunction, kStringFromWtf8Array,
-    ]);
-
-  builder.addFunction("null_array", kSig_e_v).exportFunc()
-    .addBody([
-      kExprRefNull, kArrayI8,
-      kExprI32Const, 0, kExprI32Const, 0,
-      kExprCallFunction, kStringFromWtf8Array,
-    ]);
-
-  let instance = builder.instantiate(kImports);
-  for (let [str, {offset, length}] of Object.entries(data.valid)) {
-    let start = offset;
-    let end = offset + length;
-    assertEquals(str, instance.exports.new_wtf8(start, end));
-  }
-  for (let [str, {offset, length}] of Object.entries(data.invalid)) {
-    let start = offset;
-    let end = offset + length;
-    assertThrows(() => instance.exports.new_wtf8(start, end),
-                 WebAssembly.RuntimeError, "invalid WTF-8 string");
-  }
-
-  assertEquals("ascii", instance.exports.bounds_check(0, "ascii".length));
-  assertEquals("", instance.exports.bounds_check("ascii".length,
-                                                 "ascii".length));
-  assertEquals("i", instance.exports.bounds_check("ascii".length - 1,
-                                                  "ascii".length));
-  assertThrows(() => instance.exports.bounds_check(0, 100),
-               WebAssembly.RuntimeError, "array element access out of bounds");
-  assertThrows(() => instance.exports.bounds_check(0, -1),
-               WebAssembly.RuntimeError, "array element access out of bounds");
-  assertThrows(() => instance.exports.bounds_check(-1, 0),
-               WebAssembly.RuntimeError, "array element access out of bounds");
-  assertThrows(() => instance.exports.bounds_check("ascii".length,
-                                                   "ascii".length + 1),
-               WebAssembly.RuntimeError, "array element access out of bounds");
-  assertThrows(() => instance.exports.null_array(),
-               WebAssembly.RuntimeError, "dereferencing a null pointer");
-})();
-
 function encodeWtf16LE(str) {
   // String iterator coalesces surrogate pairs.
   let out = [];
@@ -711,7 +750,7 @@ function makeWtf16TestDataSegment(strings) {
       kExprCallFunction, kStringFromWtf16Array,
     ]);
 
-  let instance = builder.instantiate(kImports);
+  let instance = builder.instantiate(kImports, kBuiltins);
   for (let [str, {offset, length}] of Object.entries(data.valid)) {
     let start = offset / 2;
     let end = start + length;
@@ -782,7 +821,7 @@ function makeWtf16TestDataSegment(strings) {
       kExprCallFunction, kStringToWtf16Array,
     ]);
 
-  let instance = builder.instantiate(kImports);
+  let instance = builder.instantiate(kImports, kBuiltins);
   for (let str of interestingStrings) {
     assertEquals(str, instance.exports.encode(str, str.length, 0));
     assertEquals(str, instance.exports.encode(str, str.length + 20, 10));
