@@ -2,8 +2,6 @@
 // Use of this source code is governed by a BSD-style license that can be
 // found in the LICENSE file.
 
-// Flags: --expose-wasm
-
 d8.file.execute('test/mjsunit/wasm/wasm-module-builder.js');
 
 function assertEq(val, expected) {
@@ -266,7 +264,7 @@ let exportingModuleBinary2 = (() => {
   builder.exportMemoryAs('b');
   builder.setTableBounds(1, 1);
   builder.addExportOfKind('c', kExternalTable, 0);
-  var o = builder.addGlobal(kWasmI32, false).exportAs('x');
+  var o = builder.addGlobal(kWasmI32, false, false).exportAs('x');
   return new Int8Array(builder.toBuffer());
 })();
 var arr = moduleExports(new Module(exportingModuleBinary2));
@@ -675,10 +673,10 @@ assertEq(get.call(tbl1, 1), null);
 assertEq(get.call(tbl1, 1.5), null);
 assertThrows(
     () => get.call(tbl1, 2), RangeError,
-    /invalid index 2 into funcref table of size 2/);
+    /invalid address 2 in funcref table of size 2/);
 assertThrows(
     () => get.call(tbl1, 2.5), RangeError,
-    /invalid index 2 into funcref table of size 2/);
+    /invalid address 2 in funcref table of size 2/);
 assertThrows(() => get.call(tbl1, -1), TypeError, /must be non-negative/);
 assertThrows(
     () => get.call(tbl1, Math.pow(2, 33)), TypeError,
@@ -705,7 +703,7 @@ assertThrows(
     /must be convertible to a valid number/);
 assertThrows(
     () => set.call(tbl1, 2, null), RangeError,
-    /invalid index 2 into funcref table of size 2/);
+    /invalid address 2 in funcref table of size 2/);
 assertThrows(
     () => set.call(tbl1, -1, null), TypeError, /must be non-negative/);
 assertThrows(
@@ -781,9 +779,8 @@ assertThrows(
 assertThrows(() => WebAssembly.validate(), TypeError);
 assertThrows(() => WebAssembly.validate('hi'), TypeError);
 assertTrue(WebAssembly.validate(emptyModuleBinary));
-// TODO: other ways for validate to return false.
-assertFalse(WebAssembly.validate(moduleBinaryImporting2Memories));
-assertFalse(WebAssembly.validate(moduleBinaryWithMemSectionAndMemImport));
+assertTrue(WebAssembly.validate(moduleBinaryImporting2Memories));
+assertTrue(WebAssembly.validate(moduleBinaryWithMemSectionAndMemImport));
 
 // 'WebAssembly.compile' data property
 let compileDesc = Object.getOwnPropertyDescriptor(WebAssembly, 'compile');
