@@ -50,6 +50,7 @@ var kPageSize = 65536;
 var kSpecMaxPages = 65536;
 var kMaxVarInt32Size = 5;
 var kMaxVarInt64Size = 10;
+var kSpecMaxFunctionParams = 1_000;
 
 let kDeclNoLocals = 0;
 
@@ -76,12 +77,15 @@ let kModuleNameCode = 0;
 let kFunctionNamesCode = 1;
 let kLocalNamesCode = 2;
 
+let kWasmSharedTypeForm = 0x65;
 let kWasmFunctionTypeForm = 0x60;
 let kWasmStructTypeForm = 0x5f;
 let kWasmArrayTypeForm = 0x5e;
 let kWasmSubtypeForm = 0x50;
 let kWasmSubtypeFinalForm = 0x4f;
 let kWasmRecursiveTypeGroupForm = 0x4e;
+let kWasmDescriptorTypeForm = 0x4d;
+let kWasmDescribesTypeForm = 0x4c;
 
 let kNoSuperType = 0xFFFFFFFF;
 
@@ -117,6 +121,7 @@ let kWasmF64 = 0x7c;
 let kWasmS128 = 0x7b;
 let kWasmI8 = 0x78;
 let kWasmI16 = 0x77;
+let kWasmF16 = 0x76;
 
 // These are defined as negative integers to distinguish them from positive type
 // indices.
@@ -131,10 +136,14 @@ let kWasmEqRef = -0x13;
 let kWasmI31Ref = -0x14;
 let kWasmStructRef = -0x15;
 let kWasmArrayRef = -0x16;
+let kWasmExnRef = -0x17;
+let kWasmNullExnRef = -0x0c;
 let kWasmStringRef = -0x19;
 let kWasmStringViewWtf8 = -0x1a;
-let kWasmStringViewWtf16 = -0x1e;
+let kWasmStringViewWtf16 = -0x20;
 let kWasmStringViewIter = -0x1f;
+const kWasmContRef = -0x18;
+const kWasmNullContRef = -0x0b;
 
 // Use the positive-byte versions inside function bodies.
 let kLeb128Mask = 0x7f;
@@ -148,19 +157,46 @@ let kNullExternRefCode = kWasmNullExternRef & kLeb128Mask;
 let kNullFuncRefCode = kWasmNullFuncRef & kLeb128Mask;
 let kStructRefCode = kWasmStructRef & kLeb128Mask;
 let kArrayRefCode = kWasmArrayRef & kLeb128Mask;
+let kExnRefCode = kWasmExnRef & kLeb128Mask;
+let kNullExnRefCode = kWasmNullExnRef & kLeb128Mask;
 let kNullRefCode = kWasmNullRef & kLeb128Mask;
 let kStringRefCode = kWasmStringRef & kLeb128Mask;
 let kStringViewWtf8Code = kWasmStringViewWtf8 & kLeb128Mask;
 let kStringViewWtf16Code = kWasmStringViewWtf16 & kLeb128Mask;
 let kStringViewIterCode = kWasmStringViewIter & kLeb128Mask;
+const kContRefCode = kWasmContRef & kLeb128Mask;
+const kNullContRefCode = kWasmNullContRef & kLeb128Mask;
 
 let kWasmRefNull = 0x63;
 let kWasmRef = 0x64;
+let kWasmExact = 0x62;
+
+// Implementation detail of `wasmRef[Null]Type`, don't use directly.
+class RefTypeBuilder {
+  constructor(opcode, heap_type) {
+    this.opcode = opcode;
+    this.heap_type = heap_type;
+    this.is_shared = false;
+    this.is_exact = false;
+  }
+  nullable() {
+    this.opcode = kWasmRefNull;
+    return this;
+  }
+  shared() {
+    this.is_shared = true;
+    return this;
+  }
+  exact() {
+    this.is_exact = true;
+    return this;
+  }
+}
 function wasmRefNullType(heap_type) {
-  return {opcode: kWasmRefNull, heap_type: heap_type};
+  return new RefTypeBuilder(kWasmRefNull, heap_type);
 }
 function wasmRefType(heap_type) {
-  return {opcode: kWasmRef, heap_type: heap_type};
+  return new RefTypeBuilder(kWasmRef, heap_type);
 }
 
 let kExternalFunction = 0;
@@ -175,17 +211,28 @@ let kSegmentZero = 0;
 
 let kExceptionAttribute = 0;
 
+const kAtomicSeqCst = 0;
+const kAtomicAcqRel = 1;
+
 // Useful signatures
 let kSig_i_i = makeSig([kWasmI32], [kWasmI32]);
 let kSig_l_l = makeSig([kWasmI64], [kWasmI64]);
 let kSig_i_l = makeSig([kWasmI64], [kWasmI32]);
 let kSig_i_ii = makeSig([kWasmI32, kWasmI32], [kWasmI32]);
 let kSig_i_iii = makeSig([kWasmI32, kWasmI32, kWasmI32], [kWasmI32]);
+let kSig_i_iiii = makeSig([kWasmI32, kWasmI32, kWasmI32, kWasmI32], [kWasmI32]);
 let kSig_v_iiii = makeSig([kWasmI32, kWasmI32, kWasmI32, kWasmI32], []);
+let kSig_l_iiii = makeSig([kWasmI32, kWasmI32, kWasmI32, kWasmI32], [kWasmI64]);
 let kSig_l_i = makeSig([kWasmI32], [kWasmI64]);
+let kSig_f_i = makeSig([kWasmI32], [kWasmF32]);
+let kSig_i_f = makeSig([kWasmF32], [kWasmI32]);
+let kSig_i_ff = makeSig([kWasmF32, kWasmF32], [kWasmI32]);
 let kSig_f_ff = makeSig([kWasmF32, kWasmF32], [kWasmF32]);
+let kSig_f_ffff = makeSig([kWasmF32, kWasmF32, kWasmF32, kWasmF32], [kWasmF32]);
 let kSig_d_dd = makeSig([kWasmF64, kWasmF64], [kWasmF64]);
+let kSig_d_dddd = makeSig([kWasmF64, kWasmF64, kWasmF64, kWasmF64], [kWasmF64]);
 let kSig_l_ll = makeSig([kWasmI64, kWasmI64], [kWasmI64]);
+let kSig_l_llll = makeSig([kWasmI64, kWasmI64, kWasmI64, kWasmI64], [kWasmI64]);
 let kSig_i_dd = makeSig([kWasmF64, kWasmF64], [kWasmI32]);
 let kSig_v_v = makeSig([], []);
 let kSig_i_v = makeSig([], [kWasmI32]);
@@ -212,6 +259,8 @@ let kSig_v_f = makeSig([kWasmF32], []);
 let kSig_f_f = makeSig([kWasmF32], [kWasmF32]);
 let kSig_f_d = makeSig([kWasmF64], [kWasmF32]);
 let kSig_d_d = makeSig([kWasmF64], [kWasmF64]);
+let kSig_d_f = makeSig([kWasmF32], [kWasmF64]);
+let kSig_d_i = makeSig([kWasmI32], [kWasmF64]);
 let kSig_r_r = makeSig([kWasmExternRef], [kWasmExternRef]);
 let kSig_a_a = makeSig([kWasmAnyFunc], [kWasmAnyFunc]);
 let kSig_i_r = makeSig([kWasmExternRef], [kWasmI32]);
@@ -262,6 +311,8 @@ const kWasmOpcodes = {
   'If': 0x04,
   'Else': 0x05,
   'Try': 0x06,
+  'TryTable': 0x1f,
+  'ThrowRef': 0x0a,
   'Catch': 0x07,
   'Throw': 0x08,
   'Rethrow': 0x09,
@@ -516,13 +567,20 @@ let kExprRefTest = 0x14;
 let kExprRefTestNull = 0x15;
 let kExprRefCast = 0x16;
 let kExprRefCastNull = 0x17;
-let kExprBrOnCastGeneric = 0x18;      // TODO(14034): Drop "Generic" name.
-let kExprBrOnCastFailGeneric = 0x19;  // TODO(14034): Drop "Generic" name.
-let kExprExternInternalize = 0x1a;
-let kExprExternExternalize = 0x1b;
+let kExprBrOnCast = 0x18;
+let kExprBrOnCastFail = 0x19;
+let kExprAnyConvertExtern = 0x1a;
+let kExprExternConvertAny = 0x1b;
 let kExprRefI31 = 0x1c;
 let kExprI31GetS = 0x1d;
 let kExprI31GetU = 0x1e;
+let kExprRefI31Shared = 0x1f;
+// Custom Descriptors proposal:
+let kExprRefGetDesc = 0x22;
+let kExprRefCastDesc = 0x23;
+let kExprRefCastDescNull = 0x24;
+let kExprBrOnCastDesc = 0x25;
+let kExprBrOnCastDescFail = 0x26;
 
 let kExprRefCastNop = 0x4c;
 
@@ -596,6 +654,7 @@ let kExprTableFill = 0x11;
 let kExprAtomicNotify = 0x00;
 let kExprI32AtomicWait = 0x01;
 let kExprI64AtomicWait = 0x02;
+let kExprAtomicFence = 0x03;
 let kExprI32AtomicLoad = 0x10;
 let kExprI32AtomicLoad8U = 0x12;
 let kExprI32AtomicLoad16U = 0x13;
@@ -661,6 +720,31 @@ let kExprI64AtomicCompareExchange8U = 0x4c;
 let kExprI64AtomicCompareExchange16U = 0x4d;
 let kExprI64AtomicCompareExchange32U = 0x4e;
 
+// Atomic GC opcodes (shared-everything-threads).
+const kExprPause = 0x04;
+const kExprStructAtomicGet = 0x5c;
+const kExprStructAtomicGetS = 0x5d;
+const kExprStructAtomicGetU = 0x5e;
+const kExprStructAtomicSet = 0x5f;
+const kExprStructAtomicAdd = 0x60;
+const kExprStructAtomicSub = 0x61;
+const kExprStructAtomicAnd = 0x62;
+const kExprStructAtomicOr = 0x63;
+const kExprStructAtomicXor = 0x64;
+const kExprStructAtomicExchange = 0x65;
+const kExprStructAtomicCompareExchange = 0x66;
+const kExprArrayAtomicGet = 0x67;
+const kExprArrayAtomicGetS = 0x68;
+const kExprArrayAtomicGetU = 0x69;
+const kExprArrayAtomicSet = 0x6a;
+const kExprArrayAtomicAdd = 0x6b;
+const kExprArrayAtomicSub = 0x6c;
+const kExprArrayAtomicAnd = 0x6d;
+const kExprArrayAtomicOr = 0x6e;
+const kExprArrayAtomicXor = 0x6f;
+const kExprArrayAtomicExchange = 0x70;
+const kExprArrayAtomicCompareExchange = 0x71;
+
 // Simd opcodes.
 let kExprS128LoadMem = 0x00;
 let kExprS128Load8x8S = 0x01;
@@ -922,14 +1006,40 @@ let kExprI16x8RelaxedQ15MulRS = wasmSignedLeb(0x111);
 let kExprI16x8DotI8x16I7x16S = wasmSignedLeb(0x112);
 let kExprI32x4DotI8x16I7x16AddS = wasmSignedLeb(0x113);
 
-// Compilation hint constants.
-let kCompilationHintStrategyDefault = 0x00;
-let kCompilationHintStrategyLazy = 0x01;
-let kCompilationHintStrategyEager = 0x02;
-let kCompilationHintStrategyLazyBaselineEagerTopTier = 0x03;
-let kCompilationHintTierDefault = 0x00;
-let kCompilationHintTierBaseline = 0x01;
-let kCompilationHintTierOptimized = 0x02;
+// FP16 SIMD
+let kExprF16x8Splat = wasmSignedLeb(0x120);
+let kExprF16x8ExtractLane = wasmSignedLeb(0x121);
+let kExprF16x8ReplaceLane = wasmSignedLeb(0x122);
+let kExprF16x8Abs = wasmSignedLeb(0x130);
+let kExprF16x8Neg = wasmSignedLeb(0x131);
+let kExprF16x8Sqrt = wasmSignedLeb(0x132);
+let kExprF16x8Ceil = wasmSignedLeb(0x133);
+let kExprF16x8Floor = wasmSignedLeb(0x134);
+let kExprF16x8Trunc = wasmSignedLeb(0x135);
+let kExprF16x8NearestInt = wasmSignedLeb(0x136);
+let kExprF16x8Eq = wasmSignedLeb(0x137);
+let kExprF16x8Ne = wasmSignedLeb(0x138);
+let kExprF16x8Lt = wasmSignedLeb(0x139);
+let kExprF16x8Gt = wasmSignedLeb(0x13a);
+let kExprF16x8Le = wasmSignedLeb(0x13b);
+let kExprF16x8Ge = wasmSignedLeb(0x13c);
+let kExprF16x8Add = wasmSignedLeb(0x13d);
+let kExprF16x8Sub = wasmSignedLeb(0x13e);
+let kExprF16x8Mul = wasmSignedLeb(0x13f);
+let kExprF16x8Div = wasmSignedLeb(0x140);
+let kExprF16x8Min = wasmSignedLeb(0x141);
+let kExprF16x8Max = wasmSignedLeb(0x142);
+let kExprF16x8Pmin = wasmSignedLeb(0x143);
+let kExprF16x8Pmax = wasmSignedLeb(0x144);
+let kExprI16x8SConvertF16x8 = wasmSignedLeb(0x145);
+let kExprI16x8UConvertF16x8 = wasmSignedLeb(0x146);
+let kExprF16x8SConvertI16x8 = wasmSignedLeb(0x147);
+let kExprF16x8UConvertI16x8 = wasmSignedLeb(0x148);
+let kExprF16x8DemoteF32x4Zero = wasmSignedLeb(0x149);
+let kExprF16x8DemoteF64x2Zero = wasmSignedLeb(0x14a);
+let kExprF32x4PromoteLowF16x8 = wasmSignedLeb(0x14b);
+let kExprF16x8Qfma = wasmSignedLeb(0x14e);
+let kExprF16x8Qfms = wasmSignedLeb(0x14f);
 
 let kTrapUnreachable = 0;
 let kTrapMemOutOfBounds = 1;
@@ -952,6 +1062,12 @@ let kAtomicWaitOk = 0;
 let kAtomicWaitNotEqual = 1;
 let kAtomicWaitTimedOut = 2;
 
+// Exception handling with exnref.
+let kCatchNoRef = 0x0;
+let kCatchRef = 0x1;
+let kCatchAllNoRef = 0x2;
+let kCatchAllRef = 0x3;
+
 let kTrapMsgs = [
   'unreachable',                                    // --
   'memory access out of bounds',                    // --
@@ -1079,7 +1195,8 @@ class Binary {
       this.emit_u8(type >= 0 ? type : type & kLeb128Mask);
     } else {
       this.emit_u8(type.opcode);
-      if ('depth' in type) this.emit_u8(type.depth);
+      if (type.is_shared) this.emit_u8(kWasmSharedTypeForm);
+      if (type.is_exact) this.emit_u8(kWasmExact);
       this.emit_heap_type(type.heap_type);
     }
   }
@@ -1140,11 +1257,6 @@ class WasmFunctionBuilder {
     return this;
   }
 
-  setCompilationHint(strategy, baselineTier, topTier) {
-    this.module.setCompilationHint(strategy, baselineTier, topTier, this.index);
-    return this;
-  }
-
   addBody(body) {
     checkExpr(body);
     // Store a copy of the body, and automatically add the end opcode.
@@ -1180,10 +1292,11 @@ class WasmFunctionBuilder {
 }
 
 class WasmGlobalBuilder {
-  constructor(module, type, mutable, init) {
+  constructor(module, type, mutable, shared, init) {
     this.module = module;
     this.type = type;
     this.mutable = mutable;
+    this.shared = shared;
     this.init = init;
   }
 
@@ -1204,7 +1317,7 @@ function checkExpr(expr) {
 }
 
 class WasmTableBuilder {
-  constructor(module, type, initial_size, max_size, init_expr) {
+  constructor(module, type, initial_size, max_size, init_expr, is_shared, is_table64) {
     // TODO(manoskouk): Add the table index.
     this.module = module;
     this.type = type;
@@ -1213,6 +1326,8 @@ class WasmTableBuilder {
     this.max_size = max_size;
     this.init_expr = init_expr;
     this.has_init = init_expr !== undefined;
+    this.is_shared = is_shared;
+    this.is_table64 = is_table64;
   }
 
   exportAs(name) {
@@ -1229,35 +1344,61 @@ function makeField(type, mutability) {
   return {type: type, mutability: mutability};
 }
 
+function MustBeNumber(x, name) {
+  if (typeof x !== 'undefined' && typeof x !== 'number') {
+    throw new Error(`${name} must be a number, was ${x}`);
+  }
+  return x;
+}
+
 class WasmStruct {
-  constructor(fields, is_final, supertype_idx) {
-    if (!Array.isArray(fields)) {
+  constructor(fields, is_final, is_shared, supertype_idx) {
+    let descriptor = undefined;
+    let describes = undefined;
+    if (Array.isArray(fields)) {
+      // Fall through.
+    } else if (fields.constructor === Object) {
+      // Options bag.
+      is_final = fields.is_final ?? fields.final ?? false;
+      is_shared = fields.is_shared ?? fields.shared ?? false;
+      supertype_idx = MustBeNumber(
+          fields.supertype_idx ?? fields.supertype ?? kNoSuperType,
+          "supertype");
+      descriptor = MustBeNumber(fields.descriptor, "'descriptor'");
+      describes = MustBeNumber(fields.describes, "'describes'");
+      fields = fields.fields ?? [];  // This must happen last.
+    } else {
       throw new Error('struct fields must be an array');
     }
     this.fields = fields;
     this.type_form = kWasmStructTypeForm;
     this.is_final = is_final;
+    this.is_shared = is_shared;
     this.supertype = supertype_idx;
+    this.descriptor = descriptor;
+    this.describes = describes;
   }
 }
 
 class WasmArray {
-  constructor(type, mutability, is_final, supertype_idx) {
+  constructor(type, mutability, is_final, is_shared, supertype_idx) {
     this.type = type;
     this.mutability = mutability;
     this.type_form = kWasmArrayTypeForm;
     this.is_final = is_final;
+    this.is_shared = is_shared;
     this.supertype = supertype_idx;
   }
 }
 
 class WasmElemSegment {
-  constructor(table, offset, type, elements, is_decl) {
+  constructor(table, offset, type, elements, is_decl, is_shared) {
     this.table = table;
     this.offset = offset;
     this.type = type;
     this.elements = elements;
     this.is_decl = is_decl;
+    this.is_shared = is_shared;
     // Invariant checks.
     if ((table === undefined) != (offset === undefined)) {
       throw new Error("invalid element segment");
@@ -1297,7 +1438,6 @@ class WasmModuleBuilder {
     this.tags = [];
     this.memories = [];
     this.functions = [];
-    this.compilation_hints = [];
     this.element_segments = [];
     this.data_segments = [];
     this.explicit = [];
@@ -1366,11 +1506,13 @@ class WasmModuleBuilder {
 
   // We use {is_final = true} so that the MVP syntax is generated for
   // signatures.
-  addType(type, supertype_idx = kNoSuperType, is_final = true) {
+  addType(type, supertype_idx = kNoSuperType, is_final = true,
+          is_shared = false) {
     var pl = type.params.length;   // should have params
     var rl = type.results.length;  // should have results
     var type_copy = {params: type.params, results: type.results,
-                     is_final: is_final, supertype: supertype_idx};
+                     is_final: is_final, is_shared: is_shared,
+                     supertype: supertype_idx};
     this.types.push(type_copy);
     return this.types.length - 1;
   }
@@ -1380,16 +1522,25 @@ class WasmModuleBuilder {
     return this.stringrefs.length - 1;
   }
 
-  addStruct(fields, supertype_idx = kNoSuperType, is_final = false) {
-    this.types.push(new WasmStruct(fields, is_final, supertype_idx));
+  // {fields} may be a list of fields, in which case the other parameters are
+  // relevant; or an options bag, which replaces the other parameters.
+  // Example: addStruct({fields: [...], supertype: 3})
+  addStruct(fields, supertype_idx = kNoSuperType, is_final = false,
+            is_shared = false) {
+    this.types.push(
+        new WasmStruct(fields, is_final, is_shared, supertype_idx));
     return this.types.length - 1;
   }
 
-  addArray(type, mutability, supertype_idx = kNoSuperType, is_final = false) {
-    this.types.push(new WasmArray(type, mutability, is_final, supertype_idx));
+  addArray(type, mutability, supertype_idx = kNoSuperType, is_final = false,
+           is_shared = false) {
+    this.types.push(
+        new WasmArray(type, mutability, is_final, is_shared, supertype_idx));
     return this.types.length - 1;
   }
 
+  nextTypeIndex() { return this.types.length; }
+
   static defaultFor(type) {
     switch (type) {
       case kWasmI32:
@@ -1402,6 +1553,10 @@ class WasmModuleBuilder {
         return wasmF64Const(0.0);
       case kWasmS128:
         return [kSimdPrefix, kExprS128Const, ...(new Array(16).fill(0))];
+      case kWasmStringViewIter:
+      case kWasmStringViewWtf8:
+      case kWasmStringViewWtf16:
+        throw new Error("String views are non-defaultable");
       default:
         if ((typeof type) != 'number' && type.opcode != kWasmRefNull) {
           throw new Error("Non-defaultable type");
@@ -1411,29 +1566,37 @@ class WasmModuleBuilder {
     }
   }
 
-  addGlobal(type, mutable, init) {
+  addGlobal(type, mutable, shared, init) {
     if (init === undefined) init = WasmModuleBuilder.defaultFor(type);
     checkExpr(init);
-    let glob = new WasmGlobalBuilder(this, type, mutable, init);
+    let glob = new WasmGlobalBuilder(this, type, mutable, shared, init);
     glob.index = this.globals.length + this.num_imported_globals;
     this.globals.push(glob);
     return glob;
   }
 
   addTable(
-      type, initial_size, max_size = undefined, init_expr = undefined) {
+      type, initial_size, max_size = undefined, init_expr = undefined,
+      is_shared = false, is_table64 = false) {
     if (type == kWasmI32 || type == kWasmI64 || type == kWasmF32 ||
         type == kWasmF64 || type == kWasmS128 || type == kWasmVoid) {
       throw new Error('Tables must be of a reference type');
     }
     if (init_expr != undefined) checkExpr(init_expr);
     let table = new WasmTableBuilder(
-        this, type, initial_size, max_size, init_expr);
+        this, type, initial_size, max_size, init_expr, is_shared, is_table64);
     table.index = this.tables.length + this.num_imported_tables;
     this.tables.push(table);
     return table;
   }
 
+  addTable64(
+      type, initial_size, max_size = undefined, init_expr = undefined,
+      is_shared = false) {
+    return this.addTable(
+        type, initial_size, max_size, init_expr, is_shared, true);
+  }
+
   addTag(type) {
     let type_index = (typeof type) == 'number' ? type : this.addType(type);
     let tag_index = this.tags.length + this.num_imported_tags;
@@ -1469,7 +1632,7 @@ class WasmModuleBuilder {
     return this.num_imported_funcs++;
   }
 
-  addImportedGlobal(module, name, type, mutable = false) {
+  addImportedGlobal(module, name, type, mutable = false, shared = false) {
     if (this.globals.length != 0) {
       throw new Error('Imported globals must be declared before local ones');
     }
@@ -1478,7 +1641,8 @@ class WasmModuleBuilder {
       name: name,
       kind: kExternalGlobal,
       type: type,
-      mutable: mutable
+      mutable: mutable,
+      shared: shared
     };
     this.imports.push(o);
     return this.num_imported_globals++;
@@ -1504,7 +1668,9 @@ class WasmModuleBuilder {
     return mem_index;
   }
 
-  addImportedTable(module, name, initial, maximum, type) {
+  addImportedTable(
+      module, name, initial, maximum, type = kWasmFuncRef, shared = false,
+      is_table64 = false) {
     if (this.tables.length != 0) {
       throw new Error('Imported tables must be declared before local ones');
     }
@@ -1514,7 +1680,9 @@ class WasmModuleBuilder {
       kind: kExternalTable,
       initial: initial,
       maximum: maximum,
-      type: type || kWasmFuncRef
+      type: type,
+      shared: !!shared,
+      is_table64: !!is_table64,
     };
     this.imports.push(o);
     return this.num_imported_tables++;
@@ -1553,29 +1721,21 @@ class WasmModuleBuilder {
     return this;
   }
 
-  setCompilationHint(strategy, baselineTier, topTier, index) {
-    this.compilation_hints[index] = {
-      strategy: strategy,
-      baselineTier: baselineTier,
-      topTier: topTier
-    };
-    return this;
-  }
-
-  // TODO(manoskouk): Refactor this to use initializer expression for {offset}.
-  addDataSegment(offset, data, is_global = false, memory_index = 0) {
+  addActiveDataSegment(memory_index, offset, data, is_shared = false) {
+    checkExpr(offset);
     this.data_segments.push({
-      offset: offset,
-      data: data,
-      is_global: is_global,
       is_active: true,
-      mem_index: memory_index
+      is_shared: is_shared,
+      mem_index: memory_index,
+      offset: offset,
+      data: data
     });
     return this.data_segments.length - 1;
   }
 
-  addPassiveDataSegment(data) {
-    this.data_segments.push({data: data, is_active: false});
+  addPassiveDataSegment(data, is_shared = false) {
+    this.data_segments.push({
+      is_active: false, is_shared: is_shared, data: data});
     return this.data_segments.length - 1;
   }
 
@@ -1596,35 +1756,35 @@ class WasmModuleBuilder {
   // {offset} is a constant expression.
   // If {type} is undefined, then {elements} are function indices. Otherwise,
   // they are constant expressions.
-  addActiveElementSegment(table, offset, elements, type) {
+  addActiveElementSegment(table, offset, elements, type, is_shared = false) {
     checkExpr(offset);
     if (type != undefined) {
       for (let element of elements) checkExpr(element);
     }
     this.element_segments.push(
-        new WasmElemSegment(table, offset, type, elements, false));
+        new WasmElemSegment(table, offset, type, elements, false, is_shared));
     return this.element_segments.length - 1;
   }
 
   // If {type} is undefined, then {elements} are function indices. Otherwise,
   // they are constant expressions.
-  addPassiveElementSegment(elements, type) {
+  addPassiveElementSegment(elements, type, is_shared = false) {
     if (type != undefined) {
       for (let element of elements) checkExpr(element);
     }
-    this.element_segments.push(
-      new WasmElemSegment(undefined, undefined, type, elements, false));
+    this.element_segments.push(new WasmElemSegment(
+      undefined, undefined, type, elements, false, is_shared));
     return this.element_segments.length - 1;
   }
 
   // If {type} is undefined, then {elements} are function indices. Otherwise,
   // they are constant expressions.
-  addDeclarativeElementSegment(elements, type) {
+  addDeclarativeElementSegment(elements, type, is_shared = false) {
     if (type != undefined) {
       for (let element of elements) checkExpr(element);
     }
-    this.element_segments.push(
-      new WasmElemSegment(undefined, undefined, type, elements, true));
+    this.element_segments.push(new WasmElemSegment(
+      undefined, undefined, type, elements, true, is_shared));
     return this.element_segments.length - 1;
   }
 
@@ -1712,6 +1872,15 @@ class WasmModuleBuilder {
             section.emit_u8(kWasmSubtypeForm);
             section.emit_u8(0);  // no supertypes
           }
+          if (type.is_shared) section.emit_u8(kWasmSharedTypeForm);
+          if (type.describes !== undefined) {
+            section.emit_u8(kWasmDescribesTypeForm);
+            section.emit_u32v(type.describes);
+          }
+          if (type.descriptor !== undefined) {
+            section.emit_u8(kWasmDescriptorTypeForm);
+            section.emit_u32v(type.descriptor);
+          }
           if (type instanceof WasmStruct) {
             section.emit_u8(kWasmStructTypeForm);
             section.emit_u32v(type.fields.length);
@@ -1751,7 +1920,8 @@ class WasmModuleBuilder {
             section.emit_u32v(imp.type_index);
           } else if (imp.kind == kExternalGlobal) {
             section.emit_type(imp.type);
-            section.emit_u8(imp.mutable);
+            let flags = (imp.mutable ? 1 : 0) | (imp.shared ? 0b10 : 0);
+            section.emit_u8(flags);
           } else if (imp.kind == kExternalMemory) {
             const has_max = imp.maximum !== undefined;
             const is_shared = !!imp.shared;
@@ -1765,8 +1935,12 @@ class WasmModuleBuilder {
             if (has_max) emit(imp.maximum);
           } else if (imp.kind == kExternalTable) {
             section.emit_type(imp.type);
-            var has_max = (typeof imp.maximum) != 'undefined';
-            section.emit_u8(has_max ? 1 : 0);             // flags
+            const has_max = (typeof imp.maximum) != 'undefined';
+            const is_shared = !!imp.shared;
+            const is_table64 = !!imp.is_table64;
+            let limits_byte =
+                (is_table64 ? 4 : 0) | (is_shared ? 2 : 0) | (has_max ? 1 : 0);
+            section.emit_u8(limits_byte);                 // flags
             section.emit_u32v(imp.initial);               // initial
             if (has_max) section.emit_u32v(imp.maximum);  // maximum
           } else if (imp.kind == kExternalTag) {
@@ -1801,9 +1975,13 @@ class WasmModuleBuilder {
             section.emit_u8(0x00);  // Reserved byte.
           }
           section.emit_type(table.type);
-          section.emit_u8(table.has_max);
-          section.emit_u32v(table.initial_size);
-          if (table.has_max) section.emit_u32v(table.max_size);
+          let limits_byte = (table.is_table64 ? 4 : 0) |
+              (table.is_shared ? 2 : 0) | (table.has_max ? 1 : 0);
+          section.emit_u8(limits_byte);
+          let emit = val => table.is_table64 ? section.emit_u64v(val) :
+                                               section.emit_u32v(val);
+          emit(table.initial_size);
+          if (table.has_max) emit(table.max_size);
           if (table.has_init) section.emit_init_expr(table.init_expr);
         }
       });
@@ -1860,7 +2038,7 @@ class WasmModuleBuilder {
         section.emit_u32v(wasm.globals.length);
         for (let global of wasm.globals) {
           section.emit_type(global.type);
-          section.emit_u8(global.mutable);
+          section.emit_u8((global.mutable ? 1 : 0) | (global.shared ? 0b10 : 0));
           section.emit_init_expr(global.init);
         }
       });
@@ -1900,23 +2078,24 @@ class WasmModuleBuilder {
           // Each case below corresponds to a flag from
           // https://webassembly.github.io/spec/core/binary/modules.html#element-section
           // (not in increasing order).
+          let shared_flag = segment.is_shared ? 0b1000 : 0;
           if (segment.is_active()) {
             if (segment.table == 0 && segment.type === undefined) {
               if (segment.expressions_as_elements()) {
-                section.emit_u8(0x04);
+                section.emit_u8(0x04 | shared_flag);
                 section.emit_init_expr(segment.offset);
               } else {
-                section.emit_u8(0x00)
+                section.emit_u8(0x00 | shared_flag)
                 section.emit_init_expr(segment.offset);
               }
             } else {
               if (segment.expressions_as_elements()) {
-                section.emit_u8(0x06);
+                section.emit_u8(0x06 | shared_flag);
                 section.emit_u32v(segment.table);
                 section.emit_init_expr(segment.offset);
                 section.emit_type(segment.type);
               } else {
-                section.emit_u8(0x02);
+                section.emit_u8(0x02 | shared_flag);
                 section.emit_u32v(segment.table);
                 section.emit_init_expr(segment.offset);
                 section.emit_u8(kExternalFunction);
@@ -1925,16 +2104,16 @@ class WasmModuleBuilder {
           } else {
             if (segment.expressions_as_elements()) {
               if (segment.is_passive()) {
-                section.emit_u8(0x05);
+                section.emit_u8(0x05 | shared_flag);
               } else {
-                section.emit_u8(0x07);
+                section.emit_u8(0x07 | shared_flag);
               }
               section.emit_type(segment.type);
             } else {
               if (segment.is_passive()) {
-                section.emit_u8(0x01);
+                section.emit_u8(0x01 | shared_flag);
               } else {
-                section.emit_u8(0x03);
+                section.emit_u8(0x03 | shared_flag);
               }
               section.emit_u8(kExternalFunction);
             }
@@ -1960,40 +2139,6 @@ class WasmModuleBuilder {
       });
     }
 
-    // If there are compilation hints add a custom section 'compilationHints'
-    // after the function section and before the code section.
-    if (wasm.compilation_hints.length > 0) {
-      if (debug) print('emitting compilation hints @ ' + binary.length);
-      // Build custom section payload.
-      let payloadBinary = new Binary();
-      let implicit_compilation_hints_count = wasm.functions.length;
-      payloadBinary.emit_u32v(implicit_compilation_hints_count);
-
-      // Defaults to the compiler's choice if no better hint was given (0x00).
-      let defaultHintByte = kCompilationHintStrategyDefault |
-          (kCompilationHintTierDefault << 2) |
-          (kCompilationHintTierDefault << 4);
-
-      // Emit hint byte for every function defined in this module.
-      for (let i = 0; i < implicit_compilation_hints_count; i++) {
-        let index = wasm.num_imported_funcs + i;
-        var hintByte;
-        if (index in wasm.compilation_hints) {
-          let hint = wasm.compilation_hints[index];
-          hintByte =
-              hint.strategy | (hint.baselineTier << 2) | (hint.topTier << 4);
-        } else {
-          hintByte = defaultHintByte;
-        }
-        payloadBinary.emit_u8(hintByte);
-      }
-
-      // Finalize as custom section.
-      let name = 'compilationHints';
-      let bytes = this.createCustomSection(name, payloadBinary.trunc_buffer());
-      binary.emit_bytes(bytes);
-    }
-
     // Add function bodies.
     if (wasm.functions.length > 0) {
       // emit function bodies
@@ -2036,24 +2181,17 @@ class WasmModuleBuilder {
       binary.emit_section(kDataSectionCode, section => {
         section.emit_u32v(wasm.data_segments.length);
         for (let seg of wasm.data_segments) {
+          let shared_flag = seg.is_shared ? 0b1000 : 0;
           if (seg.is_active) {
             if (seg.mem_index == 0) {
-              section.emit_u8(kActiveNoIndex);
+              section.emit_u8(kActiveNoIndex | shared_flag);
             } else {
-              section.emit_u8(kActiveWithIndex);
+              section.emit_u8(kActiveWithIndex | shared_flag);
               section.emit_u32v(seg.mem_index);
             }
-            if (seg.is_global) {
-              // Offset is taken from a global.
-              section.emit_u8(kExprGlobalGet);
-              section.emit_u32v(seg.offset);
-            } else {
-              // Offset is a constant.
-              section.emit_bytes(wasmI32Const(seg.offset));
-            }
-            section.emit_u8(kExprEnd);
+            section.emit_init_expr(seg.offset);
           } else {
-            section.emit_u8(kPassive);
+            section.emit_u8(kPassive | shared_flag);
           }
           section.emit_u32v(seg.data.length);
           section.emit_bytes(seg.data);
@@ -2127,8 +2265,8 @@ class WasmModuleBuilder {
     return Array.from(this.toBuffer(debug));
   }
 
-  instantiate(ffi) {
-    let module = this.toModule();
+  instantiate(ffi, options) {
+    let module = this.toModule(options);
     let instance = new WebAssembly.Instance(module, ffi);
     return instance;
   }
@@ -2138,13 +2276,13 @@ class WasmModuleBuilder {
         .then(({module, instance}) => instance);
   }
 
-  toModule(debug = false) {
-    return new WebAssembly.Module(this.toBuffer(debug));
+  toModule(options, debug = false) {
+    return new WebAssembly.Module(this.toBuffer(debug), options);
   }
 }
 
 function wasmSignedLeb(val, max_len = 5) {
-  if (val == null) throw new Error("Leb value many not be null/undefined");
+  if (val == null) throw new Error("Leb value may not be null/undefined");
   let res = [];
   for (let i = 0; i < max_len; ++i) {
     let v = val & 0x7f;
@@ -2161,7 +2299,7 @@ function wasmSignedLeb(val, max_len = 5) {
 }
 
 function wasmSignedLeb64(val, max_len = 10) {
-  if (val == null) throw new Error("Leb value many not be null/undefined");
+  if (val == null) throw new Error("Leb value may not be null/undefined");
   if (typeof val != "bigint") {
     if (val < Math.pow(2, 31)) {
       return wasmSignedLeb(val, max_len);
@@ -2250,26 +2388,49 @@ function wasmS128Const(f) {
   return result;
 }
 
-let [wasmBrOnCast, wasmBrOnCastFail] = (function() {
+let wasmEncodeHeapType = function(type) {
+  let result = wasmSignedLeb(type.heap_type, kMaxVarInt32Size);
+  if (type.is_shared) {
+    result = [kWasmSharedTypeForm].concat(result);
+  }
+  if (type.is_exact) {
+    result = [kWasmExact].concat(result);
+  }
+  return result;
+};
+
+let [wasmBrOnCast, wasmBrOnCastFail, wasmBrOnCastDesc, wasmBrOnCastDescFail] =
+(function() {
   return [
     (labelIdx, sourceType, targetType) =>
-      wasmBrOnCastImpl(labelIdx, sourceType, targetType, false),
-      (labelIdx, sourceType, targetType) =>
-      wasmBrOnCastImpl(labelIdx, sourceType, targetType, true),
+      wasmBrOnCastImpl(labelIdx, sourceType, targetType, kExprBrOnCast),
+    (labelIdx, sourceType, targetType) =>
+      wasmBrOnCastImpl(labelIdx, sourceType, targetType, kExprBrOnCastFail),
+    (labelIdx, sourceType, targetType) =>
+      wasmBrOnCastImpl(labelIdx, sourceType, targetType, kExprBrOnCastDesc),
+    (labelIdx, sourceType, targetType) =>
+      wasmBrOnCastImpl(labelIdx, sourceType, targetType, kExprBrOnCastDescFail),
   ];
-  function wasmBrOnCastImpl(labelIdx, sourceType, targetType, brOnFail) {
+  function wasmBrOnCastImpl(labelIdx, sourceType, targetType, opcode) {
     labelIdx = wasmUnsignedLeb(labelIdx, kMaxVarInt32Size);
-    let srcHeap = wasmSignedLeb(sourceType.heap_type, kMaxVarInt32Size);
-    let tgtHeap = wasmSignedLeb(targetType.heap_type, kMaxVarInt32Size);
     let srcIsNullable = sourceType.opcode == kWasmRefNull;
     let tgtIsNullable = targetType.opcode == kWasmRefNull;
     flags = (tgtIsNullable << 1) + srcIsNullable;
     return [
-      kGCPrefix, brOnFail ? kExprBrOnCastFailGeneric : kExprBrOnCastGeneric,
-      flags, ...labelIdx, ...srcHeap, ...tgtHeap];
+      kGCPrefix, opcode, flags, ...labelIdx, ...wasmEncodeHeapType(sourceType),
+      ...wasmEncodeHeapType(targetType)
+    ];
   }
 })();
 
 function getOpcodeName(opcode) {
   return globalThis.kWasmOpcodeNames?.[opcode] ?? 'unknown';
 }
+
+function wasmF32ConstSignalingNaN() {
+  return [kExprF32Const, 0xb9, 0xa1, 0xa7, 0x7f];
+}
+
+function wasmF64ConstSignalingNaN() {
+  return [kExprF64Const, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xf4, 0x7f];
+}
