@@ -29,6 +29,7 @@
 
 #include <cstdint>
 #include <cstring>
+#include <limits>
 
 #include "src/codegen/macro-assembler.h"
 #include "src/codegen/x64/assembler-x64-inl.h"
@@ -37,12 +38,14 @@
 #include "src/deoptimizer/deoptimizer.h"
 #include "src/execution/simulator.h"
 #include "src/heap/factory.h"
+#include "src/numbers/conversions.h"
 #include "src/objects/objects-inl.h"
 #include "src/objects/smi.h"
 #include "src/utils/ostreams.h"
 #include "test/common/assembler-tester.h"
 #include "test/common/value-helper.h"
 #include "test/unittests/test-utils.h"
+#include "third_party/fp16/src/include/fp16.h"
 
 namespace v8 {
 namespace internal {
@@ -57,7 +60,7 @@ using MacroAssemblerX64Test = TestWithIsolate;
 
 void PrintCode(Isolate* isolate, CodeDesc desc) {
 #ifdef OBJECT_PRINT
-  Handle<Code> code =
+  DirectHandle<Code> code =
       Factory::CodeBuilder(isolate, desc, CodeKind::FOR_TESTING).Build();
   StdoutStream os;
   Print(*code, os);
@@ -78,7 +81,8 @@ TEST_F(MacroAssemblerX64Test, TestHardAbort) {
   buffer->MakeExecutable();
   auto f = GeneratedCode<void>::FromBuffer(isolate(), buffer->start());
 
-  ASSERT_DEATH_IF_SUPPORTED({ f.Call(); }, "abort: no reason");
+  ASSERT_DEATH_IF_SUPPORTED(
+      { f.Call(); }, v8_flags.debug_code ? "abort: no reason" : "");
 }
 
 TEST_F(MacroAssemblerX64Test, TestCheck) {
@@ -90,7 +94,7 @@ TEST_F(MacroAssemblerX64Test, TestCheck) {
 
   // Fail if the first parameter is 17.
   __ movl(rax, Immediate(17));
-  __ cmpl(rax, arg_reg_1);
+  __ cmpl(rax, kCArgRegs[0]);
   __ Check(Condition::not_equal, AbortReason::kNoReason);
   __ ret(0);
 
@@ -101,7 +105,8 @@ TEST_F(MacroAssemblerX64Test, TestCheck) {
 
   f.Call(0);
   f.Call(18);
-  ASSERT_DEATH_IF_SUPPORTED({ f.Call(17); }, "abort: no reason");
+  ASSERT_DEATH_IF_SUPPORTED(
+      { f.Call(17); }, v8_flags.debug_code ? "abort: no reason" : "");
 }
 
 #undef __
@@ -130,6 +135,10 @@ using F9 = int(int16_t*, int32_t*);
 using F10 = int(int8_t*, int16_t*);
 using F11 = int(uint16_t*, uint32_t*);
 using F12 = int(uint8_t*, uint16_t*);
+using F13 = int(float*, int32_t*);
+using F14 = int(uint16_t*, int16_t*);
+using F15 = int(uint16_t*, uint16_t*);
+using F16 = int(double*, uint16_t*);
 
 #define __ masm->
 
@@ -544,7 +553,7 @@ TEST_F(MacroAssemblerX64Test, EmbeddedObj) {
 
   CodeDesc desc;
   masm->GetCode(isolate, &desc);
-  Handle<Code> code =
+  DirectHandle<Code> code =
       Factory::CodeBuilder(isolate, desc, CodeKind::FOR_TESTING).Build();
 #ifdef OBJECT_PRINT
   StdoutStream os;
@@ -960,6 +969,77 @@ TEST_F(MacroAssemblerX64Test, OperandOffset) {
   CHECK_EQ(0, result);
 }
 
+#define STORE(val, offset)                                              \
+  __ movq(kScratchRegister, Immediate(fp16_ieee_from_fp32_value(val))); \
+  __ movw(Operand(rsp, offset * kFloat16Size), kScratchRegister);
+
+#define LOAD_AND_CHECK(val, offset, op)                                     \
+  __ incq(rax);                                                             \
+  __ movq(kScratchRegister, Immediate(fp16_ieee_from_fp32_value(op(val)))); \
+  __ cmpw(kScratchRegister, Operand(rsp, offset * kFloat16Size));           \
+  __ j(not_equal, exit);
+
+void TestFloat16x8Abs(MacroAssembler* masm, Label* exit, float x, float y,
+                      float z, float w, float a, float b, float c, float d) {
+  __ AllocateStackSpace(kSimd128Size);
+
+  STORE(x, 0)
+  STORE(y, 1)
+  STORE(z, 2)
+  STORE(w, 3)
+  STORE(a, 4)
+  STORE(b, 5)
+  STORE(c, 6)
+  STORE(d, 7)
+
+  __ Movups(xmm0, Operand(rsp, 0));
+  __ Absph(xmm0, xmm0, kScratchRegister);
+  __ Movups(Operand(rsp, 0), xmm0);
+
+  LOAD_AND_CHECK(x, 0, fabsf)
+  LOAD_AND_CHECK(y, 1, fabsf)
+  LOAD_AND_CHECK(z, 2, fabsf)
+  LOAD_AND_CHECK(w, 3, fabsf)
+  LOAD_AND_CHECK(a, 4, fabsf)
+  LOAD_AND_CHECK(b, 5, fabsf)
+  LOAD_AND_CHECK(c, 6, fabsf)
+  LOAD_AND_CHECK(d, 7, fabsf)
+
+  __ addq(rsp, Immediate(kSimd128Size));
+}
+
+void TestFloat16x8Neg(MacroAssembler* masm, Label* exit, float x, float y,
+                      float z, float w, float a, float b, float c, float d) {
+  __ AllocateStackSpace(kSimd128Size);
+
+  STORE(x, 0)
+  STORE(y, 1)
+  STORE(z, 2)
+  STORE(w, 3)
+  STORE(a, 4)
+  STORE(b, 5)
+  STORE(c, 6)
+  STORE(d, 7)
+
+  __ Movups(xmm0, Operand(rsp, 0));
+  __ Negph(xmm0, xmm0, kScratchRegister);
+  __ Movups(Operand(rsp, 0), xmm0);
+
+  LOAD_AND_CHECK(x, 0, -)
+  LOAD_AND_CHECK(y, 1, -)
+  LOAD_AND_CHECK(z, 2, -)
+  LOAD_AND_CHECK(w, 3, -)
+  LOAD_AND_CHECK(a, 4, -)
+  LOAD_AND_CHECK(b, 5, -)
+  LOAD_AND_CHECK(c, 6, -)
+  LOAD_AND_CHECK(d, 7, -)
+
+  __ addq(rsp, Immediate(kSimd128Size));
+}
+
+#undef STORE
+#undef LOAD_AND_CHECK
+
 void TestFloat32x4Abs(MacroAssembler* masm, Label* exit, float x, float y,
                       float z, float w) {
   __ AllocateStackSpace(kSimd128Size);
@@ -1082,6 +1162,122 @@ void TestFloat64x2Neg(MacroAssembler* masm, Label* exit, double x, double y) {
   __ addq(rsp, Immediate(kSimd128Size));
 }
 
+template <bool dst_same_as_src>
+void TestFloat64x4Neg(MacroAssembler* masm, Label* exit, double x, double y,
+                      double z, double w) {
+  if (!CpuFeatures::IsSupported(AVX) || !CpuFeatures::IsSupported(AVX2)) return;
+
+  __ AllocateStackSpace(kSimd256Size);
+
+  CpuFeatureScope avx_scope(masm, AVX);
+  __ Move(xmm1, x);
+  __ Movsd(Operand(rsp, 0 * kDoubleSize), xmm1);
+  __ Move(xmm2, y);
+  __ Movsd(Operand(rsp, 1 * kDoubleSize), xmm2);
+  __ Move(xmm3, z);
+  __ Movsd(Operand(rsp, 2 * kDoubleSize), xmm3);
+  __ Move(xmm4, w);
+  __ Movsd(Operand(rsp, 3 * kDoubleSize), xmm4);
+  __ vmovupd(ymm0, Operand(rsp, 0));
+
+  if (dst_same_as_src) {
+    __ Negpd(ymm0, ymm0, kScratchSimd256Reg);
+    __ vmovupd(Operand(rsp, 0), ymm0);
+  } else {
+    __ Negpd(ymm1, ymm0, kScratchSimd256Reg);
+    __ vmovupd(Operand(rsp, 0), ymm1);
+  }
+
+  __ incq(rax);
+  __ Move(xmm1, -x);
+  __ Ucomisd(xmm1, Operand(rsp, 0 * kDoubleSize));
+  __ j(not_equal, exit);
+  __ incq(rax);
+  __ Move(xmm2, -y);
+  __ Ucomisd(xmm2, Operand(rsp, 1 * kDoubleSize));
+  __ j(not_equal, exit);
+  __ incq(rax);
+  __ Move(xmm2, -z);
+  __ Ucomisd(xmm2, Operand(rsp, 2 * kDoubleSize));
+  __ j(not_equal, exit);
+  __ incq(rax);
+  __ Move(xmm2, -w);
+  __ Ucomisd(xmm2, Operand(rsp, 3 * kDoubleSize));
+  __ j(not_equal, exit);
+
+  __ addq(rsp, Immediate(kSimd256Size));
+}
+
+template <bool dst_same_as_src>
+void TestFloat32x8Neg(MacroAssembler* masm, Label* exit, float x, float y,
+                      float z, float w, float a, float b, float c, float d) {
+  if (!CpuFeatures::IsSupported(AVX) || !CpuFeatures::IsSupported(AVX2)) return;
+
+  __ AllocateStackSpace(kSimd256Size);
+
+  CpuFeatureScope avx_scope(masm, AVX);
+  __ Move(xmm1, x);
+  __ Movss(Operand(rsp, 0 * kFloatSize), xmm1);
+  __ Move(xmm2, y);
+  __ Movss(Operand(rsp, 1 * kFloatSize), xmm2);
+  __ Move(xmm3, z);
+  __ Movss(Operand(rsp, 2 * kFloatSize), xmm3);
+  __ Move(xmm4, w);
+  __ Movss(Operand(rsp, 3 * kFloatSize), xmm4);
+  __ Move(xmm5, a);
+  __ Movss(Operand(rsp, 4 * kFloatSize), xmm5);
+  __ Move(xmm6, b);
+  __ Movss(Operand(rsp, 5 * kFloatSize), xmm6);
+  __ Move(xmm7, c);
+  __ Movss(Operand(rsp, 6 * kFloatSize), xmm7);
+  __ Move(xmm8, d);
+  __ Movss(Operand(rsp, 7 * kFloatSize), xmm8);
+  __ vmovups(ymm0, Operand(rsp, 0));
+
+  if (dst_same_as_src) {
+    __ Negps(ymm0, ymm0, kScratchSimd256Reg);
+    __ vmovups(Operand(rsp, 0), ymm0);
+  } else {
+    __ Negps(ymm1, ymm0, kScratchSimd256Reg);
+    __ vmovups(Operand(rsp, 0), ymm1);
+  }
+
+  __ incq(rax);
+  __ Move(xmm1, -x);
+  __ Ucomiss(xmm1, Operand(rsp, 0 * kFloatSize));
+  __ j(not_equal, exit);
+  __ incq(rax);
+  __ Move(xmm2, -y);
+  __ Ucomiss(xmm2, Operand(rsp, 1 * kFloatSize));
+  __ j(not_equal, exit);
+  __ incq(rax);
+  __ Move(xmm3, -z);
+  __ Ucomiss(xmm3, Operand(rsp, 2 * kFloatSize));
+  __ j(not_equal, exit);
+  __ incq(rax);
+  __ Move(xmm4, -w);
+  __ Ucomiss(xmm4, Operand(rsp, 3 * kFloatSize));
+  __ j(not_equal, exit);
+  __ incq(rax);
+  __ Move(xmm5, -a);
+  __ Ucomiss(xmm5, Operand(rsp, 4 * kFloatSize));
+  __ j(not_equal, exit);
+  __ incq(rax);
+  __ Move(xmm6, -b);
+  __ Ucomiss(xmm6, Operand(rsp, 5 * kFloatSize));
+  __ j(not_equal, exit);
+  __ incq(rax);
+  __ Move(xmm7, -c);
+  __ Ucomiss(xmm7, Operand(rsp, 6 * kFloatSize));
+  __ j(not_equal, exit);
+  __ incq(rax);
+  __ Move(xmm8, -d);
+  __ Ucomiss(xmm8, Operand(rsp, 7 * kFloatSize));
+  __ j(not_equal, exit);
+
+  __ addq(rsp, Immediate(kSimd256Size));
+}
+
 TEST_F(MacroAssemblerX64Test, SIMDMacros) {
   Isolate* isolate = i_isolate();
   HandleScope handles(isolate);
@@ -1093,12 +1289,25 @@ TEST_F(MacroAssemblerX64Test, SIMDMacros) {
   EntryCode(masm);
   Label exit;
 
+  float NaN = std::numeric_limits<float>::quiet_NaN();
+  float inf = std::numeric_limits<float>::infinity();
   __ xorq(rax, rax);
+  TestFloat16x8Abs(masm, &exit, 1.5, -1.5, 0.5, -0.5, NaN, -NaN, inf, -inf);
+  TestFloat16x8Neg(masm, &exit, 1.5, -1.5, 0.5, -0.5, NaN, -NaN, inf, -inf);
   TestFloat32x4Abs(masm, &exit, 1.5, -1.5, 0.5, -0.5);
   TestFloat32x4Neg(masm, &exit, 1.5, -1.5, 0.5, -0.5);
   TestFloat64x2Abs(masm, &exit, 1.75, -1.75);
   TestFloat64x2Neg(masm, &exit, 1.75, -1.75);
 
+  // There are two different paths in codegen, depend on the equality of dst and
+  // src register.
+  TestFloat64x4Neg<true>(masm, &exit, 1.75, -1.75, 0.5, -0.5);
+  TestFloat64x4Neg<false>(masm, &exit, 1.75, -1.75, 0.5, -0.5);
+  TestFloat32x8Neg<true>(masm, &exit, 1.75, -1.75, 0.5, -0.5, NaN, -NaN, inf,
+                         -inf);
+  TestFloat32x8Neg<false>(masm, &exit, 1.5, -1.5, 0.5, -0.5, NaN, -NaN, inf,
+                          -inf);
+
   __ xorq(rax, rax);  // Success.
   __ bind(&exit);
   ExitCode(masm);
@@ -1134,13 +1343,13 @@ TEST_F(MacroAssemblerX64Test, S256Select) {
   CpuFeatureScope avx2_scope(masm, AVX2);
 
   // Load src1, src2, mask
-  __ vmovdqu(src1, Operand(arg_reg_1, 0));
-  __ vmovdqu(src2, Operand(arg_reg_2, 0));
-  __ vmovdqu(mask, Operand(arg_reg_3, 0));
+  __ vmovdqu(src1, Operand(kCArgRegs[0], 0));
+  __ vmovdqu(src2, Operand(kCArgRegs[1], 0));
+  __ vmovdqu(mask, Operand(kCArgRegs[2], 0));
   // Bitselect
   __ S256Select(dst, mask, src1, src2, tmp);
   // Store result
-  __ vmovdqu(Operand(arg_reg_4, 0), dst);
+  __ vmovdqu(Operand(kCArgRegs[3], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
@@ -1226,7 +1435,11 @@ TEST_F(MacroAssemblerX64Test, DeoptExitSizeIsFixed) {
   for (int i = 0; i < kDeoptimizeKindCount; i++) {
     DeoptimizeKind kind = static_cast<DeoptimizeKind>(i);
     Label before_exit;
+
     masm.bind(&before_exit);
+    if (kind == DeoptimizeKind::kLazy) {
+      masm.CodeEntry();
+    }
     Builtin target = Deoptimizer::GetDeoptimizationEntry(kind);
     masm.CallForDeoptimization(target, 42, &before_exit, kind, &before_exit,
                                nullptr);
@@ -1251,12 +1464,12 @@ TEST_F(MacroAssemblerX64Test, I64x2Mul) {
   const XMMRegister tmp2 = xmm4;
 
   // Load array
-  __ movdqu(lhs, Operand(arg_reg_1, 0));
-  __ movdqu(rhs, Operand(arg_reg_2, 0));
+  __ movdqu(lhs, Operand(kCArgRegs[0], 0));
+  __ movdqu(rhs, Operand(kCArgRegs[1], 0));
   // Calculation
   __ I64x2Mul(dst, lhs, rhs, tmp1, tmp2);
   // Store result array
-  __ movdqu(Operand(arg_reg_3, 0), dst);
+  __ movdqu(Operand(kCArgRegs[2], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
@@ -1312,12 +1525,12 @@ TEST_F(MacroAssemblerX64Test, I64x4Mul) {
   CpuFeatureScope avx2_scope(masm, AVX2);
 
   // Load array
-  __ vmovdqu(lhs, Operand(arg_reg_1, 0));
-  __ vmovdqu(rhs, Operand(arg_reg_2, 0));
+  __ vmovdqu(lhs, Operand(kCArgRegs[0], 0));
+  __ vmovdqu(rhs, Operand(kCArgRegs[1], 0));
   // Calculation
   __ I64x4Mul(dst, lhs, rhs, tmp1, tmp2);
   // Store result array
-  __ vmovdqu(Operand(arg_reg_3, 0), dst);
+  __ vmovdqu(Operand(kCArgRegs[2], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
@@ -1375,12 +1588,12 @@ TEST_F(MacroAssemblerX64Test, I64x4Mul) {
     CpuFeatureScope avx2_scope(masm, AVX2);                                   \
                                                                               \
     /* src is register */                                                     \
-    __ name(ymm0, arg_reg_1);                                                 \
-    __ vmovdqu(Operand(arg_reg_3, 0), ymm0);                                  \
+    __ name(ymm0, kCArgRegs[0]);                                              \
+    __ vmovdqu(Operand(kCArgRegs[2], 0), ymm0);                               \
                                                                               \
     /* src is address*/                                                       \
-    __ name(ymm0, Operand(arg_reg_2, 0));                                     \
-    __ vmovdqu(Operand(arg_reg_4, 0), ymm0);                                  \
+    __ name(ymm0, Operand(kCArgRegs[1], 0));                                  \
+    __ vmovdqu(Operand(kCArgRegs[3], 0), ymm0);                               \
     __ ret(0);                                                                \
                                                                               \
     CodeDesc desc;                                                            \
@@ -1430,19 +1643,19 @@ TEST_F(MacroAssemblerX64Test, F64x4Min) {
   CpuFeatureScope avx2_scope(masm, AVX2);
 
   // Load array
-  __ vmovdqu(lhs, Operand(arg_reg_1, 0));
-  __ vmovdqu(rhs, Operand(arg_reg_2, 0));
+  __ vmovdqu(lhs, Operand(kCArgRegs[0], 0));
+  __ vmovdqu(rhs, Operand(kCArgRegs[1], 0));
   // Calculation
   __ F64x4Min(dst, lhs, rhs, tmp);
   // Store result array
-  __ vmovdqu(Operand(arg_reg_3, 0), dst);
+  __ vmovdqu(Operand(kCArgRegs[2], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
   __ GetCode(i_isolate(), &desc);
 
 #ifdef OBJECT_PRINT
-  Handle<Code> code =
+  DirectHandle<Code> code =
       Factory::CodeBuilder(i_isolate(), desc, CodeKind::FOR_TESTING).Build();
   StdoutStream os;
   Print(*code, os);
@@ -1497,19 +1710,19 @@ TEST_F(MacroAssemblerX64Test, F64x4Max) {
   CpuFeatureScope avx2_scope(masm, AVX2);
 
   // Load array
-  __ vmovdqu(lhs, Operand(arg_reg_1, 0));
-  __ vmovdqu(rhs, Operand(arg_reg_2, 0));
+  __ vmovdqu(lhs, Operand(kCArgRegs[0], 0));
+  __ vmovdqu(rhs, Operand(kCArgRegs[1], 0));
   // Calculation
   __ F64x4Max(dst, lhs, rhs, tmp);
   // Store result array
-  __ vmovdqu(Operand(arg_reg_3, 0), dst);
+  __ vmovdqu(Operand(kCArgRegs[2], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
   __ GetCode(i_isolate(), &desc);
 
 #ifdef OBJECT_PRINT
-  Handle<Code> code =
+  DirectHandle<Code> code =
       Factory::CodeBuilder(i_isolate(), desc, CodeKind::FOR_TESTING).Build();
   StdoutStream os;
   Print(*code, os);
@@ -1564,19 +1777,19 @@ TEST_F(MacroAssemblerX64Test, F32x8Min) {
   CpuFeatureScope avx2_scope(masm, AVX2);
 
   // Load array
-  __ vmovdqu(lhs, Operand(arg_reg_1, 0));
-  __ vmovdqu(rhs, Operand(arg_reg_2, 0));
+  __ vmovdqu(lhs, Operand(kCArgRegs[0], 0));
+  __ vmovdqu(rhs, Operand(kCArgRegs[1], 0));
   // Calculation
   __ F32x8Min(dst, lhs, rhs, tmp);
   // Store result array
-  __ vmovdqu(Operand(arg_reg_3, 0), dst);
+  __ vmovdqu(Operand(kCArgRegs[2], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
   __ GetCode(i_isolate(), &desc);
 
 #ifdef OBJECT_PRINT
-  Handle<Code> code =
+  DirectHandle<Code> code =
       Factory::CodeBuilder(i_isolate(), desc, CodeKind::FOR_TESTING).Build();
   StdoutStream os;
   Print(*code, os);
@@ -1634,19 +1847,19 @@ TEST_F(MacroAssemblerX64Test, F32x8Max) {
   CpuFeatureScope avx2_scope(masm, AVX2);
 
   // Load array
-  __ vmovdqu(lhs, Operand(arg_reg_1, 0));
-  __ vmovdqu(rhs, Operand(arg_reg_2, 0));
+  __ vmovdqu(lhs, Operand(kCArgRegs[0], 0));
+  __ vmovdqu(rhs, Operand(kCArgRegs[1], 0));
   // Calculation
   __ F32x8Max(dst, lhs, rhs, tmp);
   // Store result array
-  __ vmovdqu(Operand(arg_reg_3, 0), dst);
+  __ vmovdqu(Operand(kCArgRegs[2], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
   __ GetCode(i_isolate(), &desc);
 
 #ifdef OBJECT_PRINT
-  Handle<Code> code =
+  DirectHandle<Code> code =
       Factory::CodeBuilder(i_isolate(), desc, CodeKind::FOR_TESTING).Build();
   StdoutStream os;
   Print(*code, os);
@@ -1707,8 +1920,8 @@ void RunExtMulTest(Isolate* isolate, OpType expected_op) {
   CpuFeatureScope avx2_scope(masm, AVX2);
 
   // Load array
-  __ vmovdqu(lhs, Operand(arg_reg_1, 0));
-  __ vmovdqu(rhs, Operand(arg_reg_2, 0));
+  __ vmovdqu(lhs, Operand(kCArgRegs[0], 0));
+  __ vmovdqu(rhs, Operand(kCArgRegs[1], 0));
 
   bool is_signed = std::is_signed_v<T>;
   // Calculation
@@ -1727,7 +1940,7 @@ void RunExtMulTest(Isolate* isolate, OpType expected_op) {
   }
 
   // Store result array
-  __ vmovdqu(Operand(arg_reg_3, 0), dst);
+  __ vmovdqu(Operand(kCArgRegs[2], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
@@ -1817,11 +2030,11 @@ TEST_F(MacroAssemblerX64Test, I32x8ExtAddPairwiseI16x16S) {
   CpuFeatureScope avx2_scope(masm, AVX2);
 
   // Load array
-  __ vmovdqu(src, Operand(arg_reg_1, 0));
+  __ vmovdqu(src, Operand(kCArgRegs[0], 0));
   // Calculation
   __ I32x8ExtAddPairwiseI16x16S(dst, src, tmp);
   // Store result array
-  __ vmovdqu(Operand(arg_reg_2, 0), dst);
+  __ vmovdqu(Operand(kCArgRegs[1], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
@@ -1875,11 +2088,11 @@ TEST_F(MacroAssemblerX64Test, I16x16ExtAddPairwiseI8x32S) {
   CpuFeatureScope avx2_scope(masm, AVX2);
 
   // Load array
-  __ vmovdqu(src, Operand(arg_reg_1, 0));
+  __ vmovdqu(src, Operand(kCArgRegs[0], 0));
   // Calculation
   __ I16x16ExtAddPairwiseI8x32S(dst, src, tmp);
   // Store result array
-  __ vmovdqu(Operand(arg_reg_2, 0), dst);
+  __ vmovdqu(Operand(kCArgRegs[1], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
@@ -1937,11 +2150,11 @@ TEST_F(MacroAssemblerX64Test, I32x8ExtAddPairwiseI16x16U) {
   CpuFeatureScope avx2_scope(masm, AVX2);
 
   // Load array
-  __ vmovdqu(src, Operand(arg_reg_1, 0));
+  __ vmovdqu(src, Operand(kCArgRegs[0], 0));
   // Calculation
   __ I32x8ExtAddPairwiseI16x16U(dst, src, tmp);
   // Store result array
-  __ vmovdqu(Operand(arg_reg_2, 0), dst);
+  __ vmovdqu(Operand(kCArgRegs[1], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
@@ -1994,11 +2207,11 @@ TEST_F(MacroAssemblerX64Test, I16x16ExtAddPairwiseI8x32U) {
   CpuFeatureScope avx2_scope(masm, AVX2);
 
   // Load array
-  __ vmovdqu(src, Operand(arg_reg_1, 0));
+  __ vmovdqu(src, Operand(kCArgRegs[0], 0));
   // Calculation
   __ I16x16ExtAddPairwiseI8x32U(dst, src, tmp);
   // Store result array
-  __ vmovdqu(Operand(arg_reg_2, 0), dst);
+  __ vmovdqu(Operand(kCArgRegs[1], 0), dst);
   __ ret(0);
 
   CodeDesc desc;
@@ -2049,9 +2262,9 @@ TEST_F(MacroAssemblerX64Test, F64x4Splat) {
   CpuFeatureScope avx_scope(masm, AVX);
   CpuFeatureScope avx2_scope(masm, AVX2);
 
-  __ vmovsd(xmm1, Operand(arg_reg_1, 0));
+  __ vmovsd(xmm1, Operand(kCArgRegs[0], 0));
   __ F64x4Splat(ymm2, xmm1);
-  __ vmovdqu(Operand(arg_reg_2, 0), ymm2);
+  __ vmovdqu(Operand(kCArgRegs[1], 0), ymm2);
   __ ret(0);
 
   CodeDesc desc;
@@ -2083,9 +2296,9 @@ TEST_F(MacroAssemblerX64Test, F32x8Splat) {
   CpuFeatureScope avx_scope(masm, AVX);
   CpuFeatureScope avx2_scope(masm, AVX2);
 
-  __ vmovss(xmm1, Operand(arg_reg_1, 0));
+  __ vmovss(xmm1, Operand(kCArgRegs[0], 0));
   __ F32x8Splat(ymm2, xmm1);
-  __ vmovdqu(Operand(arg_reg_2, 0), ymm2);
+  __ vmovdqu(Operand(kCArgRegs[1], 0), ymm2);
   __ ret(0);
 
   CodeDesc desc;
@@ -2106,6 +2319,312 @@ TEST_F(MacroAssemblerX64Test, F32x8Splat) {
   }
 }
 
+TEST_F(MacroAssemblerX64Test, I32x8SConvertF32x8) {
+  if (!CpuFeatures::IsSupported(AVX) || !CpuFeatures::IsSupported(AVX2)) return;
+  Isolate* isolate = i_isolate();
+  HandleScope handles(isolate);
+  auto buffer = AllocateAssemblerBuffer();
+  MacroAssembler assembler(isolate, v8::internal::CodeObjectRequired::kYes,
+                           buffer->CreateView());
+
+  MacroAssembler* masm = &assembler;
+
+  __ set_root_array_available(false);
+
+  const YMMRegister dst = ymm0;
+  const YMMRegister src = ymm1;
+  const YMMRegister tmp = ymm2;
+  const Register scratch = r10;
+
+  CpuFeatureScope avx_scope(masm, AVX);
+  CpuFeatureScope avx2_scope(masm, AVX2);
+
+  // Load array
+  __ vmovdqu(src, Operand(kCArgRegs[0], 0));
+  // Calculation
+  __ I32x8SConvertF32x8(dst, src, tmp, scratch);
+  // Store result array
+  __ vmovdqu(Operand(kCArgRegs[1], 0), dst);
+  __ ret(0);
+
+  CodeDesc desc;
+  __ GetCode(i_isolate(), &desc);
+
+  PrintCode(isolate, desc);
+
+  buffer->MakeExecutable();
+  // Call the function from C++.
+  auto f = GeneratedCode<F13>::FromBuffer(i_isolate(), buffer->start());
+
+  auto convert_to_int = [=](double val) -> int32_t {
+    if (std::isnan(val)) return 0;
+    if (val < kMinInt) return kMinInt;
+    if (val > kMaxInt) return kMaxInt;
+    return static_cast<int>(val);
+  };
+
+  constexpr float float_max = std::numeric_limits<float>::max();
+  constexpr float float_min = std::numeric_limits<float>::min();
+  constexpr float NaN = std::numeric_limits<float>::quiet_NaN();
+
+  std::vector<std::array<float, 8>> test_cases = {
+      {32.4, 2.5, 12.4, 62.346, 235.6, 2.36, 1253.4, 63.46},
+      {34.5, 2.63, 234.6, 34.68, -234.6, -1.264, -23.6, -2.36},
+      {NaN, 0, 0, -0, NaN, -NaN, -0, -0},
+      {float_max, float_max, float_min, float_min, float_max + 1, float_max + 1,
+       float_min - 1, float_min - 1}};
+  float input[8];
+  int32_t output[8];
+
+  for (const auto& arr : test_cases) {
+    for (int i = 0; i < 8; i++) {
+      input[i] = arr[i];
+    }
+    f.Call(input, output);
+    for (int i = 0; i < 8; i++) {
+      CHECK_EQ(output[i], convert_to_int(input[i]));
+    }
+  }
+}
+
+TEST_F(MacroAssemblerX64Test, I16x8SConvertF16x8) {
+  if (!CpuFeatures::IsSupported(F16C) || !CpuFeatures::IsSupported(AVX) ||
+      !CpuFeatures::IsSupported(AVX2)) {
+    return;
+  }
+
+  Isolate* isolate = i_isolate();
+  HandleScope handles(isolate);
+  auto buffer = AllocateAssemblerBuffer();
+  MacroAssembler assembler(isolate, v8::internal::CodeObjectRequired::kYes,
+                           buffer->CreateView());
+
+  MacroAssembler* masm = &assembler;
+
+  __ set_root_array_available(false);
+
+  const YMMRegister dst = ymm0;
+  const YMMRegister src = ymm1;
+  const YMMRegister tmp = ymm2;
+  const Register scratch = r10;
+
+  CpuFeatureScope f16c_scope(masm, F16C);
+  CpuFeatureScope avx_scope(masm, AVX);
+  CpuFeatureScope avx2_scope(masm, AVX2);
+
+  __ vmovdqu(src, Operand(kCArgRegs[0], 0));
+  __ I16x8SConvertF16x8(dst, src, tmp, scratch);
+  __ vmovdqu(Operand(kCArgRegs[1], 0), dst);
+  __ ret(0);
+
+  CodeDesc desc;
+  __ GetCode(i_isolate(), &desc);
+
+  PrintCode(isolate, desc);
+
+  buffer->MakeExecutable();
+  auto f = GeneratedCode<F14>::FromBuffer(i_isolate(), buffer->start());
+
+  auto convert_to_int = [=](float val) -> int16_t {
+    if (std::isnan(val)) return 0;
+    if (val < kMinInt16) return kMinInt16;
+    if (val > kMaxInt16) return kMaxInt16;
+    return static_cast<int16_t>(val);
+  };
+
+  float fp16_max = 65504;
+  float fp16_min = -fp16_max;
+  float NaN = std::numeric_limits<float>::quiet_NaN();
+  float neg_zero = base::bit_cast<float>(0x80000000);
+  float inf = std::numeric_limits<float>::infinity();
+  float neg_inf = -std::numeric_limits<float>::infinity();
+
+  std::vector<std::array<float, 8>> test_cases = {
+      {32.4, 2.5, 12.4, 62.346, 235.6, 2.36, 1253.4, 63.46},
+      {34.5, 2.63, 234.6, 34.68, -234.6, -1.264, -23.6, -2.36},
+      {NaN, 0, 0, neg_zero, NaN, -NaN, neg_zero, neg_zero},
+      {fp16_max, fp16_max, fp16_min, fp16_min, fp16_max + 1, inf, fp16_min - 1,
+       neg_inf}};
+  uint16_t input[16];
+  int16_t output[16];
+
+  for (const auto& arr : test_cases) {
+    for (int i = 0; i < 8; i++) {
+      input[i] = fp16_ieee_from_fp32_value(arr[i]);
+    }
+    f.Call(input, output);
+    for (int i = 0; i < 8; i++) {
+      CHECK_EQ(output[i], convert_to_int(arr[i]));
+    }
+  }
+}
+
+TEST_F(MacroAssemblerX64Test, I16x8TruncF16x8U) {
+  if (!CpuFeatures::IsSupported(F16C) || !CpuFeatures::IsSupported(AVX) ||
+      !CpuFeatures::IsSupported(AVX2)) {
+    return;
+  }
+
+  Isolate* isolate = i_isolate();
+  HandleScope handles(isolate);
+  auto buffer = AllocateAssemblerBuffer();
+  MacroAssembler assembler(isolate, v8::internal::CodeObjectRequired::kYes,
+                           buffer->CreateView());
+
+  MacroAssembler* masm = &assembler;
+
+  __ set_root_array_available(false);
+
+  const YMMRegister dst = ymm0;
+  const YMMRegister src = ymm1;
+  const YMMRegister tmp = ymm2;
+
+  CpuFeatureScope f16c_scope(masm, F16C);
+  CpuFeatureScope avx_scope(masm, AVX);
+  CpuFeatureScope avx2_scope(masm, AVX2);
+
+  __ vmovdqu(src, Operand(kCArgRegs[0], 0));
+  __ I16x8TruncF16x8U(dst, src, tmp);
+  __ vmovdqu(Operand(kCArgRegs[1], 0), dst);
+  __ ret(0);
+
+  CodeDesc desc;
+  __ GetCode(i_isolate(), &desc);
+
+  PrintCode(isolate, desc);
+
+  buffer->MakeExecutable();
+  auto f = GeneratedCode<F15>::FromBuffer(i_isolate(), buffer->start());
+
+  auto convert_to_uint = [=](float val) -> uint16_t {
+    if (std::isnan(val)) return 0;
+    if (val < 0) return 0;
+    if (val > kMaxUInt16) return kMaxUInt16;
+    return static_cast<uint16_t>(val);
+  };
+
+  float fp16_max = 65504;
+  float fp16_min = -fp16_max;
+  float NaN = std::numeric_limits<float>::quiet_NaN();
+  float neg_zero = base::bit_cast<float>(0x80000000);
+  float inf = std::numeric_limits<float>::infinity();
+  float neg_inf = -std::numeric_limits<float>::infinity();
+
+  std::vector<std::array<float, 8>> test_cases = {
+      {32.4, 2.5, 12.4, 62.346, 235.6, 2.36, 1253.4, 63.46},
+      {34.5, 2.63, 234.6, 34.68, -234.6, -1.264, -23.6, -2.36},
+      {NaN, 0, 0, neg_zero, NaN, -NaN, neg_zero, neg_zero},
+      {fp16_max, fp16_max, fp16_min, fp16_min, fp16_max + 1, inf, fp16_min - 1,
+       neg_inf}};
+  uint16_t input[16];
+  uint16_t output[16];
+
+  for (const auto& arr : test_cases) {
+    for (int i = 0; i < 8; i++) {
+      input[i] = fp16_ieee_from_fp32_value(arr[i]);
+    }
+    f.Call(input, output);
+    for (int i = 0; i < 8; i++) {
+      CHECK_EQ(output[i], convert_to_uint(fp16_ieee_to_fp32_value(input[i])));
+    }
+  }
+}
+
+TEST_F(MacroAssemblerX64Test, Cvtpd2ph) {
+  if (!CpuFeatures::IsSupported(F16C) || !CpuFeatures::IsSupported(AVX)) {
+    return;
+  }
+
+  Isolate* isolate = i_isolate();
+  HandleScope handles(isolate);
+  auto buffer = AllocateAssemblerBuffer();
+  MacroAssembler assembler(isolate, v8::internal::CodeObjectRequired::kYes,
+                           buffer->CreateView());
+
+  MacroAssembler* masm = &assembler;
+
+  __ set_root_array_available(false);
+
+  const XMMRegister dst = xmm0;
+  const XMMRegister src = xmm1;
+  const Register tmp = r8;
+
+  CpuFeatureScope f16c_scope(masm, F16C);
+  CpuFeatureScope avx_scope(masm, AVX);
+
+  __ vmovsd(src, Operand(kCArgRegs[0], 0));
+  __ Cvtpd2ph(dst, src, tmp);
+  __ vmovss(Operand(kCArgRegs[1], 0), dst);
+  __ ret(0);
+
+  CodeDesc desc;
+  __ GetCode(i_isolate(), &desc);
+
+  PrintCode(isolate, desc);
+
+  buffer->MakeExecutable();
+  auto f = GeneratedCode<F16>::FromBuffer(i_isolate(), buffer->start());
+
+  std::vector<double> test_cases = {
+      // Float16 subnormal numbers.
+      8.940696716308592e-8, 0.000060945749282836914, 0.00006094574928283692,
+      // Float16 normal numbers.
+      0.000061035154431010596, 0.00006103515625, 0.0000610649585723877,
+      0.00006106495857238771, 0.00006112456321716307, -999.75,
+      // Underflow to zero.
+      2.980232594040899e-8, 2.9802320611338473e-8,
+      // Overflow to infinity.
+      65536,
+      // An integer which rounds down under ties-to-even when cast to
+      // float16.
+      2049,
+      // An integer which rounds up under ties-to-even when cast to
+      // float16.
+      2051,
+      // Smallest normal float16.
+      0.00006103515625,
+      // Largest subnormal float16.
+      0.00006097555160522461,
+      // Smallest float16.
+      5.960464477539063e-8,
+      // Largest double which rounds to 0 when cast to
+      // float16.
+      2.9802322387695312e-8,
+      // Smallest double which does not round to 0 when
+      // cast to float16.
+      2.980232238769532e-8,
+      // A double which rounds up to a subnormal under
+      // ties-to-even when cast to float16.
+      8.940696716308594e-8,
+      // A double which rounds down to a subnormal under
+      // ties-to-even when cast to float16.
+      1.4901161193847656e-7,
+      // The next double above the one on the previous
+      // line one.
+      1.490116119384766e-7,
+      // Max finite float16.
+      65504,
+      // Smallest double which rounds to infinity when cast to float16.
+      65520,
+      // Largest double which does not round to infinity
+      // when cast to float16.
+      65519.99999999999,
+      // Smallest double which rounds to a
+      // non-subnormal when cast to float16.
+      0.000061005353927612305,
+      // Largest double which rounds to a subnormal when
+      // cast to float16.
+      0.0000610053539276123};
+  double input;
+  uint16_t output[2];
+
+  for (const auto& val : test_cases) {
+    input = val;
+    f.Call(&input, output);
+    CHECK_EQ(output[0], DoubleToFloat16(val));
+  }
+}
+
 #undef __
 
 }  // namespace test_macro_assembler_x64
