@@ -28,12 +28,13 @@ namespace compiler {
 class RevecTest : public TestWithIsolateAndZone {
  public:
   RevecTest()
-      : TestWithIsolateAndZone(kCompressGraphZone),
-        graph_(zone()),
+      : graph_(zone()),
         common_(zone()),
         machine_(zone(), MachineRepresentation::kWord64,
                  MachineOperatorBuilder::Flag::kAllOptionalOps),
-        mcgraph_(&graph_, &common_, &machine_) {}
+        mcgraph_(&graph_, &common_, &machine_),
+        source_positions_(
+            mcgraph()->zone()->New<SourcePositionTable>(mcgraph()->graph())) {}
 
   void TestBinOp(const Operator* bin_op,
                  const IrOpcode::Value expected_simd256_op_code);
@@ -45,16 +46,18 @@ class RevecTest : public TestWithIsolateAndZone {
   void TestLoadSplat(LoadTransformation transform, const Operator* bin_op,
                      LoadTransformation expected_transform);
 
-  Graph* graph() { return &graph_; }
+  TFGraph* graph() { return &graph_; }
   CommonOperatorBuilder* common() { return &common_; }
   MachineOperatorBuilder* machine() { return &machine_; }
   MachineGraph* mcgraph() { return &mcgraph_; }
+  SourcePositionTable* source_positions() { return source_positions_; }
 
  private:
-  Graph graph_;
+  TFGraph graph_;
   CommonOperatorBuilder common_;
   MachineOperatorBuilder machine_;
   MachineGraph mcgraph_;
+  SourcePositionTable* source_positions_;
 };
 
 // Create a graph which perform binary operation on two 256 bit vectors(a, b),
@@ -110,7 +113,7 @@ void RevecTest::TestBinOp(const Operator* bin_op,
   graph()->SetSimd(true);
 
   // Test whether the graph can be revectorized
-  Revectorizer revec(zone(), graph(), mcgraph());
+  Revectorizer revec(zone(), graph(), mcgraph(), source_positions());
   EXPECT_TRUE(revec.TryRevectorize(nullptr));
 
   // Test whether the graph has been revectorized
@@ -239,7 +242,7 @@ TEST_F(RevecTest, ReorderLoadChain1) {
   graph()->SetSimd(true);
 
   // Test whether the graph can be revectorized
-  Revectorizer revec(zone(), graph(), mcgraph());
+  Revectorizer revec(zone(), graph(), mcgraph(), source_positions());
   EXPECT_TRUE(revec.TryRevectorize(nullptr));
 }
 
@@ -299,7 +302,7 @@ TEST_F(RevecTest, ReorderLoadChain2) {
   graph()->SetSimd(true);
 
   // Test whether the graph can be revectorized
-  Revectorizer revec(zone(), graph(), mcgraph());
+  Revectorizer revec(zone(), graph(), mcgraph(), source_positions());
   EXPECT_TRUE(revec.TryRevectorize(nullptr));
 }
 
@@ -357,7 +360,7 @@ void RevecTest::TestShiftOp(const Operator* shift_op,
   graph()->RecordSimdStore(store1);
   graph()->SetSimd(true);
 
-  Revectorizer revec(zone(), graph(), mcgraph());
+  Revectorizer revec(zone(), graph(), mcgraph(), source_positions());
   bool result = revec.TryRevectorize(nullptr);
 
   if (CpuFeatures::IsSupported(AVX2)) {
@@ -426,7 +429,7 @@ void RevecTest::TestSplatOp(const Operator* splat_op,
   graph()->RecordSimdStore(store1);
   graph()->SetSimd(true);
 
-  Revectorizer revec(zone(), graph(), mcgraph());
+  Revectorizer revec(zone(), graph(), mcgraph(), source_positions());
   bool result = revec.TryRevectorize(nullptr);
 
   EXPECT_TRUE(result);
@@ -513,7 +516,7 @@ TEST_F(RevecTest, ShuffleForSplat) {
   graph()->RecordSimdStore(store1);
   graph()->SetSimd(true);
 
-  Revectorizer revec(zone(), graph(), mcgraph());
+  Revectorizer revec(zone(), graph(), mcgraph(), source_positions());
   EXPECT_TRUE(revec.TryRevectorize(nullptr));
 
   // Test whether the graph has been revectorized
@@ -544,7 +547,8 @@ void RevecTest::TestLoadSplat(
                                 offset, start, start);
 
   Node* loadSplat = graph()->NewNode(
-      machine()->LoadTransform(MemoryAccessKind::kProtected, load_transform),
+      machine()->LoadTransform(MemoryAccessKind::kProtectedByTrapHandler,
+                               load_transform),
       base, a, base, start);
 
   LoadRepresentation load_rep(MachineType::Simd128());
@@ -572,7 +576,7 @@ void RevecTest::TestLoadSplat(
   graph()->RecordSimdStore(store1);
   graph()->SetSimd(true);
 
-  Revectorizer revec(zone(), graph(), mcgraph());
+  Revectorizer revec(zone(), graph(), mcgraph(), source_positions());
   bool result = revec.TryRevectorize(nullptr);
 
   EXPECT_TRUE(result);
@@ -645,7 +649,7 @@ TEST_F(RevecTest, StoreDependencyCheck) {
   graph()->SetSimd(true);
 
   // Test whether the graph can be revectorized
-  Revectorizer revec(zone(), graph(), mcgraph());
+  Revectorizer revec(zone(), graph(), mcgraph(), source_positions());
   EXPECT_FALSE(revec.TryRevectorize(nullptr));
 }
 
@@ -682,7 +686,7 @@ TEST_F(RevecTest, S128Zero) {
   graph()->SetSimd(true);
 
   // Test whether the graph can be revectorized
-  Revectorizer revec(zone(), graph(), mcgraph());
+  Revectorizer revec(zone(), graph(), mcgraph(), source_positions());
   EXPECT_TRUE(revec.TryRevectorize(nullptr));
 }
 
