@@ -27,6 +27,7 @@
 
 #include "src/handles/global-handles.h"
 
+#include "include/cppgc/macros.h"
 #include "include/v8-embedder-heap.h"
 #include "include/v8-function.h"
 #include "src/api/api-inl.h"
@@ -45,15 +46,15 @@ namespace internal {
 namespace {
 
 struct TracedReferenceWrapper {
+  void Trace(cppgc::Visitor* v) { v->Trace(handle); }
   v8::TracedReference<v8::Object> handle;
 };
 
 class NonRootingEmbedderRootsHandler final : public v8::EmbedderRootsHandler {
- public:
-  bool IsRoot(const v8::TracedReference<v8::Value>& handle) final {
-    return false;
-  }
+  CPPGC_STACK_ALLOCATED();
 
+ public:
+  NonRootingEmbedderRootsHandler() : v8::EmbedderRootsHandler() {}
   void ResetRoot(const v8::TracedReference<v8::Value>& handle) final {
     for (auto* wrapper : wrappers_) {
       if (wrapper->handle == handle) {
@@ -135,6 +136,7 @@ template <typename ConstructFunction, typename ModifierFunction,
 void WeakHandleTest(v8::Isolate* isolate, ConstructFunction construct_function,
                     ModifierFunction modifier_function, GCFunction gc_function,
                     SurvivalMode survives) {
+  ManualGCScope manual_gc_scope(reinterpret_cast<internal::Isolate*>(isolate));
   v8::HandleScope scope(isolate);
   v8::Local<v8::Context> context = v8::Context::New(isolate);
   v8::Context::Scope context_scope(context);
@@ -190,8 +192,6 @@ TEST_F(GlobalHandlesTest, EternalHandles) {
   Isolate* isolate = i_isolate();
   v8::Isolate* v8_isolate = reinterpret_cast<v8::Isolate*>(isolate);
   EternalHandles* eternal_handles = isolate->eternal_handles();
-  DisableConservativeStackScanningScopeForTesting no_stack_scanning(
-      isolate->heap());
 
   // Create a number of handles that will not be on a block boundary
   const int kArrayLength = 2048 - 1;
@@ -208,7 +208,7 @@ TEST_F(GlobalHandlesTest, EternalHandles) {
               v8::Integer::New(v8_isolate, i))
         .FromJust();
     // Create with internal api
-    eternal_handles->Create(isolate, *v8::Utils::OpenHandle(*object),
+    eternal_handles->Create(isolate, *v8::Utils::OpenDirectHandle(*object),
                             &indices[i]);
     // Create with external api
     CHECK(eternals[i].IsEmpty());
@@ -216,7 +216,13 @@ TEST_F(GlobalHandlesTest, EternalHandles) {
     CHECK(!eternals[i].IsEmpty());
   }
 
-  InvokeMemoryReducingMajorGCs(isolate);
+  {
+    // We need to invoke GC without stack, otherwise some objects may not be
+    // reclaimed because of conservative stack scanning.
+    DisableConservativeStackScanningScopeForTesting no_stack_scanning(
+        isolate->heap());
+    InvokeMemoryReducingMajorGCs(isolate);
+  }
 
   for (int i = 0; i < kArrayLength; i++) {
     for (int j = 0; j < 2; j++) {
@@ -355,16 +361,6 @@ TEST_F(GlobalHandlesTest, WeakHandleToUnmodifiedJSApiObjectDiesOnScavenge) {
       SurvivalMode::kDies);
 }
 
-TEST_F(GlobalHandlesTest,
-       TracedReferenceToUnmodifiedJSApiObjectDiesOnScavenge) {
-  if (v8_flags.single_generation) return;
-
-  ManualGCScope manual_gc(i_isolate());
-  TracedReferenceTestWithScavenge(
-      &ConstructJSApiObject<TracedReferenceWrapper>,
-      [](TracedReferenceWrapper* fp) {}, SurvivalMode::kDies);
-}
-
 TEST_F(GlobalHandlesTest,
        TracedReferenceToJSApiObjectWithIdentityHashSurvivesScavenge) {
   if (v8_flags.single_generation) return;
@@ -378,9 +374,9 @@ TEST_F(GlobalHandlesTest,
       &ConstructJSApiObject<TracedReferenceWrapper>,
       [this, &weakmap, isolate](TracedReferenceWrapper* fp) {
         v8::HandleScope scope(v8_isolate());
-        Handle<JSReceiver> key =
-            Utils::OpenHandle(*fp->handle.Get(v8_isolate()));
-        Handle<Smi> smi(Smi::FromInt(23), isolate);
+        DirectHandle<JSReceiver> key =
+            Utils::OpenDirectHandle(*fp->handle.Get(v8_isolate()));
+        DirectHandle<Smi> smi(Smi::FromInt(23), isolate);
         int32_t hash = Object::GetOrCreateHash(*key, isolate).value();
         JSWeakCollection::Set(weakmap, key, smi, hash);
       },
@@ -492,6 +488,7 @@ void ForceMajorGC1(const v8::WeakCallbackInfo<FlagAndHandles>& data) {
 
 TEST_F(GlobalHandlesTest, GCFromWeakCallbacks) {
   v8::Isolate* isolate = v8_isolate();
+  ManualGCScope manual_gc_scope(i_isolate());
   DisableConservativeStackScanningScopeForTesting no_stack_scanning(
       i_isolate()->heap());
   v8::HandleScope scope(isolate);
@@ -594,18 +591,25 @@ TEST_F(GlobalHandlesTest, TotalSizeRegularNode) {
   v8::Isolate* isolate = v8_isolate();
   v8::HandleScope scope(isolate);
 
+  // This is not necessarily zero, if the implementation of tests uses global
+  // handles.
+  size_t initial_total = i_isolate()->global_handles()->TotalSize();
+  size_t initial_used = i_isolate()->global_handles()->UsedSize();
+
   v8::Global<v8::Object>* global = new Global<v8::Object>();
-  CHECK_EQ(i_isolate()->global_handles()->TotalSize(), 0);
-  CHECK_EQ(i_isolate()->global_handles()->UsedSize(), 0);
+  CHECK_EQ(i_isolate()->global_handles()->TotalSize(), initial_total);
+  CHECK_EQ(i_isolate()->global_handles()->UsedSize(), initial_used);
   ConstructJSObject(isolate, global);
-  CHECK_GT(i_isolate()->global_handles()->TotalSize(), 0);
-  CHECK_GT(i_isolate()->global_handles()->UsedSize(), 0);
+  CHECK_GE(i_isolate()->global_handles()->TotalSize(), initial_total);
+  CHECK_GT(i_isolate()->global_handles()->UsedSize(), initial_used);
   delete global;
-  CHECK_GT(i_isolate()->global_handles()->TotalSize(), 0);
-  CHECK_EQ(i_isolate()->global_handles()->UsedSize(), 0);
+  CHECK_GE(i_isolate()->global_handles()->TotalSize(), initial_total);
+  CHECK_EQ(i_isolate()->global_handles()->UsedSize(), initial_used);
 }
 
 TEST_F(GlobalHandlesTest, TotalSizeTracedNode) {
+  DisableConservativeStackScanningScopeForTesting no_stack_scanning(
+      i_isolate()->heap());
   ManualGCScope manual_gc(i_isolate());
   v8::Isolate* isolate = v8_isolate();
   v8::HandleScope scope(isolate);
