@@ -6,7 +6,7 @@
 Convenience wrapper for compiling V8 with gn/ninja and running tests.
 Sets up build output directories if they don't exist.
 Produces simulator builds for non-Intel target architectures.
-Uses Goma by default if it is detected (at output directory setup time).
+Uses reclient by default if it is detected (at output directory setup time).
 Expects to be run from the root of a V8 checkout.
 
 Usage:
@@ -34,6 +34,9 @@ import re
 import subprocess
 import sys
 import shutil
+import time
+
+from enum import IntEnum
 from pathlib import Path
 
 USE_PTY = "linux" in sys.platform
@@ -48,9 +51,9 @@ BUILD_TARGETS_ALL = ["all"]
 
 # All arches that this script understands.
 ARCHES = [
-    "ia32", "x64", "arm", "arm64", "mips64el", "ppc", "ppc64", "riscv32",
-    "riscv64", "s390", "s390x", "android_arm", "android_arm64", "loong64",
-    "fuchsia_x64", "fuchsia_arm64"
+    "ia32", "x64", "arm", "arm64", "mips64el", "ppc64", "riscv32", "riscv64",
+    "s390x", "android_arm", "android_arm64", "loong64", "fuchsia_x64",
+    "fuchsia_arm64", "android_riscv64"
 ]
 # Arches that get built/run when you don't specify any.
 DEFAULT_ARCHES = ["ia32", "x64", "arm", "arm64"]
@@ -69,16 +72,17 @@ DEFAULT_MODES = ["release", "debug"]
 # Build targets that can be manually specified.
 TARGETS = [
     "d8", "cctest", "v8_unittests", "v8_fuzzers", "wasm_api_tests", "wee8",
-    "mkgrokdump", "generate-bytecode-expectations", "inspector-test",
-    "bigint_shell", "wami"
+    "mkgrokdump", "mksnapshot", "generate-bytecode-expectations",
+    "inspector-test", "bigint_shell", "wami", "gn_args"
 ]
 # Build targets that get built when you don't specify any (and specified tests
 # don't imply any other targets).
 DEFAULT_TARGETS = ["d8"]
 # Tests that run-tests.py would run by default that can be run with
 # BUILD_TARGETS_TESTS.
-DEFAULT_TESTS = ["cctest", "debugger", "intl", "message", "mjsunit",
-                 "unittests"]
+DEFAULT_TESTS = [
+    "cctest", "debugger", "filecheck", "intl", "message", "mjsunit", "unittests"
+]
 # These can be suffixed to any <arch>.<mode> combo, or used standalone,
 # or used as global modifiers (affecting all <arch>.<mode> combos).
 ACTIONS = {
@@ -128,6 +132,7 @@ TESTSUITES_TARGETS = {
     "bigint": "bigint_shell",
     "cctest": "cctest",
     "debugger": "d8",
+    "filecheck": "d8",
     "fuzzer": "v8_fuzzers",
     "inspector": "inspector-test",
     "intl": "d8",
@@ -142,43 +147,105 @@ TESTSUITES_TARGETS = {
     "webkit": "d8"
 }
 
-OUTDIR = Path("out")
+out_dir_override = os.getenv("V8_GM_OUTDIR")
+if out_dir_override and Path(out_dir_override).is_file:
+  OUTDIR = Path(out_dir_override)
+else:
+  OUTDIR = Path("out")
+
+build_dir_prefix = os.getenv("V8_GM_BUILD_DIR_PREFIX")
+
+V8_DIR = Path(__file__).resolve().parent.parent.parent
+GCLIENT_FILE_PATH = V8_DIR.parent / ".gclient"
+RECLIENT_CERT_CACHE = V8_DIR / ".#gm_reclient_cert_cache"
+
+BUILD_DISTRIBUTION_RE = re.compile(r"\nuse_(remoteexec|goma) = (false|true)")
+GOMA_DIR_LINE = re.compile(r"\ngoma_dir = \"[^\"]+\"")
+DEPRECATED_RBE_CFG_RE = re.compile(r"\nrbe_cfg_dir = \"[^\"]+\"")
+RECLIENT_CFG_RE = re.compile(r"\nreclient_cfg_dir = \"[^\"]+\"")
+
+class Reclient(IntEnum):
+  NONE = 0
+  GOOGLE = 1
+  CUSTOM = 2
+
+
+def get_v8_solution(solutions):
+  for solution in solutions:
+    if (solution["name"] == "v8" or
+        solution["url"] == "https://chromium.googlesource.com/v8/v8.git"):
+      return solution
+  return None
 
 
 # Note: this function is reused by update-compile-commands.py. When renaming
 # this, please update that file too!
-def detect_goma():
-  if os.environ.get("GOMA_DIR"):
-    return Path(os.environ.get("GOMA_DIR"))
-  if os.environ.get("GOMADIR"):
-    return Path(os.environ.get("GOMADIR"))
-  # There is a copy of goma in depot_tools, but it might not be in use on
-  # this machine.
-  goma = shutil.which("goma_ctl")
-  if goma is None: return None
-  cipd_bin = Path(goma).parent / ".cipd_bin"
-  if not cipd_bin.exists():
-    return None
-  # Some machines have one of these files, some have the other, some have both.
-  goma_auth = Path("~/.goma_client_oauth2_config").expanduser()
-  if goma_auth.exists():
-    return cipd_bin
-  goma_auth = Path("~/.goma_oauth2_config").expanduser()
-  if goma_auth.exists():
-    return cipd_bin
-  return None
+def detect_reclient():
+  if not GCLIENT_FILE_PATH.exists():
+    return Reclient.NONE
+  with open(GCLIENT_FILE_PATH) as f:
+    content = f.read()
+  try:
+    config_dict = {}
+    exec(content, config_dict)
+  except SyntaxError as e:
+    print("# Can't detect reclient due to .gclient syntax errors.")
+    return Reclient.NONE
+  v8_solution = get_v8_solution(config_dict["solutions"])
+  if not v8_solution:
+    print("# Can't detect reclient due to missing v8 gclient solution.")
+    return Reclient.NONE
+  custom_vars = v8_solution.get("custom_vars", {})
+  if custom_vars.get("rbe_instance"):
+    return Reclient.CUSTOM
+  if custom_vars.get("download_remoteexec_cfg"):
+    return Reclient.GOOGLE
+  return Reclient.NONE
 
 
-GOMADIR = detect_goma()
-IS_GOMA_MACHINE = GOMADIR is not None
+# Note: this function is reused by update-compile-commands.py. When renaming
+# this, please update that file too!
+def detect_reclient_cert():
+  now = int(time.time())
+  # We cache the cert expiration time in a file, because that's much faster
+  # to read than invoking `gcertstatus`.
+  if RECLIENT_CERT_CACHE.exists():
+    with open(RECLIENT_CERT_CACHE, 'r') as f:
+      cached_time = int(f.read())
+    if now < cached_time:
+      return True
+  cmd = ["gcertstatus", "-nocheck_ssh", "-format=simple"]
+  ret = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
+  if ret.returncode != 0:
+    return False
+  # Request fresh cert if less than an hour remains. Reproxy will refuse to
+  # start when the certificate is close to expiring.
+  MARGIN = 3600
+  lifetime = int(ret.stdout.decode("utf-8").strip().split(':')[1]) - MARGIN
+  if lifetime < 0:
+    return False
+  with open(RECLIENT_CERT_CACHE, 'w') as f:
+    f.write(str(now + lifetime))
+  return True
 
-USE_GOMA = "true" if IS_GOMA_MACHINE else "false"
+
+RECLIENT_MODE = detect_reclient()
+
+if platform.system() == "Linux":
+  RECLIENT_CFG_REL = "../../buildtools/reclient_cfgs/linux"
+else:
+  RECLIENT_CFG_REL = "../../buildtools/reclient_cfgs"
+
+BUILD_DISTRIBUTION_LINE = ""
+if RECLIENT_MODE:
+  BUILD_DISTRIBUTION_LINE = "\nuse_remoteexec = true"
+  if RECLIENT_MODE == Reclient.CUSTOM:
+    BUILD_DISTRIBUTION_LINE += f"\nreclient_cfg_dir = \"{RECLIENT_CFG_REL}\""
 
 RELEASE_ARGS_TEMPLATE = f"""\
 is_component_build = false
 is_debug = false
-%s
-use_goma = {USE_GOMA}
+%s{BUILD_DISTRIBUTION_LINE}
 v8_enable_backtrace = true
 v8_enable_disassembler = true
 v8_enable_object_print = true
@@ -190,8 +257,7 @@ DEBUG_ARGS_TEMPLATE = f"""\
 is_component_build = true
 is_debug = true
 symbol_level = 2
-%s
-use_goma = {USE_GOMA}
+%s{BUILD_DISTRIBUTION_LINE}
 v8_enable_backtrace = true
 v8_enable_fast_mksnapshot = true
 v8_enable_slow_dchecks = true
@@ -202,8 +268,7 @@ OPTDEBUG_ARGS_TEMPLATE = f"""\
 is_component_build = true
 is_debug = true
 symbol_level = 1
-%s
-use_goma = {USE_GOMA}
+%s{BUILD_DISTRIBUTION_LINE}
 v8_enable_backtrace = true
 v8_enable_fast_mksnapshot = true
 v8_enable_verify_heap = true
@@ -223,15 +288,20 @@ def print_help_and_exit():
   sys.exit(0)
 
 
+# Used by `tools/bash-completion.sh`
 def print_completions_and_exit():
   for a in ARCHES:
     print(str(a))
     for m in set(MODES.values()):
-      print(str(m))
       print(f"{a}.{m}")
       for t in TARGETS:
-        print(str(t))
-        print("{a}.{m}.{t}")
+        print(f"{a}.{m}.{t}")
+      for k in ACTIONS.keys():
+        print(f"{a}.{m}.{k}")
+  for t in TARGETS:
+    print(str(t))
+  for m in set(MODES.values()):
+    print(str(m))
   sys.exit(0)
 
 
@@ -268,12 +338,15 @@ def _call_with_output(cmd):
         output.append(data)
   finally:
     os.close(parent)
-    p.wait()
+    while p.poll() is None:
+      print(".", end="")
+      time.sleep(0.1)
   return p.returncode, "".join(output)
 
 
-def _write(filename, content):
-  print(f"# echo > {filename} << EOF\n{content}EOF")
+def _write(filename, content, log=True):
+  if log:
+    print(f"# echo > {filename} << EOF\n{content}EOF")
   with filename.open("w") as f:
     f.write(content)
 
@@ -291,6 +364,8 @@ def _get_machine():
 
 
 def get_path(arch, mode):
+  if build_dir_prefix:
+    return OUTDIR / f"{build_dir_prefix}-{arch}.{mode}"
   return OUTDIR / f"{arch}.{mode}"
 
 
@@ -328,7 +403,31 @@ class RawConfig:
     self.tests.update(tests)
     self.clean |= clean
 
+  def update_build_distribution_args(self):
+    args_gn = self.path / "args.gn"
+    assert args_gn.exists()
+    with open(args_gn) as f:
+      gn_args = f.read()
+    # Remove custom reclient config path (it will be added again as part of
+    # the config line below if needed).
+    new_gn_args = DEPRECATED_RBE_CFG_RE.sub("", gn_args)
+    new_gn_args = RECLIENT_CFG_RE.sub("", new_gn_args)
+    new_gn_args = BUILD_DISTRIBUTION_RE.sub(BUILD_DISTRIBUTION_LINE,
+                                            new_gn_args)
+    # Remove stale goma_dir to silence GN warnings about unused options.
+    new_gn_args = GOMA_DIR_LINE.sub("", new_gn_args)
+    if gn_args != new_gn_args:
+      print(f"# Updated gn args:{BUILD_DISTRIBUTION_LINE}")
+      _write(args_gn, new_gn_args, log=False)
+
   def build(self):
+    self.update_build_distribution_args()
+    # If the target is to just build args.gn then we are done here; otherwise
+    # drop that target because it's not something ninja can build.
+    if 'gn_args' in self.targets:
+      self.targets.remove('gn_args')
+    if len(self.targets) == 0:
+      return 0
     build_ninja = self.path / "build.ninja"
     if not build_ninja.exists():
       code = _call(f"gn gen {self.path}")
@@ -411,6 +510,8 @@ class ManagedConfig(RawConfig):
       cpu = "arm"
     elif self.arch == "android_arm64" or self.arch == "fuchsia_arm64":
       cpu = "arm64"
+    elif self.arch == "android_riscv64":
+      cpu = "riscv64"
     elif self.arch == "arm64" and _get_machine() in ("aarch64", "arm64"):
       # arm64 build host:
       cpu = "arm64"
@@ -430,15 +531,17 @@ class ManagedConfig(RawConfig):
       v8_cpu = "arm"
     elif self.arch == "android_arm64" or self.arch == "fuchsia_arm64":
       v8_cpu = "arm64"
-    elif self.arch in ("arm", "arm64", "mips64el", "ppc", "ppc64", "riscv64",
-                       "riscv32", "s390", "s390x", "loong64"):
+    elif self.arch == "android_riscv64":
+      v8_cpu = "riscv64"
+    elif self.arch in ("arm", "arm64", "mips64el", "ppc64", "riscv64",
+                       "riscv32", "s390x", "loong64"):
       v8_cpu = self.arch
     else:
       return []
     return [f"v8_target_cpu = \"{v8_cpu}\""]
 
   def get_target_os(self):
-    if self.arch in ("android_arm", "android_arm64"):
+    if self.arch in ("android_arm", "android_arm64", "android_riscv64"):
       return ["target_os = \"android\""]
     elif self.arch in ("fuchsia_x64", "fuchsia_arm64"):
       return ["target_os = \"fuchsia\""]
@@ -477,9 +580,22 @@ class ManagedConfig(RawConfig):
     return super().build()
 
   def run_tests(self):
+    host_arch = _get_machine()
+    if host_arch == "arm64":
+      if platform.system() == "Darwin" and self.arch != "arm64":
+        # MacOS-arm64 doesn't provide a good experience when users
+        # accidentally try to run x64 tests.
+        print(f"Running {self.arch} tests on Mac-arm64 isn't going to work.")
+        return 1
+      if platform.system() == "Linux" and "arm" not in self.arch:
+        # Assume that an arm64 Linux machine may have been set up to run
+        # arm32 binaries and Android on-device tests; refuse everything else.
+        print(f"Running {self.arch} tests on Linux-arm64 isn't going to work.")
+        return 1
     # Special handling for "mkgrokdump": if it was built, run it.
-    if (self.arch == "x64" and self.mode == "release" and
-        "mkgrokdump" in self.targets):
+    if ("mkgrokdump" in self.targets and self.mode == "release" and
+        ((host_arch == "x86_64" and self.arch == "x64") or
+         (host_arch == "arm64" and self.arch == "arm64"))):
       mkgrokdump_bin = self.path / "mkgrokdump"
       _call(f"{mkgrokdump_bin} > tools/v8heapconst.py")
     return super().run_tests()
@@ -547,6 +663,7 @@ class ArgumentParser(object):
       else:
         break
       path_end -= 1
+    targets = targets or DEFAULT_TARGETS
     path = Path('.'.join(words[:path_end]))
     args_gn = path / "args.gn"
     # Only accept existing build output directories, otherwise fall back
@@ -554,7 +671,8 @@ class ArgumentParser(object):
     if not args_gn.is_file():
       return False
     if path not in self.configs:
-      self.configs[path] = RawConfig(path, targets, tests, clean)
+      self.configs[path] = RawConfig(path, targets, tests, clean,
+                                     self.testrunner_args)
     else:
       self.configs[path].extend(targets, tests, clean)
     return True
@@ -570,9 +688,12 @@ class ArgumentParser(object):
     actions = []
     tests = []
     clean = False
-    # Special handling for "mkgrokdump": build it for x64.release.
+    # Special handling for "mkgrokdump": build it for (arm64|x64).release.
     if argstring == "mkgrokdump":
-      self.populate_configs(["x64"], ["release"], ["mkgrokdump"], [], False)
+      arch = "x64"
+      if _get_machine() in ("aarch64", "arm64"):
+        arch = "arm64"
+      self.populate_configs([arch], ["release"], ["mkgrokdump"], [], False)
       return
     if argstring.startswith("--"):
       # Pass all other flags to test runner.
@@ -583,7 +704,9 @@ class ArgumentParser(object):
       return
     # Specifying a single unit test looks like "unittests/Foo.Bar", test262
     # tests have names like "S15.4.4.7_A4_T1", don't split these.
-    if argstring.startswith("unittests/") or argstring.startswith("test262/"):
+    if (argstring.startswith("unittests/") or
+        argstring.startswith("test262/") or argstring.startswith("fuzzer/") or
+        argstring.startswith("wasm-api-tests/")):
       words = [argstring]
     else:
       # Assume it's a word like "x64.release" -> split at the dot.
@@ -636,6 +759,9 @@ class ArgumentParser(object):
   def parse_arguments(self, argv):
     if len(argv) == 0:
       print_help_and_exit()
+    if len(argv) >= 2 and argv[0] == "args" and os.path.exists(argv[1]):
+      code = _call(f"gn args {argv[1]}")
+      sys.exit(code)
     for argstring in argv:
       self.parse_arg(argstring)
     self.process_global_actions()
@@ -648,11 +774,11 @@ def main(argv):
   parser = ArgumentParser()
   configs = parser.parse_arguments(argv[1:])
   return_code = 0
-  # If we have Goma but it is not running, start it.
-  if (IS_GOMA_MACHINE and
-      _call("pgrep -x compiler_proxy > /dev/null", silent=True) != 0):
-    goma_ctl = GOMADIR / "goma_ctl.py"
-    _call(f"{goma_ctl} ensure_start")
+  # If we have Reclient with the Google configuration, check for current
+  # certificate.
+  if (RECLIENT_MODE == Reclient.GOOGLE and not detect_reclient_cert()):
+    print("# gcert")
+    subprocess.check_call("gcert", shell=True)
   for c in configs:
     return_code += configs[c].build()
   if return_code == 0:
